Package: DMRcaller
Type: Package
Title: Differentially Methylated Regions Caller
Version: 1.45.3
Date: 2019-01-18
Encoding: UTF-8
Authors@R: c(person(given="Nicolae Radu", family="Zabet", role =
        c("aut","cre"), email = "r.zabet@qmul.ac.uk"),
        person(given="Jonathan", family="Tsang", role = "aut", email =
        "jmft2@cam.ac.uk"), person(given="Alessandro", family="Greco",
        role = "aut", email = "apgrec@essex.ac.uk"), person(given="Ryan",
        family="Merritt", role = "aut", email = "rmerri@essex.ac.uk"), person(given="Young Jun",
        family="Kim", role = "aut", email = "qc25039@qmul.ac.uk"))
Author: Nicolae Radu Zabet <r.zabet@qmul.ac.uk>, Jonathan
        Michael Foonlan Tsang <jmft2@cam.ac.uk>, Alessandro
        Pio Greco <apgrec@essex.ac.uk>, Ryan Merritt <rmerri@essex.ac.uk> and
        Young Jun Kim <qc25039@qmul.ac.uk>
Maintainer: Nicolae Radu Zabet <r.zabet@qmul.ac.uk>
Description: Uses Bisulfite sequencing data in two conditions and
        identifies differentially methylated regions between the
        conditions in CG and non-CG context. The input is the CX report
        files produced by Bismark and the output is a list of DMRs
        stored as GRanges objects.
License: GPL-3
LazyLoad: yes
Imports: parallel, Rcpp, RcppRoll, betareg, grDevices, graphics,
        methods, stats, utils, Rsamtools, GenomicRanges,
        GenomicAlignments, Biostrings, BSgenome, BiocManager,
        S4Vectors, IRanges, InteractionSet, stringr, inflection,
        BiocParallel, Seqinfo, GenomeInfoDb, cigarillo
Depends: R (>= 3.5), GenomicRanges, IRanges, S4Vectors
Suggests: knitr, RUnit, BiocGenerics, rmarkdown, bookdown, BiocStyle,
        betareg, rtracklayer, BSgenome.Hsapiens.UCSC.hg38
biocViews: DifferentialMethylation, DNAMethylation, Software,
        Sequencing, Coverage
VignetteBuilder: knitr
NeedsCompilation: no
Config/roxygen2/version: 8.0.0
git_url: https://git.bioconductor.org/packages/DMRcaller
git_branch: devel
git_last_commit: cae718d
git_last_commit_date: 2026-07-27
Repository: Bioconductor 3.24
Date/Publication: 2026-07-27
Packaged: 2026-07-27 21:25:35 UTC; biocbuild

# Generated by roxygen2: do not edit by hand

export(analyseReadsInsideRegionsForCondition)
export(analyseReadsInsideRegionsForConditionPMD)
export(computeCoMethylatedPositions)
export(computeCoMethylatedRegions)
export(computeDMRs)
export(computeDMRsReplicates)
export(computeMethylationDataCoverage)
export(computeMethylationDataSpatialCorrelation)
export(computeMethylationProfile)
export(computeOverlapProfile)
export(computePMDs)
export(computeVMDs)
export(filterDMRs)
export(filterPMDs)
export(filterVMDs)
export(filterVMRsONT)
export(getWholeChromosomes)
export(joinReplicates)
export(mergeDMRsIteratively)
export(mergePMDsIteratively)
export(plotLocalMethylationProfile)
export(plotMethylationDataCoverage)
export(plotMethylationDataSpatialCorrelation)
export(plotMethylationProfile)
export(plotMethylationProfileFromData)
export(plotOverlapProfile)
export(poolMethylationDatasets)
export(poolTwoMethylationDatasets)
export(readBismark)
export(readBismarkPool)
export(readONTbam)
export(saveBismark)
export(selectCytosine)
import(BiocParallel)
import(Biostrings)
import(GenomicAlignments)
import(GenomicRanges)
import(IRanges)
import(InteractionSet)
import(Rcpp)
import(Rsamtools)
import(S4Vectors)
import(betareg)
import(inflection)
import(stringr)
importFrom(BiocParallel,SerialParam)
importFrom(BiocParallel,bplapply)
importFrom(Biostrings,getSeq)
importFrom(Biostrings,matchPattern)
importFrom(Biostrings,reverseComplement)
importFrom(GenomicRanges,GRanges)
importFrom(GenomicRanges,GRangesList)
importFrom(GenomicRanges,findOverlaps)
importFrom(GenomicRanges,mcols)
importFrom(IRanges,IRanges)
importFrom(IRanges,IntegerList)
importFrom(RcppRoll,roll_sum)
importFrom(grDevices,colorRampPalette)
importFrom(grDevices,rainbow)
importFrom(graphics,axis)
importFrom(graphics,legend)
importFrom(graphics,lines)
importFrom(graphics,mtext)
importFrom(graphics,par)
importFrom(graphics,points)
importFrom(graphics,rect)
importFrom(graphics,segments)
importFrom(graphics,text)
importFrom(methods,is)
importFrom(parallel,mclapply)
importFrom(stats,cor.test)
importFrom(stats,na.pass)
importFrom(stats,pnorm)
importFrom(stats,setNames)
importFrom(stats,var.test)
importFrom(stats,wilcox.test)
importFrom(utils,capture.output)
importFrom(utils,combn)
importFrom(utils,write.table)

# merge.regions - Get rid of small gaps between regions. Optionally, only glue
# together regions which have the same sign. Also merge together regions which
# overlap, regardless of whether they have the same sign.
.mergeRegions <- function(regions, minGap = 1, respectSigns = TRUE) {
  if (!respectSigns) {
    output <- reduce(regions, drop.empty.ranges=TRUE, min.gapwidth=minGap, 
                     ignore.strand=TRUE)
  } else {
    pos  <- which(regions$direction == 1)
    zero <- which(regions$direction == 0)
    neg  <- which(regions$direction == -1)
    
    preg <- reduce(regions[pos], drop.empty.ranges=TRUE, min.gapwidth=minGap, 
                   ignore.strand=TRUE)
    zreg <- reduce(regions[zero], drop.empty.ranges=TRUE, min.gapwidth=minGap, 
                   ignore.strand=TRUE)
    nreg <- reduce(regions[neg], drop.empty.ranges=TRUE, min.gapwidth=minGap, 
                   ignore.strand=TRUE)
    
    if(length(preg)>0){
      preg$direction <- 1 
    }
    if(length(zreg)>0){ 
      zreg$direction <- 0 
    }
    if(length(nreg)>0){
      nreg$direction <- -1
    }
    
    output <- c(preg,zreg,nreg)        
    if(!isDisjoint(output)){
      output <- disjoin(output)
      output$direction <- rep(0, times=length(output))
      output$direction[overlapsAny(output, regions[regions$direction == -1], 
                                   ignore.strand = TRUE)] <- -1
      output$direction[overlapsAny(output, regions[regions$direction == 1], 
                                   ignore.strand = TRUE)] <- 1
    }
    
    output <- output[order(start(output))]
    output$context <- rep(unique(regions$context), length.out=length(output))
  }
  return(output)
}


#' This takes a list of DMRs and attempts to merge DMRs while keeping the new 
#' DMRs statistically significant
.mergeDMRsIteratively <- function(DMRs, 
                                  minGap, 
                                  respectSigns = TRUE, 
                                  methylationData,
                                  minProportionDifference=0.4, 
                                  minReadsPerCytosine = 4, 
                                  pValueThreshold=0.01,
                                  test="fisher",
                                  alternative = "two.sided"){
  
  overlaps <- countOverlaps(DMRs, DMRs, maxgap = minGap, ignore.strand = TRUE)
  notToJoin <- DMRs[overlaps == 1]
  DMRs <- DMRs[overlaps > 1]
  
  joinedAny <- TRUE
  iteration <- 1
  while(joinedAny){
    joinedAny <- FALSE
    bufferDMRs <-GRanges()    
    index <- 1
    localIndex <- index
    joinedCount <- 0
    while(localIndex < length(DMRs)){
      localDMRs <- DMRs[index]
      canJoin <- TRUE
      while(canJoin){
        localIndex <- localIndex + 1
        newDMRs <-.joinDMRs(c(localDMRs, DMRs[localIndex]),
                            minGap = minGap, 
                            respectSigns = respectSigns, 
                            methylationData = methylationData,
                            minProportionDifference=minProportionDifference, 
                            minReadsPerCytosine = minReadsPerCytosine, 
                            pValueThreshold=pValueThreshold,
                            test=test,
                            alternative = alternative)
        if(length(newDMRs) == 1){
          joinedCount <- joinedCount + 1          
          localDMRs <- newDMRs
          joinedAny <- TRUE
          if(localIndex == length(DMRs)){
            canJoin <- FALSE
            bufferDMRs <- c(bufferDMRs, newDMRs)
          }
        } else{
          bufferDMRs <- c(bufferDMRs, localDMRs)
          canJoin <- FALSE
          index <- localIndex
          if(localIndex == length(DMRs)){
            bufferDMRs <- c(bufferDMRs, DMRs[localIndex])
          }
        }
      }
    }
    DMRs <- bufferDMRs
    iteration <- iteration + 1
  }
  
  DMRs <- c(DMRs, notToJoin)
  
  DMRs <- DMRs[order(DMRs)]
  
  return(DMRs)
}

# join if possible a set of DMRs
.joinDMRs <- function(DMRs,
                      minGap = minGap, 
                      respectSigns = TRUE, 
                      methylationData = methylationData,
                      minProportionDifference=0.4, 
                      minReadsPerCytosine = 4, 
                      pValueThreshold=0.01,
                      test="fisher",
                      alternative = "two.sided"){

  #are within the requeired distance
  if(length(reduce(DMRs, drop.empty.ranges=TRUE, min.gapwidth=minGap, 
                   ignore.strand=TRUE)) == 1){
    #all have the same direction
    if(!respectSigns | length(unique(DMRs$direction)) == 1){
      direction <- unique(DMRs$direction)
      if(length(unique(strand(DMRs))) == 1){
        localDMR <- DMRs[1]
        end(localDMR) <- max(end(DMRs))
        start(localDMR) <- min(start(DMRs))
        localDMR <- .analyseReadsInsideRegions(methylationData, localDMR)
        localDMR$pValue <-.computeaAjustedPValuesInDMRs(test, localDMR, 
                                                        alternative = alternative)
        if(abs(localDMR$proportion1 - localDMR$proportion2) >= minProportionDifference & 
             localDMR$pValue <= pValueThreshold &
             (localDMR$sumReadsN1 / localDMR$cytosinesCount) >= minReadsPerCytosine &
             (localDMR$sumReadsN2 / localDMR$cytosinesCount) >= minReadsPerCytosine){
          DMRs <- localDMR
        }
      }
    }
  }
  return(DMRs)
}

.getLongestDMRs <- function(DMRs,
                            minGap = minGap, 
                            respectSigns = TRUE, 
                            methylationData = methylationData,
                            minProportionDifference=0.4, 
                            minReadsPerCytosine = 4, 
                            pValueThreshold=0.01,
                            test="fisher",
                            alternative = "two.sided"){
  newDMRs <-.joinDMRs(DMRs,
                      minGap = minGap, 
                      respectSigns = respectSigns, 
                      methylationData = methylationData,
                      minProportionDifference=minProportionDifference, 
                      minReadsPerCytosine = minReadsPerCytosine, 
                      pValueThreshold=pValueThreshold,
                      test=test,
                      alternative = alternative)
  if(length(newDMRs) == 1){
    result <- newDMRs
  } else{
    result <- .mergeDMRsIteratively(DMRs, 
                                    minGap = minGap, 
                                    respectSigns = respectSigns, 
                                    methylationData = methylationData,
                                    minProportionDifference=minProportionDifference, 
                                    minReadsPerCytosine = minReadsPerCytosine, 
                                    pValueThreshold=pValueThreshold,
                                    test=test,
                                    alternative = alternative)
      
  }
  return(result)
}

#' This takes a list of DMRs and attempts to merge DMRs while keeping the new 
#' DMRs statistically significant
.smartMergeDMRs <- function(DMRs, 
                            minGap, 
                            respectSigns = TRUE, 
                            methylationData,
                            minProportionDifference=0.4, 
                            minReadsPerCytosine = 4, 
                            pValueThreshold=0.01,
                            test="fisher",
                            alternative = "two.sided",
                            cores = 1,
                            BPPARAM = BPPARAM){  
  overlaps <- countOverlaps(DMRs, DMRs, maxgap = minGap, ignore.strand = TRUE)
  notToJoin <- DMRs[overlaps == 1]
  
  DMRs <- DMRs[overlaps > 1]
  

  if(length(DMRs) > 0){
    overlaps <- findOverlaps(DMRs, 
                             reduce(DMRs, min.gapwidth = minGap, 
                                    ignore.strand=TRUE), 
                             maxgap = minGap, ignore.strand = TRUE)
    DMRsList <- S4Vectors::splitAsList(DMRs[queryHits(overlaps)],  
                                     subjectHits(overlaps))
    
    if(cores > 1){
      bufferDMRs <- BiocParallel::bplapply(1:length(DMRsList), function(i){ .getLongestDMRs(DMRsList[[i]],
                                                                                        minGap = minGap, 
                                                                                        respectSigns = respectSigns, 
                                                                                        methylationData = methylationData,
                                                                                        minProportionDifference=minProportionDifference, 
                                                                                        minReadsPerCytosine = minReadsPerCytosine, 
                                                                                        pValueThreshold=pValueThreshold,
                                                                                        test=test,
                                                                                        alternative = alternative)}, 
                                       BPPARAM = BPPARAM)
      bufferDMRs <- unlist(GRangesList(bufferDMRs))
     } else{
       bufferDMRs <- GRanges()
       for(i in 1:length(DMRsList)){
         bufferDMRs <- c(bufferDMRs, .getLongestDMRs(DMRsList[[i]],
                                                     minGap = minGap, 
                                                     respectSigns = respectSigns, 
                                                     methylationData = methylationData,
                                                     minProportionDifference=minProportionDifference, 
                                                     minReadsPerCytosine = minReadsPerCytosine, 
                                                     pValueThreshold=pValueThreshold,
                                                     test=test,
                                                     alternative = alternative))
       }
    } 

    DMRs <- c(bufferDMRs, notToJoin)
  } else{
    DMRs <- notToJoin
  }
  
  DMRs <- DMRs[order(DMRs)]
  
  return(DMRs)
}





#returns the sum of values in a range
.getSumInRange <- function(vector, regions, start){
  sums <- rep(0, length(regions))
  for(i in 1:length(regions)){
    sums[i] <- sum(vector[(start(regions[i]) - start + 1):(end(regions[i]) - start + 1)])
  }
  return(sums)
}



#' Performs the analysis in all regions in a \code{\link[GenomicRanges]{GRanges}} object
#'
#' @title Analyse reads inside regions
#' @param methylationData a \code{\link[GenomicRanges]{GRanges}} object with five metadata 
#' columns see \code{\link{methylationDataList}}
#' @param regions a \code{\link[GenomicRanges]{GRanges}} object with the identified regions
#' @return a \code{\link[GenomicRanges]{GRanges}} object with eaual sized tiles of the regions. 
#' The object consists of the following metadata 
#' \describe{
#'  \item{sumReadsM1}{the number of methylated reads in condition 1}
#'  \item{sumReadsN1}{the total number of reads in condition 1}
#'  \item{proportion1}{the proportion of methylated reads in condition 1}
#'  \item{sumReadsM2}{the number of methylated reads in condition 2} 
#'  \item{sumReadsN2}{the total number of reads in condition 2} 
#'  \item{proportion2}{the proportion of methylated reads in condition 2} 
#'  \item{cytosinesCount}{the number of cytosines in the correct context} 
#' }  
#'       
#' @author Radu Zabet
.analyseReadsInsideRegions <- function(methylationData, regions){
  
  overlaps <- findOverlaps(methylationData, regions, ignore.strand = TRUE)
  methylationDataContextList <- S4Vectors::splitAsList(methylationData[queryHits(overlaps)],  subjectHits(overlaps))
  regionsIndexes <- as.integer(names(methylationDataContextList))
  
  regions$sumReadsM1 <- rep(0, times=length(regions))
  regions$sumReadsN1 <- rep(0, times=length(regions))    
  regions$proportion1 <- rep(0, times=length(regions))        
  regions$sumReadsM2 <- rep(0, times=length(regions))            
  regions$sumReadsN2 <- rep(0, times=length(regions))            
  regions$proportion2 <- rep(0, times=length(regions))        
  regions$cytosinesCount <- rep(0, times=length(regions))        
  
 
  if(length(regionsIndexes) > 0){  
    regions$sumReadsM1[regionsIndexes] <- sapply(methylationDataContextList,.sumReadsM1)
    regions$sumReadsN1[regionsIndexes] <- sapply(methylationDataContextList,.sumReadsN1)               
    regions$sumReadsM2[regionsIndexes] <- sapply(methylationDataContextList,.sumReadsM2)
    regions$sumReadsN2[regionsIndexes] <- sapply(methylationDataContextList,.sumReadsN2)
    regions$cytosinesCount[regionsIndexes] <- sapply(methylationDataContextList,length)
    
    valid <- regions$cytosinesCount[regionsIndexes] > 0
    regions$proportion1[regionsIndexes[valid]] <- regions$sumReadsM1[regionsIndexes[valid]]/regions$sumReadsN1[regionsIndexes[valid]]      
    regions$proportion2[regionsIndexes[valid]] <- regions$sumReadsM2[regionsIndexes[valid]]/regions$sumReadsN2[regionsIndexes[valid]]
  }
  return(regions)
}


.sumReadsM1 <- function(methylationData){
  return(sum(methylationData$readsM1))
}

.sumReadsN1 <- function(methylationData){
  return(sum(methylationData$readsN1))
}

.sumReadsM2 <- function(methylationData){
  return(sum(methylationData$readsM2))
}

.sumReadsN2 <- function(methylationData){
  return(sum(methylationData$readsN2))
}

#' Performs the analysis in all regions in a \code{\link[GenomicRanges]{GRanges}} object
#'
#' @title Analyse reads inside regions
#' @param methylationData a \code{\link[GenomicRanges]{GRanges}} object with five metadata 
#' columns see \code{\link{methylationDataList}}
#' @param regions a \code{\link[GenomicRanges]{GRanges}} object with the identified regions
#' @return a \code{\link[GenomicRanges]{GRanges}} object with eaual sized tiles of the regions. 
#' The object consists of the following metadata 
#' \describe{
#'  \item{sumReadsM}{the number of methylated reads in condition 1}
#'  \item{sumReadsN}{the total number of reads in condition 1}
#'  \item{proportion}{the proportion of methylated reads in condition 1}
#'  \item{cytosinesCount}{the number of cytosines in the correct context} 
#' }  
#'       
#' @author Radu Zabet
.analyseReadsInsideRegionsForCondition <- function(regions, methylationData, label="", context=""){
  
  overlaps <- findOverlaps(methylationData, regions, ignore.strand = TRUE)
  methylationDataContextList <- S4Vectors::splitAsList(methylationData[queryHits(overlaps)],  subjectHits(overlaps))
  regionsIndexes <- as.integer(names(methylationDataContextList))
  
  sumReadsM <- rep(0, times=length(regions))
  sumReadsN <- rep(0, times=length(regions))    
  proportion <- rep(0, times=length(regions))        
  cytosinesCount <- rep(0, times=length(regions))        
    
  
  sumReadsM[regionsIndexes] <- sapply(methylationDataContextList,.sumReadsM)
  sumReadsN[regionsIndexes] <- sapply(methylationDataContextList,.sumReadsN)               
  cytosinesCount[regionsIndexes] <- sapply(methylationDataContextList,length)
  
  proportion[regionsIndexes] <- sumReadsM[regionsIndexes]/sumReadsN[regionsIndexes]      
  
  
  done <- FALSE
  metadata <- DataFrame("sumReadsM" = sumReadsM, 
                         "sumReadsN" = sumReadsN, 
                         "proportion" = proportion, 
                         "cytosinesCount" = cytosinesCount)
  if(!is.null(label)){
    if(length(label) == 1){
      colnames(metadata) <- c(paste0("sumReadsM", label, context),
                              paste0("sumReadsN", label, context),
                              paste0("proportion", label, context),
                              paste0("cytosinesCount", context))
      done <- TRUE
    }
  }
  
  if(!done){
    colnames(metadata) <- c(paste0("sumReadsM", context),
                            paste0("sumReadsN", context),
                            paste0("proportion", context),
                            paste0("cytosinesCount", context))
  }
  values(regions) <- cbind(values(regions), metadata)
  return(regions)
}


.sumReadsM <- function(methylationData){
  return(sum(methylationData$readsM))
}

.sumReadsN <- function(methylationData){
  return(sum(methylationData$readsN))
}



#' This function counts the number of cytosines in each DMR
#'
#' @title Count cytosines inside
#' @param methylationData a \code{\link[GenomicRanges]{GRanges}} object with five metadata 
#' columns; see \code{\link{methylationDataList}}
#' @param regions a \code{\link[GenomicRanges]{GRanges}} with the computed regions
#' @return a \code{vector} with the number of cytosines in each DMR
#'  
#' @author Jonathan Michael Foonlan Tsang
.countCytosinesInside <- function(methylationData, regions) {
  return(countSubjectHits(findOverlaps(methylationData, regions, ignore.strand = TRUE)))
}


#' Performs the analysis in equal width regions of an \code{\link[GenomicRanges]{GRanges}} 
#' object
#'
#' @title Analyse reads inside regions
#' @param methylationData a \code{\link[GenomicRanges]{GRanges}} object with five metadata 
#' columns; see \code{\link{methylationDataList}}
#' @param regions a \code{\link[GenomicRanges]{GRanges}} object with the identified regions
#' @return a \code{\link[GenomicRanges]{GRanges}} object with eaual sized tiles of the regions. 
#' The object consists of the following metadata 
#' \describe{
#'  \item{sumReadsM1}{the number of methylated reads in condition 1}
#'  \item{sumReadsN1}{the total number of reads in condition 1}
#'  \item{proportion1}{the proportion of methylated reads in condition 1}
#'  \item{sumReadsM2}{the number of methylated reads in condition 2} 
#'  \item{sumReadsN2}{the total number of reads in condition 2} 
#'  \item{proportion2}{the proportion of methylated reads in condition 2} 
#'  \item{cytosinesCount}{the number of cytosines in the correct context} 
#' }  
#'       
#' @author Radu Zabet
.analyseReadsInsideBins <- function(methylationData, bins, currentRegion){
  
  binSize <- min(unique(width(bins)))
  #Rcpp
  readsM1 <- .movingSum(start(currentRegion), end(currentRegion), start(methylationData), methylationData$readsM1, windowSize = binSize)
  sumReadsM1 <- readsM1[seq(1,length(readsM1)-binSize, by=binSize)]
  
  readsN1 <- .movingSum(start(currentRegion), end(currentRegion), start(methylationData), methylationData$readsN1, windowSize = binSize)
  sumReadsN1 <- readsN1[seq(1,length(readsN1)-binSize, by=binSize)]
  
  proportion1 <- sumReadsM1/sumReadsN1
  
  readsM2 <- .movingSum(start(currentRegion), end(currentRegion), start(methylationData), methylationData$readsM2, windowSize = binSize)
  sumReadsM2 <- readsM2[seq(1,length(readsM2)-binSize, by=binSize)]
  
  readsN2 <- .movingSum(start(currentRegion), end(currentRegion), start(methylationData), methylationData$readsN2, windowSize = binSize)
  sumReadsN2 <- readsN2[seq(1,length(readsN2)-binSize, by=binSize)]
  
  proportion2 <- sumReadsM2/sumReadsN2
  
  
  cytosines <- .movingSum(start(currentRegion), end(currentRegion), start(methylationData), rep(1, length(start(methylationData))), windowSize = binSize)
  cytosinesCount <- cytosines[seq(1,length(cytosines)-binSize, by=binSize)]
  
  bins$sumReadsM1 <- sumReadsM1
  bins$sumReadsN1 <- sumReadsN1    
  bins$proportion1 <- proportion1        
  bins$sumReadsM2 <- sumReadsM2           
  bins$sumReadsN2 <- sumReadsN2            
  bins$proportion2 <- proportion2       
  bins$cytosinesCount <- cytosinesCount 

  return(bins)
}


#' This function splits a set of GRanges into a list of equally width GRanges 
#' objects
#'
#' @title .splitGRangesEqualy
#' @param regions a \code{\link[GenomicRanges]{GRanges}} object 
#' @param breaks number of elements to break the GRanges
#' @return a \code{list} consisting of GRanges objects of same total width
#'  
#' @author Radu Zabet
.splitGRangesEqualy <- function(regions, breaks=1){
  result <- lapply(split(unlist(tile(regions, n=breaks)), rep(1:breaks, each=length(regions))), reduce, ignore.strand=TRUE)
  return(result)
}




#' Performs the analysis in equal width regions of an \code{\link[GenomicRanges]{GRanges}} 
#' object
#'
#' @title Analyse reads inside regions for one sample
#' @param methylationData a \code{\link[GenomicRanges]{GRanges}} object with five metadata 
#' columns; see \code{\link{methylationDataList}}.
#' @param bins a \code{\link[GenomicRanges]{GRanges}} object with the bins
#' @return a \code{\link[GenomicRanges]{GRanges}} object with eaual sized tiles of the regions. 
#' The object consists of the following metadata 
#' \describe{
#'  \item{sumReadsM}{the number of methylated reads}
#'  \item{sumReadsN}{the total number of reads}
#'  \item{Proportion}{the proportion of methylated reads}
#' }  
#'       
#' @author Radu Zabet
.analyseReadsInsideBinsOneSample <- function(methylationData, bins, currentRegion){
  
  binSize <- min(unique(width(bins)))
  #Rcpp
  readsM <- .movingSum(start(currentRegion), end(currentRegion), start(methylationData), methylationData$readsM, windowSize = binSize)
  sumReadsM <- readsM[seq(1,length(readsM)-binSize, by=binSize)]
  
  readsN <- .movingSum(start(currentRegion), end(currentRegion), start(methylationData), methylationData$readsN, windowSize = binSize)
  sumReadsN <- readsN[seq(1,length(readsN)-binSize, by=binSize)]
  
  Proportion <- sumReadsM/sumReadsN

  cytosines <- .movingSum(start(currentRegion), end(currentRegion), start(methylationData), rep(1, length(start(methylationData))), windowSize = binSize)
  cytosinesCount <- cytosines[seq(1,length(cytosines)-binSize, by=binSize)]
  
  bins$sumReadsM <- sumReadsM
  bins$sumReadsN <- sumReadsN    
  bins$Proportion <- Proportion        
  bins$cytosinesCount <- cytosinesCount 
  
  return(bins)
}




#' Counts the number of cytosines in all the regions
#'
#' @title Count total number of Cytosines
#' @param methylationData a \code{\link[GenomicRanges]{GRanges}} object with five metadata 
#' columns: see \code{\link{methylationDataList}}
#' @param regions a \code{\link[GenomicRanges]{GRanges}} object with the identified regions
#' @return a the total number of Cytosines
#'       
#' @author Radu Zabet
.countTotalNumberOfCytosines <- function(methylationData, regions){
  
  total_cytocines <- 0
  for (index in 1:length(regions)) {
    currentRegion <- regions[index]
    total_cytocines <- total_cytocines + length(methylationData[queryHits(findOverlaps(methylationData, currentRegion, ignore.strand = TRUE))])
  }
  return(total_cytocines)
}


#' Performs the analysis in all regions in a \code{\link[GenomicRanges]{GRanges}} object
#'
#' @title Analyse reads inside regions
#' @param methylationData a \code{\link[GenomicRanges]{GRanges}} object with four metadata 
#' columns; see \code{\link{MethylationDataList}}.
#' @param regions a \code{\link[GenomicRanges]{GRanges}} object with the identified regions
#' @return a \code{\link[GenomicRanges]{GRanges}} object with eaual sized tiles of the regions. 
#' The object consists of the following metadata.
#' \describe{
#'  \item{sumReadsM}{the number of methylated reads in condition 1}
#'  \item{sumReadsN}{the total number of reads in condition 1}
#'  \item{Proportion}{the proportion of methylated reads in condition 1}
#' }  
#'       
#' @author Radu Zabet
.analyseReadsInsideRegionsOneSample <- function(methylationData, regions){
  overlaps <- findOverlaps(methylationData, regions, ignore.strand = TRUE)
  methylationDataContextList <- S4Vectors::splitAsList(methylationData[queryHits(overlaps)],  subjectHits(overlaps))
  regionsIndexes <- as.integer(names(methylationDataContextList))
  
  regions$sumReadsM <- rep(0, times=length(regions))
  regions$sumReadsN <- rep(0, times=length(regions))    
  regions$Proportion <- rep(0, times=length(regions))        
  regions$cytosinesCount <- rep(0, times=length(regions))
  
  regions$sumReadsM[regionsIndexes] <- sapply(methylationDataContextList,.sumReadsM)
  regions$sumReadsN[regionsIndexes] <- sapply(methylationDataContextList,.sumReadsN)  
  regions$Proportion[regionsIndexes] <- regions$sumReadsM[regionsIndexes]/regions$sumReadsN[regionsIndexes]      
  regions$cytosinesCount[regionsIndexes] <- sapply(methylationDataContextList, length)
  
  return(regions)
}

.sumReadsM <- function(methylationData){
  return(sum(methylationData$readsM))
}

.sumReadsN <- function(methylationData){
  return(sum(methylationData$readsN))
}
 

#' This takes a list of PMDs and attempts to merge PMDs while keeping the new 
#' PMDs filtering by the preset parameter
.mergePMDsIteratively <- function(PMDs, 
                                  minGap, 
                                  respectSigns = TRUE, 
                                  methylationData,
                                  minReadsPerCytosine = 4, 
                                  minMethylation = 0.4,
                                  maxMethylation = 0.6){
  
  overlaps <- countOverlaps(PMDs, PMDs, maxgap = minGap, ignore.strand = TRUE)
  notToJoin <- PMDs[overlaps == 1]
  PMDs <- PMDs[overlaps > 1]
  
  joinedAny <- TRUE
  iteration <- 1
  while(joinedAny){
    joinedAny <- FALSE
    bufferPMDs <-GRanges()    
    index <- 1
    localIndex <- index
    joinedCount <- 0
    while(localIndex < length(PMDs)){
      localPMDs <- PMDs[index]
      canJoin <- TRUE
      while(canJoin){
        localIndex <- localIndex + 1
        newPMDs <-.joinPMDs(c(localPMDs, PMDs[localIndex]),
                            minGap = minGap, 
                            respectSigns = respectSigns, 
                            methylationData = methylationData,
                            minReadsPerCytosine = minReadsPerCytosine, 
                            minMethylation = minMethylation,
                            maxMethylation = maxMethylation)
        if(length(newPMDs) == 1){
          joinedCount <- joinedCount + 1          
          localPMDs <- newPMDs
          joinedAny <- TRUE
          if(localIndex == length(PMDs)){
            canJoin <- FALSE
            bufferPMDs <- c(bufferPMDs, newPMDs)
          }
        } else{
          bufferPMDs <- c(bufferPMDs, localPMDs)
          canJoin <- FALSE
          index <- localIndex
          if(localIndex == length(PMDs)){
            bufferPMDs <- c(bufferPMDs, PMDs[localIndex])
          }
        }
      }
    }
    PMDs <- bufferPMDs
    iteration <- iteration + 1
  }
  
  PMDs <- c(PMDs, notToJoin)
  
  PMDs <- PMDs[order(PMDs)]
  
  return(PMDs)
}

# join if possible a set of PMDs
.joinPMDs <- function(PMDs,
                      minGap = minGap, 
                      respectSigns = TRUE, 
                      methylationData,
                      minReadsPerCytosine = 4, 
                      minMethylation = 0.4,
                      maxMethylation = 0.6){
  
  #are within the requeired distance
  if(length(reduce(PMDs, drop.empty.ranges=TRUE, min.gapwidth=minGap, 
                   ignore.strand=TRUE)) == 1){
    #all have the same direction
    if(!respectSigns | length(unique(PMDs$direction)) == 1){
      direction <- unique(PMDs$direction)
      if(length(unique(strand(PMDs))) == 1){
        localPMD <- PMDs[1]
        end(localPMD) <- max(end(PMDs))
        start(localPMD) <- min(start(PMDs))
        localPMD <- .analyseReadsInsideRegionsPMDs(methylationData, localPMD)
        if(localPMD$proportion >= minMethylation & localPMD$proportion <= maxMethylation &
           localPMD$sumReadsN / localPMD$cytosinesCount >= minReadsPerCytosine){
          PMDs <- localPMD
        }
      }
    }
  }
  return(PMDs)
}

.getLongestPMDs <- function(PMDs,
                            minGap = minGap, 
                            respectSigns = TRUE, 
                            methylationData = methylationData,
                            minReadsPerCytosine = 4,
                            minMethylation = 0.4,
                            maxMethylation = 0.6){
  newPMDs <-.joinPMDs(PMDs,
                      minGap = minGap, 
                      respectSigns = respectSigns, 
                      methylationData = methylationData,
                      minReadsPerCytosine = minReadsPerCytosine, 
                      minMethylation = minMethylation,
                      maxMethylation = maxMethylation)
  if(length(newPMDs) == 1){
    result <- newPMDs
  } else{
    result <- .mergePMDsIteratively(PMDs, 
                                    minGap = minGap, 
                                    respectSigns = respectSigns, 
                                    methylationData = methylationData,
                                    minReadsPerCytosine = minReadsPerCytosine, 
                                    minMethylation = minMethylation,
                                    maxMethylation = maxMethylation)
    
  }
  return(result)
}

#' This takes a list of PMDs and attempts to merge PMDs while keeping the new 
#' PMDs statistically significant
.smartMergePMDs <- function(PMDs, 
                            minGap, 
                            respectSigns = TRUE, 
                            methylationData,
                            minReadsPerCytosine = 4, 
                            minMethylation = 0.4,
                            maxMethylation = 0.6,
                            cores = 1,
                            BPPARAM = BPPARAM){  
  overlaps <- countOverlaps(PMDs, PMDs, maxgap = minGap, ignore.strand = TRUE)
  notToJoin <- PMDs[overlaps == 1]
  
  PMDs <- PMDs[overlaps > 1]
  
  
  if(length(PMDs) > 0){
    overlaps <- findOverlaps(PMDs, 
                             reduce(PMDs, min.gapwidth = minGap, 
                                    ignore.strand=TRUE), 
                             maxgap = minGap, ignore.strand = TRUE)
    PMDsList <- S4Vectors::splitAsList(PMDs[queryHits(overlaps)],  
                                       subjectHits(overlaps))
    
    if(cores > 1){
      bufferPMDs <- BiocParallel::bplapply(1:length(PMDsList), function(i){ .getLongestPMDs(PMDsList[[i]],
                                                                                        minGap = minGap, 
                                                                                        respectSigns = respectSigns, 
                                                                                        methylationData = methylationData,
                                                                                        minReadsPerCytosine = minReadsPerCytosine, 
                                                                                        minMethylation = minMethylation,
                                                                                        maxMethylation = maxMethylation)}, 
                                       BPPARAM = BPPARAM)
      bufferPMDs <- unlist(GRangesList(bufferPMDs))
    } else{
      bufferPMDs <- GRanges()
      for(i in 1:length(PMDsList)){
        bufferPMDs <- c(bufferPMDs, .getLongestPMDs(PMDsList[[i]],
                                                    minGap = minGap, 
                                                    respectSigns = respectSigns, 
                                                    methylationData = methylationData,
                                                    minReadsPerCytosine = minReadsPerCytosine, 
                                                    minMethylation = minMethylation,
                                                    maxMethylation = maxMethylation))
      }
    } 
    
    PMDs <- c(bufferPMDs, notToJoin)
  } else{
    PMDs <- notToJoin
  }
  
  PMDs <- PMDs[order(PMDs)]
  
  return(PMDs)
}

#' Performs the analysis in all regions in a \code{\link[GenomicRanges]{GRanges}} object for 
#' PMD calculation
#'
#' @title Analyse reads inside regions
#' @param methylationData a \code{\link[GenomicRanges]{GRanges}} object with five metadata 
#' columns see \code{\link{methylationDataList}}
#' @param regions a \code{\link[GenomicRanges]{GRanges}} object with the identified regions
#' @return a \code{\link[GenomicRanges]{GRanges}} object with eaual sized tiles of the regions. 
#' The object consists of the following metadata 
#' \describe{
#'  \item{sumReadsM}{the number of methylated reads}
#'  \item{sumReadsN}{the total number of reads}
#'  \item{proportion}{the proportion of methylated reads}
#'  \item{ONT_Cm}{comma-delimited modified read‐indices}
#'  \item{ONT_C}{comma-delimited read‐indices covering but unmodified}
#'  \item{cytosinesCount}{the number of cytosines in the correct context} 
#' }  
#'       
#' @author Radu Zabet and Young Jun Kim
#' 
.analyseReadsInsideRegionsPMDs <- function(methylationData, regions){
  
  overlaps <- findOverlaps(methylationData, regions, ignore.strand = TRUE)
  methylationDataContextList <- S4Vectors::splitAsList(methylationData[queryHits(overlaps)],  subjectHits(overlaps))
  regionsIndexes <- as.integer(names(methylationDataContextList))
  
  regions$sumReadsM <- rep(0, times=length(regions))
  regions$sumReadsN <- rep(0, times=length(regions))    
  regions$proportion <- rep(0, times=length(regions)) 
  regions$cytosinesCount <- rep(0, times=length(regions))        
  
  
  if(length(regionsIndexes) > 0){  
    regions$sumReadsM[regionsIndexes] <- sapply(methylationDataContextList,.sumReadsM)
    regions$sumReadsN[regionsIndexes] <- sapply(methylationDataContextList,.sumReadsN)               
    regions$cytosinesCount[regionsIndexes] <- sapply(methylationDataContextList,length)
    
    valid <- regions$cytosinesCount[regionsIndexes] > 0
    regions$proportion[regionsIndexes[valid]] <- regions$sumReadsM[regionsIndexes[valid]]/regions$sumReadsN[regionsIndexes[valid]]     
  }
  return(regions)
}


.sumReadsM <- function(methylationData){
  return(sum(methylationData$readsM))
}
.sumReadsN <- function(methylationData){
  return(sum(methylationData$readsN))
}

#' Performs the analysis in equal width regions of an \code{\link[GenomicRanges]{GRanges}} 
#' object
#'
#' @title Analyse reads inside regions
#' @param methylationData a \code{\link[GenomicRanges]{GRanges}} object with five metadata 
#' columns; see \code{\link{methylationDataList}}
#' @param regions a \code{\link[GenomicRanges]{GRanges}} object with the identified regions
#' @return a \code{\link[GenomicRanges]{GRanges}} object with eaual sized tiles of the regions. 
#' The object consists of the following metadata 
#' \describe{
#'  \item{sumReadsM1}{the number of methylated reads in condition 1}
#'  \item{sumReadsN1}{the total number of reads in condition 1}
#'  \item{proportion1}{the proportion of methylated reads in condition 1}
#'  \item{sumReadsM2}{the number of methylated reads in condition 2} 
#'  \item{sumReadsN2}{the total number of reads in condition 2} 
#'  \item{proportion2}{the proportion of methylated reads in condition 2} 
#'  \item{cytosinesCount}{the number of cytosines in the correct context} 
#' }  
#'       
#' @author Radu Zabet and Young Jun Kim
.analyseReadsInsideBinsPMDs <- function(methylationData, bins, currentRegion){
  
  binSize <- min(unique(width(bins)))
  #Rcpp
  readsM <- .movingSum(start(currentRegion), end(currentRegion), start(methylationData), methylationData$readsM, windowSize = binSize)
  sumReadsM <- readsM[seq(1,length(readsM)-binSize, by=binSize)]
  
  readsN <- .movingSum(start(currentRegion), end(currentRegion), start(methylationData), methylationData$readsN, windowSize = binSize)
  sumReadsN <- readsN[seq(1,length(readsN)-binSize, by=binSize)]
  
  proportion <- sumReadsM/sumReadsN
  
  cytosines <- .movingSum(start(currentRegion), end(currentRegion), start(methylationData), rep(1, length(start(methylationData))), windowSize = binSize)
  cytosinesCount <- cytosines[seq(1,length(cytosines)-binSize, by=binSize)]
  
  bins$sumReadsM <- sumReadsM
  bins$sumReadsN <- sumReadsN   
  bins$proportion <- proportion        
  bins$cytosinesCount <- cytosinesCount 
  
  return(bins)
}

#' This function computes the partially methylated domains between pre-set
#' min and max proportion values.  
#'
#' @title Compute PMDs
#' @param methylationData the methylation data in condition
#' (see \code{\link{ontSampleGRangesList}}).
#' @param regions a \code{\link[GenomicRanges]{GRanges}} object with the regions where to 
#' compute the PMDs. If \code{NULL}, the PMDs are computed genome-wide.
#' @param context the context in which the PMDs are computed (\code{"CG"}, 
#' \code{"CHG"} or \code{"CHH"}).
#' @param method Character string specifying the algorithm for PMD detection. 
#' If \code{"noise_filter"}, a sliding window of size \code{windowSize} is applied 
#' with the specified \code{kernelFunction} (and \code{lambda} for a Gaussian kernel) 
#' to smooth methylation counts before calling and merging PMDs. 
#' If \code{"neighbourhood"}, individual partially methylated cytosines are 
#' identified first and then merged into PMDs. 
#' If \code{"bins"}, the genome is partitioned into fixed bins of size \code{binSize}, 
#' partially methylation is sorted per bin, and significant bins are merged. 
#' @param windowSize the size of the triangle base measured in nucleotides. 
#' This parameter is required only if the selected method is 
#' \code{"noise_filter"}. 
#' @param kernelFunction a \code{character} indicating which kernel function to 
#' be used. Can be one of \code{"uniform"}, \code{"triangular"}, 
#' \code{"gaussian"} or \code{"epanechnicov"}. This is required only if the 
#' selected method is \code{"noise_filter"}. 
#' @param lambda numeric value required for the Gaussian filter 
#' (\code{K(x) = exp(-lambda*x^2)}). This is required only if the selected 
#' method is \code{"noise_filter"} and the selected kernel function is 
#' \code{"gaussian"}. 
#' @param binSize the size of the tiling bins in nucleotides. This parameter is 
#' required only if the selected method is \code{"bins"}.
#' @param minCytosinesCount PMDs with less cytosines in the specified context 
#' than \code{minCytosinesCount} will be discarded.
#' @param minMethylation Numeric [0,1]; minimum mean methylation proportion.
#' @param maxMethylation Numeric [0,1]; maximum mean methylation proportion.
#' @param minGap PMDs separated by a gap of at least \code{minGap} are not 
#' merged. Note that only PMDs where the change in methylation is in the same 
#' direction are joined.
#' @param minSize PMDs with a size smaller than \code{minSize} are discarded.
#' @param minReadsPerCytosine  PMDs with the average number of reads lower than 
#' \code{minReadsPerCytosine} are discarded. 
#' @param parallel Logical; run in parallel if \code{TRUE}.
#' @param BPPARAM A \code{BiocParallelParam} object controlling parallel execution.
#'    This value will automatically set when parallel is \code{TRUE}, also able to set as manually.
#' @param cores Integer number of workers (must not exceed BPPARAM$workers).
#'    This value will automatically set as the maximum number of system workers,
#'    also able to set as manually.
#' @return the PMDs stored as a \code{\link[GenomicRanges]{GRanges}} object with the following 
#' metadata columns:
#' \describe{
#'  \item{context}{the context in which the PMDs was computed (\code{"CG"}, 
#'  \code{"CHG"} or \code{"CHH"}).}
#'  \item{sumReadsM}{the number of methylated reads.}
#'  \item{sumReadsN}{the total number of reads.} 
#'  \item{proportion}{the proportion methylated reads filtered between
#'  \code{minMethylation} and \code{maxMethylation}.} 
#'  \item{cytosinesCount}{the number of cytosines in the PMDs.} 
#' }
#' @seealso \code{\link{readONTbam}}, \code{\link{filterPMDs}}, \code{\link{mergePMDsIteratively}}, 
#' \code{\link{analyseReadsInsideRegionsForConditionPMD}} and 
#' \code{\link{PMDsNoiseFilterCG}}
#' @examples
#' # load the ONT methylation data 
#' data(ontSampleGRangesList)
#' 
#' # the regions where to compute the PMDs
#' chr1_ranges <- GRanges(seqnames = Rle("chr1"), ranges = IRanges(1E6+5E5,2E6))
#' 
#' # compute the PMDs in CG context with noise_filter method
#' PMDsNoiseFilterCG <- computePMDs(ontSampleGRangesList[["GM18501"]],
#'                                  regions = chr1_ranges,
#'                                  context = "CG",
#'                                  windowSize = 100,
#'                                  method = "noise_filter",
#'                                  kernelFunction = "triangular", 
#'                                  lambda = 0.5,
#'                                  minCytosinesCount = 4, 
#'                                  minMethylation = 0.4, 
#'                                  maxMethylation = 0.6, 
#'                                  minGap = 200, 
#'                                  minSize = 50, 
#'                                  minReadsPerCytosine = 4, 
#'                                  cores = 1,
#'                                  parallel = FALSE)
#'  
#' \dontrun{                                
#' # compute the PMDs in CG context with neighbourhood method
#' PMDsNeighbourhoodCG <- computePMDs(ontSampleGRangesList[["GM18501"]],
#'                                    regions = chr1_ranges,
#'                                    context = "CG",
#'                                    method = "neighbourhood"
#'                                    minCytosinesCount = 4, 
#'                                    minMethylation = 0.4, 
#'                                    maxMethylation = 0.6, 
#'                                    minGap = 200, 
#'                                    minSize = 50, 
#'                                    minReadsPerCytosine = 4, 
#'                                    cores = 1,
#'                                    parallel = FALSE)
#'                                    
#' # compute the PMDs in CG context with bins method
#' PMDsBinsCG <- computePMDs(ontSampleGRangesList[["GM18501"]],
#'                           regions = chr1_ranges,
#'                           context = "CG",
#'                           method = "bins",
#'                           binSize = 100,
#'                           minCytosinesCount = 4, 
#'                           minMethylation = 0.4, 
#'                           maxMethylation = 0.6, 
#'                           minGap = 200, 
#'                           minSize = 50, 
#'                           minReadsPerCytosine = 4, 
#'                           cores = 1,
#'                           parallel = FALSE)
#' }
#' @author Nicolae Radu Zabet, Jonathan Michael Foonlan Tsang and Young Jun Kim
#' @export
computePMDs <- function(methylationData, 
                        regions = NULL, 
                        context = "CG", 
                        method="noise_filter",
                        windowSize = 100,
                        kernelFunction = "triangular", 
                        lambda = 0.5,
                        binSize = 100,
                        minCytosinesCount = 4, 
                        minMethylation = 0.4, 
                        maxMethylation = 0.6, 
                        minGap = 200, 
                        minSize = 50, 
                        minReadsPerCytosine = 4, 
                        parallel = FALSE,
                        BPPARAM = NULL,
                        cores = NULL) {
  ##Parameters checking
  cat("Parameters checking ...\n")
  
  # generate the BPPARAM value if set as parallel 
  if (parallel == TRUE){
    BPPARAM <- suppressWarnings(.validateBPPARAM(BPPARAM, progressbar = TRUE)) 
  }else{
    # Force serial execution
    BPPARAM <- BiocParallel::SerialParam(progressbar = TRUE)
  }
  # If cores argument is specified
  if (!is.null(cores)) {
    .stopIfNotAll(.isInteger(cores, positive = TRUE), 
                  "the number of cores used when computing the DMRs needs to be an integer higher or equal to 1.")
    
    # Check if user requested more cores than available
    if (cores > BPPARAM$workers) {
      warning(paste0("The number of requested cores (", cores, 
                     ") exceeds the available system cores (", BPPARAM$workers, 
                     "). Automatically setting cores to the maximum available (", 
                     BPPARAM$workers, ")."))
      cores <- BPPARAM$workers
    } else {
      message(paste0("Using user-specified core count: ", cores))
    }
    
    # Apply the final core number
    BPPARAM$workers <- cores
  } else {
    cores <- BPPARAM$workers
  }
  cat("Current parallel setting, BPPARAM: ", capture.output(BPPARAM),sep = "\n")
  
  .validateMethylationData(methylationData, variableName="methylationData")
  
  
  regions <- .validateGRanges(regions, methylationData)
  
  .validateContext(context)
  
  .stopIfNotAll(c(!is.null(method), 
                  all(is.character(method)),
                  length(method) == 1,
                  all(method %in% c("noise_filter","neighbourhood","bins"))),
                " method can be only noise_filter, neighbourhood or bins")  
  
  if(method == "noise_filter"){
    .stopIfNotAll(c(.isInteger(windowSize, positive=TRUE)), 
                  " the window size used by the interpolation method is an integer higher than 0")
    
    .stopIfNotAll(c(!is.null(kernelFunction), 
                    kernelFunction%in%c("uniform", "triangular", "gaussian", "epanechnicov")), 
                  paste("Unknown kernel function: ", kernelFunction, ". 
                        kernelFunction should be one of \"uniform\", \"triangular\", 
                        \"gaussian\", \"epanechnicov\"",sep=""))
    
    if(kernelFunction == "gaussian"){
      .stopIfNotAll(c(!is.null(lambda),
                      is.numeric(lambda)), 
                    " lambda needs to be a numeric value.")
    }
    
  }
  
  if(method == "bins"){
    .stopIfNotAll(c(.isInteger(binSize, positive=TRUE)), 
                  " the bin size used by the method is an integer higher than 0")
    
  }
  
  .stopIfNotAll(c(.isInteger(minCytosinesCount, positive=TRUE)), 
                " the minCytosinesCount is an integer higher or equal to 0")
  
  .stopIfNotAll(c(!is.null(minMethylation), is.numeric(minMethylation), minMethylation >= 0), 
                " minMethylation needs to be a numeric value higher or equal to 0")
  
  .stopIfNotAll(c(!is.null(maxMethylation), is.numeric(maxMethylation), maxMethylation >= 0), 
                " maxMethylation needs to be a numeric value higher or equal to 0")
  
  .stopIfNotAll(c(maxMethylation >= minMethylation), 
                " maxMethylation should be higher than minMethylation value")
  
  .stopIfNotAll(c(.isInteger(minGap, positive=TRUE)),
                " the minimum gap between PMDs is an integer higher or equal to 0")
  
  .stopIfNotAll(c(.isInteger(minSize, positive=TRUE)),
                " the minimum size of a DMR is an integer higher or equal to 0")
  
  .stopIfNotAll(c(.isInteger(minReadsPerCytosine, positive=TRUE)), 
                " the minimum average number of reads in a DMR is an integer higher or equal to 0")
  
  computedPMDs <- GRanges()
  
  if(method == "noise_filter"){
    computedPMDs <- .computePMDsNoiseFilter(methylationData = methylationData, 
                                            regions = regions, 
                                            context = context,
                                            windowSize = windowSize,
                                            kernelFunction = kernelFunction, 
                                            lambda = lambda,
                                            minCytosinesCount = minCytosinesCount,
                                            minMethylation = minMethylation,
                                            maxMethylation = maxMethylation,
                                            minGap = minGap, 
                                            minSize = minSize, 
                                            minReadsPerCytosine = minReadsPerCytosine, 
                                            cores = cores,
                                            BPPARAM = BPPARAM)
  } else if(method == "neighbourhood"){
    computedPMDs <- .computePMDsNeighbourhood(methylationData = methylationData, 
                                              regions = regions, 
                                              context = context, 
                                              minCytosinesCount = minCytosinesCount, 
                                              minMethylation = minMethylation,
                                              maxMethylation = maxMethylation,
                                              minGap = minGap, 
                                              minSize = minSize, 
                                              minReadsPerCytosine = minReadsPerCytosine, 
                                              cores = cores,
                                              BPPARAM = BPPARAM)
  } else if(method == "bins"){
    computedPMDs <- .computePMDsBins(methylationData = methylationData, 
                                     regions = regions, 
                                     context = context, 
                                     binSize = binSize,
                                     minCytosinesCount = minCytosinesCount, 
                                     minMethylation = minMethylation,
                                     maxMethylation = maxMethylation,
                                     minGap = minGap, 
                                     minSize = minSize, 
                                     minReadsPerCytosine = minReadsPerCytosine, 
                                     cores = cores,
                                     BPPARAM = BPPARAM)
    
  } else{
    cat("Unknown method: ",method," \n")
  } 
  
  return(computedPMDs)  
}

#' This function computes the partially methylated regions 
#' using the noise filter method.  
.computePMDsNoiseFilter <- function(methylationData, 
                                    regions = NULL, 
                                    context = "CG",
                                    windowSize = 100,
                                    kernelFunction="triangular", 
                                    lambda=0.5,                                    
                                    minCytosinesCount = 4, 
                                    minMethylation = 0.4, 
                                    maxMethylation = 0.6,
                                    minGap = 200, 
                                    minSize = 50, 
                                    minReadsPerCytosine = 4, 
                                    cores = 1,
                                    BPPARAM = BPPARAM) {
  
  regions <- reduce(regions)
  
  # extract the methylation data in the correct context
  cat("Extract methylation in the corresponding context \n")
  
  contextMethylationData <- methylationData[methylationData$context%in%context]
  rm(methylationData)
  localContextMethylationData <- contextMethylationData[queryHits(findOverlaps(contextMethylationData, regions))]
  rm(contextMethylationData)  
  
  regionsList <- .splitGRangesEqualy(regions, cores)
  
  # inner loop function for BiocParallel::bplapply
  .computePMDsInterpolationLoop = function(i){
    computedPMDs <- GRanges()    
    for(index in 1:length(regionsList[[i]])){
      currentRegion <- regionsList[[i]][index]
      
      cat("Computing PMDs at ",.printGenomicRanges(currentRegion),"\n")
      
      # Select the points in methylationData that we're interested in. These are the 
      # points that lie within 'regions', as well as any that lie within 
      # window.size of them. 
      windowSizeHalf <- floor((windowSize - 1)/2)
      extendedRegion <- currentRegion
      start(extendedRegion) <- start(currentRegion) - windowSizeHalf
      end(extendedRegion) <- end(currentRegion) + windowSizeHalf
      
      overlaps <- findOverlaps(localContextMethylationData, extendedRegion)
      if(length(overlaps) > 0){
        localMethylationData <- localContextMethylationData[queryHits(overlaps)]
        
        cat("Calculating interpolations...\n")
        
        #Rcpp
        movingAverageMethylReads <- round(.movingAverage(start(currentRegion), 
                                                          end(currentRegion), 
                                                          start(localMethylationData), 
                                                          localMethylationData$readsM, 
                                                          windowSizeHalf = windowSizeHalf))
        movingAverageTotalReads <- round(.movingAverage(start(currentRegion), 
                                                         end(currentRegion), 
                                                         start(localMethylationData), 
                                                         localMethylationData$readsN, 
                                                         windowSizeHalf = windowSizeHalf))
        movingAverageProportion <- movingAverageMethylReads / movingAverageTotalReads
        
        
        cat("Identifying PMDs...\n")  
        # filtering the expected PMD window
        
        
        # compute the partially methylated cytosines
 
        bufferIndex <- 
          movingAverageTotalReads >= minReadsPerCytosine &
          movingAverageProportion > minMethylation &
          movingAverageProportion < maxMethylation  
        
        bufferIndex[is.na(bufferIndex)] <- FALSE
        PMPs <- sign(bufferIndex)
        
        #join the partially methylated cytosines into regions
        rle <- rle(PMPs)
        rle$cumulative <- cumsum(rle$lengths)
        endOfRuns <- rle$cumulative + start(currentRegion) - 1
        
        PMDs <- GRanges(
          seqnames    = seqnames(currentRegion),
          ranges      = IRanges(endOfRuns - rle$lengths + 1, endOfRuns),
          strand      = strand(currentRegion),
          direction   = rle$values,
          context     = paste(context, collapse = "_")
        )
        
        # Select the crude list of PMDs
        PMDs <- PMDs[!is.na(PMDs$direction) & (PMDs$direction == 1)]
        
        # append current PMDs to the global list of PMDs 
        if(length(computedPMDs) == 0){
          computedPMDs <- PMDs
        } else{
          computedPMDs <- c(computedPMDs,PMDs)
        }
        
      }
    }
    
    return(computedPMDs)
    
  }
  
  # compute the PMDs
  if(cores > 1){
    cat("Compute the PMDs using ", cores, "cores\n")
    computedPMDs <- BiocParallel::bplapply(1:length(regionsList), .computePMDsInterpolationLoop, BPPARAM = BPPARAM)
  } else {
    computedPMDs <- lapply(1:length(regionsList), .computePMDsInterpolationLoop)
  }
  
  
  computedPMDs <- unlist(GRangesList(computedPMDs))
  
  
  if(length(computedPMDs) > 0){
    computedPMDs <- computedPMDs[order(computedPMDs)]
    cat("Analysed reads inside PMDs\n")
    overlaps <- countOverlaps(localContextMethylationData, computedPMDs)
    localContextMethylationDataPMDs <- localContextMethylationData[overlaps > 0]
    if(cores > 1){
      computedPMDsList <- S4Vectors::splitAsList(computedPMDs,  rep(1:cores, length.out=length(computedPMDs)))
      buffercomputedPMDsList <- BiocParallel::bplapply(1:length(computedPMDsList), function(i){ 
        .analyseReadsInsideRegionsPMDs(localContextMethylationDataPMDs, computedPMDsList[[i]])}, 
        BPPARAM = BPPARAM)
      computedPMDs <- unlist(GRangesList(buffercomputedPMDsList))
      computedPMDs <- computedPMDs[order(computedPMDs)]
      
    } else{
      computedPMDs <- .analyseReadsInsideRegionsPMDs(localContextMethylationDataPMDs, computedPMDs)
    }
    computedPMDs <- computedPMDs[!is.na(computedPMDs$cytosinesCount) & computedPMDs$cytosinesCount > 0]

    computedPMDs <- computedPMDs[!is.na(computedPMDs$proportion) &
                                 computedPMDs$proportion >= minMethylation & 
                                   computedPMDs$proportion <= maxMethylation]
    
    cat("Merge PMDs iteratively\n")    
    # Get rid of small gaps between PMDs.
    if(minGap > 0){
      computedPMDs <- .smartMergePMDs(computedPMDs, 
                                      minGap = minGap, 
                                      respectSigns = TRUE, 
                                      methylationData = localContextMethylationData,
                                      minReadsPerCytosine = minReadsPerCytosine,
                                      minMethylation = minMethylation,
                                      maxMethylation = maxMethylation,
                                      cores = cores,
                                      BPPARAM = BPPARAM)
    }  
    cat("Filter PMDs \n")    
    if(length(computedPMDs) > 0){
      #remove small PMDs 
      computedPMDs <- computedPMDs[width(computedPMDs) >= minSize]
      if(length(computedPMDs) > 0){
        #remove PMDswith few cytosines
        computedPMDs <- computedPMDs[!is.na(computedPMDs$cytosinesCount) & computedPMDs$cytosinesCount >= minCytosinesCount]
      }
    }
  } 
  if (length(computedPMDs) >0 & length(computedPMDs$direction) != 0){
    computedPMDs$direction <- NULL
  }
  return(computedPMDs)  
}

#' This function computes the partially methylated domains
#' using the neighbourhood method. This assumes the computation of partially methylated 
#' cytosines followed by smart merging of these cytosines while keeping the new PMDs 
#' statistically significant.   
.computePMDsNeighbourhood <- function(methylationData, 
                                      regions = NULL, 
                                      context = "CG",
                                      minCytosinesCount = 4, 
                                      minMethylation = 0.4, 
                                      maxMethylation = 0.6,
                                      minGap = 200, 
                                      minSize = 50, 
                                      minReadsPerCytosine = 4, 
                                      cores = 1,
                                      BPPARAM = BPPARAM) {  
  
  regions <- reduce(regions)
  
  # extract the methylation data in the correct context
  cat("Extract methylation in the corresponding context \n")
  
  
  contextMethylationData <- methylationData[methylationData$context%in%context]
  rm(methylationData)
  localContextMethylationData <- contextMethylationData[queryHits(findOverlaps(contextMethylationData, regions))]
  rm(contextMethylationData)  
  
  cat("Computing PMDs \n")
  PMPs <- GRanges()
  if(length(localContextMethylationData) > 0){
    PMPs <- localContextMethylationData
    
    PMPs$sumReadsM <- PMPs$readsM
    PMPs$sumReadsN <- PMPs$readsN
    PMPs$proportion <- PMPs$readsM / PMPs$readsN
    PMPs$cytosinesCount <- 1
    PMPs$direction <- sign(PMPs$proportion >= minMethylation &
                             PMPs$proportion <= maxMethylation)
    
    bufferIndex <- 
      PMPs$sumReadsN >=minReadsPerCytosine & PMPs$proportion >= minMethylation &
      PMPs$proportion <= maxMethylation
    PMPs <- PMPs[bufferIndex]
    strand(PMPs) <- "*" 
  }
  
  computedPMDs <- GRanges()
  if(length(PMPs) > 0){    
    cat("Merge PMDs iteratively\n")    
    # Get rid of small gaps between PMDs.
    if(minGap > 0){
      computedPMDs <- .smartMergePMDs(PMPs, 
                                      minGap = minGap, 
                                      respectSigns = TRUE, 
                                      methylationData = localContextMethylationData,
                                      minReadsPerCytosine = minReadsPerCytosine, 
                                      minMethylation = minMethylation,
                                      maxMethylation = maxMethylation,
                                      cores = cores,
                                      BPPARAM = BPPARAM)
    } else{
      computedPMDs <- PMPs
    } 
    computedPMDs$ONT_Cm = NULL
    computedPMDs$ONT_C  = NULL
    cat("Filter PMDs \n")    
    if(length(computedPMDs) > 0){
      #remove small PMDs 
      computedPMDs <- computedPMDs[width(computedPMDs) >= minSize]
      if(length(computedPMDs) > 0){
        #remove PMDs with few cytosines
        computedPMDs <- computedPMDs[!is.na(computedPMDs$cytosinesCount) & computedPMDs$cytosinesCount >= minCytosinesCount]
        
      }
      
    }
  }  
  if (length(computedPMDs) >0 & length(computedPMDs$direction) != 0){
    computedPMDs$direction <- NULL
  }
  return(computedPMDs)  
}


#' This function computes the partially methylated regions 
#' using the bins method. 
.computePMDsBins <- function(methylationData,
                             regions = NULL, 
                             context = "CG", 
                             binSize = 100,
                             minCytosinesCount = 4,
                             minMethylation = 0.4, 
                             maxMethylation = 0.6,
                             minGap = 200, 
                             minSize = 50, 
                             minReadsPerCytosine = 4, 
                             cores = 1,
                             BPPARAM = BPPARAM) {
  
  regions <- reduce(regions)
  
  # extract the methylation data in the correct context
  cat("Extract methylation in the corresponding context \n")
  
  contextMethylationData <- methylationData[methylationData$context%in%context]
  rm(methylationData)
  localContextMethylationData <- contextMethylationData[queryHits(findOverlaps(contextMethylationData, regions))]
  rm(contextMethylationData)  

  regionsList <- .splitGRangesEqualy(regions, cores)
  
  # inner loop function for BiocParallel::bplapply
  .computePMDsBinsLoop = function(i){
    computedPMDs <- GRanges()
    for(index in 1:length(regionsList[[i]])){
      currentRegion <- regionsList[[i]][index]
      
      
      cat("Computing PMDs at ",.printGenomicRanges(currentRegion),"\n")
      
      seqs <- seq(start(currentRegion), (end(currentRegion)-binSize), by = binSize);
      
      bins <- GRanges(seqnames(currentRegion), IRanges(seqs, (seqs+binSize-1)))
      
      overlapsBins <- findOverlaps(localContextMethylationData, currentRegion)
      
      if(length(overlapsBins) > 0){
        localMethylationData <- localContextMethylationData[queryHits(overlapsBins)]
        
        cat("Count inside each bin...\n")
        #bins <- .analyseReadsInsideRegions(localMethylationData, bins, context, cores)
        bins <- .analyseReadsInsideBinsPMDs(localMethylationData, bins, currentRegion)
        
        cat("Filter the bins...\n")
        # Get rid of the bins with fewer than minCytosinesCount cytosines inside.  
        bins  <- bins[bins$cytosinesCount >= minCytosinesCount]
        
        # Get rid of the bins with fewer than minReadsPerCytosine reads per cytosine.  
        bins  <- bins[(bins$sumReadsN/bins$cytosinesCount >= minReadsPerCytosine)]
        
        # Get rid of the bins with outlier about ranges for PMD parameter
        bins  <- bins[bins$proportion >= minMethylation & bins$proportion <= maxMethylation]
        
        
        cat("Identifying PMDs...\n")    
        bins$context <- rep(paste(context, collapse = "_"), length(bins))
        bins$direction <- rep(NA, length(bins))
        bins$direction <- sign(bins$proportion)
        
        # Select the crude list of PMDs
        PMDs <- bins[!is.na(bins$direction) & bins$direction == 1]
        
        # append current PMDs to the global list of PMDs 
        if(length(computedPMDs) == 0){
          computedPMDs <- PMDs
        } else{
          computedPMDs <- c(computedPMDs,PMDs)
        }
      }
    }
    return(computedPMDs)
  }
  
  # compute the PMDs
  if(cores > 1){
    cat("Compute the PMDs using ", cores, "cores\n")
    computedPMDs <- BiocParallel::bplapply(1:length(regionsList), .computePMDsBinsLoop, BPPARAM = BPPARAM)
  } else {
    computedPMDs <- lapply(1:length(regionsList), .computePMDsBinsLoop)
  }
  
  
  computedPMDs <- unlist(GRangesList(computedPMDs))
  
  if(length(computedPMDs) > 0){
    
    cat("Merge adjacent PMDs\n")    
    computedPMDs <- computedPMDs[order(computedPMDs)]
   
    cat("Merge PMDs iteratively\n")    
    # Get rid of small gaps between PMDs.
    if(minGap > 0){
      computedPMDs <- .smartMergePMDs(computedPMDs, 
                                      minGap = minGap, 
                                      respectSigns = TRUE, 
                                      methylationData = localContextMethylationData,
                                      minReadsPerCytosine = minReadsPerCytosine,
                                      minMethylation = minMethylation,
                                      maxMethylation = maxMethylation,
                                      cores = cores,
                                      BPPARAM = BPPARAM)
    }  
    
    cat("Filter PMDs \n")    
    if(length(computedPMDs) > 0){
      #remove small PMDs 
      computedPMDs <- computedPMDs[width(computedPMDs) >= minSize]
      if(length(computedPMDs) > 0){
        #remove PMDswith few cytosines
        computedPMDs <- computedPMDs[!is.na(computedPMDs$cytosinesCount) & computedPMDs$cytosinesCount >= minCytosinesCount]
      }
    }
  }  
  if (length(computedPMDs) >0 & length(computedPMDs$direction) != 0){
    computedPMDs$direction <- NULL
  }
  return(computedPMDs)
}

#' This function verifies whether a set of potential PMDs (e.g. genes, 
#' transposons, CpG islands) are partially methylated or not.
#'
#' @title Filter PMDs 
#' @param methylationData the methylation data in condition
#' (see \code{\link{ontSampleGRangesList}}).
#' @param potentialPMDs a \code{\link[GenomicRanges]{GRanges}} object with potential PMDs 
#' where to compute the PMDs. This can be a a list of gene and/or transposable 
#' elements coordinates.
#' @param context the context in which the PMDs are computed (\code{"CG"}, 
#' \code{"CHG"} or \code{"CHH"}).
#' @param minCytosinesCount PMDs with less cytosines in the specified context 
#' than \code{minCytosinesCount} will be discarded.
#' @param minMethylation Numeric [0,1]; minimum mean methylation proportion.
#' @param maxMethylation Numeric [0,1]; maximum mean methylation proportion.
#' @param minReadsPerCytosine  PMDs with the average number of reads lower than 
#' \code{minReadsPerCytosine} are discarded. 
#' @param parallel Logical; run in parallel if \code{TRUE}.
#' @param BPPARAM A \code{BiocParallelParam} object controlling parallel execution.
#'    This value will automatically set when parallel is \code{TRUE}, also able to set as manually.
#' @param cores Integer number of workers (must not exceed BPPARAM$workers).
#'    This value will automatically set as the maximum number of system workers,
#'    also able to set as manually.
#' @return a \code{\link[GenomicRanges]{GRanges}} object with 5 metadata columns that contain 
#' the PMDs; see \code{\link{computePMDs}}.
#' @seealso \code{\link{PMDsNoiseFilterCG}}, \code{\link{computePMDs}}, 
#' \code{\link{analyseReadsInsideRegionsForCondition}}  
#' and \code{\link{mergePMDsIteratively}}
#' @examples
#' # load the ONT methylation data 
#' data(ontSampleGRangesList)
#' # load the gene annotation data
#' data(GEs_hg38)
#' 
#' # select the transcript
#' transcript <- GEs_hg38[which(GEs_hg38$type == "transcript")]
#' 
#' # the regions where to compute the PMDs
#' regions <- GRanges(seqnames = Rle("chr1"), ranges = IRanges(1E6+5E5,2E6))
#' transcript <- transcript[overlapsAny(transcript, regions)]
#' 
#' # filter genes that are partially methylated in the two conditions
#' PMDsGenesCG <- filterPMDs(ontSampleGRangesList[["GM18501"]], 
#'                potentialPMDs = transcript, 
#'                context = "CG", minMethylation = 0.4, maxMethylation = 0.6,
#'                minCytosinesCount = 4, minReadsPerCytosine = 3, cores = 1)
#'
#' @author Nicolae Radu Zabet and Young Jun Kim
#' @export
filterPMDs <- function(methylationData, 
                       potentialPMDs, 
                       context = "CG",   
                       minCytosinesCount = 4, 
                       minMethylation = 0.4, 
                       maxMethylation = 0.6,
                       minReadsPerCytosine = 3, 
                       parallel = FALSE,
                       BPPARAM = NULL,
                       cores = NULL) {

  ##Parameters checking
  cat("Parameters checking ...\n")
  
  # generate the BPPARAM value if set as parallel 
  if (parallel == TRUE){
    BPPARAM <- suppressWarnings(.validateBPPARAM(BPPARAM, progressbar = TRUE)) 
  }else{
    # Force serial execution
    BPPARAM <- BiocParallel::SerialParam(progressbar = TRUE)
  }
  # If cores argument is specified
  if (!is.null(cores)) {
    .stopIfNotAll(.isInteger(cores, positive = TRUE), 
                  "the number of cores used when computing the DMRs needs to be an integer higher or equal to 1.")
    
    # Check if user requested more cores than available
    if (cores > BPPARAM$workers) {
      warning(paste0("The number of requested cores (", cores, 
                     ") exceeds the available system cores (", BPPARAM$workers, 
                     "). Automatically setting cores to the maximum available (", 
                     BPPARAM$workers, ")."))
      cores <- BPPARAM$workers
    } else {
      message(paste0("Using user-specified core count: ", cores))
    }
    
    # Apply the final core number
    BPPARAM$workers <- cores
  } else {
    cores <- BPPARAM$workers
  }
  cat("Current parallel setting, BPPARAM: ", capture.output(BPPARAM),sep = "\n")
  
  .validateMethylationData(methylationData, variableName="methylationData")
  
  regions <- getWholeChromosomes(methylationData)
  
  .validateContext(context)
  
  .validateGRanges(potentialPMDs, generateGenomeWide=FALSE, variableName="potentialPMDs", minLength=NULL)
  
  regions <- .validateGRanges(regions, methylationData)
  
  .stopIfNotAll(c(.isInteger(minCytosinesCount, positive=TRUE)), 
                " the minCytosinesCount is an integer higher or equal to 0")
  
  .stopIfNotAll(c(.isInteger(minReadsPerCytosine, positive=TRUE)), 
                " the minimum number of reads in a bin is an integer higher or equal to 0")
  
  .stopIfNotAll(c(!is.null(minMethylation), is.numeric(minMethylation), minMethylation >= 0), 
                " minMethylation needs to be a numeric value higher or equal to 0")
  
  .stopIfNotAll(c(!is.null(maxMethylation), is.numeric(maxMethylation), maxMethylation >= 0), 
                " maxMethylation needs to be a numeric value higher or equal to 0")
  
  .stopIfNotAll(c(maxMethylation >= minMethylation), 
                " maxMethylation should be higher than minMethylation value")
  
  regions <- reduce(regions)
  
  if(length(potentialPMDs) > 0){
    
    
    
    # extract the methylation data in the correct context
    cat("Extract methylation in the corresponding context \n")
    
    contextMethylationData <- methylationData[methylationData$context%in%context]
    rm(methylationData)
    localContextMethylationData <- contextMethylationData[queryHits(findOverlaps(contextMethylationData, regions))]
    rm(contextMethylationData)  
    
    regionsList <- .splitGRangesEqualy(regions, cores)
    
    # inner loop function for BiocParallel::bplapply
    .filterPMDsLoop = function(i){
      computedPMDs <- GRanges()  
      for(index in 1:length(regionsList[[i]])){
        currentRegion <- regionsList[[i]][index]
        
        
        cat("Computing PMDs at ",.printGenomicRanges(currentRegion),"\n")
        
        cat("Selecting data...\n")
        
        # Select the points in methylationData that we're interested in. These are the 
        # points that lie within 'regions', as well as any that lie within 
        # window.size of them. 
        
        overlapsPotentialPMDs <- findOverlaps(potentialPMDs, currentRegion)
        if(length(overlapsPotentialPMDs) > 0){
          potentialPMDsLocal <- potentialPMDs[queryHits(overlapsPotentialPMDs)]
          
          localMethylationData <- localContextMethylationData[queryHits(findOverlaps(localContextMethylationData, currentRegion))]
          potentialPMDsLocal <- .analyseReadsInsideRegionsPMDs(localMethylationData, potentialPMDsLocal)
          
          if(length(computedPMDs) == 0){
            computedPMDs <- potentialPMDsLocal
          } else{
            computedPMDs <- c(computedPMDs,potentialPMDsLocal)
          }
        } 
      }
      return(computedPMDs)
    }
    
    # compute the PMDs
    if(cores > 1){
      cat("Compute the PMDs using ", cores, "cores\n")
      computedPMDs <- BiocParallel::bplapply(1:length(regionsList), .filterPMDsLoop, BPPARAM = BPPARAM)
    } else {
      computedPMDs <- lapply(1:length(regionsList), .filterPMDsLoop)
    }
    
    
    computedPMDs <-  unlist(GRangesList(computedPMDs))
    
    if(length(computedPMDs) > 0){
      cat("Identifying PMDs...\n")    
      
      bufferIndex <- !is.na(computedPMDs$proportion) &
        computedPMDs$proportion >= minMethylation & 
        computedPMDs$proportion <= maxMethylation &
        computedPMDs$sumReadsN/computedPMDs$cytosinesCount >= minReadsPerCytosine &
        computedPMDs$cytosinesCount >= minCytosinesCount
      
      computedPMDs <- computedPMDs[bufferIndex]  
    }
  } else{
    computedPMDs <- GRanges() 
  }
  
  if(length(computedPMDs) > 0){
    computedPMDs <- computedPMDs[order(computedPMDs)]
  }
  
  return(computedPMDs)
  
}



#' This function takes a list of PMDs and attempts to merge PMDs while keeping 
#' the new PMDs statistically significant.
#'
#' @title Merge PMDs iteratively
#' @param PMDs the list of PMDs as a \code{\link[GenomicRanges]{GRanges}} object; 
#' e.g. see \code{\link{computePMDs}}
#' @param minGap PMDs separated by a gap of at least \code{minGap} are not 
#' merged.
#' @param respectSigns logical value indicating whether to respect the sign when 
#' joining PMDs.
#' @param methylationData the methylation data in GRanges
#' (see \code{\link{ontSampleGRangesList}}).
#' @param context the context in which the PMDs are computed (\code{"CG"}, 
#' \code{"CHG"} 
#' or \code{"CHH"}).
#' @param minReadsPerCytosine two adjacent PMDs are merged only if the number of 
#' reads per cytosine of the new DMR is higher than \code{minReadsPerCytosine}.
#' @param minMethylation Numeric [0,1]; minimum mean methylation proportion.
#' @param maxMethylation Numeric [0,1]; maximum mean methylation proportion.
#' @param parallel Logical; run in parallel if \code{TRUE}.
#' @param BPPARAM A \code{BiocParallelParam} object controlling parallel execution.
#'    This value will automatically set when parallel is \code{TRUE}, also able to set as manually.
#' @param cores Integer number of workers (must not exceed BPPARAM$workers).
#'    This value will automatically set as the maximum number of system workers,
#'    also able to set as manually.
#' @return the reduced list of PMDs as a \code{\link[GenomicRanges]{GRanges}} object; 
#' e.g. see \code{\link{computePMDs}}
#' @seealso \code{\link{filterPMDs}}, \code{\link{computePMDs}}, 
#' \code{\link{analyseReadsInsideRegionsForCondition}} and 
#' \code{\link{PMDsNoiseFilterCG}}
#' @examples
#' # load the ONT methylation data 
#' data(ontSampleGRangesList)
#' 
#' # load the PMDs in CG context they were computed with minGap = 200
#' data(PMDsNoiseFilterCG)
#' 
#'
#' # merge the PMDs 
#' PMDsNoiseFilterCGLarger <- mergePMDsIteratively(PMDsNoiseFilterCG[1:100], 
#'                            minGap = 500, respectSigns = TRUE, 
#'                            ontSampleGRangesList[["GM18501"]], context = "CG", 
#'                            minReadsPerCytosine = 4, minMethylation = 0.4, 
#'                            maxMethylation = 0.6, cores = 1)
#' 
#' \dontrun{
#' # set genomic coordinates where to compute PMDs
#' chr1_ranges <- GRanges(seqnames = Rle("chr1"), ranges = IRanges(1E6+5E5,2E6))
#' 
#' # compute PMDs and remove gaps smaller than 200 bp
#' PMDsNoiseFilterCG200 <- computePMDs(ontSampleGRangesList[["GM18501"]], 
#'                        regions = chr1_ranges, context = "CG", method = "noise_filter", 
#'                        windowSize = 100, kernelFunction = "triangular",  
#'                        minCytosinesCount = 1, minMethylation = 0.4, 
#'                        maxMethylation = 0.6, minGap = 0, minSize = 200, 
#'                        minReadsPerCytosine = 1, cores = 1)
#' PMDsNoiseFilterCG0 <- computePMDs(ontSampleGRangesList[["GM18501"]], 
#'                        regions = chr1_ranges, context = "CG", method = "noise_filter", 
#'                        windowSize = 100, kernelFunction = "triangular", 
#'                        minCytosinesCount = 1, minMethylation = 0.4, 
#'                        maxMethylation = 0.6, minGap = 0, minSize = 0, 
#'                        minReadsPerCytosine = 1, cores = 1)
#' PMDsNoiseFilterCG0Merged200 <- mergePMDsIteratively(PMDsNoiseFilterCG0, 
#'                              minGap = 200, respectSigns = TRUE, 
#'                              ontSampleGRangesList[["GM18501"]], context = "CG",  
#'                              minReadsPerCytosine = 4, minMethylation = 0.4, 
#'                              maxMethylation = 0.6, cores = 1)                      
#' 
#' #check that all newley computed PMDs are identical
#' print(all(PMDsNoiseFilterCG200 == PMDsNoiseFilterCG0Merged200))
#' 
#' }
#' 
#' @author Nicolae Radu Zabet and Young Jun Kim
#' 
#' @export
mergePMDsIteratively <- function(PMDs, 
                                 minGap = 200, 
                                 respectSigns = TRUE, 
                                 methylationData,
                                 context = "CG",
                                 minReadsPerCytosine = 4, 
                                 minMethylation = 0.4, 
                                 maxMethylation = 0.6,
                                 parallel = FALSE,
                                 BPPARAM = NULL,
                                 cores = NULL){
  
  ##Parameters checking
  cat("Parameters checking ...\n")
  
  # generate the BPPARAM value if set as parallel 
  if (parallel == TRUE){
    BPPARAM <- suppressWarnings(.validateBPPARAM(BPPARAM, progressbar = TRUE)) 
  }else{
    # Force serial execution
    BPPARAM <- BiocParallel::SerialParam(progressbar = TRUE)
  }
  # If cores argument is specified
  if (!is.null(cores)) {
    .stopIfNotAll(.isInteger(cores, positive = TRUE), 
                  "the number of cores used when computing the DMRs needs to be an integer higher or equal to 1.")
    
    # Check if user requested more cores than available
    if (cores > BPPARAM$workers) {
      warning(paste0("The number of requested cores (", cores, 
                     ") exceeds the available system cores (", BPPARAM$workers, 
                     "). Automatically setting cores to the maximum available (", 
                     BPPARAM$workers, ")."))
      cores <- BPPARAM$workers
    } else {
      message(paste0("Using user-specified core count: ", cores))
    }
    
    # Apply the final core number
    BPPARAM$workers <- cores
  } else {
    cores <- BPPARAM$workers
  }
  cat("Current parallel setting, BPPARAM: ", capture.output(BPPARAM),sep = "\n")
  
  .validateMethylationData(methylationData, variableName="methylationData")
  .validateContext(context)
  .validateGRanges(PMDs, generateGenomeWide=FALSE, variableName="PMDs", minLength=NULL)

  .stopIfNotAll(c(.isInteger(minGap, positive=TRUE)),
                " the minimum gap between PMDs is an integer higher or equal to 0")
  
  .stopIfNotAll(c(.isInteger(minReadsPerCytosine, positive=TRUE)), 
                " the minimum average number of reads in a DMR is an integer higher or equal to 0")
  
  .stopIfNotAll(c(!is.null(minMethylation), is.numeric(minMethylation), minMethylation >= 0), 
                " minMethylation needs to be a numeric value higher or equal to 0")
  
  .stopIfNotAll(c(!is.null(maxMethylation), is.numeric(maxMethylation), maxMethylation >= 0), 
                " maxMethylation needs to be a numeric value higher or equal to 0")
  
  .stopIfNotAll(c(maxMethylation >= minMethylation), 
                " maxMethylation should be higher than minMethylation value")
  
  totalRegion <- reduce(PMDs, drop.empty.ranges=FALSE, min.gapwidth=minGap, ignore.strand=TRUE)
  
  contextMethylationData <- methylationData[methylationData$context%in%context]
  rm(methylationData)
  localContextMethylationData <- contextMethylationData[queryHits(findOverlaps(contextMethylationData, totalRegion))]
  rm(contextMethylationData)  
  
  cat("Merge PMDs iteratively ...\n")
  
  
  return(.smartMergePMDs(PMDs, 
                         minGap = minGap, 
                         respectSigns = respectSigns, 
                         methylationData = localContextMethylationData,
                         minReadsPerCytosine = minReadsPerCytosine,
                         minMethylation = minMethylation,
                         maxMethylation = maxMethylation,
                         cores = cores,
                         BPPARAM = BPPARAM))                                               
}


#' This function extracts from the methylation data the total number of reads, 
#' the number of methylated reads and the number of cytosines in the specific 
#' context from a region (e.g. PMDs)
#'
#' @title Analyse reads inside regions for condition
#' @param regions a \code{\link[GenomicRanges]{GRanges}} object with a list of regions on the 
#' genome; e.g. could be a list of PMDs
#' @param methylationData the methylation data in one condition
#' (see \code{\link{ontSampleGRangesList}}).
#' @param context the context in which to extract the reads (\code{"CG"}, 
#' \code{"CHG"} or \code{"CHH"}).
#' @param label a string to be added to the columns to identify the condition
#' @param parallel Logical; run in parallel if \code{TRUE}.
#' @param BPPARAM A \code{BiocParallelParam} object controlling parallel execution.
#'    This value will automatically set when parallel is \code{TRUE}, also able to set as manually.
#' @param cores Integer number of workers (must not exceed BPPARAM$workers).
#'    This value will automatically set as the maximum number of system workers,
#'    also able to set as manually.
#' @return a \code{\link[GenomicRanges]{GRanges}} object with additional four metadata columns
#' \describe{
#'  \item{sumReadsM}{the number of methylated reads}
#'  \item{sumReadsN}{the total number of reads} 
#'  \item{proportion}{the proportion methylated reads} 
#'  \item{cytosinesCount}{the number of cytosines in the regions} 
#' }
#' @seealso \code{\link{filterPMDs}}, \code{\link{computePMDs}}, 
#' \code{\link{PMDsNoiseFilterCG}}, and \code{\link{mergePMDsIteratively}}
#' @examples
#' 
#' # load the ONT methylation data
#' data(ontSampleGRangesList)
#'  
#' #load the PMDs in CG context. These PMDs were computed with minGap = 200.
#' data(PMDsNoiseFilterCG)
#' 
#' #retrive the number of reads in CG context in GM18501
#' PMDsNoiseFilterCGreadsCG <- analyseReadsInsideRegionsForConditionPMD(
#'                              PMDsNoiseFilterCG[1:10], 
#'                              ontSampleGRangesList[["GM18501"]], context = "CG", 
#'                              label = "GM18501")
#' 
#' 
#' @author Nicolae Radu Zabet and Young Jun Kim
#' 
#' @export
analyseReadsInsideRegionsForConditionPMD <- function(regions,
                                                  methylationData,
                                                  context,
                                                  label = "",
                                                  parallel = FALSE,
                                                  BPPARAM = NULL,
                                                  cores = NULL){
  
  ##Parameters checking
  cat("Parameters checking ...\n")
  
  # generate the BPPARAM value if set as parallel 
  if (parallel == TRUE){
    BPPARAM <- suppressWarnings(.validateBPPARAM(BPPARAM, progressbar = TRUE)) 
  }else{
    # Force serial execution
    BPPARAM <- BiocParallel::SerialParam(progressbar = TRUE)
  }
  # If cores argument is specified
  if (!is.null(cores)) {
    .stopIfNotAll(.isInteger(cores, positive = TRUE), 
                  "the number of cores used when computing the DMRs needs to be an integer higher or equal to 1.")
    
    # Check if user requested more cores than available
    if (cores > BPPARAM$workers) {
      warning(paste0("The number of requested cores (", cores, 
                     ") exceeds the available system cores (", BPPARAM$workers, 
                     "). Automatically setting cores to the maximum available (", 
                     BPPARAM$workers, ")."))
      cores <- BPPARAM$workers
    } else {
      message(paste0("Using user-specified core count: ", cores))
    }
    
    # Apply the final core number
    BPPARAM$workers <- cores
  } else {
    cores <- BPPARAM$workers
  }
  cat("Current parallel setting, BPPARAM: ", capture.output(BPPARAM),sep = "\n")
  
  .validateGRanges(regions, generateGenomeWide=FALSE, variableName="regions", minLength=NULL)
  .validateMethylationData(methylationData, variableName="methylationData")
  .validateContext(context)
  
  cat("Extract methylation levels in corresponding context ...\n")
  contextMethylationData <- methylationData[methylationData$context%in%context]
  rm(methylationData)
  
  cat("Compute reads inside each region ...\n")
  if(length(regions) > 0){
    if(cores > 1){
      cat("Analyse reads in regions using ", cores, "cores\n")
      regionsList <- split(regions, rep(1:cores, length.out=length(regions)))
      .analyseReadsInsideRegionsForConditionLoop = function(i){
        regionsLocal <- .analyseReadsInsideRegionsForCondition(regionsList[[i]], 
                                                               contextMethylationData, 
                                                               label = label,
                                                               context = context)
        return(regionsLocal)
      }
      regions <- BiocParallel::bplapply(1:length(regionsList), 
                                    .analyseReadsInsideRegionsForConditionLoop, BPPARAM = BPPARAM)
      regions <- unlist(GRangesList(regions))
    } else{
      regions <- .analyseReadsInsideRegionsForCondition(regions, 
                                                        contextMethylationData, 
                                                        label = label,
                                                        context = context)
    }
    regions <- regions[order(regions)]
  }
  return(regions)
}  

#' This function computes the moving average for a vector that has missing 
#' values. 
#'
#' @title Moving average 
#' @param minPos the first position of the input fragment
#' @param maxPos the last position of the input fragment
#' @param pos the positions where the vector has values
#' @param val the values of the input vector
#' @param weights the weights associated with the moving average 
#' @param windowSize the size of the moving average
#' @param normalize \code{logical} value that indicates whether the moving 
#' average is normalised
#' @param kernelFunction a \code{character} indicating which kernel function to 
#' be used. Can be one of \code{"uniform"}, \code{"triangular"}, 
#' \code{"gaussian"} or \code{"epanechnicov"}.
#' @param lambda numeric value required for the Gaussian filter 
#' (\code{K(x) = exp(-lambda*x^2)})
#' @return A numeric vector with the values computed by the moving average
#'  
#' @author Radu Zabet
.movingAverage <- function(minPos, maxPos, pos, val, weights=1, windowSizeHalf=150, normalize = FALSE, kernelFunction="triangular", lambda=0.5) {
  
  .stopIfNotAll(c(length(pos)==length(val)), "pos and val vectors need to have the same length")
  
  if(length(weights) < length(val)){
    weights <- rep(weights, length.out=length(val))
  } else if(length(weights) > length(val)){
    weights <- weights[1:length(val)]
  }
  
  # Filter out NAs
  keepIndexes <- which(!is.na(pos) & !is.na(val) & !is.na(weights))
  pos <- pos[keepIndexes]
  weights <- weights[keepIndexes]  
  val <- val[keepIndexes]
    
  #set the values
  rawVector <- rep(0, maxPos - minPos + 2*windowSizeHalf + 1)
  rawVector[(pos - minPos + windowSizeHalf + 1)] <-weights*val
  
  normVector <- rep(0, maxPos - minPos + 2*windowSizeHalf + 1)
  normVector[(pos - minPos + windowSizeHalf + 1)] <-weights
  
  
  
  # Define the (triangular) kernel.
  if(kernelFunction =="uniform"){
    kernel <- rep(1, times=(2*windowSizeHalf + 1))  
  } else if(kernelFunction =="triangular"){
    kernel <- 1 - abs(-windowSizeHalf:windowSizeHalf)/windowSizeHalf 
  } else if(kernelFunction =="gaussian"){
    kernel <- .gaussianKernel(lambda, - windowSizeHalf:windowSizeHalf)
  } else if(kernelFunction =="epanechnicov"){
    kernel <- .epanechnicovKernel(-windowSizeHalf:windowSizeHalf, windowSizeHalf)
  } else{
    stop(paste("Unknown kernel function: ", kernelFunction, ". 
               It should be one of \"uniform\", \"triangular\", \"gaussian\", \"epanechnicov\"",sep=""))
  }
  
  kernel <- kernel / sum(kernel)
  
  if(windowSizeHalf >= 1){
    smoothedVector <- RcppRoll::roll_sum(rawVector, length(kernel), weights = kernel, normalize = normalize) / RcppRoll::roll_sum(normVector, length(kernel), weights = kernel, normalize = normalize)
  } else{
    smoothedVector <- rawVector 
  }
  
  return(smoothedVector);
}


.gaussianKernel <- function(lambda,x){
  return(exp(-lambda*x^2))
}

.epanechnicovKernel <- function(x, max){
  return((3*(1-(x/max)^2))/4)
}

#' This function computes the moving sum for a vector that has missing values. 
#'
#' @title Moving sum 
#' @param minPos the first position of the input fragment
#' @param maxPos the last position of the input fragment
#' @param pos the positions where the vector has values
#' @param val the values of the input vector
#' @param weights the weights associated with the moving sum 
#' @param windowSize the size of the moving sum
#' @param normalize \code{logical} value that indicates whether the moving sum is normalised
#' @return A numeric vector with the values computed by the moving sum
#'  
#' @author Radu Zabet
.movingSum <- function(minPos, maxPos, pos, val, weights=1, windowSize=150, normalize = FALSE) {
  
  .stopIfNotAll(c(length(pos)==length(val)), "pos and val vectors need to have the same length")
  
  if(length(weights) < length(val)){
    weights <- rep(weights, length.out=length(val))
  } else if(length(weights) > length(val)){
    weights <- weights[1:length(val)]
  }
  
  # Filter out NAs
  keepIndexes <- which(!is.na(pos) & !is.na(val) & !is.na(weights))
  pos <- pos[keepIndexes]
  weights <- weights[keepIndexes]  
  val <- val[keepIndexes]
  
  
  #set the values
  rawVector <- rep(0, (maxPos - minPos + 1))
  rawVector[pos - minPos + 1] <-weights*val
  rawVector <- c(rawVector,rep(0,(windowSize-1)))
  
  # Define the (triangular) kernel.
  kernel <- c(rep(1, (windowSize)))
  
  
  smoothedVector <- RcppRoll::roll_sum(rawVector, length(kernel), weights = kernel, normalize = normalize) 
  
  #smoothedVector <- smoothedVector[seq(1,length(smoothedVector), by=windowSize)]
  
  return(smoothedVector);
}

#' Filter VMRs with ONT-specific variance tests and CI filters
#'
#' This function verifies whether a set of potential VMRs (e.g., genes,
#' transposons, CpG islands) are differentially methylated or not in ONT data,
#' adding per-read Wilcoxon and F-tests on per-site proportions, confidence interval
#' filtering, and optional variance-fold change cutoffs.
#'
#' @title Filter VMRs for ONT Data
#'
#' @param methylationData1 A \code{GRanges} of methylation calls for condition 1
#'   (see \code{\link{ontSampleGRangesList}}).
#' @param methylationData2 A \code{GRanges} of methylation calls for condition 2.
#' @param potentialVMRs A \code{GRanges} of candidate VMR regions (genes, TEs,
#'   CpG islands, etc.).
#' @param context Character string specifying cytosine context ("CG", "CHG",
#'   or "CHH").
#' @param pValueThreshold Numeric p-value threshold (0<value<1) for both
#'   Wilcoxon and F-tests after FDR adjustment.
#' @param minCytosinesCount Integer minimum number of cytosines per region.
#' @param minProportionDifference Numeric minimum methylation difference
#'   between conditions (0<value<1).
#' @param minReadsPerCytosine Integer minimum average coverage per cytosine.
#' @param ciExcludesOne Logical; if \code{TRUE}, filter out regions whose
#'   F-test 95\% confidence interval spans 1 (i.e., no significant variance change).
#' @param varRatioFc Optional; numeric fold-change cutoff on variance ratio
#'   (e.g., 2 for twofold variance difference). Regions with variance ratio
#'   outside \code{[1/varRatioFc, varRatioFc]} are kept when set.
#' @param parallel Logical; run in parallel if \code{TRUE}.
#' @param BPPARAM A \code{BiocParallelParam} object controlling parallel execution.
#'    This value will automatically set when parallel is \code{TRUE}, also able to set as manually.
#' @param cores Integer number of workers (must not exceed \code{BPPARAM$workers}).
#'    This value will automatically set as the maximum number of system workers,
#'    also able to set as manually.
#'
#' @return A \code{GRanges} with the same ranges as \code{regions}, plus these metadata:
#' \describe{
#'   \item{sumReadsM1}{total methylated reads in condition 1}
#'   \item{sumReadsN1}{total reads in condition 1}
#'   \item{proportion1}{methylation proportion (sumReadsM1/sumReadsN1)}
#'   \item{variance1}{variance of per-read methylation proportions in condition 1}
#'   \item{sumReadsM2}{total methylated reads in condition 2}
#'   \item{sumReadsN2}{total reads in condition 2}
#'   \item{proportion2}{methylation proportion (sumReadsM2/sumReadsN2)}
#'   \item{variance2}{variance of per-read methylation proportions in condition 2}
#'   \item{cytosinesCount}{number of cytosines observed in each region}
#'   \item{wilcox_pvalue}{FDR adjusted p-value from Wilcoxon rank-sum test comparing per-read proportions}
#'   \item{f_pvalue}{FDR adjusted p-value from F-test comparing variances of per-read proportions}
#'   \item{var_ratio}{Ratio of variances (variance1 / variance2)}
#'   \item{wilcox_result}{Full \code{htest} object returned by \code{wilcox.test}}
#'   \item{F_test_result}{Full \code{htest} object returned by \code{var.test}}
#'   \item{direction}{a number indicating whether the region lost (-1)  or gain
#'     (+1) methylation in condition 2 compared to condition 1.}
#'   \item{regionType}{a string indicating whether the region lost (\code{"loss"})
#'     or gained (\code{"gain"}) methylation in condition 2 compared to condition 1.}
#'   \item{is_DMR}{logical; \code{TRUE} if region passed the \code{wilcox.test}}
#'   \item{is_VMR}{logical; \code{TRUE} if region passed the \code{var.test}}
#' }
#'
#' @details
#' For each potential VMR, per-site methylation proportions are aggregated per read,
#' then a two-sample Wilcoxon rank-sum test compares means (\code{wilcox_pvalue}), and
#' an F-test compares variances (\code{f_pvalue}). You may further filter by requiring
#' the 95% confidence interval from the F-test to exclude 1 (\code{ciExcludesOne}) and/or
#' apply a fold-change cutoff on the variance ratio (\code{varRatioFc}).
#'
#' @seealso \code{\link{readONTbam}},
#' \code{\link{computePMDs}}, \code{\link{computeCoMethylatedPositions}},
#' \code{\link{ontSampleGRangesList}}, \code{\link{GEs_hg38}}
#'
#' @examples 
#' \dontrun{
#' # load the ONT methylation data
#' data(ontSampleGRangesList)
#' # load the gene annotation data
#' data(GEs_hg38)
#'
#' # select the transcript
#' transcript <- GEs_hg38[which(GEs_hg38$type == "transcript")]
#'
#' # the regions where to compute the PMDs
#' regions <- GRanges(seqnames = Rle("chr1"), ranges = IRanges(1E6+5E5,2E6))
#' transcript <- transcript[overlapsAny(transcript, regions)]
#'
#' # filter genes that are differntially methylated in the two conditions
#' VMRsGenesCG <- filterVMRsONT(ontSampleGRangesList[["GM18501"]],
#'                ontSampleGRangesList[["GM18876"]], potentialVMRs = transcript,
#'                context = "CG", pValueThreshold = 0.01,
#'                minCytosinesCount = 4, minProportionDifference = 0.01,
#'                minReadsPerCytosine = 3, ciExcludesOne = TRUE,
#'                varRatioFc = NULL, parallel = TRUE) # parallel recommended
#' }
#' @author Nicolae Radu Zabet and Young Jun Kim
#' @import stringr
#' @export

filterVMRsONT <- function(methylationData1,
                       methylationData2,
                       potentialVMRs,
                       context = "CG",
                       pValueThreshold = 0.01,
                       minCytosinesCount = 4,
                       minProportionDifference = 0.4,
                       minReadsPerCytosine = 3,
                       ciExcludesOne = TRUE,
                       varRatioFc = NULL, ## set as 2 if you want 2 fold-change varience cut-off
                       parallel = FALSE,
                       BPPARAM = NULL,
                       cores = NULL) {
  ##Parameters checking
  cat("Parameters checking ...\n")

  # generate the BPPARAM value if set as parallel
  if (parallel == TRUE){
    BPPARAM <- suppressWarnings(.validateBPPARAM(BPPARAM, progressbar = TRUE))
  }else{
    # Force serial execution
    BPPARAM <- BiocParallel::SerialParam(progressbar = TRUE)
  }
  # If cores argument is specified
  if (!is.null(cores)) {
    .stopIfNotAll(.isInteger(cores, positive = TRUE), 
                  "the number of cores used when computing the DMRs needs to be an integer higher or equal to 1.")
    
    # Check if user requested more cores than available
    if (cores > BPPARAM$workers) {
      warning(paste0("The number of requested cores (", cores, 
                     ") exceeds the available system cores (", BPPARAM$workers, 
                     "). Automatically setting cores to the maximum available (", 
                     BPPARAM$workers, ")."))
      cores <- BPPARAM$workers
    } else {
      message(paste0("Using user-specified core count: ", cores))
    }
    
    # Apply the final core number
    BPPARAM$workers <- cores
  } else {
    cores <- BPPARAM$workers
  }
  cat("Current parallel setting, BPPARAM: ", capture.output(BPPARAM),sep = "\n")


  .validateMethylationData(methylationData1, variableName="methylationData1")
  .validateMethylationData(methylationData2, variableName="methylationData2")

  regions <- union(getWholeChromosomes(methylationData1),
                   getWholeChromosomes(methylationData2))

  .validateContext(context)

  .validateGRanges(potentialVMRs, generateGenomeWide=FALSE, variableName="potentialVMRs", minLength=NULL)

  .stopIfNotAll(c(!is.null(pValueThreshold), is.numeric(pValueThreshold), pValueThreshold > 0, pValueThreshold < 1),
                " the p-value threshold needs to be in the interval (0,1)")

  .stopIfNotAll(c(.isInteger(minCytosinesCount, positive=TRUE)),
                " the minCytosinesCount is an integer higher or equal to 0")

  .stopIfNotAll(c(!is.null(minProportionDifference), is.numeric(minProportionDifference), minProportionDifference > 0, minProportionDifference < 1),
                " the minimum difference in methylation needs to be in the interval (0,1)")

  .stopIfNotAll(c(.isInteger(minReadsPerCytosine, positive=TRUE)),
                " the minimum number of reads in a bin is an integer higher or equal to 0")

  .stopIfNotAll(is.logical(ciExcludesOne),
                " choose ciExcludesOne as TRUE if want to filter out regions whose F-test 95% confidence interval spans 1")

  if (!is.null(varRatioFc)) {
    .stopIfNotAll(c(is.numeric(varRatioFc), length(varRatioFc) == 1, varRatioFc >= 1),
    "`varRatioFc` must be a single numeric value >= 1 (e.g. 2 for a 2 times variance cutoff)")
  }


  regions <- reduce(regions)

  if(length(potentialVMRs) > 0){

    # extract the methylation data in the correct context
    cat("Extract methylation in the corresponding context \n")

    contextMethylationData1 <- methylationData1[methylationData1$context%in%context]
    rm(methylationData1)
    localContextMethylationData1 <- contextMethylationData1[queryHits(findOverlaps(contextMethylationData1, regions))]
    rm(contextMethylationData1)

    contextMethylationData2 <- methylationData2[methylationData2$context%in%context]
    rm(methylationData2)
    localContextMethylationData2 <- contextMethylationData2[queryHits(findOverlaps(contextMethylationData2, regions))]
    rm(contextMethylationData2)

    localContextMethylationData <- .joinMethylationData(localContextMethylationData1, localContextMethylationData2)
    rm(localContextMethylationData1, localContextMethylationData2)


    regionsList <- .splitGRangesEqualy(regions, cores)

    .filterVMRsLoop = function(i){
      computedVMRs <- GRanges()
      for(index in 1:length(regionsList[[i]])){
        currentRegion <- regionsList[[i]][index]


        cat("Computing VMRs at ",.printGenomicRanges(currentRegion),"\n")

        cat("Selecting data...\n")

        # Select the points in methylationData that we're interested in. These are the
        # points that lie within 'regions', as well as any that lie within
        # window.size of them.

        overlapsPotentialVMRs <- findOverlaps(potentialVMRs, currentRegion)
        if(length(overlapsPotentialVMRs) > 0){
          potentialVMRsLocal <- potentialVMRs[queryHits(overlapsPotentialVMRs)]

          localMethylationData <- localContextMethylationData[queryHits(findOverlaps(localContextMethylationData, currentRegion))]
          potentialVMRsLocal <- .analyseReadsInsideRegionsVMR(localMethylationData, potentialVMRsLocal)

          if(length(computedVMRs) == 0){
            computedVMRs <- potentialVMRsLocal
          } else{
            computedVMRs <- c(computedVMRs,potentialVMRsLocal)
          }
        }
      }
      return(computedVMRs)
    }

    # compute the VMRs
    if(cores > 1){
      cat("Compute the VMRs using ", cores, "cores\n")
      computedVMRs <- BiocParallel::bplapply(1:length(regionsList), .filterVMRsLoop, BPPARAM = BPPARAM)
    } else {
      computedVMRs <- lapply(1:length(regionsList), .filterVMRsLoop)
    }

    computedVMRs <-  unlist(GRangesList(computedVMRs))

    if(length(computedVMRs) > 0){
      computedVMRs$regionType <- rep("loss", length(computedVMRs))
      computedVMRs$regionType[which(computedVMRs$proportion1 < computedVMRs$proportion2)] <- "gain"
      computedVMRs$direction <- rep(-1, length(computedVMRs))
      computedVMRs$direction[which(computedVMRs$proportion1 < computedVMRs$proportion2)] <- 1
      computedVMRs <- computedVMRs[order(computedVMRs)]
    } else{
      computedVMRs <- GRanges()
    }

    if(length(computedVMRs) > 0){
      cat("Identifying VMRs...\n")
      computedVMRs <- computedVMRs[which(!is.na(computedVMRs$f_pvalue))] ## REMOVE this part if not want to discard NA value
      computedVMRs$wilcox_pvalue <- stats::p.adjust(computedVMRs$wilcox_pvalue, method = "fdr")
      computedVMRs$f_pvalue <- stats::p.adjust(computedVMRs$f_pvalue, method = "fdr")

      bufferIndex <- !is.na(computedVMRs$f_pvalue) &
        abs(computedVMRs$proportion1 - computedVMRs$proportion2) >= minProportionDifference &
        computedVMRs$sumReadsN1/computedVMRs$cytosinesCount >= minReadsPerCytosine &
        computedVMRs$sumReadsN2/computedVMRs$cytosinesCount >= minReadsPerCytosine &
        computedVMRs$cytosinesCount >= minCytosinesCount
      if (ciExcludesOne == TRUE){
        bufferIndex <- bufferIndex & (sapply(computedVMRs$F_test_result, function(x) x$conf.int[1]) > 1 |
                                        sapply(computedVMRs$F_test_result, function(x) x$conf.int[2]) < 1)
      } else {
        bufferIndex <- bufferIndex & !(sapply(computedVMRs$F_test_result, function(x) x$conf.int[1]) > 1 |
                                         sapply(computedVMRs$F_test_result, function(x) x$conf.int[2]) < 1)
      }

      if (!is.null(varRatioFc)){
        bufferIndex <- bufferIndex & (computedVMRs$var_ratio >= varRatioFc | computedVMRs$var_ratio <= 1/varRatioFc)
      }

      computedVMRs <- computedVMRs[bufferIndex]

      computedVMRs$is_DMR <- computedVMRs$wilcox_pvalue <= pValueThreshold
      computedVMRs$is_VMR <- computedVMRs$f_pvalue <= pValueThreshold
    }
  } else{
    computedVMRs <- GRanges()
  }

  return(computedVMRs)

}

#' Performs the analysis in all regions in a \code{\link[GenomicRanges]{GRanges}} object,
#' computing per-region summary statistics and statistical tests.
#'
#' @title Analyse reads inside regions (with VMR testing)
#'
#' @param methylationData A \code{GRanges} of methylation calls. Must include metadata:
#'   \describe{
#'     \item{readsM1, readsN1}{methylated and total read counts in condition 1}
#'     \item{readsM2, readsN2}{methylated and total read counts in condition 2}
#'     \item{ONT_Cm, ONT_C}{per-read methylation and coverage indices (character vectors)}
#'   }
#'
#' @param regions A \code{GRanges} of genomic intervals to summarise.
#'
#' @return A \code{GRanges} with the same ranges as \code{regions}, plus these metadata:
#' \describe{
#'   \item{sumReadsM1}{total methylated reads in condition 1}
#'   \item{sumReadsN1}{total reads in condition 1}
#'   \item{proportion1}{methylation proportion (sumReadsM1/sumReadsN1)}
#'   \item{variance1}{variance of per-read methylation proportions in condition 1}
#'   \item{sumReadsM2}{total methylated reads in condition 2}
#'   \item{sumReadsN2}{total reads in condition 2}
#'   \item{proportion2}{methylation proportion (sumReadsM2/sumReadsN2)}
#'   \item{variance2}{variance of per-read methylation proportions in condition 2}
#'   \item{cytosinesCount}{number of cytosines observed in each region}
#'   \item{wilcox_pvalue}{p-value from Wilcoxon rank-sum test comparing per-read proportions}
#'   \item{f_pvalue}{p-value from F-test comparing variances of per-read proportions}
#'   \item{var_ratio}{Ratio of variances (variance1 / variance2)}
#'   \item{wilcox_result}{Full \code{htest} object returned by \code{wilcox.test}}
#'   \item{F_test_result}{Full \code{htest} object returned by \code{var.test}}
#' }
#'
#' @author Radu Zabet and Young Jun Kim
.analyseReadsInsideRegionsVMR <- function(methylationData, regions){

  overlaps <- findOverlaps(methylationData, regions, ignore.strand = TRUE)
  methylationDataContextList <- S4Vectors::splitAsList(methylationData[queryHits(overlaps)],  subjectHits(overlaps))
  regionsIndexes <- as.integer(names(methylationDataContextList))

  regions$sumReadsM1 <- rep(0, times=length(regions))
  regions$sumReadsN1 <- rep(0, times=length(regions))
  regions$proportion1 <- rep(0, times=length(regions))
  regions$variance1 <- rep(0, times=length(regions))
  regions$sumReadsM2 <- rep(0, times=length(regions))
  regions$sumReadsN2 <- rep(0, times=length(regions))
  regions$proportion2 <- rep(0, times=length(regions))
  regions$variance2 <- rep(0, times=length(regions))
  regions$cytosinesCount <- rep(0, times=length(regions))
  regions$wilcox_pvalue <- rep(0, times=length(regions))
  regions$f_pvalue <- rep(0, times=length(regions))
  ### run the wilcoxon test for comparing the proportion per reads between two dataset
  if (length(regionsIndexes) > 0){
    regions$sumReadsM1[regionsIndexes] <- sapply(methylationDataContextList,.sumReadsM1)
    regions$sumReadsN1[regionsIndexes] <- sapply(methylationDataContextList,.sumReadsN1)
    regions$sumReadsM2[regionsIndexes] <- sapply(methylationDataContextList,.sumReadsM2)
    regions$sumReadsN2[regionsIndexes] <- sapply(methylationDataContextList,.sumReadsN2)
    regions$cytosinesCount[regionsIndexes] <- sapply(methylationDataContextList,length)

    valid <- regions$cytosinesCount[regionsIndexes] > 0
    regions$proportion1[regionsIndexes[valid]] <- regions$sumReadsM1[regionsIndexes[valid]]/regions$sumReadsN1[regionsIndexes[valid]]
    regions$proportion2[regionsIndexes[valid]] <- regions$sumReadsM2[regionsIndexes[valid]]/regions$sumReadsN2[regionsIndexes[valid]]

    # get the wilcoxon test and f-test result
    test_result_list <- lapply(methylationDataContextList, .wilcox_ftestPerRead)

    wilcox_pvals <- sapply(test_result_list, function(res) res$wilcoxTest$p.value)
    var1        <- sapply(test_result_list, function(res) res$varience1)
    var2        <- sapply(test_result_list, function(res) res$varience2)
    f_pvals     <- sapply(test_result_list, function(res) {
      if (!is.null(res$fTest)) res$fTest$p.value else NA_real_
    })
    wilcox_objs <- lapply(test_result_list, function(res) res$wilcoxTest)
    f_objs     <- lapply(test_result_list, function(res) res$fTest)

    regions$wilcox_pvalue[regionsIndexes] <- wilcox_pvals
    regions$variance1[regionsIndexes]        <- var1
    regions$variance2[regionsIndexes]        <- var2
    regions$var_ratio[regionsIndexes]        <- var1 / var2
    regions$f_pvalue[regionsIndexes]         <- f_pvals
    regions$wilcox_result[regionsIndexes]    <- wilcox_objs
    regions$F_test_result[regionsIndexes]    <- f_objs
  }
  return(regions)
}


.sumReadsM1 <- function(methylationData){
  return(sum(methylationData$readsM1))
}

.sumReadsN1 <- function(methylationData){
  return(sum(methylationData$readsN1))
}

.sumReadsM2 <- function(methylationData){
  return(sum(methylationData$readsM2))
}

.sumReadsN2 <- function(methylationData){
  return(sum(methylationData$readsN2))
}

.wilcox_ftestPerRead <- function(methylationData1){
  results <- list()
  # collect the sequence index from GRanges (ONT_Cm, ONT_C)
  read_Cm_idx_list <- strsplit(unlist(methylationData1$ONT_Cm),c("_"))
  read_C_idx_list <- strsplit(unlist(methylationData1$ONT_C),c("_"))

  sample1_read_Cm_idx <- sapply(read_Cm_idx_list, function(x) x[1] == "Sample1")
  sample2_read_Cm_idx <- sapply(read_Cm_idx_list, function(x) x[1] == "Sample2")
  read_Cm_idx1_list <- read_Cm_idx_list[sample1_read_Cm_idx]
  read_Cm_idx2_list <- read_Cm_idx_list[sample2_read_Cm_idx]

  sample1_read_C_idx <- sapply(read_C_idx_list, function(x) x[1] == "Sample1")
  sample2_read_C_idx <- sapply(read_C_idx_list, function(x) x[1] == "Sample2")
  read_C_idx1_list <- read_C_idx_list[sample1_read_C_idx]
  read_C_idx2_list <- read_C_idx_list[sample2_read_C_idx]

  .make_id_vector <- function(list) {
    # list is e.g. read_Cm_idx1_list
    if (length(list)==0) return(character(0))
    # extract the "second field", coerce to character, drop any NAs
    tmp <- as.character(sapply(list, function(x) x[2], USE.NAMES=FALSE))
    tmp <- tmp[!is.na(tmp)]
    if (length(tmp)==0) return(character(0))
    unique(unlist(strsplit(tmp, ",", fixed=TRUE), use.names=FALSE))
  }
  read_Cm_idx1 <- .make_id_vector(read_Cm_idx1_list)
  read_Cm_idx2 <- .make_id_vector(read_Cm_idx2_list)

  read_C_idx1  <- .make_id_vector(read_C_idx1_list)
  read_C_idx2  <- .make_id_vector(read_C_idx2_list)

  total_idx1 <- sort(unique(c(read_Cm_idx1,read_C_idx1)))
  total_idx2 <- sort(unique(c(read_Cm_idx2,read_C_idx2)))

  # compute proportions per reads using the GRanges (ONT_Cm, ONT_C)
  proportions_S1 <- rep(0, length(total_idx1))
  proportions_S2 <- rep(0, length(total_idx1))

  for (k in seq_along(total_idx1)){
    id <- total_idx1[k]
    per_read_Cm_S1 <- sapply(mcols(methylationData1)$ONT_Cm, function(x) {
      s1 <- x[grepl("^Sample1_", x)]
      if (length(s1) == 0) return(0L)
      sum(str_count(s1, fixed(id)))
    })

    per_read_C_S1 <- sapply(mcols(methylationData1)$ONT_C, function(x) {
      s1 <- x[grepl("^Sample1_", x)]
      if (length(s1) == 0) return(0L)
      sum(str_count(s1, fixed(id)))
    })
    per_read_S1_M <-sum(per_read_Cm_S1)
    per_read_S1_N <-sum(per_read_Cm_S1)+sum(per_read_C_S1)
    proportions_S1[k] <- per_read_S1_M/per_read_S1_N
  }

  for (k in seq_along(total_idx2)){
    id <- total_idx2[k]
    per_read_Cm_S2 <- sapply(mcols(methylationData1)$ONT_Cm, function(x) {
      s2 <- x[grepl("^Sample2_", x)]
      if (length(s2) == 0) return(0L)
      sum(str_count(s2, fixed(id)))
    })

    per_read_C_S2 <- sapply(mcols(methylationData1)$ONT_C, function(x) {
      s2 <- x[grepl("^Sample2_", x)]
      if (length(s2) == 0) return(0L)
      sum(str_count(s2, fixed(id)))
    })

    per_read_S2_M <-sum(per_read_Cm_S2)
    per_read_S2_N <-sum(per_read_Cm_S2)+sum(per_read_C_S2)
    proportions_S2[k] <- per_read_S2_M/per_read_S2_N
    # ## DEBUG
    # cat(c("per_read_S2_M:",per_read_S2_M," per_read_S2_N:",per_read_S2_N," per_read_S2_P:",per_read_S2_M/per_read_S2_N,"\n"))
  }

  # run Wilcoxon test and f-test
  if (length(proportions_S1) > 0 & length(proportions_S2) > 0) {
    wt <- try(wilcox.test(proportions_S1, proportions_S2), silent=TRUE)
    if (inherits(wt, "try-error") || !inherits(wt, "htest")) {
      wt <- structure(
        list(
          statistic   = setNames(NA_real_, "W"),
          p.value     = NA_real_,
          alternative = "true location shift is not equal to 0",
          method      = "Wilcoxon rank sum test",
          data.name   = paste0("proportions_S1 and proportions_S2")
        ),
        class = "htest"
      )
    }
  }
  results$wilcoxTest<- wt
  results$varience1 <- var(proportions_S1)
  results$varience2 <- var(proportions_S2)

  if(length(proportions_S1)>=2 & length(proportions_S2)>=2 &
     var(proportions_S1, na.rm=TRUE) > 0 & var(proportions_S2, na.rm=TRUE) > 0){
    results$fTest <- var.test(proportions_S1,proportions_S2)
  } else {
    results$fTest <- NULL
  }
 return(results)
}

#' This function computes the p-values of the Score test.
#'
#' @title Score test
#' @param m1 the number of methylated reads in condition 1
#' @param n1 the total number of methylated reads in condition 1
#' @param m2 the number of methylated reads in condition 2
#' @param n2 the total number of methylated reads in condition 2
#' @return The p-values of the Score test
#'
#' @author Radu Zabet
.scoreTest <-  function(m1, n1, m2, n2){
  p <- (m1 + m2)/(n1 + n2)

  zScore <- rep(NA, times=length(m1))
  bufferIndexes <- which((p * (1-p))!=0)
  p1 <- m1/n1
  p2 <- m2/n2
  zScore[bufferIndexes]  <- (p1[bufferIndexes] - p2[bufferIndexes]) / (sqrt(p[bufferIndexes] * (1-p[bufferIndexes])*(1/n1[bufferIndexes]+1/n2[bufferIndexes])))
  bufferIndexes <- which((p * (1-p))==0)
  zScore[bufferIndexes] <- 0

  pValue <- rep(NA, times=length(zScore))
  bufferIndexes <- which(!is.na(zScore))
  pValue[bufferIndexes] <- 2*pnorm(-abs(zScore[bufferIndexes]))

  return(pValue)
}


#' This function computes the p-values of the Fisher's exact test.
#'
#' @title Fisher test
#' @param m1 the number of methylated reads in condition 1
#' @param n1 the total number of methylated reads in condition 1
#' @param m2 the number of methylated reads in condition 2
#' @param n2 the total number of methylated reads in condition 2
#' @param alternative indicates the alternative hypothesis and must be one of
#' \code{"two.sided"}, \code{"greater"} or \code{"less"}. You can specify just
#' the initial letter. Only used in the 2 by 2 case.
#' @return The p-values of Fisher's exact test.
#'
#' @author Radu Zabet
.fisherTest <-  function(m1, n1, m2, n2, alternative = c("two.sided", "less", "greater")){

  u1 <- n1 - m1
  u2 <- n2 - m2
  dataM <- cbind(m1, u1, m2, u2)

  pValue <- rep(NA, times=length(m1))

  bufferIndexes <- which(!(is.na(u1) | is.na(u2) | n1 <= 0 | n2 <= 0 | m1 < 0 | m2 < 0 | m1 > n1 | m2 > n2))

  if(length(bufferIndexes) > 0){
    pValue[bufferIndexes] <- unlist(lapply( apply(matrix(dataM[bufferIndexes, ], ncol=4), 1, list),
                                            .fisherTestPValue,
                                            alternative = alternative))
  }

  pValue[pValue > 1] <- 1
  pValue[pValue < 0] <- 0

  return(pValue)
}

#' This function computes the p-values of the Fisher's exact test.
#'
#' @title Fisher test
#' @param x a vector with 4 values
#' @return The p-values of Fisher's exact test.
#'
#' @author Radu Zabet
.fisherTestPValue <-  function(x, alternative = c("two.sided", "less", "greater")){
  return(stats::fisher.test(matrix(unlist(x),nrow=2, byrow=TRUE), alternative = alternative)$p.value)
}

#' This function computes the adjusted p-values (using Benjamini & Hochberg
#' method)
#'
#'
#' @title Compute adjusted p-values
#' @param test a \code{character} indicating the statistical test used.
#' (\code{"fisher"} for Fisher's exact test or \code{"score"} for Score test).
#' @param m1 the number of methylated reads in condition 1.
#' @param n1 the total number of methylated reads in condition 1.
#' @param m2 the number of methylated reads in condition 2.
#' @param n2 the total number of methylated reads in condition 2.
#' @param alternative indicates the alternative hypothesis and must be one of
#' \code{"two.sided"}, \code{"greater"} or \code{"less"}. You can specify just
#' the initial letter. Only used in the 2 by 2 case. This is used only for
#' Fisher's test
#' @return The adjusted p-values of the statistical test.
#' @author Radu Zabet
.computeAdjuestedPValues <- function(test, m1, n1, m2, n2, alternative = "two.sided"){
  if(test == "score"){
    pValue <- .scoreTest(m1,n1,m2,n2)
  } else{
    pValue <- .fisherTest(m1,n1,m2,n2, alternative="two.sided")
  }

  # convert p-values to FDR
  adjPValue <- rep(NA, times=length(pValue))

  adjPValue[which(!is.na(pValue))] <- stats::p.adjust(pValue[which(!is.na(pValue))], method="fdr")

  return(adjPValue)
}



#' This function computes the adjusted p-values (using Benjamini & Hochberg
#' method) using one the statistical tests
#'
#' @title Compute adjusted p-values in DMRs
#' @param test a \code{character} indicating the statistical test used .
#' (\code{"fisher"} for Fisher's exact test or \code{"score"} for Score test).
#' @param DMRs a \code{\link{GRanges}} object with the DMRs.
#' @param alternative indicates the alternative hypothesis and must be one of
#' \code{"two.sided"}, \code{"greater"} or \code{"less"}. You can specify just
#' the initial letter. Only used in the 2 by 2 case. This is used only for
#' Fisher's test.
#' @return The adjuested p-values of the statistical test.
#'
#' @author Radu Zabet
.computeaAjustedPValuesInDMRs <- function(test, DMRs ,alternative = "two.sided"){


  m1 <- DMRs$sumReadsM1
  n1 <- DMRs$sumReadsN1
  m2 <- DMRs$sumReadsM2
  n2 <- DMRs$sumReadsN2

  if(test == "score"){
    pValue <- .scoreTest(m1,n1,m2,n2)
  } else{
    pValue <- .fisherTest(m1,n1,m2,n2, alternative=alternative)
  }



  # convert p-values to FDR
  adjPValue <- rep(NA, times=length(pValue))

  adjPValue[which(!is.na(pValue))] <- stats::p.adjust(pValue[which(!is.na(pValue))], method="fdr")

  return(adjPValue)
}

#' Compute Co-Methylated Regions (CMRs)
#'
#' @title Compute pairwise co-methylation statistics between regions
#' @description
#' \code{computeCoMethylatedRegions()} calculates pairwise correlation statistics for methylation levels
#' across defined genomic regions (e.g., PMDs, Enhancer binding sites). For each region pair within the
#' specified distance range, the function computes per-read methylation proportions and
#' performs correlation testing (Pearson, Spearman, or Kendall). Pairs with strong correlations
#' (beyond user-defined thresholds) and significant p-values (FDR-adjusted) are returned.
#'
#' @param methylationData A \code{GRanges} object containing cytosine sites, annotated with
#' ONT methylation metadata (columns \code{ONT_Cm}, \code{ONT_C}, etc.).
#' @param regions A \code{GRanges} object defining genomic regions (e.g., PMDs, Enhancer binding sites) to
#' evaluate for CMRs.
#' @param minDistance Minimum genomic distance (in base pairs) between two regions to be considered (default: 500).
#' @param maxDistance Maximum genomic distance (in base pairs) between two regions (default: 50,000).
#' @param minCoverage Minimum number of shared reads (per region pair) required to compute correlation (default: 4).
#' @param pValueThreshold Significance threshold for FDR-adjusted p-values (default: 0.05).
#' @param correlation_test Statistical method to compute correlation; must be one of \code{"pearson"}, \code{"spearman"}, or \code{"kendall"} (default: \code{"pearson"}).
#' @param minCorrelation Minimum allowed correlation value for a significant result (must be in between -1 and 0; default: -0.5).
#' @param maxCorrelation Maximum allowed correlation value for a significant result (must be in between 0 and 1; default: 0.5).
#' @param parallel Logical; run in parallel if \code{TRUE}.
#' @param BPPARAM A \code{BiocParallelParam} object controlling parallel execution.
#' This value will automatically set when parallel is \code{TRUE}, also able to set as manually.
#' @param cores Integer number of workers (must not exceed BPPARAM$workers).
#'    This value will automatically set as the maximum number of system workers,
#'    also able to set as manually.
#' 
#' @return A \code{GInteractions} object, where each row represents a significantly correlated pair of genomic regions
#' from the input \code{regions}. The anchors of each interaction correspond to original regions,
#' and their genomic coordinates are retained in the \code{anchor1} and \code{anchor2} slots.
#'
#' Additionally, a \code{genomic_position} meta-column is included to indicate the original coordinate ranges
#' (in UCSC/IGV format) for each interaction, aiding downstream interpretation or visualisation.
#'
#' Each interaction is annotated with:
#' \describe{
#'   \item{correlation}{Correlation coefficient between the two regions}
#'   \item{coverage}{Number of shared reads used for correlation}
#'   \item{p.value}{FDR-adjusted p-value for the correlation test}
#' }
#' @details
#' The function first identifies all region pairs within the user-defined distance range. For each pair,
#' it calculates methylation proportions per read across both regions, extracts common reads,
#' and tests correlation using the selected method. FDR correction is applied globally across all region pairs.
#'
#' @seealso \code{\link{readONTbam}}, \code{\link{computePMDs}},
#' \code{\link{ontSampleGRangesList}}
#'
#' @examples
#' # Load methylation data and PMD regions
#' data("ont_gr_GM18870_chr1_sorted_bins_1k")
#' data("ont_gr_GM18870_chr1_PMD_bins_1k")
#' 
#' # Compute highly correlated regions (Pearson)
#' coMethylationRegion_pearson <- computeCoMethylatedRegions(ont_gr_GM18870_chr1_sorted_bins_1k,
#'                                                           ont_gr_GM18870_chr1_PMD_bins_1k[1:5],
#'                                                           minDistance = 500,
#'                                                           maxDistance = 50000,
#'                                                           minCoverage = 4,
#'                                                           pValueThreshold = 0.05,
#'                                                           correlation_test = "pearson",
#'                                                           minCorrelation = -0.5,
#'                                                           maxCorrelation = 0.5,
#'                                                           parallel = FALSE,
#'                                                           BPPARAM  = NULL)
#'
#' @author Nicolae Radu Zabet and Young Jun Kim
#' @import GenomicRanges
#' @import GenomicAlignments
#' @import InteractionSet
#' @import BiocParallel
#' @importFrom stats cor.test
#' @export

# correlation between the methylated regions
## Correlation between Methylated Regions
computeCoMethylatedRegions <- function(methylationData,
                                       regions,
                                       minDistance = 500,
                                       maxDistance = 50000,
                                       minCoverage = 4,
                                       pValueThreshold = 0.05,
                                       correlation_test = "pearson",
                                       minCorrelation = -0.5,
                                       maxCorrelation = 0.5,
                                       parallel = FALSE,
                                       BPPARAM  = NULL,
                                       cores = NULL){
  ##Parameters checking
  cat("[computeCoMethylatedRegions] Parameters checking ...\n")
  
  # generate the BPPARAM value if set as parallel 
  if (parallel == TRUE){
    BPPARAM <- suppressWarnings(.validateBPPARAM(BPPARAM, progressbar = TRUE)) 
  }else{
    # Force serial execution
    BPPARAM <- BiocParallel::SerialParam(progressbar = TRUE)
  }
  # If cores argument is specified
  if (!is.null(cores)) {
    .stopIfNotAll(.isInteger(cores, positive = TRUE), 
                  "the number of cores used when computing the DMRs needs to be an integer higher or equal to 1.")
    
    # Check if user requested more cores than available
    if (cores > BPPARAM$workers) {
      warning(paste0("The number of requested cores (", cores, 
                     ") exceeds the available system cores (", BPPARAM$workers, 
                     "). Automatically setting cores to the maximum available (", 
                     BPPARAM$workers, ")."))
      cores <- BPPARAM$workers
    } else {
      message(paste0("Using user-specified core count: ", cores))
    }
    
    # Apply the final core number
    BPPARAM$workers <- cores
  } else {
    cores <- BPPARAM$workers
  }
  cat("Current parallel setting, BPPARAM: ", capture.output(BPPARAM),sep = "\n")
  
  .validateMethylationData(methylationData, variableName="methylationData")
  regions <- .validateGRanges(regions, methylationData)
  
  .validateCorrelationStatTest(correlation_test)
  
  .stopIfNotAll(c(!is.null(minDistance), is.numeric(minDistance), minDistance >= 0),
                " minDistance needs to be a numeric value higher or equal to 0")
  
  .stopIfNotAll(c(!is.null(maxDistance), is.numeric(maxDistance), maxDistance >= 0),
                " maxDistance needs to be a numeric value higher or equal to 0")
  
  .stopIfNotAll(c(maxDistance >= minDistance),
                " maxDistance should be higher than minDistance value")
  
  .stopIfNotAll(c(!is.null(minCorrelation), is.numeric(minCorrelation), minCorrelation <= 0, minCorrelation >= -1),
                " minCorrelation needs to be a numeric value between -1 and 0")
  
  .stopIfNotAll(c(!is.null(maxCorrelation), is.numeric(maxCorrelation), maxCorrelation <= 1, maxCorrelation >= 0),
                " maxCorrelation needs to be a numeric value between 0 and 1")
  
  .stopIfNotAll(c(maxCorrelation >= minCorrelation),
                " maxCorrelation should be higher than minCorrelation value")
  
  .stopIfNotAll(c(.isInteger(minCoverage, positive=TRUE)),
                " the minimum gap between PMDs is an integer higher or equal to 0")
  
  .stopIfNotAll(c(!is.null(pValueThreshold),
                  is.numeric(pValueThreshold),
                  pValueThreshold > 0,
                  pValueThreshold < 1),
                " the p-value threshold needs to be in the interval (0,1)")
  
  t_start <- Sys.time()
  
  # Find the cases for comparison within minDistance and maxDistance
  cat("[computeCoMethylatedRegions] Count all case of pairs...\n")
  allPairs <- .makeRegionPairs(regions, minDistance=500, maxDistance=50000)

  n_total <- length(allPairs$distance)
  results_gi =  GInteractions()
  
  # Compute the correlation of proportion per read by pairs
  cat("[computeCoMethylatedRegions] Compute Correlated Methylated Regions...\n")
  results_list <- bplapply(seq_len(n_total), function(i) {
    bin1 <- regions[allPairs[i,1]]
    bin2 <- regions[allPairs[i,2]]
    
    comet_result <- .computeMoleculeCoMeth(
      methylationData,
      bin1,
      bin2,
      method = correlation_test, 
      minCoverage = minCoverage
    )
    
    if (!is.null(comet_result)) {
      result <- GInteractions(bin1, bin2)
      result$pre_p.value <- comet_result$pvalue
      result$coverage <- comet_result$n
      result$correlation <- comet_result$correlation
      return(result)
    } else {
      return(NULL)
    }
  }, BPPARAM = BPPARAM)
  
  results_gi <- do.call(c, results_list[!vapply(results_list, is.null, logical(1))])
  
  cat("[computeCoMethylatedRegions] Filter Correlated Methylated Regions \n")
  # Adjust the p.value using the FDR method
  results_gi$p.value <- stats::p.adjust(results_gi$pre_p.value, method = 'fdr')
  mcols(results_gi)$pre_p.value <- NULL
  
  results_gi <- results_gi[
    which(
      !is.na(results_gi$p.value) & results_gi$p.value < pValueThreshold &
      !is.na(results_gi$correlation) &
        (results_gi$correlation < minCorrelation |
        results_gi$correlation > maxCorrelation)
    )
  ]
  
  total_elapsed <- difftime(Sys.time(), t_start, units = "secs")
  cat(sprintf("[computeCoMethylatedRegions] Done! Total elapsed time: %.1f sec\n", as.numeric(total_elapsed)))
  return(results_gi)
}

# --------------
#  Helper: compute the correlation between the regions
# --------------
# 1. Find out which regions make pairs within specific distance
.makeRegionPairs <- function(regions, minDistance=500, maxDistance=50000) {
  # Generate all possible unique pairs of region indices (without repetition, no self-pairs)
  combs <- combn(seq_along(regions), 2, simplify=TRUE)  # matrix of 2 x N pairs
  
  # Compute distances for each pair
  dists <- apply(combs, 2, function(x) distance(regions[x[1]], regions[x[2]]))
  
  # Keep only valid pairs
  keep <- !is.na(dists) & dists >= minDistance & dists <= maxDistance
  
  # Return a data.frame with indices and distance
  data.frame(
    regionA = combs[1, keep],
    regionB = combs[2, keep],
    distance = dists[keep]
  )
}

# 2. Calculate Correlation from the two regions
.computeMoleculeCoMeth <- function(methylationData,
                                   regionA,
                                   regionB,
                                   method = "pearson", 
                                   minCoverage = 5){
  # subset reads overlapping region A and region B
  readsA <- methylationData[queryHits(findOverlaps(methylationData, regionA))]
  readsB <- methylationData[queryHits(findOverlaps(methylationData, regionB))]
  
  indexPropA <- .getProportionPerRead(readsA)
  indexPropB <- .getProportionPerRead(readsB)
  
  commonIds <- intersect(indexPropA$index_id,indexPropB$index_id)
  
  propCommonA =indexPropA$proportion[match(commonIds, indexPropA$index_id)]
  propCommonB =indexPropB$proportion[match(commonIds, indexPropB$index_id)]
  
  # compute correlation
  if(length(commonIds) >= minCoverage){
    cor_result <- cor.test(propCommonA,propCommonB, method = method)
    
      return(list(correlation=cor_result$estimate,
              pvalue=cor_result$p.value,
              n=length(commonIds)))
  }
}

# 3. Get the proportions and index from per-read
# Returns a list with:
# - index_id: character vector of unique read identifiers
# - proportion: numeric vector of methylation proportion per read (M / (M + U))
.getProportionPerRead <- function(methylationData){
  if (!("ONT_Cm" %in% names(mcols(methylationData))) |
      !("ONT_C"  %in% names(mcols(methylationData)))) {
    return(list(mean = NA_real_, sd = NA_real_, mean_w = NA_real_, sd_w = NA_real_))
  }
  
  # collect the sequence index from GRanges (ONT_Cm, ONT_C)
  split_ids <- function(x) {
    if (length(x) == 0L) return(character(0))
    x <- as.character(x)
    x <- x[!is.na(x) & nzchar(x)]
    if (length(x) == 0L) return(character(0))
    unlist(strsplit(x, ",", fixed = TRUE), use.names = FALSE)
  }
  
  read_Cm_ids <- split_ids(unlist(methylationData$ONT_Cm))
  read_C_ids  <- split_ids(unlist(methylationData$ONT_C))
  total_idx   <- sort(unique(c(read_Cm_ids, read_C_ids)))
  total_idx   <- total_idx[nzchar(total_idx)]
  
  if (length(total_idx) == 0L) {
    return(list(mean = NA_real_, sd = NA_real_, mean_w = NA_real_, sd_w = NA_real_))
  }
  
  Cm_list <- lapply(mcols(methylationData)$ONT_Cm, split_ids)
  C_list  <- lapply(mcols(methylationData)$ONT_C,  split_ids)
  
  proportions_S <- numeric(length(total_idx))
  N_vec         <- numeric(length(total_idx))

  for (k in seq_along(total_idx)) {
    id <- total_idx[k]
    per_read_Cm_S <- vapply(Cm_list, function(vec) sum(vec == id), integer(1))
    per_read_C_S  <- vapply(C_list,  function(vec) sum(vec == id), integer(1))
    M <- sum(per_read_Cm_S)
    N <- M + sum(per_read_C_S)
    proportions_S[k] <- if (N > 0L) M / N else NA_real_
    N_vec[k]         <- N
  }

  # remove NA/NaN
  ok <- is.finite(proportions_S)
  total_idx <- total_idx[ok]
  p  <- proportions_S[ok]
  
  return(list(index_id = total_idx, proportion = p))
}

#' Checks whether the passed parameter is the statistical test for detecting Correlation in Regions
#'
#' @title Validate statistial test for Correlation between two Regions
#' @param test the statistical test used to call Correlation Between Methylated Regions (CMRs) (\code{"pearson"},
#'  \code{"kendall"} or \code{"spearman"} for Correlation test).
#' @keywords internal
#' @author Radu Zabet and Young Jun Kim
.validateCorrelationStatTest <- function(test){
  .stopIfNotAll(c(!is.null(test), is.character(test), length(test) == 1, test %in% c("pearson","kendall","spearman")),
                " Test needs to be one of the following \"pearson\", \"kendall\" or \"spearman\" for Correlation test" )
}

## ----setup, include=FALSE-----------------------------------------------------
local_caption <- paste(
  "Local methylation profile.",
  "The points on the graph represent methylation proportion of individual cytosines,",
  "their colour (red or blue) which sample they belong to and the intensity of the colour how many reads that particular cytosine had.",
  "This means that darker colours indicate stronger evidence that the corresponding cytosine has the corresponding methylation proportion,",
  "while lighter colours indicate a weaker evidence.",
  "The solid lines represent the smoothed profiles and the intensity of the colour the coverage at the corresponding position (darker colours indicate more reads while lighter ones less reads).",
  "The boxes on top represent the DMRs, where a filled box will represent a DMR which gained methylation while a box with a pattern represent a DMR that lost methylation.",
  "The DMRs need to have a metadata column `regionType` which can be either `gain` (where there is more methylation in condition 2 compared to condition 1) or `loss` (where there is less methylation in condition 2 compared to condition 1).",
  "In case this metadata column is missing all DMRs are drawn using filled boxes.",
  "Finally we also allow annotation of the DNA sequence.",
  "We represent by black boxes all the exons, which are joined by a horizontal black line, thus, marking the full body of the gene.",
  "With grey boxes we mark the transposable elements.",
  "Both for genes and transposable elements we plot them over a mid line if they are on the positive strand and under the mid line if they are on the negative strand."
)

knitr::opts_chunk$set(
  echo = TRUE,
  message = FALSE,
  warning = FALSE,
  fig.width = 7,
  fig.height = 5)

## ----pvalue, echo=TRUE, message=FALSE, eval=FALSE-----------------------------
# pValue <- 2*pnorm(-abs(zScore))

## ----pvalue_adjust, echo=TRUE, message=FALSE, eval=FALSE----------------------
# pValue <- p.adjust(pValue, method="fdr")

## ----load_presaved, message=FALSE,cache=FALSE---------------------------------
library(DMRcaller)

#load presaved data
data(methylationDataList)

## ----load, echo=TRUE, message=FALSE, cache=FALSE, eval=FALSE------------------
# # specify the Bismark CX report files
# saveBismark(methylationDataList[["WT"]],
#             "chr3test_a_thaliana_wt.CX_report")
# saveBismark(methylationDataList[["met1-3"]],
#             "chr3test_a_thaliana_met13.CX_report")
# 
# # load the data
# methylationDataWT <- readBismark("chr3test_a_thaliana_wt.CX_report")
# methylationDataMet13 <- readBismark("chr3test_a_thaliana_met13.CX_report")
# methylationDataList <- GRangesList("WT" = methylationDataWT,
#                                    "met1-3" = methylationDataMet13)

## ----pool_data, echo=TRUE, message=FALSE, cache=FALSE, eval=FALSE-------------
# # load the data
# methylationDataAll <- poolMethylationDatasets(methylationDataList)
# 
# # In the case of 2 elements, this is equivalent to
# methylationDataAll <- poolTwoMethylationDatasets(methylationDataList[[1]],
#                                                  methylationDataList[[2]])

## ----read_pool_data, echo=TRUE, message=FALSE, cache=FALSE, eval=FALSE--------
# # load the data
# methylationDataAll <- readBismarkPool(c(file_wt, file_met13))

## ----selectCytosine_example, echo=TRUE, message=FALSE, cache=FALSE,eval=FALSE----
# library(BSgenome.Hsapiens.UCSC.hg38)
# 
# # Select cytosines in CG context on chromosome 1
# gr_cpg <- selectCytosine(
#   genome = BSgenome.Hsapiens.UCSC.hg38,
#   context = "CG",
#   chr     = "chr1"
# )
# gr_cpg

## ----read_ONT, echo=TRUE,message=FALSE,cache=FALSE,eval=FALSE-----------------
# # Path to ONT BAM (with MM/ML tags)
# bamfile <- system.file("extdata",
#                        "scanBamChr1Random5.bam",
#                        package="DMRcaller")
# 
# # Decode methylation and compute read-level counts
# OntChr1CG <- readONTbam(
#   bamfile    = bamfile,
#   ref_gr     = gr_cpg,
#   genome     = BSgenome.Hsapiens.UCSC.hg38,
#   modif      = "C+m?",
#   chr        = "chr1",
#   region     = FALSE,
#   prob_thresh= 0.5,
#   parallel   = FALSE,
# )
# 
# # View results
# OntChr1CG

## ----BSgenome_available, echo=TRUE, cache=FALSE, message=FALSE, eval=FALSE----
# head(BSgenome::available.genomes())
# 

## ----load_ontSampleGRangesList, message=FALSE, cache=FALSE--------------------
# load presaved ont GRanges data
data(ontSampleGRangesList)

## ----lowResProfile, fig.height=4, fig.width=12, fig.cap="Low resolution profile in CG context for WT and met1-3", out.width='95%',echo=TRUE, fig.align='center', message=FALSE, cache=FALSE----
plotMethylationProfileFromData(methylationDataList[["WT"]],
                               methylationDataList[["met1-3"]],
                               conditionsNames = c("WT","met1-3"),
                               windowSize = 10000,
                               autoscale = FALSE,
                               context = c("CG"))

## ----profile_fine, echo=TRUE, message=FALSE, cache=FALSE,  fig.width=5,fig.height=5, fig.align='center', eval=FALSE----
# regions <- GRanges(seqnames = Rle("Chr3"), ranges = IRanges(1,1E6))
# 
# # compute low resolution profile in 10 Kb windows
# profileCGWT <- computeMethylationProfile(methylationDataList[["WT"]],
#                                          regions,
#                                          windowSize = 10000,
#                                          context = "CG")
# 
# profileCGMet13 <- computeMethylationProfile(methylationDataList[["met1-3"]],
#                                             regions,
#                                             windowSize = 10000,
#                                             context = "CG")
# 
# profilesCG <- GRangesList("WT" = profileCGWT, "met1-3" = profileCGMet13)
# 
# #plot the low resolution profile
# par(mar=c(4, 4, 3, 1)+0.1)
# par(mfrow=c(1,1))
# plotMethylationProfile(profilesCG,
#                        autoscale = FALSE,
#                        labels = NULL,
#                        title="CG methylation on Chromosome 3",
#                        col=c("#D55E00","#E69F00"),
#                        pch = c(1,0),
#                        lty = c(4,1))

## ----coverage, fig.width=6,fig.height=5,out.width='60%', fig.cap="Coverage. For example, this figure shows that in WT only $30%$ of the cytosines in CHH context have at least 10 reads", fig.align='center', message=FALSE,cache=FALSE----
# plot the coverage in the two contexts
par(mar=c(4, 4, 3, 1)+0.1)
plotMethylationDataCoverage(methylationDataList[["WT"]],
                            methylationDataList[["met1-3"]],
                            breaks = c(1,5,10,15),
                            regions = NULL,
                            conditionsNames=c("WT","met1-3"),
                            context = c("CHH"),
                            proportion = TRUE,
                            labels=LETTERS,
                            contextPerRow = FALSE)

## ----coverage_compute, message=FALSE,cache=FALSE, eval=FALSE------------------
# coverageCGWT <- computeMethylationDataCoverage(methylationDataList[["WT"]],
#                                                context="CG",
#                                                breaks = c(1,5,10,15))

## ----correlationPlot, fig.width=6,fig.height=5, fig.cap="Spatial correlation of methylation levels", fig.align='center', out.width='60%', message=FALSE,cache=FALSE----
# compute the spatial correlation of methylation levels
plotMethylationDataSpatialCorrelation(methylationDataList[["WT"]],
                            distances = c(1,100,10000), regions = NULL,
                            conditionsNames = c("WT"),
                            context = c("CG"),
                            labels = LETTERS, col = NULL,
                            pch = c(1,0,16,2,15,17), lty = c(4,1,3,2,6,5),
                            contextPerRow = FALSE,
                            log = "x")

## ----correlation_compute, message=FALSE,cache=FALSE, eval=FALSE---------------
# correlation_CG_wt <- computeMethylationDataSpatialCorrelation(methylationDataList[["WT"]],
#                                                               context="CG",
#                                                               distances=c(1,10,100,1000,10000))

## ----compute_DMRs_CG_noise_filter, message=FALSE,cache=FALSE------------------
chr_local <- GRanges(seqnames = Rle("Chr3"), ranges = IRanges(5E5,6E5))

# compute the DMRs in CG context with noise_filter method
DMRsNoiseFilterCG <- computeDMRs(methylationDataList[["WT"]],
                      methylationDataList[["met1-3"]],
                      regions = chr_local,
                      context = "CG",
                      method = "noise_filter",
                      windowSize = 100,
                      kernelFunction = "triangular",
                      test = "score",
                      pValueThreshold = 0.01,
                      minCytosinesCount = 4,
                      minProportionDifference = 0.4,
                      minGap = 0,
                      minSize = 50,
                      minReadsPerCytosine = 4,
                      cores = 1)

print(DMRsNoiseFilterCG)

## ----compute_DMRs_CG_neighbourhood, message=FALSE,cache=FALSE-----------------
# compute the DMRs in CG context with neighbourhood method
DMRsNeighbourhoodCG <- computeDMRs(methylationDataList[["WT"]],
                                   methylationDataList[["met1-3"]],
                                   regions = chr_local,
                                   context = "CG",
                                   method = "neighbourhood",
                                   test = "score",
                                   pValueThreshold = 0.01,
                                   minCytosinesCount = 4,
                                   minProportionDifference = 0.4,
                                   minGap = 200,
                                   minSize = 1,
                                   minReadsPerCytosine = 4,
                                   cores = 1)

print(DMRsNeighbourhoodCG)

## ----compute_DMRs_CG_bins, message=FALSE,cache=FALSE--------------------------
# compute the DMRs in CG context with bins method
DMRsBinsCG <- computeDMRs(methylationDataList[["WT"]],
                          methylationDataList[["met1-3"]],
                          regions = chr_local,
                          context = "CG",
                          method = "bins",
                          binSize = 100,
                          test = "score",
                          pValueThreshold = 0.01,
                          minCytosinesCount = 4,
                          minProportionDifference = 0.4,
                          minGap = 200,
                          minSize = 50,
                          minReadsPerCytosine = 4,
                          cores = 1)

print(DMRsBinsCG)

## ----compute_DMRs_CG_GE, message=FALSE,cache=FALSE----------------------------
# load the gene annotation data
data(GEs)

#select the genes
genes <- GEs[which(GEs$type == "gene")]

# compute the DMRs in CG context over genes
DMRsGenesCG <- filterDMRs(methylationDataList[["WT"]],
                          methylationDataList[["met1-3"]],
                          potentialDMRs = genes[overlapsAny(genes, chr_local)],
                          context = "CG",
                          test = "score",
                          pValueThreshold = 0.01,
                          minCytosinesCount = 4,
                          minProportionDifference = 0.4,
                          minReadsPerCytosine = 3,
                          cores = 1)
print(DMRsGenesCG)

## ----merge_DMRs_CG_noise_filter, message=FALSE,cache=FALSE--------------------
DMRsNoiseFilterCGMerged <- mergeDMRsIteratively(DMRsNoiseFilterCG,
                                                minGap = 200,
                                                respectSigns = TRUE,
                                                methylationDataList[["WT"]],
                                                methylationDataList[["met1-3"]],
                                                context = "CG",
                                                minProportionDifference = 0.4,
                                                minReadsPerCytosine = 4,
                                                pValueThreshold = 0.01,
                                                test="score")
print(DMRsNoiseFilterCGMerged)

## ----differenceMethylationReplicates, fig.width=12,fig.height=5,out.width='95%', fig.cap="Methylation proportions in the synthetic dataset.", fig.align='center',message=FALSE,cache=FALSE----
# loading synthetic data
data("syntheticDataReplicates")

# create vector with colours for plotting
cbbPalette <- c("#000000", "#E69F00", "#56B4E9", "#009E73", "#F0E442",
                "#0072B2", "#D55E00", "#CC79A7")

# plotting the difference in proportions
plot(start(syntheticDataReplicates),
     syntheticDataReplicates$readsM1/syntheticDataReplicates$readsN1,
     ylim=c(0,1), col=cbbPalette[2], xlab="Position in Chr3 (bp)",
     ylab="Methylation proportion")
points(start(syntheticDataReplicates),
       syntheticDataReplicates$readsM2/syntheticDataReplicates$readsN2,
       col=cbbPalette[7], pch=4)
points(start(syntheticDataReplicates),
       syntheticDataReplicates$readsM3/syntheticDataReplicates$readsN3,
       col=cbbPalette[3], pch=2)
points(start(syntheticDataReplicates),
       syntheticDataReplicates$readsM4/syntheticDataReplicates$readsN4,
       col=cbbPalette[6], pch=3)
legend(x = "topleft", legend=c("Treatment 1", "Treatment 2", "Control 1",
                               "Control 2"), pch=c(1,4,2,3),
       col=cbbPalette[c(2,7,3,6)], bty="n", cex=1.0)

## ----computingBiologicalReplicates, message=FALSE,cache=FALSE, eval=TRUE------
if (requireNamespace("GenomeInfoDb", quietly = TRUE)) {
# starting with data joined using joinReplicates

# creating condition vector
condition <- c("a", "a", "b", "b")

# computing DMRs using the neighbourhood method
DMRsReplicatesBins <- computeDMRsReplicates(methylationData = syntheticDataReplicates,
                                            condition = condition,
                                            regions = NULL,
                                            context = "CG",
                                            method = "bins",
                                            binSize = 100,
                                            test = "betareg",
                                            pseudocountM = 1,
                                            pseudocountN = 2,
                                            pValueThreshold = 0.01,
                                            minCytosinesCount = 4,
                                            minProportionDifference = 0.4,
                                            minGap = 0,
                                            minSize = 50,
                                            minReadsPerCytosine = 4,
                                            cores = 1)

print(DMRsReplicatesBins)
}

## ----analyse_reads_inside_regions_for_condition, message=FALSE,cache=FALSE----
#retrive the number of reads in CHH context in WT in CG DMRs
DMRsNoiseFilterCGreadsCHH <- analyseReadsInsideRegionsForCondition(
                              DMRsNoiseFilterCGMerged,
                              methylationDataList[["WT"]], context = "CHH",
                              label = "WT")
print(DMRsNoiseFilterCGreadsCHH)

## ----compute_PMDs_CG_noise_filter, message=FALSE, cache=FALSE,eval=FALSE------
# # load the ONT methylation data
# data(ontSampleGRangesList)
# 
# # define the region of interest on chr1
# chr1_ranges <- GRanges(seqnames = Rle("chr1"), ranges = IRanges(1E6+5E5,2E6))
# 
# # compute the PMDs in CG context using the noise_filter method
# PMDsNoiseFilterCG <- computePMDs(ontSampleGRangesList[["GM18501"]],
#                                  regions = chr1_ranges,
#                                  context = "CG",
#                                  method = "noise_filter",
#                                  windowSize = 100,
#                                  kernelFunction = "triangular",
#                                  lambda = 0.5,
#                                  minCytosinesCount = 4,
#                                  minMethylation = 0.4,
#                                  maxMethylation = 0.6,
#                                  minGap = 200,
#                                  minSize = 50,
#                                  minReadsPerCytosine = 4,
#                                  cores = 1,
#                                  parallel = FALSE)
# 
# PMDsNoiseFilterCG

## ----compute_PMDs_CG_neighbourhood, message=FALSE, cache=FALSE, eval=FALSE----
# # compute the PMDs using the neighbourhood method
# PMDsNeighbourhoodCG <- computePMDs(ontSampleGRangesList[["GM18501"]],
#                                    regions = chr1_ranges,
#                                    context = "CG",
#                                    method = "neighbourhood",
#                                    minCytosinesCount = 4,
#                                    minMethylation = 0.4,
#                                    maxMethylation = 0.6,
#                                    minGap = 200,
#                                    minSize = 50,
#                                    minReadsPerCytosine = 4,
#                                    cores = 1,
#                                    parallel = FALSE)
# 
# PMDsNeighbourhoodCG

## ----compute_PMDs_CG_bins, message=FALSE, cache=FALSE, eval=FALSE-------------
# # compute the PMDs using the bins method
# PMDsBinsCG <- computePMDs(ontSampleGRangesList[["GM18501"]],
#                           regions = chr1_ranges,
#                           context = "CG",
#                           method = "bins",
#                           binSize = 100,
#                           minCytosinesCount = 4,
#                           minMethylation = 0.4,
#                           maxMethylation = 0.6,
#                           minGap = 200,
#                           minSize = 50,
#                           minReadsPerCytosine = 4,
#                           cores = 1,
#                           parallel = FALSE)
# 
# PMDsBinsCG

## ----compute_VMDs_CG, message=FALSE, cache=FALSE, eval=FALSE------------------
# # load the ONT methylation data
# data(ontSampleGRangesList)
# 
# # define the region of interest on chr1
# chr1_ranges <- GRanges(seqnames = Rle("chr1"), ranges = IRanges(1E6+5E5, 1E6+6E5))
# 
# # compute the VMDs in CG context using the "EDE.high" variance threshold
# VMDsBinsCG <- computeVMDs(ontSampleGRangesList[["GM18501"]],
#                           regions = chr1_ranges,
#                           context = "CG",
#                           binSize = 100,
#                           minCytosinesCount = 4,
#                           sdCutoffMethod = "per.high",  # elbow-detected high variance
#                           percentage = 0.05,            # only used if sdCutoffMethod = "per.high" or "per.low"
#                           minGap = 200,
#                           minSize = 50,
#                           minReadsPerCytosine = 4,
#                           parallel = FALSE,
#                           cores = 1)
# 
# VMDsBinsCG

## ----filter_VMRs, message=FALSE, cache=FALSE, eval=FALSE----------------------
# # load the ONT methylation data
# data(ontSampleGRangesList)
# # load the gene annotation data
# data(GEs_hg38)
# 
# # select the transcript
# transcript <- GEs_hg38[which(GEs_hg38$type == "transcript")]
# 
# # the regions where to compute the PMDs
# regions <- GRanges(seqnames = Rle("chr1"), ranges = IRanges(1E6+5E5,2E6))
# transcript <- transcript[overlapsAny(transcript, regions)]
# 
# # filter genes that are differntially methylated in the two conditions
# VMRsGenesCG <- filterVMRsONT(ontSampleGRangesList[["GM18501"]],
#                ontSampleGRangesList[["GM18876"]], potentialVMRs = transcript,
#                context = "CG", pValueThreshold = 0.01,
#                minCytosinesCount = 4, minProportionDifference = 0.01,
#                minReadsPerCytosine = 3, ciExcludesOne = TRUE,
#                varRatioFc = NULL, parallel = TRUE) # parallel recommended

## ----compute_comethylation_fisher, message=FALSE, cache=FALSE, eval=TRUE------
# load ONT methylation and PMD data
data("ont_gr_GM18870_chr1_PMD_bins_1k")
data("ont_gr_GM18870_chr1_sorted_bins_1k")

# compute co-methylation using Fisher's exact test
coMethylationFisher <- computeCoMethylatedPositions(
  ont_gr_GM18870_chr1_sorted_bins_1k,
  regions = ont_gr_GM18870_chr1_PMD_bins_1k[1:4],
  minDistance = 150,
  maxDistance = 1000,
  minCoverage = 1,
  pValueThreshold = 0.01,
  test = "fisher",
  parallel = FALSE)

coMethylationFisher

## ----compute_cmrs_pearson, message=FALSE, cache=FALSE, eval=TRUE--------------
# load ONT methylation and PMD data
data("ont_gr_GM18870_chr1_sorted_bins_1k")
data("ont_gr_GM18870_chr1_PMD_bins_1k")

# compute highly correlated PMD region pairs using Pearson method
coMethRegions <- computeCoMethylatedRegions(
  methylationData = ont_gr_GM18870_chr1_sorted_bins_1k,
  regions = ont_gr_GM18870_chr1_PMD_bins_1k[1:5],
  minDistance = 500,
  maxDistance = 50000,
  minCoverage = 4,
  pValueThreshold = 0.05,
  correlation_test = "pearson",
  minCorrelation = -0.5,
  maxCorrelation = 0.5,
  parallel = FALSE
)

coMethRegions

## ----DMRdensity, fig.width=15,fig.height=2.5,out.width='95%', fig.cap= "Distribution of DMRs. Darker colour indicates higher density, while lighter colour lower density.", fig.align='center', message=FALSE,cache=FALSE----
# compute the distribution of DMRs
hotspots <- computeOverlapProfile(DMRsNoiseFilterCGMerged, chr_local,
                                  windowSize=5000, binary=TRUE)
# plot the distribution of DMRs
plotOverlapProfile(GRangesList("Chr3"=hotspots))

## ----localMethylationProfile, fig.width=15,fig.height=5,out.width='95%',  fig.cap=local_caption, fig.align='center',message=FALSE,cache=FALSE----
# select a 20 Kb location on the Chr3
chr3Reg <- GRanges(seqnames = Rle("Chr3"), ranges = IRanges(510000,530000))

# create a list with all DMRs
DMRsCGList <- list("noise filter" = DMRsNoiseFilterCGMerged,
                   "neighbourhood" = DMRsNeighbourhoodCG,
                   "bins" = DMRsBinsCG,
                   "genes" = DMRsGenesCG)
# plot the local profile
par(cex=0.9)
par(mar=c(4, 4, 3, 1)+0.1)
plotLocalMethylationProfile(methylationDataList[["WT"]],
                            methylationDataList[["met1-3"]],
                            chr3Reg,
                            DMRsCGList,
                            conditionsNames = c("WT", "met1-3"),
                            GEs,
                            windowSize = 300,
                            main="CG methylation")

## ----DMRinParallel, message=FALSE,cache=FALSE---------------------------------
# Parallel; compute the DMRs in CG context with noise_filter method
DMRsNoiseFilterCG_3cores <- computeDMRs(methylationDataList[["WT"]],
                      methylationDataList[["met1-3"]],
                      regions = chr_local,
                      context = "CG",
                      method = "noise_filter",
                      windowSize = 100,
                      kernelFunction = "triangular",
                      test = "score",
                      pValueThreshold = 0.01,
                      minCytosinesCount = 4,
                      minProportionDifference = 0.4,
                      minGap = 0,
                      minSize = 50,
                      minReadsPerCytosine = 4,
                      parallel = TRUE, # default: FALSE
                      BPPARAM = NULL, # default: NULL
                      cores = 3 # default: 1
                      ) 

print(DMRsNoiseFilterCG_3cores)

## ----session_info, echo=TRUE--------------------------------------------------
sessionInfo()

#' This function computes the variance methylated domains between pre-set
#' min and max proportion values.  
#'
#' @title Compute VMDs
#' @param methylationData the methylation data in condition
#' (see \code{\link{ontSampleGRangesList}}).
#' @param regions a \code{\link{GRanges}} object with the regions where to 
#' compute the VMDs. If \code{NULL}, the VMDs are computed genome-wide.
#' @param context the context in which the VMDs are computed (\code{"CG"}, 
#' \code{"CHG"} or \code{"CHH"}).
#' @param binSize the size of the tiling bins in nucleotides. This parameter is 
#' required only if the selected method is \code{"bins"}.
#' @param minCytosinesCount VMDs with less cytosines in the specified context 
#' than \code{minCytosinesCount} will be discarded.
#' @param sdCutoffMethod Character string specifying how to determine the cutoff
#' for filtering VMDs based on their methylation variance (standard deviation).
#' Available options are:
#' \describe{
#'   \item{\code{"per.high"}}{Selects the top \code{percentage} of regions with the highest variance (standard deviation).}
#'   \item{\code{"per.low"}}{Selects the bottom \code{percentage} of regions with the lowest variance.}
#'   \item{\code{"EDE.high"}}{Uses the elbow point (inflection/knee) from the descendingly sorted variance values to determine a data-driven high-variance cutoff. Retains regions with SD above this elbow point.}
#'   \item{\code{"EDE.low"}}{Uses the elbow point from the ascendingly sorted variance values to define a low-variance cutoff. Retains regions with SD below this point.}
#' }
#' This allows either quantile-based filtering or automatic detection of variance thresholds based on distribution shape.
#' @param percentage Numeric cutoff used when \code{sdCutoffMethod} is set to
#' \code{"per.high"} or \code{"per.low"}. Represents the quantile threshold:
#' for example, \code{percentage = 0.05} keeps the top 5\% or bottom 5\% of
#' bins based on standard deviation, depending on the selected method.
#' @param minGap VMDs separated by a gap of at least \code{minGap} are not 
#' merged. Note that only VMDs where the change in methylation is in the same 
#' direction are joined.
#' @param minSize VMDs with a size smaller than \code{minSize} are discarded.
#' @param minReadsPerCytosine  VMDs with the average number of reads lower than 
#' \code{minReadsPerCytosine} are discarded. 
#' @param parallel Logical; run in parallel if \code{TRUE}.
#' @param BPPARAM A \code{BiocParallelParam} object controlling parallel execution.
#'    This value will automatically set when parallel is \code{TRUE}, also able to set as manually.
#' @param cores Integer number of workers (must not exceed BPPARAM$workers).
#'    This value will automatically set as the maximum number of system workers,
#'    also able to set as manually.
#' @return the VMDs stored as a \code{\link{GRanges}} object with the following 
#' metadata columns:
#' \describe{
#'  \item{context}{the context in which the VMDs was computed (\code{"CG"}, 
#'  \code{"CHG"} or \code{"CHH"}).}
#'  \item{sumReadsM}{the number of methylated reads.}
#'  \item{sumReadsN}{the total number of reads.} 
#'  \item{proportion}{the proportion from total methylated reads.} 
#'  \item{cytosinesCount}{the number of cytosines in the VMDs.}
#'  \item{mean}{mean value comparing per‐read proportions}
#'  \item{sd}{standard deviation comparing per‐read proportions}
#'  \item{w_mean}{weighted mean value comparing per‐read proportions}
#'  \item{w_sd}{weighted standard deviation comparing per‐read proportions} 
#' }
#' @seealso \code{\link{readONTbam}}, \code{\link{filterVMDs}}
#' and \code{\link{analyseReadsInsideRegionsForCondition}}
#' @examples
#' \dontrun{
#' # load the ONT methylation data 
#' data(ontSampleGRangesList)
#' 
#' # the regions where to compute the VMDs
#' chr1_ranges <- GRanges(seqnames = Rle("chr1"), ranges = IRanges(1E6+5E5,1E6+6E5))
#' 
#' # compute the VMDs in CG context with bins method
#' VMDsBinsCG <- computeVMDs(ontSampleGRangesList[["GM18501"]],
#'                           regions = NULL,
#'                           context = "CG",
#'                           binSize = 100,
#'                           minCytosinesCount = 4,
#'                           sdCutoffMethod = "EDE.high",
#'                           percentage = 0.05,
#'                           minGap = 200,
#'                           minSize = 50,
#'                           minReadsPerCytosine = 4,
#'                           parallel = FALSE,
#'                           BPPARAM = NULL,
#'                           cores = 1)
#' }
#' @author Nicolae Radu Zabet, Jonathan Michael Foonlan Tsang and Young Jun Kim
#' @import inflection 
#' @export
 

computeVMDs <- function(methylationData, 
                        regions = NULL, 
                        context = "CG", 
                        binSize = 100,
                        minCytosinesCount = 4, 
                        sdCutoffMethod = "per.high",
                        percentage = 0.05,
                        minGap = 200, 
                        minSize = 50, 
                        minReadsPerCytosine = 4, 
                        parallel = FALSE,
                        BPPARAM = NULL,
                        cores = NULL) {
  ##Parameters checking
  cat("Parameters checking ...\n")
  
  # generate the BPPARAM value if set as parallel 
  if (parallel == TRUE){
    BPPARAM <- suppressWarnings(.validateBPPARAM(BPPARAM, progressbar = TRUE)) 
  }else{
    # Force serial execution
    BPPARAM <- BiocParallel::SerialParam(progressbar = TRUE)
  }
  # If cores argument is specified
  if (!is.null(cores)) {
    .stopIfNotAll(.isInteger(cores, positive = TRUE), 
                  "the number of cores used when computing the DMRs needs to be an integer higher or equal to 1.")
    
    # Check if user requested more cores than available
    if (cores > BPPARAM$workers) {
      warning(paste0("The number of requested cores (", cores, 
                     ") exceeds the available system cores (", BPPARAM$workers, 
                     "). Automatically setting cores to the maximum available (", 
                     BPPARAM$workers, ")."))
      cores <- BPPARAM$workers
    } else {
      message(paste0("Using user-specified core count: ", cores))
    }
    
    # Apply the final core number
    BPPARAM$workers <- cores
  } else {
    cores <- BPPARAM$workers
  }
  cat("Current parallel setting, BPPARAM: ", capture.output(BPPARAM),sep = "\n")
  
  .validateMethylationData(methylationData, variableName="methylationData")
  
  regions <- .validateGRanges(regions, methylationData)
  
  .validateContext(context)
  
  .stopIfNotAll(c(.isInteger(binSize, positive=TRUE)),
                " the bin size used by the method is an integer higher than 0")
  
  .stopIfNotAll(c(.isInteger(minCytosinesCount, positive=TRUE)), 
                " the minCytosinesCount is an integer higher or equal to 0")
  
  .stopIfNotAll(c(!is.null(sdCutoffMethod),
                  all(is.character(sdCutoffMethod)),
                  length(sdCutoffMethod) == 1,
                  all(sdCutoffMethod %in% c("per.high","per.low","EDE.high","EDE.low"))),
                " sdCutoffMethod can be only per.high, per.low, EDE.high or EDE.low")
  
  if (sdCutoffMethod %in% c("per.high", "per.low")){
    .stopIfNotAll(c(!is.null(percentage),is.numeric(percentage), percentage >= 0, percentage <= 1),
                  " percentage needs to be a numeric value between 0 and 1.")
  }
  
  .stopIfNotAll(c(.isInteger(minGap, positive=TRUE)),
                " the minimum gap between VMDs is an integer higher or equal to 0")
  
  .stopIfNotAll(c(.isInteger(minSize, positive=TRUE)),
                " the minimum size of a DMR is an integer higher or equal to 0")
  
  .stopIfNotAll(c(.isInteger(minReadsPerCytosine, positive=TRUE)), 
                " the minimum average number of reads in a DMR is an integer higher or equal to 0")
  
  computedVMDs <- GRanges()
  
  computedVMDs <- .computeVMDsBins(methylationData = methylationData, 
                                   regions = regions, 
                                   context = context, 
                                   binSize = binSize,
                                   minCytosinesCount = minCytosinesCount, 
                                   minGap = minGap, 
                                   minSize = minSize, 
                                   minReadsPerCytosine = minReadsPerCytosine,
                                   sdCutoffMethod = sdCutoffMethod,
                                   percentage = percentage,
                                   cores = cores,
                                   parallel = parallel,
                                   BPPARAM = BPPARAM)
  
  return(computedVMDs)  
}


#' This function computes the variance methylated regions 
#' using the bins method. 
.computeVMDsBins <- function(methylationData,
                             regions = NULL, 
                             context = "CG", 
                             binSize = 100,
                             minCytosinesCount = 4,
                             minGap = 200, 
                             minSize = 50, 
                             minReadsPerCytosine = 4, 
                             sdCutoffMethod = "per.high",
                             percentage = 0.05,
                             cores = 1,
                             parallel = parallel,
                             BPPARAM = BPPARAM) {
  
  regions <- reduce(regions)
  
  # extract the methylation data in the correct context
  cat("Extract methylation in the corresponding context \n")
  
  contextMethylationData <- methylationData[methylationData$context%in%context]
  rm(methylationData)
  localContextMethylationData <- contextMethylationData[queryHits(findOverlaps(contextMethylationData, regions))]
  rm(contextMethylationData)  
  
  regionsList <- .splitGRangesEqualy(regions, cores)
  
  # inner loop function for BiocParallel::bplapply
  .computeVMDsBinsLoop = function(i){
    computedVMDs <- GRanges()
    for(index in 1:length(regionsList[[i]])){
      currentRegion <- regionsList[[i]][index]
      
      
      cat("Computing VMDs at ",.printGenomicRanges(currentRegion),"\n")
      
      seqs <- seq(start(currentRegion), (end(currentRegion)-binSize), by = binSize);
      
      bins <- GRanges(seqnames(currentRegion), IRanges(seqs, (seqs+binSize-1)))
      
      overlapsBins <- findOverlaps(localContextMethylationData, currentRegion)
      
      if(length(overlapsBins) > 0){
        localMethylationData <- localContextMethylationData[queryHits(overlapsBins)]
        
        cat("Count inside each bin...\n")
        #bins <- .analyseReadsInsideRegions(localMethylationData, bins, context, cores)
        bins <- .analyseReadsInsideBinsVMDs(localMethylationData, bins, currentRegion)
        
        cat("Filter the bins...\n")
        # Get rid of the bins with fewer than minCytosinesCount cytosines inside.  
        bins  <- bins[bins$cytosinesCount >= minCytosinesCount]
        
        # Get rid of the bins with fewer than minReadsPerCytosine reads per cytosine.  
        bins  <- bins[(bins$sumReadsN/bins$cytosinesCount >= minReadsPerCytosine)]
        
        # filtering VMD in this part
        
        cat("Identifying VMDs...\n")    
        bins$context <- rep(paste(context, collapse = "_"), length(bins))
        bins$direction <- rep(NA, length(bins))
        bins$direction <- sign(bins$proportion)
        
        # Select the crude list of VMDs
        VMDs <- bins[!is.na(bins$direction) & bins$direction == 1]
        
        # append current VMDs to the global list of VMDs 
        if(length(computedVMDs) == 0){
          computedVMDs <- VMDs
        } else{
          computedVMDs <- c(computedVMDs,VMDs)
        }
      }
    }
    return(computedVMDs)
  }
  
  # compute the VMDs
  if(cores > 1){
    cat("Compute the VMDs using ", cores, "cores\n")
    computedVMDs <- BiocParallel::bplapply(1:length(regionsList), .computeVMDsBinsLoop, BPPARAM = BPPARAM)
  } else {
    computedVMDs <- lapply(1:length(regionsList), .computeVMDsBinsLoop)
  }
  
  computedVMDs <- unlist(GRangesList(computedVMDs))
  
  if(length(computedVMDs) > 0){
    
    cat("Merge adjacent VMDs\n")    
    computedVMDs <- computedVMDs[order(computedVMDs)]
    
    cat("Merge VMDs iteratively\n")    
    # Get rid of small gaps between VMDs.
    if(minGap > 0){
      computedVMDs <- .smartMergePMDs(computedVMDs, 
                                      minGap = minGap, 
                                      respectSigns = TRUE, 
                                      methylationData = localContextMethylationData,
                                      minReadsPerCytosine = minReadsPerCytosine,
                                      minMethylation = 0,
                                      maxMethylation = 1,
                                      cores = cores,
                                      BPPARAM = BPPARAM)
    }  
    computedVMDs <- computedVMDs[!duplicated(granges(computedVMDs))]
    hits <- findOverlaps(computedVMDs, computedVMDs, type = "within", ignore.strand = FALSE)
    computedVMDs <- computedVMDs[-unique(queryHits(hits)[queryHits(hits) != subjectHits(hits)])]
    
    cat("Filter VMDs \n") 
    
    computedVMDs <- DMRcaller::filterVMDs(localContextMethylationData,
                               computedVMDs,
                               context = context,
                               minCytosinesCount = minCytosinesCount,
                               minReadsPerCytosine = minReadsPerCytosine,
                               sdCutoffMethod = sdCutoffMethod,
                               percentage = percentage,
                               parallel = parallel,
                               BPPARAM = BPPARAM,
                               cores = cores)
    
   
    if(length(computedVMDs) > 0){
      #remove small VMDs 
      computedVMDs <- computedVMDs[width(computedVMDs) >= minSize]
      if(length(computedVMDs) > 0){
        #remove VMDswith few cytosines
        computedVMDs <- computedVMDs[!is.na(computedVMDs$cytosinesCount) & computedVMDs$cytosinesCount >= minCytosinesCount]
      }
    }
  }  
  if (length(computedVMDs) >0 & length(computedVMDs$direction) != 0){
    computedVMDs$direction <- NULL
  }
  return(computedVMDs)
}

.analyseReadsInsideBinsVMDs <- function(methylationData, bins, currentRegion){
  
  binSize <- min(unique(width(bins)))
  #Rcpp
  readsM <- .movingSum(start(currentRegion), end(currentRegion), start(methylationData), methylationData$readsM, windowSize = binSize)
  sumReadsM <- readsM[seq(1,length(readsM)-binSize, by=binSize)]
  
  readsN <- .movingSum(start(currentRegion), end(currentRegion), start(methylationData), methylationData$readsN, windowSize = binSize)
  sumReadsN <- readsN[seq(1,length(readsN)-binSize, by=binSize)]
  
  proportion <- sumReadsM/sumReadsN
  
  cytosines <- .movingSum(start(currentRegion), end(currentRegion), start(methylationData), rep(1, length(start(methylationData))), windowSize = binSize)
  cytosinesCount <- cytosines[seq(1,length(cytosines)-binSize, by=binSize)]
  
  
  overlaps <- findOverlaps(methylationData, bins, ignore.strand = TRUE)
  methylationDataContextList <- S4Vectors::splitAsList(methylationData[queryHits(overlaps)],  subjectHits(overlaps))
  regionsIndexes <- as.integer(names(methylationDataContextList))
  
  bin_result_list <- lapply(methylationDataContextList, .mean_sdPerRead)
  
  mean1 <- sapply(bin_result_list, function(res) res$mean)
  sd1   <- sapply(bin_result_list, function(res) res$sd)
  w_mean1 <- sapply(bin_result_list, function(res) res$mean_w)
  w_sd1   <- sapply(bin_result_list, function(res) res$sd_w)
  
  bins$sumReadsM <- sumReadsM
  bins$sumReadsN <- sumReadsN   
  bins$proportion <- proportion        
  bins$cytosinesCount <- cytosinesCount 
  
  bins$mean[regionsIndexes] <- mean1
  bins$sd[regionsIndexes]   <- sd1
  bins$w_mean[regionsIndexes] <- w_mean1
  bins$w_sd[regionsIndexes]   <- w_sd1
  
  return(bins)
}


#' This function verifies whether a set of potential VMDs (e.g. genes, 
#' transposons, CpG islands) are variance methylated or not.
#'
#' @title Filter VMDs 
#' @param methylationData the methylation data in condition
#' (see \code{\link{ontSampleGRangesList}}).
#' @param potentialVMDs a \code{\link{GRanges}} object with potential VMDs 
#' where to compute the VMDs. This can be a a list of gene and/or transposable 
#' elements coordinates.
#' @param context the context in which the VMDs are computed (\code{"CG"}, 
#' \code{"CHG"} or \code{"CHH"}).
#' @param minCytosinesCount VMDs with less cytosines in the specified context 
#' than \code{minCytosinesCount} will be discarded.
#' @param minReadsPerCytosine  VMDs with the average number of reads lower than 
#' \code{minReadsPerCytosine} are discarded. 
#' @param sdCutoffMethod Character string specifying how to determine the cutoff
#' for filtering VMDs based on their methylation variance (weighted standard deviation).
#' Available options are:
#' \describe{
#'   \item{\code{"per.high"}}{Selects the top \code{percentage} of regions with the highest variance (standard deviation).}
#'   \item{\code{"per.low"}}{Selects the bottom \code{percentage} of regions with the lowest variance.}
#'   \item{\code{"EDE.high"}}{Uses the elbow point (inflection/knee) from the descendingly sorted variance values to determine a data-driven high-variance cutoff. Retains regions with SD above this elbow point.}
#'   \item{\code{"EDE.low"}}{Uses the elbow point from the ascendingly sorted variance values to define a low-variance cutoff. Retains regions with SD below this point.}
#' }
#' This allows either quantile-based filtering or automatic detection of variance thresholds based on distribution shape.
#' @param percentage Numeric cutoff used when \code{sdCutoffMethod} is set to
#' \code{"per.high"} or \code{"per.low"}. Represents the quantile threshold:
#' for example, \code{percentage = 0.05} keeps the top 5\% or bottom 5\% of
#' bins based on weighted standard deviation, depending on the selected method.
#' @param parallel Logical; run in parallel if \code{TRUE}.
#' @param BPPARAM A \code{BiocParallelParam} object controlling parallel execution.
#'    This value will automatically set when parallel is \code{TRUE}, also able to set as manually.
#' @param cores Integer number of workers (must not exceed BPPARAM$workers).
#'    This value will automatically set as the maximum number of system workers,
#'    also able to set as manually.
#' @return a \code{\link{GRanges}} object with 9 metadata columns that contain 
#' the VMDs; see \code{\link{computeVMDs}}.
#' @seealso \code{\link{computeVMDs}} 
#' and \code{\link{analyseReadsInsideRegionsForCondition}}  
#' @examples
#' # load the ONT methylation data 
#' data(ontSampleGRangesList)
#' # load the gene annotation data
#' data(GEs_hg38)
#' 
#' # select the transcript
#' transcript <- GEs_hg38[which(GEs_hg38$type == "transcript")]
#' 
#' # the regions where to compute the VMDs
#' regions <- GRanges(seqnames = Rle("chr1"), ranges = IRanges(1E6+5E5,2E6))
#' transcript <- transcript[overlapsAny(transcript, regions)]
#' 
#' # filter genes that are variance methylated in the two conditions
#' VMDsGenesCG <- filterVMDs(ontSampleGRangesList[["GM18501"]], 
#'                potentialVMDs = transcript, 
#'                context = "CG", sdCutoffMethod = "per.high", percentage = 0.05,
#'                minCytosinesCount = 4, minReadsPerCytosine = 3, cores = 1)
#'
#' @author Nicolae Radu Zabet and Young Jun Kim
#' @export
filterVMDs <- function(methylationData, 
                       potentialVMDs, 
                       context = "CG",   
                       minCytosinesCount = 4, 
                       minReadsPerCytosine = 3, 
                       sdCutoffMethod = "per.high",
                       percentage = 0.05,
                       parallel = FALSE,
                       BPPARAM = NULL,
                       cores = NULL) {
  
  ##Parameters checking
  cat("Parameters checking ...\n")
  
  # generate the BPPARAM value if set as parallel 
  if (parallel == TRUE){
    BPPARAM <- suppressWarnings(.validateBPPARAM(BPPARAM, progressbar = TRUE)) 
  }else{
    # Force serial execution
    BPPARAM <- BiocParallel::SerialParam(progressbar = TRUE)
  }
  # If cores argument is specified
  if (!is.null(cores)) {
    .stopIfNotAll(.isInteger(cores, positive = TRUE), 
                  "the number of cores used when computing the DMRs needs to be an integer higher or equal to 1.")
    
    # Check if user requested more cores than available
    if (cores > BPPARAM$workers) {
      warning(paste0("The number of requested cores (", cores, 
                     ") exceeds the available system cores (", BPPARAM$workers, 
                     "). Automatically setting cores to the maximum available (", 
                     BPPARAM$workers, ")."))
      cores <- BPPARAM$workers
    } else {
      message(paste0("Using user-specified core count: ", cores))
    }
    
    # Apply the final core number
    BPPARAM$workers <- cores
  } else {
    cores <- BPPARAM$workers
  }
  cat("Current parallel setting, BPPARAM: ", capture.output(BPPARAM),sep = "\n")
  
  .validateMethylationData(methylationData, variableName="methylationData")
  
  regions <- getWholeChromosomes(methylationData)
  
  .validateContext(context)
  
  .validateGRanges(potentialVMDs, generateGenomeWide=FALSE, variableName="potentialVMDs", minLength=NULL)
  
  regions <- .validateGRanges(regions, methylationData)
  
  .stopIfNotAll(c(.isInteger(minCytosinesCount, positive=TRUE)), 
                " the minCytosinesCount is an integer higher or equal to 0")
  
  .stopIfNotAll(c(.isInteger(minReadsPerCytosine, positive=TRUE)), 
                " the minimum number of reads in a bin is an integer higher or equal to 0")
  
  regions <- reduce(regions)
  
  if(length(potentialVMDs) > 0){
    
    
    
    # extract the methylation data in the correct context
    cat("Extract methylation in the corresponding context \n")
    
    contextMethylationData <- methylationData[methylationData$context%in%context]
    rm(methylationData)
    localContextMethylationData <- contextMethylationData[queryHits(findOverlaps(contextMethylationData, regions))]
    rm(contextMethylationData)  
    
    regionsList <- .splitGRangesEqualy(regions, cores)
    
    # inner loop function for BiocParallel::bplapply
    .filterVMDsLoop = function(i){
      computedVMDs <- GRanges()  
      for(index in 1:length(regionsList[[i]])){
        currentRegion <- regionsList[[i]][index]
        
        
        cat("Computing VMDs at ",.printGenomicRanges(currentRegion),"\n")
        
        cat("Selecting data...\n")
        
        # Select the points in methylationData that we're interested in. These are the 
        # points that lie within 'regions', as well as any that lie within 
        # window.size of them. 
        
        overlapsPotentialVMDs <- findOverlaps(potentialVMDs, currentRegion)
        if(length(overlapsPotentialVMDs) > 0){
          potentialVMDsLocal <- potentialVMDs[queryHits(overlapsPotentialVMDs)]
          
          localMethylationData <- localContextMethylationData[queryHits(findOverlaps(localContextMethylationData, currentRegion))]
          potentialVMDsLocal <- .analyseReadsInsideRegionsVMDs(localMethylationData, potentialVMDsLocal)
          
          if(length(computedVMDs) == 0){
            computedVMDs <- potentialVMDsLocal
          } else{
            computedVMDs <- c(computedVMDs,potentialVMDsLocal)
          }
        } 
      }
      return(computedVMDs)
    }
    
    # compute the VMDs
    if(cores > 1){
      cat("Compute the VMDs using ", cores, "cores\n")
      computedVMDs <- BiocParallel::bplapply(1:length(regionsList), .filterVMDsLoop, BPPARAM = BPPARAM)
    } else {
      computedVMDs <- lapply(1:length(regionsList), .filterVMDsLoop)
    }
    
    
    computedVMDs <-  unlist(GRangesList(computedVMDs))
    
    if(length(computedVMDs) > 0){
      cat("Identifying VMDs...\n")    
      
      bufferIndex <- !is.na(computedVMDs$proportion) &
        computedVMDs$sumReadsN/computedVMDs$cytosinesCount >= minReadsPerCytosine &
        computedVMDs$cytosinesCount >= minCytosinesCount
      
      computedVMDs <- computedVMDs[bufferIndex]  
    }
  } else{
    computedVMDs <- GRanges() 
  }
  
  if(length(computedVMDs) > 0){
    computedVMDs <- computedVMDs[order(computedVMDs)]
  }
  ## cutoff by percentage method
  if (sdCutoffMethod %in% c("per.high","per.low")){
    # # Convert VMD result to data frame
    # df <- as.data.frame(computedVMDs)
    
    # # Sort by SD in descending order
    # df_sorted <- df[order(-df$w_sd), ]
    # 
    # # Create x-axis index
    # df_sorted$bin_index <- seq_len(nrow(df_sorted))
    
    # Setting the cutoff of sd
    cutoff_high <- quantile(computedVMDs$w_sd, 1-percentage, na.rm = TRUE)
    cutoff_low <- quantile(computedVMDs$w_sd, percentage, na.rm = TRUE)
    
    # col <- c("#D55E00","#E69F00", "#0072B2", "#56B4E9", "#F0E442", "#009E73")
    # # Plot for visual inspection 
    # plot(df_sorted$bin_index, df_sorted$w_sd, type="l", xlab="Bins (sorted by SD)", col = "black",
    #      ylab="Standard Deviation of Methylation", main = "SD distribution across bins")
    # abline(h = cutoff_high, lty = 3, col = col[1])  # horizontal dashed line at SD threshold
    # abline(h = cutoff_low, lty = 3, col = col[6])
    # legend(x = "topright", legend=c("cutoff_high", "cutoff_low"), lty = c(3,3),
    #        col = c(col[1],col[6]),   bty = "n",  cex = 0.7,inset = c(0.01, 0.01)
    # )
    if (sdCutoffMethod == "per.high"){
      computedVMDs <- computedVMDs[computedVMDs$w_sd >= cutoff_high]
    } else if (sdCutoffMethod == "per.low"){
      computedVMDs <- computedVMDs[computedVMDs$w_sd <= cutoff_low]
    }
  }
  
  ## cutoff by EDE method
  if (sdCutoffMethod %in% c("EDE.high","EDE.low")){
    # Sort them in decreasing order so the largest variance bins come first
    y <- sort(computedVMDs$w_sd, decreasing = TRUE)   
    x <- seq_along(y)   
    
    # Find the elbow point (knee) in the curve
    ede_res <- inflection::ede(x, y, 0)
    k_idx2 <- ede_res[1]
    k_idx3 <- ede_res[2]
    k_idx4 <- ede_res[3]
    knee_y2 <- y[k_idx2]
    knee_y3 <- y[k_idx3]
    knee_y4 <- y[k_idx4]
    
    # col <- c("#D55E00","#E69F00", "#0072B2", "#56B4E9", "#F0E442", "#009E73")
    # # Plot for visual inspection 
    # plot(x, y, type="l", xlab="Bins (sorted by SD)", col = "black",
    #      ylab="Standard Deviation of Methylation", main = "SD distribution across bins")
    # abline(h = knee_y2, lty = 3, col = col[2])  # horizontal dashed line at SD threshold
    # abline(h = knee_y4, lty = 3, col = col[3])
    # abline(h = knee_y3, lty = 3, col = col[4])
    # legend(x = "topright", legend=c("EDE j1", "EDE j2", "EDE j3"), lty = c(3,3,3),
    #        col = col[2:4],   bty = "n",  cex = 0.7,inset = c(0.01, 0.01)
    # )
    if (sdCutoffMethod == "EDE.high"){
      computedVMDs <- computedVMDs[computedVMDs$w_sd >= knee_y2]
    } else if (sdCutoffMethod == "EDE.low"){
      computedVMDs <- computedVMDs[computedVMDs$w_sd <= knee_y3]
    }
  }
  return(computedVMDs)
  
}

#' Performs the analysis in all regions in a \code{\link{GRanges}} object,
#' computing per-region mean and standard deviation of proportions.
#'
#' @title Analyse reads inside regions with computing mean and standard deviation per-read proportions
#'
#' @param methylationData A \code{GRanges} of methylation calls. Must include metadata:
#'   \describe{
#'     \item{readsM, readsN}{methylated and total read counts in condition}
#'     \item{ONT_Cm, ONT_C}{per-read methylation and coverage indices (character vectors)}
#'   }
#'
#' @param regions A \code{GRanges} of genomic intervals to summarise.
#'
#' @return A \code{GRanges} with the same ranges as \code{regions}, plus these metadata:
#' \describe{
#'   \item{sumReadsM}{total methylated reads in condition}
#'   \item{sumReadsN}{total reads in condition}
#'   \item{proportion}{methylation proportion (sumReadsM/sumReadsN)}
#'   \item{cytosinesCount}{number of cytosines observed in each region}
#'   \item{mean}{mean value comparing per‐read proportions}
#'   \item{sd}{standard deviation comparing per‐read proportions}
#'   \item{w_mean}{weighted mean value comparing per‐read proportions}
#'   \item{w_sd}{weighted standard deviation comparing per‐read proportions}
#' } 
#'       
#' @author Radu Zabet and Young Jun Kim
.analyseReadsInsideRegionsVMDs <- function(methylationData, regions){
  
  overlaps <- findOverlaps(methylationData, regions, ignore.strand = TRUE)
  methylationDataContextList <- S4Vectors::splitAsList(methylationData[queryHits(overlaps)],  subjectHits(overlaps))
  regionsIndexes <- as.integer(names(methylationDataContextList))
  
  regions$sumReadsM <- rep(0, times=length(regions))
  regions$sumReadsN <- rep(0, times=length(regions))    
  regions$proportion <- rep(0, times=length(regions)) 
  regions$cytosinesCount <- rep(0, times=length(regions))
  regions$mean <- rep(0, times=length(regions))
  regions$sd <- rep(0, times=length(regions))
  regions$w_mean <- rep(0, times=length(regions))
  regions$w_sd <- rep(0, times=length(regions))
  
  
  if(length(regionsIndexes) > 0){  
    regions$sumReadsM[regionsIndexes] <- sapply(methylationDataContextList,.sumReadsM)
    regions$sumReadsN[regionsIndexes] <- sapply(methylationDataContextList,.sumReadsN)               
    regions$cytosinesCount[regionsIndexes] <- sapply(methylationDataContextList,length)
    
    valid <- regions$cytosinesCount[regionsIndexes] > 0
    regions$proportion[regionsIndexes[valid]] <- regions$sumReadsM[regionsIndexes[valid]]/regions$sumReadsN[regionsIndexes[valid]]
    
    # get the mean and sd result per reads
    test_result_list <- lapply(methylationDataContextList, .mean_sdPerRead)
    
    mean1 <- sapply(test_result_list, function(res) res$mean)
    sd1   <- sapply(test_result_list, function(res) res$sd)
    w_mean1 <- sapply(test_result_list, function(res) res$mean_w)
    w_sd1   <- sapply(test_result_list, function(res) res$sd_w)
    
    regions$mean[regionsIndexes] <- mean1
    regions$sd[regionsIndexes]   <- sd1
    regions$w_mean[regionsIndexes] <- w_mean1
    regions$w_sd[regionsIndexes]   <- w_sd1
  }
  return(regions)
}


.sumReadsM <- function(methylationData){
  return(sum(methylationData$readsM))
}
.sumReadsN <- function(methylationData){
  return(sum(methylationData$readsN))
}

.mean_sdPerRead <- function(methylationData1){
  if (!("ONT_Cm" %in% names(mcols(methylationData1))) |
      !("ONT_C"  %in% names(mcols(methylationData1)))) {
    return(list(mean = NA_real_, sd = NA_real_, mean_w = NA_real_, sd_w = NA_real_))
  }

  # collect the sequence index from GRanges (ONT_Cm, ONT_C)
  split_ids <- function(x) {
    if (length(x) == 0L) return(character(0))
    x <- as.character(x)
    x <- x[!is.na(x) & nzchar(x)]
    if (length(x) == 0L) return(character(0))
    unlist(strsplit(x, ",", fixed = TRUE), use.names = FALSE)
  }
  
  read_Cm_ids <- split_ids(unlist(methylationData1$ONT_Cm))
  read_C_ids  <- split_ids(unlist(methylationData1$ONT_C))
  total_idx   <- sort(unique(c(read_Cm_ids, read_C_ids)))
  total_idx   <- total_idx[nzchar(total_idx)]
  
  if (length(total_idx) == 0L) {
    return(list(mean = NA_real_, sd = NA_real_, mean_w = NA_real_, sd_w = NA_real_))
  }

  Cm_list <- lapply(mcols(methylationData1)$ONT_Cm, split_ids)
  C_list  <- lapply(mcols(methylationData1)$ONT_C,  split_ids)
  
  proportions_S <- numeric(length(total_idx))
  N_vec         <- numeric(length(total_idx))
  
  for (k in seq_along(total_idx)) {
    id <- total_idx[k]
    per_read_Cm_S <- vapply(Cm_list, function(vec) sum(vec == id), integer(1))
    per_read_C_S  <- vapply(C_list,  function(vec) sum(vec == id), integer(1))
    M <- sum(per_read_Cm_S)
    N <- M + sum(per_read_C_S)
    proportions_S[k] <- if (N > 0L) M / N else NA_real_
    N_vec[k]         <- N
  }
  
  # remove NA/NaN
  ok <- is.finite(proportions_S)
  p  <- proportions_S[ok]
  w  <- N_vec[ok]
  
  
  if (length(p) == 0L) {
    return(list(mean = NA_real_, sd = NA_real_, mean_w = NA_real_, sd_w = NA_real_))
  }
  
  mean_u <- mean(p)
  sd_u   <- sd(p)
  
  if (length(w) > 0 && sum(w) > 0) {
    mu_w  <- sum(w * p) / sum(w)
    var_w <- sum(w * (p - mu_w)^2) / sum(w)   
    sd_w  <- sqrt(var_w)
  } else {
    mu_w <- NA_real_
    sd_w <- NA_real_
  }
  
  list(mean = mean_u, sd = sd_u, mean_w = mu_w, sd_w = sd_w)
}

#' This function extracts GC sites in the genome 
#'
#' @title Extract GC
#' @param methylationData the methylation data stored as a \code{\link{GRanges}}
#' object with four metadata columns (see \code{\link{methylationDataList}}).
#' 
#' @param genome a BSgenome with the DNA sequence of the organism
#' 
#' @param contexts the context in which the DMRs are computed (\code{"ALL"}, 
#' \code{"CG"}, \code{"CHG"} or \code{"CHH"}).
#' @return the a subset of \code{methylationData} consisting of all GC sites. 
#'
#' @examples
#' 
#' \dontrun{
#' # load the genome sequence
#' if(!require("BSgenome.Athaliana.TAIR.TAIR9", character.only = TRUE)){
#'   if (!requireNamespace("BiocManager", quietly=TRUE))
    #'   install.packages("BiocManager")
#'   BiocManager::install("BSgenome.Athaliana.TAIR.TAIR9")
#' }
#' library(BSgenome.Athaliana.TAIR.TAIR9)
#' 
#' # load the methylation data
#' data(methylationDataList)
#' 
#' methylationDataWTGpCpG <- extractGC(methylationDataList[["WT"]], 
#'                                     BSgenome.Athaliana.TAIR.TAIR9,
#'                                     "CG")
#' 
#' }
#' 
#' @author Ryan Merritt
extractGC <- function(methylationData, 
                      genome, 
                      contexts=c("ALL","CG","CHG","CHH")){
  # Splitting data according to the context specified
  message("Splitting data according to context \n")
  contexts <- match.arg(contexts)
  switch(contexts,
         "ALL" = metData <- methylationData,
         "CG" = metData <- methylationData[which(methylationData$context=="CG")],
         "CHH" = metData <- methylationData[which(methylationData$context=="CHH")],
         "CHG" = metData <- methylationData[which(methylationData$context=="CHG")]
  )
  # Shifting strands to get the base before the cytosine
  message("Shifting Strands \n")
  start(metData[strand(metData)=="+"]) <- start(metData[strand(metData)=="+"])-1
  end(metData[strand(metData)=="-"]) <- end(metData[strand(metData)=="-"])+1
  # Removing cytosines that are at the ends of each chromosome
  message("Removing data outside of chromosome length ranges \n")
  chromlength <- width(getSeq(genome))
  names(chromlength) <- seqlevels(metData)
  chromlength <- chromlength[which(seqlevels(metData)%in%seqlevels(genome))]
  for(i in 1:length(chromlength)){
    metData <- metData[!(seqnames(metData)==names(chromlength)[i] &
                           end(metData)>chromlength[i])]
  }
  for(i in 1:length(chromlength)){
    metData <- metData[!(seqnames(metData)==names(chromlength)[i] &
                           start(metData)<=0)]
  }
  message("Extracting GC \n")
  metSeq <- getSeq(genome,metData)
  metDataGC <- which(metSeq =="GC")
  metDataGC <- metData[metDataGC]
  message("Unshifting strands \n")
  start(metDataGC[strand(metDataGC)=="+"]) <- start(metDataGC[strand(metDataGC)
                                                              =="+"])+1
  end(metDataGC[strand(metDataGC)=="-"]) <- end(metDataGC[strand(metDataGC)
                                                          =="-"])-1
  if(contexts=="ALL"){
    metDataGC <- split(metDataGC,metDataGC$context)
  }
  return(metDataGC)
}

#' This function computes the differentially methylated regions between two 
#' conditions.  
#'
#' @title Compute DMRs
#' @param methylationData1 the methylation data in condition 1
#' (see \code{\link{methylationDataList}}).
#' @param methylationData2 the methylation data in condition 2
#' (see \code{\link{methylationDataList}}).
#' @param regions a \code{\link{GRanges}} object with the regions where to 
#' compute the DMRs. If \code{NULL}, the DMRs are computed genome-wide.
#' @param context the context in which the DMRs are computed (\code{"CG"}, 
#' \code{"CHG"} or \code{"CHH"}).
#' @param method the method used to compute the DMRs (\code{"noise_filter"}, 
#' \code{"neighbourhood"} or \code{"bins"}). The \code{"noise_filter"} method 
#' uses a triangular kernel to smooth the number of reads and then performs a 
#' statistical test to determine which regions dispay different levels of 
#' methylation in the two conditions. The \code{"neighbourhood"} method 
#' computates differentially methylated cytosines. Finally, the \code{"bins"} 
#' method partiones the genome into equal sized tilling bins and performs the 
#' statistical test between the two conditions in each bin. For all three 
#' methods, the cytosines or bins are then merged into DMRs without affecting 
#' the inital parameters used when calling the differentiall methylated 
#' cytosines/bins (p-value, difference in methylation levels, minimum number of 
#' reads per cytosine).
#' @param windowSize the size of the triangle base measured in nucleotides. 
#' This parameter is required only if the selected method is 
#' \code{"noise_filter"}. 
#' @param kernelFunction a \code{character} indicating which kernel function to 
#' be used. Can be one of \code{"uniform"}, \code{"triangular"}, 
#' \code{"gaussian"} or \code{"epanechnicov"}. This is required only if the 
#' selected method is \code{"noise_filter"}. 
#' @param lambda numeric value required for the Gaussian filter 
#' (\code{K(x) = exp(-lambda*x^2)}). This is required only if the selected 
#' method is \code{"noise_filter"} and the selected kernel function is 
#' \code{"gaussian"}. 
#' @param binSize the size of the tiling bins in nucleotides. This parameter is 
#' required only if the selected method is \code{"bins"}.
#' @param test the statistical test used to call DMRs (\code{"fisher"} for 
#' Fisher's exact test or \code{"score"} for Score test). 
#' @param pValueThreshold DMRs with p-values (when performing the statistical 
#' test; see \code{test}) higher or equal than \code{pValueThreshold} are 
#' discarded. Note that we adjust the p-values using the Benjamini and 
#' Hochberg's method to control the false discovery rate.
#' @param minCytosinesCount DMRs with less cytosines in the specified context 
#' than \code{minCytosinesCount} will be discarded.
#' @param minProportionDifference DMRs where the difference in methylation 
#' proportion between the two conditions is lower than 
#' \code{minProportionDifference} are discarded.
#' @param minGap DMRs separated by a gap of at least \code{minGap} are not 
#' merged. Note that only DMRs where the change in methylation is in the same 
#' direction are joined.
#' @param minSize DMRs with a size smaller than \code{minSize} are discarded.
#' @param minReadsPerCytosine  DMRs with the average number of reads lower than 
#' \code{minReadsPerCytosine} are discarded. 
#' @param parallel Logical; run in parallel if \code{TRUE}.
#' @param BPPARAM A \code{BiocParallelParam} object controlling parallel execution.
#'    This value will automatically set when parallel is \code{TRUE}, also able to set as manually.
#' @param cores Integer number of workers (must not exceed BPPARAM$workers).
#'    This value will automatically set as the maximum number of system workers,
#'    also able to set as manually.
#' @return the DMRs stored as a \code{\link{GRanges}} object with the following 
#' metadata columns:
#' \describe{
#'  \item{direction}{a number indicating whether the region lost (-1)  or gain 
#'  (+1) methylation in condition 2 compared to condition 1.}
#'  \item{context}{the context in which the DMRs was computed (\code{"CG"}, 
#'  \code{"CHG"} or \code{"CHH"}).}
#'  \item{sumReadsM1}{the number of methylated reads in condition 1.}
#'  \item{sumReadsN1}{the total number of reads in condition 1.} 
#'  \item{proportion1}{the proportion methylated reads in condition 1.} 
#'  \item{sumReadsM2}{the number of methylated reads in condition 2.} 
#'  \item{sumReadsN2}{the total number reads in condition 2.}
#'  \item{proportion2}{the proportion methylated reads in condition 2.} 
#'  \item{cytosinesCount}{the number of cytosines in the DMR.} 
#'  \item{regionType}{a string indicating whether the region lost (\code{"loss"})  
#'  or gained (\code{"gain"}) methylation in condition 2 compared to condition 1.} 
#'  \item{pValue}{the p-value (adjusted to control the false discovery rate with 
#'  the Benjamini and Hochberg's method) of the statistical test when the DMR was 
#'  called.} 
#' }
#' @seealso \code{\link{filterDMRs}}, \code{\link{mergeDMRsIteratively}}, 
#' \code{\link{analyseReadsInsideRegionsForCondition}} and 
#' \code{\link{DMRsNoiseFilterCG}}
#' @examples
#' 
#' # load the methylation data
#' data(methylationDataList)
#' 
#' # the regions where to compute the DMRs
#' regions <- GRanges(seqnames = Rle("Chr3"), ranges = IRanges(1,1E5))
#' 
#' # compute the DMRs in CG context with noise_filter method
#' DMRsNoiseFilterCG <- computeDMRs(methylationDataList[["WT"]], 
#'                      methylationDataList[["met1-3"]], regions = regions, 
#'                      context = "CG", method = "noise_filter", 
#'                      windowSize = 100, kernelFunction = "triangular",  
#'                      test = "score", pValueThreshold = 0.01, 
#'                      minCytosinesCount = 4, minProportionDifference = 0.4, 
#'                      minGap = 200, minSize = 50, minReadsPerCytosine = 4, 
#'                      cores = 1)
#' 
#' \dontrun{
#' # compute the DMRs in CG context with neighbourhood method
#' DMRsNeighbourhoodCG <- computeDMRs(methylationDataList[["WT"]], 
#'                        methylationDataList[["met1-3"]], regions = regions, 
#'                        context = "CG", method = "neighbourhood", 
#'                        test = "score", pValueThreshold = 0.01, 
#'                        minCytosinesCount = 4, minProportionDifference = 0.4, 
#'                        minGap = 200, minSize = 50, minReadsPerCytosine = 4, 
#'                        cores = 1)
#' 
#' # compute the DMRs in CG context with bins method
#' DMRsBinsCG <- computeDMRs(methylationDataList[["WT"]], 
#'                methylationDataList[["met1-3"]], regions = regions, 
#'                context = "CG", method = "bins", binSize = 100, 
#'                test = "score", pValueThreshold = 0.01, minCytosinesCount = 4, 
#'                minProportionDifference = 0.4, minGap = 200, minSize = 50, 
#'                minReadsPerCytosine = 4, cores = 1)
#' 
#' }
#' @author Nicolae Radu Zabet and Jonathan Michael Foonlan Tsang
#' @export
computeDMRs <- function(methylationData1, 
                        methylationData2, 
                        regions = NULL, 
                        context = "CG", 
                        method="noise_filter",
                        windowSize = 100,
                        kernelFunction = "triangular", 
                        lambda = 0.5,
                        binSize = 100,
                        test = "fisher", 
                        pValueThreshold = 0.01, 
                        minCytosinesCount = 4, 
                        minProportionDifference = 0.4,
                        minGap = 200, 
                        minSize = 50, 
                        minReadsPerCytosine = 4, 
                        parallel = FALSE,
                        BPPARAM = NULL,
                        cores = NULL) {
  ##Parameters checking
  cat("Parameters checking ...\n")
  
  # generate the BPPARAM value if set as parallel 
  if (parallel == TRUE){
    BPPARAM <- suppressWarnings(.validateBPPARAM(BPPARAM, progressbar = TRUE)) 
  }else{
    # Force serial execution
    BPPARAM <- BiocParallel::SerialParam(progressbar = TRUE)
  }
  # If cores argument is specified
  if (!is.null(cores)) {
    .stopIfNotAll(.isInteger(cores, positive = TRUE), 
                  "the number of cores used when computing the DMRs needs to be an integer higher or equal to 1.")
    
    # Check if user requested more cores than available
    if (cores > BPPARAM$workers) {
      warning(paste0("The number of requested cores (", cores, 
                     ") exceeds the available system cores (", BPPARAM$workers, 
                     "). Automatically setting cores to the maximum available (", 
                     BPPARAM$workers, ")."))
      cores <- BPPARAM$workers
    } else {
      message(paste0("Using user-specified core count: ", cores))
    }
    
    # Apply the final core number
    BPPARAM$workers <- cores
  } else {
    cores <- BPPARAM$workers
  }
  cat("Current parallel setting, BPPARAM: ", capture.output(BPPARAM),sep = "\n")
  
  .validateMethylationData(methylationData1, variableName="methylationData1")
  .validateMethylationData(methylationData2, variableName="methylationData2")
  
  
  regions <- union(.validateGRanges(regions, methylationData1), .validateGRanges(regions, methylationData2))
  
  .validateContext(context)
  
  .stopIfNotAll(c(!is.null(method), 
                  all(is.character(method)),
                  length(method) == 1,
                  all(method %in% c("noise_filter","neighbourhood","bins"))),
                " method can be only noise_filter, neighbourhood or bins")  
  
  if(method == "noise_filter"){
    .stopIfNotAll(c(.isInteger(windowSize, positive=TRUE)), 
                  " the window size used by the interpolation method is an integer higher than 0")
    
    .stopIfNotAll(c(!is.null(kernelFunction), 
                    kernelFunction%in%c("uniform", "triangular", "gaussian", "epanechnicov")), 
                  paste("Unknown kernel function: ", kernelFunction, ". 
                        kernelFunction should be one of \"uniform\", \"triangular\", 
                        \"gaussian\", \"epanechnicov\"",sep=""))
    
    if(kernelFunction == "gaussian"){
      .stopIfNotAll(c(!is.null(lambda),
                      is.numeric(lambda)), 
                    " lambda needs to be a numeric value.")
    }
    
  }
  
  if(method == "bins"){
    .stopIfNotAll(c(.isInteger(binSize, positive=TRUE)), 
                  " the bin size used by the method is an integer higher than 0")
    
  }
  
  .validateStatisticalTest(test)
  
  .stopIfNotAll(c(!is.null(pValueThreshold), 
                  is.numeric(pValueThreshold), 
                  pValueThreshold > 0, 
                  pValueThreshold < 1),
                " the p-value threshold needs to be in the interval (0,1)")
  
  .stopIfNotAll(c(.isInteger(minCytosinesCount, positive=TRUE)), 
                " the minCytosinesCount is an integer higher or equal to 0")
  
  .stopIfNotAll(c(!is.null(minProportionDifference), 
                  is.numeric(minProportionDifference), 
                  minProportionDifference > 0, 
                  minProportionDifference < 1), 
                " the minimum difference in methylation needs to be in the interval (0,1)")
  
  .stopIfNotAll(c(.isInteger(minGap, positive=TRUE)),
                " the minimum gap between DMRs is an integer higher or equal to 0")
  
  .stopIfNotAll(c(.isInteger(minSize, positive=TRUE)),
                " the minimum size of a DMR is an integer higher or equal to 0")
  
  .stopIfNotAll(c(.isInteger(minReadsPerCytosine, positive=TRUE)), 
                " the minimum average number of reads in a DMR is an integer higher or equal to 0")

  
  computedDMRs <- GRanges()
  
  if(method == "noise_filter"){
    computedDMRs <- .computeDMRsNoiseFilter(methylationData1 = methylationData1, 
                                            methylationData2 = methylationData2, 
                                            regions = regions, 
                                            context = context,
                                            windowSize = windowSize,
                                            kernelFunction = kernelFunction, 
                                            lambda = lambda,
                                            test = test, 
                                            pValueThreshold = pValueThreshold, 
                                            minCytosinesCount = minCytosinesCount, 
                                            minProportionDifference =minProportionDifference, 
                                            minGap = minGap, 
                                            minSize = minSize, 
                                            minReadsPerCytosine = minReadsPerCytosine, 
                                            cores = cores,
                                            BPPARAM = BPPARAM)
  } else if(method == "neighbourhood"){
    computedDMRs <- .computeDMRsNeighbourhood(methylationData1 = methylationData1, 
                                              methylationData2 = methylationData2, 
                                              regions = regions, 
                                              context = context, 
                                              test = test, 
                                              pValueThreshold = pValueThreshold, 
                                              minCytosinesCount = minCytosinesCount, 
                                              minProportionDifference =minProportionDifference, 
                                              minGap = minGap, 
                                              minSize = minSize, 
                                              minReadsPerCytosine = minReadsPerCytosine, 
                                              cores = cores,
                                              BPPARAM = BPPARAM)
  } else if(method == "bins"){
    computedDMRs <- .computeDMRsBins(methylationData1 = methylationData1, 
                                     methylationData2 = methylationData2, 
                                     regions = regions, 
                                     context = context, 
                                     binSize = binSize,
                                     test = test, 
                                     pValueThreshold = pValueThreshold, 
                                     minCytosinesCount = minCytosinesCount, 
                                     minProportionDifference =minProportionDifference, 
                                     minGap = minGap, 
                                     minSize = minSize, 
                                     minReadsPerCytosine = minReadsPerCytosine, 
                                     cores = cores,
                                     BPPARAM = BPPARAM)
    
  } else{
    cat("Unknown method: ",method," \n")
  } 
  
  return(computedDMRs)  
}

#' This function computes the differentially methylated regions between two conditions
#' using the noise filter method.  
.computeDMRsNoiseFilter <- function(methylationData1, 
                                    methylationData2, 
                                    regions = NULL, 
                                    context = "CG",
                                    windowSize = 100,
                                    kernelFunction="triangular", 
                                    lambda=0.5,                                    
                                    test = "fisher", 
                                    pValueThreshold = 0.01, 
                                    minCytosinesCount = 4, 
                                    minProportionDifference = 0.4, 
                                    minGap = 200, 
                                    minSize = 50, 
                                    minReadsPerCytosine = 4, 
                                    cores = 1,
                                    BPPARAM = BPPARAM) {
  
  regions <- reduce(regions)
  
  # extract the methylation data in the correct context
  cat("Extract methylation in the corresponding context \n")
  
  
  contextMethylationData1 <- methylationData1[methylationData1$context%in%context]
  rm(methylationData1)
  localContextMethylationData1 <- contextMethylationData1[queryHits(findOverlaps(contextMethylationData1, regions))]
  rm(contextMethylationData1)  
  
  contextMethylationData2 <- methylationData2[methylationData2$context%in%context]
  rm(methylationData2)
  localContextMethylationData2 <- contextMethylationData2[queryHits(findOverlaps(contextMethylationData2, regions))]
  rm(contextMethylationData2)  
  
  localContextMethylationData <- .joinMethylationData(localContextMethylationData1, localContextMethylationData2)
  rm(localContextMethylationData1, localContextMethylationData2)
  
  regionsList <- .splitGRangesEqualy(regions, cores)
  
  # inner loop function for BiocParallel::bplapply
  .computeDMRsInterpolationLoop = function(i){
    computedDMRs <- GRanges()    
    for(index in 1:length(regionsList[[i]])){
      currentRegion <- regionsList[[i]][index]
      
      cat("Computing DMRs at ",.printGenomicRanges(currentRegion),"\n")
      
      # Select the points in methylationData that we're interested in. These are the 
      # points that lie within 'regions', as well as any that lie within 
      # window.size of them. 
      windowSizeHalf <- floor((windowSize - 1)/2)
      extendedRegion <- currentRegion
      start(extendedRegion) <- start(currentRegion) - windowSizeHalf
      end(extendedRegion) <- end(currentRegion) + windowSizeHalf
      
      overlaps <- findOverlaps(localContextMethylationData, extendedRegion)
      if(length(overlaps) > 0){
        localMethylationData <- localContextMethylationData[queryHits(overlaps)]
        
        cat("Calculating interpolations...\n")
        
        #Rcpp
        movingAverageMethylReads1 <- round(.movingAverage(start(currentRegion), 
                                                          end(currentRegion), 
                                                          start(localMethylationData), 
                                                          localMethylationData$readsM1, 
                                                          windowSizeHalf = windowSizeHalf))
        movingAverageTotalReads1 <- round(.movingAverage(start(currentRegion), 
                                                         end(currentRegion), 
                                                         start(localMethylationData), 
                                                         localMethylationData$readsN1, 
                                                         windowSizeHalf = windowSizeHalf))
        movingAverageProportion1 <- movingAverageMethylReads1 / movingAverageTotalReads1
        
        #Rcpp
        movingAverageMethylReads2 <- round(.movingAverage(start(currentRegion), 
                                                          end(currentRegion), 
                                                          start(localMethylationData), 
                                                          localMethylationData$readsM2, 
                                                          windowSizeHalf = windowSizeHalf))
        movingAverageTotalReads2 <- round(.movingAverage(start(currentRegion), 
                                                         end(currentRegion), 
                                                         start(localMethylationData), 
                                                         localMethylationData$readsN2, 
                                                         windowSizeHalf = windowSizeHalf))
        movingAverageProportion2 <- movingAverageMethylReads2 / movingAverageTotalReads2
        
        cat("Identifying DMRs...\n")    
        pValue <- .computeAdjuestedPValues(test, 
                                           movingAverageMethylReads1, 
                                           movingAverageTotalReads1, 
                                           movingAverageMethylReads2, 
                                           movingAverageTotalReads2, 
                                           alternative = "two.sided")
        
        
        # compute the differentially methylated cytosines
        DMPs <- rep(0, times=width(currentRegion))
        DMPs[is.na(pValue)] <- -2
        bufferIndex <- !is.na(pValue) & 
          pValue < pValueThreshold & 
          abs(movingAverageProportion1 - movingAverageProportion2) >= minProportionDifference & 
          movingAverageTotalReads1 >=minReadsPerCytosine &
          movingAverageTotalReads2 >=minReadsPerCytosine
        DMPs[bufferIndex] <- sign(movingAverageProportion2[bufferIndex] - movingAverageProportion1[bufferIndex])
        
        #join the differentially methylated cytosines into regions
        rle <- rle(DMPs)
        rle$cumulative <- cumsum(rle$lengths)
        endOfRuns <- rle$cumulative + start(currentRegion) - 1
        
        DMRs <- GRanges(
          seqnames    = seqnames(currentRegion),
          ranges      = IRanges(endOfRuns - rle$lengths + 1, endOfRuns),
          strand      = strand(currentRegion),
          direction   = rle$values,
          context     = paste(context, collapse = "_")
        )
        
        
        
        DMRs$direction[DMRs$direction == -2] <- NA
        
        # Select the crude list of DMRs
        DMRs <- DMRs[!is.na(DMRs$direction) & (DMRs$direction == 1 | DMRs$direction == -1)]
        
        
        # append current DMRs to the global list of DMRs 
        if(length(computedDMRs) == 0){
          computedDMRs <- DMRs
        } else{
          computedDMRs <- c(computedDMRs,DMRs)
        }
        
      }
    }
    
    return(computedDMRs)
    
  }
  
  # compute the DMRs (TEST: using bplapply )
  if(cores > 1){
    cat("Compute the DMRs using ", cores, "cores\n")
    computedDMRs <- BiocParallel::bplapply(1:length(regionsList), .computeDMRsInterpolationLoop, BPPARAM = BPPARAM)
  } else {
    computedDMRs <- lapply(1:length(regionsList), .computeDMRsInterpolationLoop)
  }
  
  
  computedDMRs <- unlist(GRangesList(computedDMRs))
  
  
  if(length(computedDMRs) > 0){
    computedDMRs <- computedDMRs[order(computedDMRs)]
    cat("Analysed reads inside DMRs\n")
    overlaps <- countOverlaps(localContextMethylationData, computedDMRs)
    localContextMethylationDataDMRs <- localContextMethylationData[overlaps > 0]
    if(cores > 1){
      computedDMRsList <- S4Vectors::splitAsList(computedDMRs,  rep(1:cores, length.out=length(computedDMRs)))
      bufferComputedDMRsList <- BiocParallel::bplapply(1:length(computedDMRsList), function(i){ 
        .analyseReadsInsideRegions(localContextMethylationDataDMRs, computedDMRsList[[i]])}, 
        BPPARAM = BPPARAM)
      computedDMRs <- unlist(GRangesList(bufferComputedDMRsList))
      computedDMRs <- computedDMRs[order(computedDMRs)]
      
    } else{
      computedDMRs <- .analyseReadsInsideRegions(localContextMethylationDataDMRs, computedDMRs)
    }
    computedDMRs <- computedDMRs[computedDMRs$cytosinesCount > 0]
    
    computedDMRs$pValue <- .computeaAjustedPValuesInDMRs(test, computedDMRs ,alternative = "two.sided")
    
    computedDMRs <- computedDMRs[computedDMRs$pValue < pValueThreshold & 
                                   abs(computedDMRs$proportion1 - computedDMRs$proportion2) >= minProportionDifference & 
                                   computedDMRs$sumReadsN1/computedDMRs$cytosinesCount >=minReadsPerCytosine &
                                   computedDMRs$sumReadsN2/computedDMRs$cytosinesCount >=minReadsPerCytosine]
    
    cat("Merge DMRs iteratively\n")    
    # Get rid of small gaps between DMRs.
    #computedDMRs <- .mergeRegions(computedDMRs, minGap = minGap, respectSigns = TRUE)
    if(minGap > 0){
      computedDMRs <- .smartMergeDMRs(computedDMRs, 
                                      minGap = minGap, 
                                      respectSigns = TRUE, 
                                      methylationData = localContextMethylationData,
                                      minProportionDifference=minProportionDifference, 
                                      minReadsPerCytosine = minReadsPerCytosine, 
                                      pValueThreshold=pValueThreshold,
                                      test=test, 
                                      alternative = "two.sided",
                                      cores = cores,
                                      BPPARAM = BPPARAM)
    }  
    
    cat("Filter DMRs \n")    
    if(length(computedDMRs) > 0){
      #remove small DMRs 
      computedDMRs <- computedDMRs[width(computedDMRs) >= minSize]
      if(length(computedDMRs) > 0){
        #remove DMRswith few cytosines
        computedDMRs <- computedDMRs[computedDMRs$cytosinesCount >= minCytosinesCount]
        #recompute the adjusted p-values
        if(length(computedDMRs) > 0){
          computedDMRs$pValue <- .computeaAjustedPValuesInDMRs(test, computedDMRs ,alternative = "two.sided")
          computedDMRs$regionType <- rep("loss", length(computedDMRs))
          computedDMRs$regionType[which(computedDMRs$proportion1 < computedDMRs$proportion2)] <- "gain"
          
        }
      }
      
    }
  }  
  
  return(computedDMRs)  
}

#' This function computes the differentially methylated regions between two conditions
#' using the neighbourhood method. This assumes the computation of differentially methylated 
#' cytosines followed by smart merging of these cytosines while keeping the new DMRs 
#' statistically significant.   
.computeDMRsNeighbourhood <- function(methylationData1, 
                                      methylationData2, 
                                      regions = NULL, 
                                      context = "CG",
                                      test = "fisher", 
                                      pValueThreshold = 0.01, 
                                      minCytosinesCount = 4, 
                                      minProportionDifference = 0.4,
                                      minGap = 200, 
                                      minSize = 50, 
                                      minReadsPerCytosine = 4, 
                                      cores = 1,
                                      BPPARAM = BPPARAM) {  
  
  regions <- reduce(regions)
  
  # extract the methylation data in the correct context
  cat("Extract methylation in the corresponding context \n")
  
  
  contextMethylationData1 <- methylationData1[methylationData1$context%in%context]
  rm(methylationData1)
  localContextMethylationData1 <- contextMethylationData1[queryHits(findOverlaps(contextMethylationData1, regions))]
  rm(contextMethylationData1)  
  
  contextMethylationData2 <- methylationData2[methylationData2$context%in%context]
  rm(methylationData2)
  localContextMethylationData2 <- contextMethylationData2[queryHits(findOverlaps(contextMethylationData2, regions))]
  rm(contextMethylationData2)  
  
  localContextMethylationData <- .joinMethylationData(localContextMethylationData1, localContextMethylationData2)
  rm(localContextMethylationData1, localContextMethylationData2)
  
  cat("Computing DMRs \n")
  DMPs <- GRanges()
  if(length(localContextMethylationData) > 0){
    DMPs <- localContextMethylationData
    DMPs$pValue <- .computeAdjuestedPValues(test, 
                                            DMPs$readsM1, 
                                            DMPs$readsN1, 
                                            DMPs$readsM2, 
                                            DMPs$readsN2, 
                                            alternative = "two.sided")
    DMPs <- DMPs[!is.na(DMPs$pValue)]
    DMPs$sumReadsM1 <- DMPs$readsM1
    DMPs$sumReadsN1 <- DMPs$readsN1
    DMPs$proportion1 <- DMPs$readsM1 / DMPs$readsN1
    DMPs$sumReadsM2 <- DMPs$readsM2
    DMPs$sumReadsN2 <- DMPs$readsN2
    DMPs$proportion2 <- DMPs$readsM2 / DMPs$readsN2
    DMPs$cytosinesCount <- 1
    DMPs$direction <- sign(DMPs$proportion2 - DMPs$proportion1)
    
    bufferIndex <- DMPs$pValue < pValueThreshold & 
      abs(DMPs$proportion2 - DMPs$proportion1) >= minProportionDifference & 
      DMPs$sumReadsN1 >=minReadsPerCytosine &
      DMPs$sumReadsN2 >=minReadsPerCytosine
    DMPs <- DMPs[bufferIndex]
    strand(DMPs) <- "*" 
  }
  
  computedDMRs <- GRanges()
  if(length(DMPs) > 0){    
    cat("Merge DMRs iteratively\n")    
    # Get rid of small gaps between DMRs.
    if(minGap > 0){
      computedDMRs <- .smartMergeDMRs(DMPs, 
                                      minGap = minGap, 
                                      respectSigns = TRUE, 
                                      methylationData = localContextMethylationData,
                                      minProportionDifference=minProportionDifference, 
                                      minReadsPerCytosine = minReadsPerCytosine, 
                                      pValueThreshold=pValueThreshold,
                                      test=test, 
                                      alternative = "two.sided",
                                      cores = cores,
                                      BPPARAM = BPPARAM)
    } else{
      computedDMRs <- DMPs
    } 
    
    cat("Filter DMRs \n")    
    if(length(computedDMRs) > 0){
      #remove small DMRs 
      computedDMRs <- computedDMRs[width(computedDMRs) >= minSize]
      if(length(computedDMRs) > 0){
        #remove DMRs with few cytosines
        computedDMRs <- computedDMRs[computedDMRs$cytosinesCount >= minCytosinesCount]
        #recompute the adjusted p-values
        if(length(computedDMRs) > 0){
          computedDMRs$pValue <- .computeaAjustedPValuesInDMRs(test, computedDMRs ,alternative = "two.sided")
          computedDMRs$regionType <- rep("loss", length(computedDMRs))
          computedDMRs$regionType[which(computedDMRs$proportion1 < computedDMRs$proportion2)] <- "gain"
          
        }
      }
      
    }
  }  
  
  return(computedDMRs)  
}


#' This function computes the differentially methylated regions between two conditions
#' using the bins method.  
.computeDMRsBins <- function(methylationData1, 
                             methylationData2, 
                             regions = NULL, 
                             context = "CG", 
                             binSize = 100,
                             test = "fisher", 
                             pValueThreshold = 0.01, 
                             minCytosinesCount = 4,
                             minProportionDifference = 0.4, 
                             minGap = 200, 
                             minSize = 50, 
                             minReadsPerCytosine = 4, 
                             cores = 1,
                             BPPARAM = BPPARAM) {
  
  regions <- reduce(regions)
  
  # extract the methylation data in the correct context
  cat("Extract methylation in the corresponding context \n")
  
  contextMethylationData1 <- methylationData1[methylationData1$context%in%context]
  rm(methylationData1)
  localContextMethylationData1 <- contextMethylationData1[queryHits(findOverlaps(contextMethylationData1, regions))]
  rm(contextMethylationData1)  
  
  contextMethylationData2 <- methylationData2[methylationData2$context%in%context]
  rm(methylationData2)
  localContextMethylationData2 <- contextMethylationData2[queryHits(findOverlaps(contextMethylationData2, regions))]
  rm(contextMethylationData2)  
  
  localContextMethylationData <- .joinMethylationData(localContextMethylationData1, localContextMethylationData2)
  rm(localContextMethylationData1, localContextMethylationData2)
  
  
  regionsList <- .splitGRangesEqualy(regions, cores)
  
  # inner loop function for BiocParallel::bplapply
  .computeDMRsBinsLoop = function(i){
    computedDMRs <- GRanges()
    for(index in 1:length(regionsList[[i]])){
      currentRegion <- regionsList[[i]][index]
      
      
      cat("Computing DMRs at ",.printGenomicRanges(currentRegion),"\n")
      
      seqs <- seq(start(currentRegion), (end(currentRegion)-binSize), by = binSize);
      
      bins <- GRanges(seqnames(currentRegion), IRanges(seqs, (seqs+binSize-1)))
      
      overlapsBins <- findOverlaps(localContextMethylationData, currentRegion)
      
      if(length(overlapsBins) > 0){
        localMethylationData <- localContextMethylationData[queryHits(overlapsBins)]
        
        cat("Count inside each bin...\n")
        #bins <- .analyseReadsInsideRegions(localMethylationData, bins, context, cores)
        bins <- .analyseReadsInsideBins(localMethylationData, bins, currentRegion)
        
        cat("Filter the bins...\n")
        # Get rid of the bins with fewer than minCytosinesCount cytosines inside.  
        bins  <- bins[bins$cytosinesCount >= minCytosinesCount]
        
        # Get rid of the bins with fewer than minReadsPerCytosine reads per cytosine.  
        bins  <- bins[(bins$sumReadsN1/bins$cytosinesCount >= minReadsPerCytosine) & 
                        (bins$sumReadsN2/bins$cytosinesCount >= minReadsPerCytosine)]
        
        # Get rid of the bins with small difference in proportion of methylation 
        bins  <- bins[(abs(bins$proportion1 - bins$proportion2) >= minProportionDifference)]
        
        
        cat("Identifying DMRs...\n")    
        pValue <- .computeAdjuestedPValues(test, bins$sumReadsM1, bins$sumReadsN1, bins$sumReadsM2, bins$sumReadsN2, alternative = "two.sided")
        
        bins <- bins[!is.na(pValue) & pValue < pValueThreshold ]  
        bins$context <- rep(paste(context, collapse = "_"), length(bins))
        bins$direction <- rep(NA, length(bins))
        bins$direction <- sign(bins$proportion2 - bins$proportion1)
        
        # Select the crude list of DMRs
        DMRs <- bins[!is.na(bins$direction) & (bins$direction == 1 | bins$direction == -1)]
        
        # append current DMRs to the global list of DMRs 
        if(length(computedDMRs) == 0){
          computedDMRs <- DMRs
        } else{
          computedDMRs <- c(computedDMRs,DMRs)
        }
      }
    }
    return(computedDMRs)
  }
  
  # compute the DMRs
  if(cores > 1){
    cat("Compute the DMRs using ", cores, "cores\n")
    computedDMRs <- BiocParallel::bplapply(1:length(regionsList), .computeDMRsBinsLoop, BPPARAM = BPPARAM)
  } else {
    computedDMRs <- lapply(1:length(regionsList), .computeDMRsBinsLoop)
  }
  
  
  computedDMRs <- unlist(GRangesList(computedDMRs))
  
  if(length(computedDMRs) > 0){
    
    cat("Merge adjacent DMRs\n")    
    computedDMRs <- computedDMRs[order(computedDMRs)]
    computedDMRs$pValue <- .computeaAjustedPValuesInDMRs(test, computedDMRs ,alternative = "two.sided")
    
    cat("Merge DMRs iteratively\n")    
    # Get rid of small gaps between DMRs.
    if(minGap > 0){
      computedDMRs <- .smartMergeDMRs(computedDMRs, 
                                      minGap = minGap, 
                                      respectSigns = TRUE, 
                                      methylationData = localContextMethylationData,
                                      minProportionDifference=minProportionDifference, 
                                      minReadsPerCytosine = minReadsPerCytosine, 
                                      pValueThreshold=pValueThreshold,
                                      test=test, 
                                      alternative = "two.sided",
                                      cores = cores,
                                      BPPARAM = BPPARAM)
    }  
    
    cat("Filter DMRs \n")    
    if(length(computedDMRs) > 0){
      #remove small DMRs 
      computedDMRs <- computedDMRs[width(computedDMRs) >= minSize]
      
      if(length(computedDMRs) > 0){
        #remove DMRs with few cytosines
        computedDMRs <- computedDMRs[computedDMRs$cytosinesCount >= minCytosinesCount]
        #recompute the adjusted p-values
        if(length(computedDMRs) > 0){
          computedDMRs$pValue <- .computeaAjustedPValuesInDMRs(test, computedDMRs ,alternative = "two.sided")
          computedDMRs$regionType <- rep("loss", length(computedDMRs))
          computedDMRs$regionType[which(computedDMRs$proportion1 < computedDMRs$proportion2)] <- "gain"
        }
      }
    }
  }  
  
  return(computedDMRs)
}

#' This function verifies whether a set of pottential DMRs (e.g. genes, 
#' transposons, CpG islands) are differentially methylated or not.
#'
#' @title Filter DMRs 
#' @param methylationData1 the methylation data in condition 1
#' (see \code{\link{methylationDataList}}).
#' @param methylationData2 the methylation data in condition 2
#' (see \code{\link{methylationDataList}}).
#' @param potentialDMRs a \code{\link{GRanges}} object with potential DMRs 
#' where to compute the DMRs. This can be a a list of gene and/or transposable 
#' elements coordinates.
#' @param context the context in which the DMRs are computed (\code{"CG"}, 
#' \code{"CHG"} or \code{"CHH"}).
#' @param test the statistical test used to call DMRs (\code{"fisher"} for 
#' Fisher's exact test or \code{"score"} for Score test). 
#' @param pValueThreshold DMRs with p-values (when performing the statistical 
#' test; see \code{test}) higher or equal than \code{pValueThreshold} are 
#' discarded. Note that we adjust the p-values using the Benjamini and 
#' Hochberg's method to control the false discovery rate.
#' @param minCytosinesCount DMRs with less cytosines in the specified context 
#' than \code{minCytosinesCount} will be discarded.
#' @param minProportionDifference DMRs where the difference in methylation 
#' proportion between the two conditions is lower than 
#' \code{minProportionDifference} are discarded.
#' @param minReadsPerCytosine  DMRs with the average number of reads lower than 
#' \code{minReadsPerCytosine} are discarded. 
#' @param parallel Logical; run in parallel if \code{TRUE}.
#' @param BPPARAM A \code{BiocParallelParam} object controlling parallel execution.
#'    This value will automatically set when parallel is \code{TRUE}, also able to set as manually.
#' @param cores Integer number of workers (must not exceed BPPARAM$workers).
#'    This value will automatically set as the maximum number of system workers,
#'    also able to set as manually. 
#' @return a \code{\link{GRanges}} object with 11 metadata columns that contain 
#' the DMRs; see \code{\link{computeDMRs}}.
#' @seealso \code{\link{DMRsNoiseFilterCG}}, \code{\link{computeDMRs}}, 
#' \code{\link{analyseReadsInsideRegionsForCondition}}  
#' and \code{\link{mergeDMRsIteratively}}
#' @examples
#' # load the methylation data
#' data(methylationDataList)
#' # load the gene annotation data
#' data(GEs)
#' 
#' #select the genes
#' genes <- GEs[which(GEs$type == "gene")]
#' 
#' # the regions where to compute the DMRs
#' regions <- GRanges(seqnames = Rle("Chr3"), ranges = IRanges(1,1E5))
#' genes <- genes[overlapsAny(genes, regions)]
#' 
#' # filter genes that are differntially methylated in the two conditions
#' DMRsGenesCG <- filterDMRs(methylationDataList[["WT"]], 
#'                methylationDataList[["met1-3"]], potentialDMRs = genes, 
#'                context = "CG", test = "score", pValueThreshold = 0.01, 
#'                minCytosinesCount = 4, minProportionDifference = 0.4, 
#'                minReadsPerCytosine = 3, cores = 1)
#'
#' @author Nicolae Radu Zabet
#' @export
filterDMRs <- function(methylationData1, 
                       methylationData2, 
                       potentialDMRs, 
                       context = "CG",   
                       test = "fisher", 
                       pValueThreshold = 0.01, 
                       minCytosinesCount = 4, 
                       minProportionDifference = 0.4, 
                       minReadsPerCytosine = 3, 
                       parallel = FALSE,
                       BPPARAM = NULL,
                       cores = NULL) {
  ##Parameters checking
  cat("Parameters checking ...\n")
  
  # generate the BPPARAM value if set as parallel 
  if (parallel == TRUE){
    BPPARAM <- suppressWarnings(.validateBPPARAM(BPPARAM, progressbar = TRUE)) 
  }else{
    # Force serial execution
    BPPARAM <- BiocParallel::SerialParam(progressbar = TRUE)
  }
  # If cores argument is specified
  if (!is.null(cores)) {
    .stopIfNotAll(.isInteger(cores, positive = TRUE), 
                  "the number of cores used when computing the DMRs needs to be an integer higher or equal to 1.")
    
    # Check if user requested more cores than available
    if (cores > BPPARAM$workers) {
      warning(paste0("The number of requested cores (", cores, 
                     ") exceeds the available system cores (", BPPARAM$workers, 
                     "). Automatically setting cores to the maximum available (", 
                     BPPARAM$workers, ")."))
      cores <- BPPARAM$workers
    } else {
      message(paste0("Using user-specified core count: ", cores))
    }
    
    # Apply the final core number
    BPPARAM$workers <- cores
  } else {
    cores <- BPPARAM$workers
  }
  cat("Current parallel setting, BPPARAM: ", capture.output(BPPARAM),sep = "\n")
  
  
  .validateMethylationData(methylationData1, variableName="methylationData1")
  .validateMethylationData(methylationData2, variableName="methylationData2")
  
  regions <- union(getWholeChromosomes(methylationData1), 
                   getWholeChromosomes(methylationData2))
  
  .validateContext(context)
  
  
  .validateGRanges(potentialDMRs, generateGenomeWide=FALSE, variableName="potentialDMRs", minLength=NULL)
  
  .validateStatisticalTest(test)
  
  .stopIfNotAll(c(!is.null(pValueThreshold), is.numeric(pValueThreshold), pValueThreshold > 0, pValueThreshold < 1),
                " the p-value threshold needs to be in the interval (0,1)")
  
  .stopIfNotAll(c(.isInteger(minCytosinesCount, positive=TRUE)), 
                " the minCytosinesCount is an integer higher or equal to 0")
  
  .stopIfNotAll(c(!is.null(minProportionDifference), is.numeric(minProportionDifference), minProportionDifference > 0, minProportionDifference < 1), 
                " the minimum difference in methylation needs to be in the interval (0,1)")
  
  .stopIfNotAll(c(.isInteger(minReadsPerCytosine, positive=TRUE)), 
                " the minimum number of reads in a bin is an integer higher or equal to 0")
  
  regions <- reduce(regions)
  
  if(length(potentialDMRs) > 0){
    
    
    
    # extract the methylation data in the correct context
    cat("Extract methylation in the corresponding context \n")
    
    contextMethylationData1 <- methylationData1[methylationData1$context%in%context]
    rm(methylationData1)
    localContextMethylationData1 <- contextMethylationData1[queryHits(findOverlaps(contextMethylationData1, regions))]
    rm(contextMethylationData1)  
    
    contextMethylationData2 <- methylationData2[methylationData2$context%in%context]
    rm(methylationData2)
    localContextMethylationData2 <- contextMethylationData2[queryHits(findOverlaps(contextMethylationData2, regions))]
    rm(contextMethylationData2)  
    
    localContextMethylationData <- .joinMethylationData(localContextMethylationData1, localContextMethylationData2)
    rm(localContextMethylationData1, localContextMethylationData2)
    
    
    regionsList <- .splitGRangesEqualy(regions, cores)
    
    # inner loop function for BiocParallel::bplapply
    .filterDMRsLoop = function(i){
      computedDMRs <- GRanges()  
      for(index in 1:length(regionsList[[i]])){
        currentRegion <- regionsList[[i]][index]
        
        
        cat("Computing DMRs at ",.printGenomicRanges(currentRegion),"\n")
        
        cat("Selecting data...\n")
        
        # Select the points in methylationData that we're interested in. These are the 
        # points that lie within 'regions', as well as any that lie within 
        # window.size of them. 
        
        overlapsPotentialDMRs <- findOverlaps(potentialDMRs, currentRegion)
        if(length(overlapsPotentialDMRs) > 0){
          potentialDMRsLocal <- potentialDMRs[queryHits(overlapsPotentialDMRs)]
          
          localMethylationData <- localContextMethylationData[queryHits(findOverlaps(localContextMethylationData, currentRegion))]
          potentialDMRsLocal <- .analyseReadsInsideRegions(localMethylationData, potentialDMRsLocal)
          
          if(length(computedDMRs) == 0){
            computedDMRs <- potentialDMRsLocal
          } else{
            computedDMRs <- c(computedDMRs,potentialDMRsLocal)
          }
        } 
      }
      return(computedDMRs)
    }
    
    # compute the DMRs
    if(cores > 1){
      cat("Compute the DMRs using ", cores, "cores\n")
      computedDMRs <- BiocParallel::bplapply(1:length(regionsList), .filterDMRsLoop, BPPARAM = BPPARAM)
    } else {
      computedDMRs <- lapply(1:length(regionsList), .filterDMRsLoop)
    }
    
    
    computedDMRs <-  unlist(GRangesList(computedDMRs))
    
    if(length(computedDMRs) > 0){
      cat("Identifying DMRs...\n")    
      pValue <- .computeAdjuestedPValues(test, computedDMRs$sumReadsM1, computedDMRs$sumReadsN1, computedDMRs$sumReadsM2, computedDMRs$sumReadsN2, alternative = "two.sided")
      
      
      bufferIndex <- !is.na(pValue) & 
        pValue < pValueThreshold & 
        abs(computedDMRs$proportion1 - computedDMRs$proportion2) >= minProportionDifference & 
        computedDMRs$sumReadsN1/computedDMRs$cytosinesCount >= minReadsPerCytosine &
        computedDMRs$sumReadsN2/computedDMRs$cytosinesCount >= minReadsPerCytosine &
        computedDMRs$cytosinesCount >= minCytosinesCount
      
      computedDMRs <- computedDMRs[bufferIndex]  
    }
  } else{
    computedDMRs <- GRanges() 
  }
  
  if(length(computedDMRs) > 0){
    computedDMRs$pValue <- .computeaAjustedPValuesInDMRs(test, computedDMRs ,alternative = "two.sided")
    computedDMRs$regionType <- rep("loss", length(computedDMRs))
    computedDMRs$regionType[which(computedDMRs$proportion1 < computedDMRs$proportion2)] <- "gain"
    computedDMRs$direction <- rep(-1, length(computedDMRs))
    computedDMRs$direction[which(computedDMRs$proportion1 < computedDMRs$proportion2)] <- 1
    computedDMRs <- computedDMRs[order(computedDMRs)]
  }
  
  return(computedDMRs)
  
}
#' This function takes a list of DMRs and attempts to merge DMRs while keeping 
#' the new DMRs statistically significant.
#'
#' @title Merge DMRs iteratively
#' @param DMRs the list of DMRs as a \code{\link{GRanges}} object; 
#' e.g. see \code{\link{computeDMRs}}
#' @param minGap DMRs separated by a gap of at least \code{minGap} are not 
#' merged.
#' @param respectSigns logical value indicating whether to respect the sign when 
#' joining DMRs.
#' @param methylationData1 the methylation data in condition 1
#' (see \code{\link{methylationDataList}}).
#' @param methylationData2 the methylation data in condition 2
#' (see \code{\link{methylationDataList}}).
#' @param context the context in which the DMRs are computed (\code{"CG"}, 
#' \code{"CHG"} 
#' or \code{"CHH"}).
#' @param minProportionDifference two adjacent DMRs are merged only if the 
#' difference in methylation proportion of the new DMR is higher than 
#' \code{minProportionDifference}.
#' @param minReadsPerCytosine two adjacent DMRs are merged only if the number of 
#' reads per cytosine of the new DMR is higher than \code{minReadsPerCytosine}.
#' @param pValueThreshold two adjacent DMRs are merged only if the p-value of 
#' the new DMR (see \code{test} below) is lower than \code{pValueThreshold}. 
#' Note that we adjust the p-values using the Benjamini and Hochberg's method to 
#' control the false discovery rate.
#' @param test the statistical test used to call DMRs (\code{"fisher"} for 
#' Fisher's exact test or \code{"score"} for Score test). 
#' @param alternative indicates the alternative hypothesis and must be one of 
#' \code{"two.sided"}, \code{"greater"} or \code{"less"}.
#' @param parallel Logical; run in parallel if \code{TRUE}.
#' @param BPPARAM A \code{BiocParallelParam} object controlling parallel execution.
#'    This value will automatically set when parallel is \code{TRUE}, also able to set as manually.
#' @param cores Integer number of workers (must not exceed BPPARAM$workers).
#'    This value will automatically set as the maximum number of system workers,
#'    also able to set as manually.
#' @return the reduced list of DMRs as a \code{\link{GRanges}} object; 
#' e.g. see \code{\link{computeDMRs}}
#' @seealso \code{\link{filterDMRs}}, \code{\link{computeDMRs}}, 
#' \code{\link{analyseReadsInsideRegionsForCondition}} and 
#' \code{\link{DMRsNoiseFilterCG}}
#' @examples
#' # load the methylation data
#' data(methylationDataList)
#' 
#' #load the DMRs in CG context they were computed with minGap = 200
#' data(DMRsNoiseFilterCG)
#' 
#'
#' #merge the DMRs 
#' DMRsNoiseFilterCGLarger <- mergeDMRsIteratively(DMRsNoiseFilterCG[1:100], 
#'                            minGap = 500, respectSigns = TRUE, 
#'                            methylationDataList[["WT"]], 
#'                            methylationDataList[["met1-3"]],
#'                            context = "CG", minProportionDifference=0.4, 
#'                            minReadsPerCytosine = 1, pValueThreshold=0.01, 
#'                            test="score",alternative = "two.sided")
#' 
#' 
#' \dontrun{
#' #set genomic coordinates where to compute DMRs
#' regions <- GRanges(seqnames = Rle("Chr3"), ranges = IRanges(1,1E5))
#' 
#' # compute DMRs and remove gaps smaller than 200 bp
#' DMRsNoiseFilterCG200 <- computeDMRs(methylationDataList[["WT"]], 
#'                        methylationDataList[["met1-3"]], regions = regions, 
#'                        context = "CG", method = "noise_filter", 
#'                        windowSize = 100, kernelFunction = "triangular",  
#'                        test = "score", pValueThreshold = 0.01, 
#'                        minCytosinesCount = 1, minProportionDifference = 0.4, 
#'                        minGap = 200, minSize = 0, minReadsPerCytosine = 1, 
#'                        cores = 1)
#'                        
#' DMRsNoiseFilterCG0 <- computeDMRs(methylationDataList[["WT"]], 
#'                        methylationDataList[["met1-3"]], regions = regions, 
#'                        context = "CG", method = "noise_filter", 
#'                        windowSize = 100, kernelFunction = "triangular",  
#'                        test = "score", pValueThreshold = 0.01, 
#'                        minCytosinesCount = 1, minProportionDifference = 0.4, 
#'                        minGap = 0, minSize = 0, minReadsPerCytosine = 1, 
#'                        cores = 1)
#' DMRsNoiseFilterCG0Merged200 <- mergeDMRsIteratively(DMRsNoiseFilterCG0, 
#'                              minGap = 200, respectSigns = TRUE, 
#'                              methylationDataList[["WT"]], 
#'                              methylationDataList[["met1-3"]],
#'                              context = "CG", minProportionDifference=0.4, 
#'                              minReadsPerCytosine = 1, pValueThreshold=0.01, 
#'                              test="score",alternative = "two.sided")                      
#' 
#' #check that all newley computed DMRs are identical
#' print(all(DMRsNoiseFilterCG200 == DMRsNoiseFilterCG0Merged200))
#' 
#' }
#' 
#' @author Nicolae Radu Zabet 
#' 
#' @export
mergeDMRsIteratively <- function(DMRs, 
                                 minGap, 
                                 respectSigns = TRUE, 
                                 methylationData1,
                                 methylationData2,
                                 context = "CG",
                                 minProportionDifference=0.4, 
                                 minReadsPerCytosine = 4, 
                                 pValueThreshold=0.01,
                                 test="fisher",
                                 alternative = "two.sided",
                                 parallel = FALSE,
                                 BPPARAM = NULL,
                                 cores = NULL){
  
  ##Parameters checking
  cat("Parameters checking ...\n")
  
  # generate the BPPARAM value if set as parallel 
  if (parallel == TRUE){
    BPPARAM <- suppressWarnings(.validateBPPARAM(BPPARAM, progressbar = TRUE)) 
  }else{
    # Force serial execution
    BPPARAM <- BiocParallel::SerialParam(progressbar = TRUE)
  }
  # If cores argument is specified
  if (!is.null(cores)) {
    .stopIfNotAll(.isInteger(cores, positive = TRUE), 
                  "the number of cores used when computing the DMRs needs to be an integer higher or equal to 1.")
    
    # Check if user requested more cores than available
    if (cores > BPPARAM$workers) {
      warning(paste0("The number of requested cores (", cores, 
                     ") exceeds the available system cores (", BPPARAM$workers, 
                     "). Automatically setting cores to the maximum available (", 
                     BPPARAM$workers, ")."))
      cores <- BPPARAM$workers
    } else {
      message(paste0("Using user-specified core count: ", cores))
    }
    
    # Apply the final core number
    BPPARAM$workers <- cores
  } else {
    cores <- BPPARAM$workers
  }
  cat("Current parallel setting, BPPARAM: ", capture.output(BPPARAM),sep = "\n")
  
  .validateMethylationData(methylationData1, variableName="methylationData1")
  .validateMethylationData(methylationData2, variableName="methylationData2")
  .validateContext(context)
  .validateGRanges(DMRs, generateGenomeWide=FALSE, variableName="DMRs", minLength=NULL)
  .validateStatisticalTest(test)
  .stopIfNotAll(c(!is.null(pValueThreshold), 
                  is.numeric(pValueThreshold), 
                  pValueThreshold > 0, 
                  pValueThreshold < 1),
                " the p-value threshold needs to be in the interval (0,1)")
  .stopIfNotAll(c(!is.null(minProportionDifference), 
                  is.numeric(minProportionDifference), 
                  minProportionDifference > 0, 
                  minProportionDifference < 1), 
                " the minimum difference in methylation needs to be in the interval (0,1)")
  
  .stopIfNotAll(c(.isInteger(minGap, positive=TRUE)),
                " the minimum gap between DMRs is an integer higher or equal to 0")
  
  .stopIfNotAll(c(.isInteger(minReadsPerCytosine, positive=TRUE)), 
                " the minimum average number of reads in a DMR is an integer higher or equal to 0")
  
  totalRegion <- reduce(DMRs, drop.empty.ranges=FALSE, min.gapwidth=minGap, ignore.strand=TRUE)
  
  contextMethylationData1 <- methylationData1[methylationData1$context%in%context]
  rm(methylationData1)
  localContextMethylationData1 <- contextMethylationData1[queryHits(findOverlaps(contextMethylationData1, totalRegion))]
  rm(contextMethylationData1)  
  
  contextMethylationData2 <- methylationData2[methylationData2$context%in%context]
  rm(methylationData2)
  localContextMethylationData2 <- contextMethylationData2[queryHits(findOverlaps(contextMethylationData2, totalRegion))]
  rm(contextMethylationData2)  
  
  localContextMethylationData <- .joinMethylationData(localContextMethylationData1, localContextMethylationData2)
  rm(localContextMethylationData1, localContextMethylationData2)
  
  cat("Merge DMRs iteratively ...\n")
  
  
  return(.smartMergeDMRs(DMRs, 
                         minGap = minGap, 
                         respectSigns = respectSigns, 
                         methylationData = localContextMethylationData,
                         minProportionDifference=minProportionDifference, 
                         minReadsPerCytosine = minReadsPerCytosine, 
                         pValueThreshold=pValueThreshold,
                         test=test, 
                         alternative = alternative, 
                         cores = cores,
                         BPPARAM = BPPARAM))                                               
}


#' This function extracts from the methylation data the total number of reads, 
#' the number of methylated reads and the number of cytosines in the specific 
#' context from a region (e.g. DMRs)
#'
#' @title Analyse reads inside regions for condition
#' @param regions a \code{\link{GRanges}} object with a list of regions on the 
#' genome; e.g. could be a list of DMRs
#' @param methylationData the methylation data in one condition
#' (see \code{\link{methylationDataList}}).
#' @param context the context in which to extract the reads (\code{"CG"}, 
#' \code{"CHG"} or \code{"CHH"}).
#' @param label a string to be added to the columns to identify the condition
#' @param parallel Logical; run in parallel if \code{TRUE}.
#' @param BPPARAM A \code{BiocParallelParam} object controlling parallel execution.
#'    This value will automatically set when parallel is \code{TRUE}, also able to set as manually.
#' @param cores Integer number of workers (must not exceed BPPARAM$workers).
#'    This value will automatically set as the maximum number of system workers,
#'    also able to set as manually.
#' @return a \code{\link{GRanges}} object with additional four metadata columns
#' \describe{
#'  \item{sumReadsM}{the number of methylated reads}
#'  \item{sumReadsN}{the total number of reads} 
#'  \item{proportion}{the proportion methylated reads} 
#'  \item{cytosinesCount}{the number of cytosines in the regions} 
#' }
#' @seealso \code{\link{readONTbam}}, \code{\link{filterDMRs}}, \code{\link{computeDMRs}}, 
#' \code{\link{DMRsNoiseFilterCG}}, and \code{\link{mergeDMRsIteratively}}
#' @examples
#' 
#' # load the methylation data
#' data(methylationDataList)
#'  
#' #load the DMRs in CG context. These DMRs were computed with minGap = 200.
#' data(DMRsNoiseFilterCG)
#' 
#' #retrive the number of reads in CHH context in WT
#' DMRsNoiseFilterCGreadsCHH <- analyseReadsInsideRegionsForCondition(
#'                              DMRsNoiseFilterCG[1:10], 
#'                              methylationDataList[["WT"]], context = "CHH", 
#'                              label = "WT")
#' 
#' 
#' @author Nicolae Radu Zabet 
#' 
#' @export
analyseReadsInsideRegionsForCondition <- function(regions,
                                                  methylationData,
                                                  context,
                                                  label = "",
                                                  parallel = FALSE,
                                                  BPPARAM = NULL,
                                                  cores = NULL){
  
  ##Parameters checking
  cat("Parameters checking ...\n")
                                                    
  # generate the BPPARAM value if set as parallel 
  if (parallel == TRUE){
    BPPARAM <- suppressWarnings(.validateBPPARAM(BPPARAM, progressbar = TRUE)) 
  }else{
    # Force serial execution
    BPPARAM <- BiocParallel::SerialParam(progressbar = TRUE)
  }
  # If cores argument is specified
  if (!is.null(cores)) {
    .stopIfNotAll(.isInteger(cores, positive = TRUE), 
                  "the number of cores used when computing the DMRs needs to be an integer higher or equal to 1.")
    
    # Check if user requested more cores than available
    if (cores > BPPARAM$workers) {
      warning(paste0("The number of requested cores (", cores, 
                     ") exceeds the available system cores (", BPPARAM$workers, 
                     "). Automatically setting cores to the maximum available (", 
                     BPPARAM$workers, ")."))
      cores <- BPPARAM$workers
    } else {
      message(paste0("Using user-specified core count: ", cores))
    }
    
    # Apply the final core number
    BPPARAM$workers <- cores
  } else {
    cores <- BPPARAM$workers
  }
  cat("Current parallel setting, BPPARAM: ", capture.output(BPPARAM),sep = "\n")
  
  .validateGRanges(regions, generateGenomeWide=FALSE, variableName="regions", minLength=NULL)
  .validateMethylationData(methylationData, variableName="methylationData")
  .validateContext(context)
  
  cat("Extract methylation levels in corresponding context ...\n")
  contextMethylationData <- methylationData[methylationData$context%in%context]
  rm(methylationData)
  
  cat("Compute reads inside each region ...\n")
  if(length(regions) > 0){
    if(cores > 1){
      cat("Analyse reads in regions using ", cores, "cores\n")
      regionsList <- split(regions, rep(1:cores, length.out=length(regions)))
      .analyseReadsInsideRegionsForConditionLoop = function(i){
        regionsLocal <- .analyseReadsInsideRegionsForCondition(regionsList[[i]], 
                                                               contextMethylationData, 
                                                               label = label,
                                                               context = context)
        return(regionsLocal)
      }
      regions <- BiocParallel::bplapply(1:length(regionsList), 
                                    .analyseReadsInsideRegionsForConditionLoop, BPPARAM = BPPARAM)
      regions <- unlist(GRangesList(regions))
    } else{
      regions <- .analyseReadsInsideRegionsForCondition(regions, 
                                                        contextMethylationData, 
                                                        label = label,
                                                        context = context)
    }
    regions <- regions[order(regions)]
  }
  return(regions)
}  

library(DMRcaller)
library(RUnit)



# create synthetic data
step <- 25
numberOfCytosines <- 200
startPos <- seq(from = 1, by=step, length.out=numberOfCytosines)

syntheticMethylationData1 <- GRanges(
  seqnames              = factor(rep("Chr1", length(startPos))),
  ranges                = IRanges(startPos, startPos),
  strand                = rep("+", length(startPos)),
  context               = rep("CG", length(startPos)),
  readsM                = rep(0, length(startPos)),
  readsN                = rep(0, length(startPos)),
  trinucleotide_context = rep("CGH", length(startPos)))

syntheticMethylationData1$readsN <- sample(50:150,
                                           length(syntheticMethylationData1$readsN),
                                           replace = TRUE)
syntheticMethylationData1$readsN <- c(rep(50,50),rep(100, 100), rep(50,50))


proportion1 <- c(rep(0.1, 10), rep(0.5, 30), rep(0.2, 3), rep(0.45, 7),
                 rep(0.2, 50), rep(0.8, 20), rep(0.3, 10), rep(0.045, 20),
                 rep(0.3, 20), rep(0.7, 20), rep(0.3, 10))
syntheticMethylationData1$readsM <- round(syntheticMethylationData1$readsN*proportion1)

syntheticMethylationData2 <- syntheticMethylationData1
proportion2 <- c(rep(0.045, 130), rep(0.6, 20), rep(0.045, 50))
syntheticMethylationData2$readsM <- round(proportion2*syntheticMethylationData2$readsN)


#DMRs
vals <- rep(0, length(proportion1))
vals[(syntheticMethylationData1$readsM/syntheticMethylationData1$readsN -
        syntheticMethylationData2$readsM/syntheticMethylationData2$readsN)>0.4] <- 1
vals[(syntheticMethylationData2$readsM/syntheticMethylationData2$readsN -
        syntheticMethylationData1$readsM/syntheticMethylationData1$readsN)>0.4] <- -1



#join the differentially methylated cytosines into regions
rle <- rle(vals)
rle$cumulative <- cumsum(rle$lengths)
endOfRuns <- rle$cumulative

DMRs <- GRanges(
  seqnames    = "Chr1",
  ranges      = IRanges(((endOfRuns - rle$lengths + 1)*step), (endOfRuns*step)),
  strand      = "+",
  direction   = rle$values
)
DMRs <- DMRs[DMRs$direction!=0]


genes <- GRanges(
  seqnames    = "Chr1",
  ranges      = IRanges(c(300, 1800),c(1200,2100)),
  strand      = "+"
)



# check if compute coverage works
test_computeMethylationDataCoverage <- function() {
  breaks <- c(1,10,40,90,110)
  answer <- c(1,1,1,0.5,0)
  checkEqualsNumeric(computeMethylationDataCoverage(syntheticMethylationData1,
                                                    context="CG",
                                                    breaks=breaks),
                     answer, tolerance=1.0e-8)
  checkEqualsNumeric(computeMethylationDataCoverage(syntheticMethylationData2,
                                                    context="CG",
                                                    breaks=breaks),
                     answer, tolerance=1.0e-8)
}


# check if compute coverage works
test_computeDMRs <- function() {
  DMRs_CG_noise_filter <-  computeDMRs(syntheticMethylationData1,
                                       syntheticMethylationData2,
                                       NULL,
                                       method = "noise_filter",
                                       context = "CG",
                                       pValueThreshold = 0.01,
                                       minReadsPerCytosine = 3,
                                       minProportionDifference = 0.41,
                                       minGap = 50,
                                       minSize = 50,
                                       test = "score",
                                       windowSize=50)

  checkTrue(all(overlapsAny(DMRs_CG_noise_filter, DMRs)))
  checkTrue(all(overlapsAny(DMRs, DMRs_CG_noise_filter)))

  DMRs_CG_bins <-  computeDMRs(syntheticMethylationData1,
                               syntheticMethylationData2,
                               NULL,
                               method = "bins",
                               context = "CG",
                               binSize = 100,
                               test = "score",
                               pValueThreshold = 0.01,
                               minCytosinesCount = 3,
                               minProportionDifference = 0.41,
                               minGap = 100,
                               minReadsPerCytosine = 3,
                               cores = 1)
  checkTrue(all(overlapsAny(DMRs_CG_bins, DMRs)))
  checkTrue(all(overlapsAny(DMRs, DMRs_CG_bins)))
}

test_computeDMRs <- function() {
}


test_filterDMRs <- function() {
  DMRs_CG_genes <-  filterDMRs(syntheticMethylationData1,
                               syntheticMethylationData2,
                               genes,
                               context = "CG",
                               test = "score",
                               pValueThreshold=0.01,
                               minCytosinesCount = 1,
                               minProportionDifference = 0.41,
                               minReadsPerCytosine = 3)
  checkTrue(DMRs_CG_genes == genes[overlapsAny(genes, DMRs)])
}

test_filterDMRs <- function() {
  DMRs_CG_noise_filter <-  computeDMRs(syntheticMethylationData1,
                                       syntheticMethylationData2,
                                       NULL,
                                       method = "noise_filter",
                                       context="CG",
                                       pValueThreshold = 0.01,
                                       minReadsPerCytosine = 3,
                                       minProportionDifference = 0.41,
                                       minGap = 50,
                                       minSize = 50,
                                       test = "score",
                                       windowSize=50)

  DMRs_CG_noise_filter_filtered <- filterDMRs(syntheticMethylationData1,
                                              syntheticMethylationData2,
                                              DMRs_CG_noise_filter,
                                              context = "CG",
                                              minCytosinesCount = 10,
                                              minProportionDifference = 0.55,
                                              minReadsPerCytosine = 20,
                                              pValueThreshold = 0.01)

  checkTrue(all(overlapsAny(DMRs_CG_noise_filter_filtered, DMRs[2:4])))
  checkTrue(all(overlapsAny(DMRs[2:4], DMRs_CG_noise_filter_filtered)))
}

test_mergeDMRsIteratively <- function() {
  DMRs_CG_bins <-  computeDMRs(syntheticMethylationData1,
                               syntheticMethylationData2,
                               NULL,
                               context = "CG",
                               binSize = 100,
                               method = "bins",
                               test = "score",
                               pValueThreshold = 0.01,
                               minCytosinesCount = 3,
                               minProportionDifference = 0.4,
                               minGap = 0,
                               minReadsPerCytosine = 3,
                               cores = 1)

  DMRs_CG_bins_merged <- mergeDMRsIteratively(DMRs_CG_bins,
                                              minGap = 200,
                                              respectSigns = TRUE,
                                              syntheticMethylationData1,
                                              syntheticMethylationData2,
                                              context = "CG",
                                              minProportionDifference = 0.4,
                                              minReadsPerCytosine = 3,
                                              pValueThreshold = 0.01,
                                              test = "score",
                                              alternative = "two.sided")
  checkTrue(length(DMRs_CG_bins_merged) < length(DMRs_CG_bins))
  checkTrue(all(overlapsAny(DMRs_CG_bins_merged, DMRs_CG_bins)))
  checkTrue(all(overlapsAny(DMRs_CG_bins, DMRs_CG_bins_merged)))
}

# testing replicates
data("methylationDataList")

# testing joinReplicates
test_joinReplicates <- function(){
  methylationDataReplicates <- joinReplicates(methylationData1 = methylationDataList$WT,
                 methylationData2 = methylationDataList$`met1-3`,
                 usecomplete = FALSE)
  checkTrue(length(methylationDataList$WT) == length(methylationDataReplicates))
  checkTrue(length(methylationDataList$`met1-3`) == length(methylationDataReplicates))
  # checks that the number of reads columns is even
  checkTrue(length(grep("reads", colnames(mcols(methylationDataReplicates)))) %% 2 == 0)
}

# testing computeDMRsReplicates
data("syntheticDataReplicates")

condition <- c("a", "a", "b", "b")

test_computeDMRsReplicates <- function(){
  DMRsReplicatesNeighbourhood <- computeDMRsReplicates(methylationData = syntheticDataReplicates,   ### CHANGE; methylationData -> syntheticDataReplicates
                                                       condition = condition,
                                                       regions = NULL,
                                                       context = "CHH",
                                                       method = "neighbourhood",
                                                       test = "betareg",
                                                       pseudocountM = 1,
                                                       pseudocountN = 2,
                                                       pValueThreshold = 0.01,
                                                       minCytosinesCount = 4,
                                                       minProportionDifference = 0.4,
                                                       minGap = 200,
                                                       minSize = 50,
                                                       minReadsPerCytosine = 4,
                                                       cores = 1)
  checkTrue(DMRsReplicatesNeighbourhood$pValue < 0.01)
  checkTrue(DMRsReplicatesNeighbourhood$proportion2 -
            DMRsReplicatesNeighbourhood$proportion1 < 0.4)
  checkTrue(DMRsReplicatesNeighbourhood$context == "CHH")
  checkTrue(overlapsAny(ranges(DMRsReplicatesNeighbourhood),
                        ranges(syntheticDataReplicates)))

}

data(ontSampleGRangesList)

test_readONTbam_chr1_sample <- function() {
  if (!requireNamespace("BSgenome.Hsapiens.UCSC.hg38", quietly = TRUE)) {
    message("BSgenome.Hsapiens.UCSC.hg38 not installed; skipping test")
    return(TRUE)
  }
  if (!requireNamespace("GenomeInfoDb", quietly = TRUE)) {
    message("GenomeInfoDb not installed; skipping test")
    return(TRUE)
  }
  library(BSgenome.Hsapiens.UCSC.hg38)
  ## 1) Locate the example BAM and its index in extdata/
  bamfile <- system.file("extdata", "scanBamChr1Random5.bam",
                         package = "DMRcaller")
  stopifnot(file.exists(bamfile),
            file.exists(paste0(bamfile, ".bai")))


  ## 2) Call readONTbam with ref_gr = NULL so it builds its own
  ##    cytosine reference for CG sites on chr1
  gr <- readONTbam(
    bamfile = bamfile,
    ref_gr  = NULL,
    context = "CG",
    chr     = "chr1"
  )

  ## test1. Sanity check: total coverage should be at least as large
  ##    as the count of modified reads at every site
  checkTrue(
    all(gr$readsN >= gr$readsM)
  )

  ## test2. selectCytosine function check (reference GRanges)
  checkTrue(unique(seqnames(gr))=="chr1")
  checkTrue(unique(gr$context)=="CG")
}

#' Compute Co-Methylation Positions within Regions (CMPs)
#'
#' @title Compute pairwise co-methylation statistics for cytosine sites within regions
#' @description
#' \code{computeCoMethylatedPositions()} calculates pairwise co-methylation between all cytosine sites
#' within each given region, using ONT methylation calls annotated to each site.
#' For each pair of cytosines within the same strand and PMD, it builds a 2x2 contingency table
#' reflecting the overlap state of reads (both methylated, only one methylated, or neither),
#' performs a statistical test (Fisher's exact by default), and reports FDR-adjusted p-values.
#'
#' @param methylationData A \code{GRanges} object containing cytosine sites, annotated with
#'   per-site ONT methylation calls (columns \code{ONT_Cm}, \code{ONT_C}, \code{readsN}, etc).
#' @param regions A \code{GRanges} object with list including genomic context such as gene and/or transposable
#' elements coordinates which possibly have DMRs, VMRs or PMDs.
#' @param minDistance Minimum distance (in bp) between two cytosines to consider for co-methylation (default: 150).
#' @param maxDistance Maximum distance (in bp) between two cytosines to consider (default: 1000).
#' @param minCoverage Minimum read coverage required for both cytosines in a pair (default: 4).
#' @param pValueThreshold FDR-adjusted p-value threshold for reporting significant co-methylation (default: 0.01).
#' @param test Statistical test to use for co-methylation (\code{"fisher"}
#' for Fisher's exact [default], or \code{"permutation"} for chi-squared).
#' NOTE: highly recommended to do parallel when use permutation test.
#' @param alternative indicates the alternative hypothesis and must be one of
#' \code{"two.sided"}, \code{"greater"} or \code{"less"}. You can specify just
#' the initial letter. Only used in the 2 by 2 case. This is used only for
#' Fisher's test.
#' @param parallel Logical; run in parallel if \code{TRUE}.
#' @param BPPARAM A \code{BiocParallelParam} object controlling parallel execution.
#' This value will automatically set when parallel is \code{TRUE}, also able to set as manually.
#' @param cores Integer number of workers (must not exceed BPPARAM$workers).
#'    This value will automatically set as the maximum number of system workers,
#'    also able to set as manually.
#' @return A list of length equal to \code{regions}, where each entry is a \code{GInteractions} object
#'   of significant cytosine pairs (by strand), annotated with:
#'   \describe{
#'     \item{C1_C2}{number of reads methylated at both cytosines}
#'     \item{C1_only}{number methylated at only first cytosine}
#'     \item{C2_only}{number methylated at only second cytosine}
#'     \item{neither}{number methylated at neither cytosines}
#'     \item{strand}{The DNA strand ("+" or "-") on which the two CpGs reside.}
#'     \item{genomic_position}{The original region (from \code{regions}) containing
#'     this cytosines pair, formatted in UCSC or IGV style, e.g. \code{"chr1:1522971-1523970"}.}
#'     \item{p.value}{FDR-adjusted p-value for co-methylation association}
#'   }
#'
#' @details
#' Pairwise tests are performed separately for each strand (+ and -) within each region.
#' FDR correction is performed for all pairs within each region and strand.
#'
#' @seealso \code{\link{readONTbam}}, \code{\link{computePMDs}},
#' \code{\link{ontSampleGRangesList}}
#'
#' @examples
#' \dontrun{
#' # load the ONT methylation data and PMD data
#' data(ont_gr_GM18870_chr1_PMD_bins_1k)
#' data(ont_gr_GM18870_chr1_sorted_bins_1k)
#'
#' # compute the co-methylations with Fisher's exact test
#' coMetylationFisher <- computeCoMethylatedPositions(
#'   ont_gr_GM18870_chr1_sorted_bins_1k,
#'   regions = ont_gr_GM18870_chr1_PMD_bins_1k[1:4],
#'   minDistance = 150,
#'   maxDistance = 1000,
#'   minCoverage = 4,
#'   pValueThreshold = 0.01,
#'   test = "fisher",
#'   parallel = FALSE)
#'
#' # compute the co-methylations with Permuation test
#' coMetylationPermutation <- computeCoMethylatedPositions(
#'   ont_gr_GM18870_chr1_sorted_bins_1k,
#'   regions = ont_gr_GM18870_chr1_PMD_bins_1k[1:4],
#'   minDistance = 150,
#'   maxDistance = 1000,
#'   minCoverage = 4,
#'   pValueThreshold = 0.01,
#'   test = "permutation",
#'   parallel = FALSE) # highly recommended to set as TRUE
#' }
#'
#' @author Nicolae Radu Zabet and Young Jun Kim
#' @import GenomicRanges
#' @import GenomicAlignments
#' @import InteractionSet
#' @import BiocParallel
#' @export
computeCoMethylatedPositions <- function(methylationData,
                                          regions,
                                          minDistance = 150,
                                          maxDistance = 1000,
                                          minCoverage = 4,
                                          pValueThreshold = 0.01,
                                          alternative = "two.sided",
                                          test = "fisher",
                                          parallel = FALSE,
                                          BPPARAM  = NULL,
                                          cores = NULL){
  ##Parameters checking
  cat("Parameters checking ...\n")

  # generate the BPPARAM value if set as parallel 
  if (parallel == TRUE){
    BPPARAM <- suppressWarnings(.validateBPPARAM(BPPARAM, progressbar = TRUE)) 
  }else{
    # Force serial execution
    BPPARAM <- BiocParallel::SerialParam(progressbar = TRUE)
  }
  # If cores argument is specified
  if (!is.null(cores)) {
    .stopIfNotAll(.isInteger(cores, positive = TRUE), 
                  "the number of cores used when computing the DMRs needs to be an integer higher or equal to 1.")
    
    # Check if user requested more cores than available
    if (cores > BPPARAM$workers) {
      warning(paste0("The number of requested cores (", cores, 
                     ") exceeds the available system cores (", BPPARAM$workers, 
                     "). Automatically setting cores to the maximum available (", 
                     BPPARAM$workers, ")."))
      cores <- BPPARAM$workers
    } else {
      message(paste0("Using user-specified core count: ", cores))
    }
    
    # Apply the final core number
    BPPARAM$workers <- cores
  } else {
    cores <- BPPARAM$workers
  }
  cat("Current parallel setting, BPPARAM: ", capture.output(BPPARAM),sep = "\n")

  .validateMethylationData(methylationData, variableName="methylationData")
  regions <- .validateGRanges(regions, methylationData)

  .validateCoMethylationStatTest(test)

  .validateCoMethylationAlternative(alternative)

  .stopIfNotAll(c(!is.null(minDistance), is.numeric(minDistance), minDistance >= 0),
                " minDistance needs to be a numeric value higher or equal to 0")

  .stopIfNotAll(c(!is.null(maxDistance), is.numeric(maxDistance), maxDistance >= 0),
                " maxDistance needs to be a numeric value higher or equal to 0")

  .stopIfNotAll(c(maxDistance >= minDistance),
                " maxMethylation should be higher than minMethylation value")

  .stopIfNotAll(c(.isInteger(minCoverage, positive=TRUE)),
                " the minimum gap between PMDs is an integer higher or equal to 0")

  .stopIfNotAll(c(!is.null(pValueThreshold),
                  is.numeric(pValueThreshold),
                  pValueThreshold > 0,
                  pValueThreshold < 1),
                " the p-value threshold needs to be in the interval (0,1)")

  results_gil =  vector(mode = "list", length = length(regions))
  regions <- .generateGRangesName(regions)
  t_start <- Sys.time()
  n_total <- length(regions)
  for (i in seq_along(regions)) {
    elapsed <- difftime(Sys.time(), t_start, units = "secs")
    cat(sprintf("[computeCoMethylatedPositions] PMD %d/%d (%.1f%%) at %s | Elapsed: %.1f sec\n",
                i, n_total, 100*i/n_total, format(Sys.time(), "%H:%M:%S"), as.numeric(elapsed)))
    bin <- regions[i]
    in_bin <- methylationData[queryHits(findOverlaps(methylationData, bin))]
    if (length(in_bin) < 2) next
    all_strands <- unique(as.character(strand(in_bin)))
    bin_results <- list()
    for (s in all_strands) {
      in_bin_strand <- in_bin[strand(in_bin) == s]
      bin_results[[s]] <- .processStrandPairs(
        in_bin_strand,
        minDistance = minDistance,
        maxDistance = maxDistance,
        minCoverage = minCoverage,
        pValueThreshold = pValueThreshold,
        test = test,
        alternative = alternative
      )
    }
    merged <- do.call(c, bin_results)
    results_gil[[i]] <- merged
  }
  names(results_gil) <- regions$genomic_position

  # keep only those bins where we got a valid GInteractions back
  keep <- vapply(results_gil,
                 function(x) !is.null(x) & length(x) > 0L,
                 logical(1))
  results_gil <- results_gil[keep]

  results_gi <- do.call(c, Map(function(strand_list, region_name) {
    # 1) pull out the two strand‐specific GInteractions
    gi_pos <- strand_list[["+"]]
    gi_neg <- strand_list[["-"]]
    # guard against NULL
    if (is.null(gi_pos)) gi_pos <- GInteractions()
    if (is.null(gi_neg)) gi_neg <- GInteractions()

    mcols(gi_pos)$strand <- rep("+",length(gi_pos))
    mcols(gi_neg)$strand <- rep("-",length(gi_neg))
    # 2) concatenate them
    gi <- c(gi_pos, gi_neg)
    # 3) add a column ‘genomic_position’ with the region’s name
    mcols(gi)$genomic_position <- rep(region_name, length(gi))
    gi
  }, results_gil, names(results_gil)))

  result <- Reduce(c, results_gi)

  # Adjust the p.value using the FDR method
  mcols(result)$p.value <- stats::p.adjust(result$pre_p.value, method = 'fdr')
  mcols(result)$pre_p.value <- NULL
  keep_final <- result$p.value < pValueThreshold
  result <- result[which(result$p.value < pValueThreshold & !is.na(result$p.value))]


  total_elapsed <- difftime(Sys.time(), t_start, units = "secs")
  cat(sprintf("[computeCoMethylatedPositions] Done! Total elapsed time: %.1f sec\n", as.numeric(total_elapsed)))
  return(result)
}

# --------------
#  Helper: compute the co-methylation between the cpg in regions
# --------------
# 1. Calculate contingency table for two CpGs
.getCoMethylationContingencyTable <- function(c1, c2) {
  split_ids <- function(x) {
    v <- as.character(unlist(x))
    if (length(v) == 0L) return(character(0))
    v <- v[!is.na(v) & nzchar(v)]
    if (length(v) == 0L) return(character(0))
    out <- unlist(strsplit(v, ",", fixed = TRUE), use.names = FALSE)
    unique(out[nzchar(out)])
  }

  # Per-CpG sets: methylated (M) and unmethylated (U) read IDs
  M1 <- split_ids(c1$ONT_Cm)
  U1 <- split_ids(c1$ONT_C)
  M2 <- split_ids(c2$ONT_Cm)
  U2 <- split_ids(c2$ONT_C)

  R <- intersect(union(M1, U1), union(M2, U2))

  C1_C2   <- length(intersect(R, intersect(M1, M2)))
  C1_only <- length(intersect(R, intersect(M1, U2)))
  C2_only <- length(intersect(R, intersect(U1, M2)))
  neither <- length(intersect(R, intersect(U1, U2)))

  list(C1_C2 = C1_C2, C1_only = C1_only, C2_only = C2_only, neither = neither)

}

.getMethylationStatusVectors <- function(c1, c2) {
  reads1_meth <- unlist(strsplit(unlist(c1$ONT_Cm), split = ","))
  reads1_unmeth <- unlist(strsplit(unlist(c1$ONT_C), split = ","))
  reads2_meth <- unlist(strsplit(unlist(c2$ONT_Cm), split = ","))
  reads2_unmeth <- unlist(strsplit(unlist(c2$ONT_C), split = ","))
  all_reads <- unique(c(reads1_meth, reads1_unmeth, reads2_meth, reads2_unmeth))
  status1 <- ifelse(all_reads %in% reads1_meth, 1,
                    ifelse(all_reads %in% reads1_unmeth, 0, NA))
  status2 <- ifelse(all_reads %in% reads2_meth, 1,
                    ifelse(all_reads %in% reads2_unmeth, 0, NA))
  keep <- !is.na(status1) & !is.na(status2)
  status1_filt <- status1[keep]
  status2_filt <- status2[keep]
  return(list(status1 = status1_filt, status2 = status2_filt))
}

# 2. Statistical test block
.coMethylationFisherTest <- function(c1_c2, c1_only, c2_only,
                                     neither, alternative= "two.sided"){
    table <- matrix(
      c(c1_c2, c1_only, c2_only, neither),
      nrow = 2, byrow = TRUE,
      dimnames = list(c("c1_yes", "c1_no"), c("c2_yes", "c2_no"))
    )
    return(stats::fisher.test(table, alternative = alternative)$p.value)
}

### score test need to change to other methods as detecting the discordant methylation
.coMethylationScoreTest <- function(c1_c2, c1_only, c2_only, neither){
    n_total <- c1_c2 + c1_only + c2_only + neither
    m1 <- c1_c2
    n1 <- c1_c2 + c1_only
    m2 <- c1_c2
    n2 <- c1_c2 + c2_only
    return(.scoreTest(m1, n1, m2, n2))
}

.coMethylationBinomTest <- function(c1_c2, c1_only, c2_only, neither, alternative= "two.sided"){
  n_total <- c1_c2 + c1_only + c2_only + neither
  concordance <- c1_c2 + neither
  return(stats::binom.test(concordance,n_total,p = 0.5, alternative = alternative)$p.value)
}

.coMethylationPermutationTest <- function(
    status1, status2,
    nperm = 1000,
    statfun = function(tab) stats::chisq.test(tab)$statistic){
  # Create the observed contingency table from the two status vectors
  obs_tab <- table(status1, status2)
  if (!all(dim(obs_tab) == c(2,2))) {
    return(NA_real_)
  }
  obs_stat <- suppressWarnings(statfun(obs_tab))

  # Permute status2 nperm times and recalculate the test statistic for each permutation
  perm_stats <- replicate(nperm, {
    perm_status2 <- sample(status2)
    perm_tab <- table(status1, perm_status2)
    if (!all(dim(perm_tab) == c(2,2))) return(0)
    suppressWarnings(statfun(perm_tab))
  })
  perm_stats <- perm_stats[!is.na(perm_stats)]
  pval <- (sum(perm_stats >= obs_stat) + 1) / (length(perm_stats) + 1)
  return(pval)
}

# 3. Process pairs in one strand; only output adjusted p-values in the results
.processStrandPairs <- function(in_bin_strand,
                                minDistance,
                                maxDistance,
                                minCoverage,
                                pValueThreshold,
                                test = "fisher",
                                alternative = "two.sided",
                                nperm = 1000,
                                BPPARAM = BiocParallel::bpparam()) {
  if (length(in_bin_strand) < 2) return(GInteractions())
  cpg_pos <- start(in_bin_strand)
  coverage <- in_bin_strand$readsN
  pair_idx <- combn(seq_along(in_bin_strand), 2)
  dist_bp <- abs(cpg_pos[pair_idx[1,]] - cpg_pos[pair_idx[2,]])
  keep <- which(
    dist_bp >= minDistance & dist_bp <= maxDistance &
      coverage[pair_idx[1,]] >= minCoverage & coverage[pair_idx[2,]] >= minCoverage
  )
  if (length(keep) == 0) return(GInteractions())

  idx1 <- pair_idx[1, keep]
  idx2 <- pair_idx[2, keep]
  n_pairs <- length(idx1)
  results <- GInteractions(
    granges(in_bin_strand[idx1]),
    granges(in_bin_strand[idx2])
  )

  # Updated meta columns
  C1_C2 <- integer(n_pairs)
  C1_only <- integer(n_pairs)
  C2_only <- integer(n_pairs)
  neither <- integer(n_pairs)
  pvals <- numeric(n_pairs)


  # Parallel computation of contingency table statistics for all CpG pairs
  pair_stats <- bplapply(seq_len(n_pairs), function(i) {
    cont <- .getCoMethylationContingencyTable(in_bin_strand[idx1[i]], in_bin_strand[idx2[i]])
    if (test == "permutation") {
      metStatus <- .getMethylationStatusVectors(in_bin_strand[idx1[i]], in_bin_strand[idx2[i]])
      pval <- .coMethylationPermutationTest(metStatus$status1, metStatus$status2, nperm = nperm)
    } else if (test == "fisher"){
      pval <- .coMethylationFisherTest(cont$C1_C2, cont$C1_only, cont$C2_only, cont$neither, alternative)
    } else if (test == "score"){
      pval <- .coMethylationScoreTest(cont$C1_C2, cont$C1_only, cont$C2_only, cont$neither)
    } else if (test == "binom"){
      pval <- .coMethylationBinomTest(cont$C1_C2, cont$C1_only, cont$C2_only, cont$neither, alternative)
    }
    c(C1_C2 = cont$C1_C2, C1_only = cont$C1_only, C2_only = cont$C2_only, neither = cont$neither, pval = pval)
  }, BPPARAM = BPPARAM)

  pair_stats_mat <- do.call(rbind, pair_stats)
  if (is.null(dim(pair_stats_mat))) pair_stats_mat <- t(as.matrix(pair_stats_mat))

  # Extract each statistic as a numeric vector for all pairs
  C1_C2    <- pair_stats_mat[, "C1_C2"]
  C1_only  <- pair_stats_mat[, "C1_only"]
  C2_only  <- pair_stats_mat[, "C2_only"]
  neither  <- pair_stats_mat[, "neither"]
  pvals    <- pair_stats_mat[, "pval"]

  # Subset the results and assign metadata columns to GInteractions object
  results$pre_p.value  <- pvals
  results$C1_C2    <- C1_C2
  results$C1_only  <- C1_only
  results$C2_only  <- C2_only
  results$neither  <- neither

  return(results)
}

### generate the genomic_position formatted by the UCSC or IGV style
.generateGRangesName <- function(regions) {
  mcols(regions)$genomic_position <- paste0(seqnames(regions), ":",start(regions), "-", end(regions))
  return(regions)
}

#' This function computes the low resolution profiles for the bisulfite
#' sequencing data.
#'
#' @title Compute methylation profile
#' @param methylationData the methylation data stored as a \code{\link{GRanges}}
#' object with four metadata columns (see \code{\link{methylationDataList}}).
#' @param region a \code{\link{GRanges}} object with the regions where to
#' compute the DMRs.
#' @param windowSize a \code{numeric} value indicating the size of the window in
#' which methylation is averaged.
#' @param context the context in which the DMRs are computed (\code{"CG"},
#' \code{"CHG"} or \code{"CHH"}).
#' @return a \code{\link{GRanges}} object with equal sized tiles of the
#' \code{region}. The object consists of the following metadata
#' \describe{
#'  \item{sumReadsM}{the number of methylated reads.}
#'  \item{sumReadsN}{the total number of reads.}
#'  \item{Proportion}{the proportion of methylated reads.}
#'  \item{cytosinesCount}{the number of cytosines in the regions}
#'  \item{context}{the context (\code{"CG"}, \code{"CHG"} or \code{"CHH"}).}
#' }
#' @seealso \code{\link{plotMethylationProfileFromData}},
#' \code{\link{plotMethylationProfile}}, \code{\link{methylationDataList}}
#' @examples
#'
#' # load the methylation data
#' data(methylationDataList)
#'
#' # the region where to compute the profile
#' region <- GRanges(seqnames = Rle("Chr3"), ranges = IRanges(1,1E6))
#'
#' # compute low resolution profile in 20 Kb windows
#' lowResProfileWTCHH <- computeMethylationProfile(methylationDataList[["WT"]],
#'                      region, windowSize = 20000, context = "CHH")
#'
#' \dontrun{
#' # compute low resolution profile in 10 Kb windows
#' lowResProfileWTCG <- computeMethylationProfile(methylationDataList[["WT"]],
#'                      region, windowSize = 10000, context = "CG")
#'
#' lowResProfileMet13CG <- computeMethylationProfile(
#'                      methylationDataList[["met1-3"]],  region,
#'                      windowSize = 10000, context = "CG")
#' }
#'
#' @author Nicolae Radu Zabet and Jonathan Michael Foonlan Tsang
#' @export
computeMethylationProfile <- function(methylationData,
                                      region,
                                      windowSize = floor(width(region)/500),
                                      context = "CG") {
  .validateMethylationData(methylationData)

  .validateGRanges(region, methylationData, length=1, generateGenomeWide=FALSE)

  .validateContext(context)

  .stopIfNotAll(c(.isInteger(windowSize, positive=TRUE)),
                " the window size used to compute the methylation profile is an integer higher or equal to 0")


  seqname <- seqnames(region)
  minPos <- start(region)
  maxPos <- end(region)

  hits <- findOverlaps(methylationData, region)



  localMethylationData <- methylationData[queryHits(hits)];
  rm(methylationData)
  contextMethylationData <- localMethylationData[localMethylationData$context%in%context];
  rm(localMethylationData)

  cat("Calculating methylation profile for ", .printGenomicRanges(region), " using a window of ", windowSize, " bp \n")
  seqs = seq(minPos, maxPos - windowSize, windowSize);

  ranges <- GRanges(seqname, IRanges(seqs, seqs+windowSize-1))

  #for windowSize <= 2000 Bins method is faster, but for anything else, regions method is faster
  if(windowSize <= 2000){
    ranges <- .analyseReadsInsideBinsOneSample(contextMethylationData, ranges, region)
  } else{
    ranges <- .analyseReadsInsideRegionsOneSample(contextMethylationData, ranges)
  }

  ranges$context <- paste(context, collapse = "_")

  return(ranges)
}


#' This function plots the low resolution profiles for the bisulfite sequencing
#' data.
#'
#' @title Plot Methylation Profile
#' @param methylationProfiles a \code{GRangesList} object. Each
#' \code{\link{GRanges}} object in the list is generated by calling  the
#' function \code{\link{computeMethylationProfile}}.
#' @param autoscale a \code{logical} value indicating whether the values are
#' autoscalled for each context or not.
#' @param labels a \code{vector} of \code{character} used to add a subfigure
#' characters to the plot. If \code{NULL} nothing is added.
#' @param title the plot title.
#' @param col a \code{character} vector with the colours. It needs to contain a
#' minimum of 2 colours per context. If not or if NULL, the defalut colours will
#' be used.
#' @param pch the R symbols used to plot the data.
#' @param lty the line types used to plot the data.
#' @return Invisibly returns \code{NULL}
#' @seealso \code{\link{plotMethylationProfileFromData}},
#' \code{\link{computeMethylationProfile}} and \code{\link{methylationDataList}}
#' @examples
#'
#'
#' # load the methylation data
#' data(methylationDataList)
#'
#' # the region where to compute the profile
#' region <- GRanges(seqnames = Rle("Chr3"), ranges = IRanges(1,1E6))
#'
#' # compute low resolution profile in 20 Kb windows
#' lowResProfileWTCG <- computeMethylationProfile(methylationDataList[["WT"]],
#'                      region, windowSize = 20000, context = "CG")
#'
#' lowResProfilsCG <- GRangesList("WT" = lowResProfileWTCG)
#'
#' #plot the low resolution profile
#' par(mar=c(4, 4, 3, 1)+0.1)
#' par(mfrow=c(1,1))
#' plotMethylationProfile(lowResProfilsCG, autoscale = FALSE,
#'                        title="CG methylation on Chromosome 3",
#'                        col=c("#D55E00","#E69F00"),  pch = c(1,0),
#'                        lty = c(4,1))
#'
#' \dontrun{
#' # compute low resolution profile in 10 Kb windows in CG context
#' lowResProfileWTCG <- computeMethylationProfile(methylationDataList[["WT"]],
#'                      region, windowSize = 10000, context = "CG")
#'
#' lowResProfileMet13CG <- computeMethylationProfile(
#'                      methylationDataList[["met1-3"]], region,
#'                      windowSize = 10000, context = "CG")
#'
#' lowResProfileCG <- GRangesList("WT" = lowResProfileWTCG,
#'                    "met1-3" = lowResProfileMet13CG)
#'
#' # compute low resolution profile in 10 Kb windows in CHG context
#' lowResProfileWTCHG <- computeMethylationProfile(methylationDataList[["WT"]],
#'                      region, windowSize = 10000, context = "CHG")
#'
#' lowResProfileMet13CHG <- computeMethylationProfile(
#'                      methylationDataList[["met1-3"]], region,
#'                      windowSize = 10000, context = "CHG")
#'
#' lowResProfileCHG <- GRangesList("WT" = lowResProfileWTCHG,
#'                    "met1-3" = lowResProfileMet13CHG)
#'
#' # plot the low resolution profile
#' par(mar=c(4, 4, 3, 1)+0.1)
#' par(mfrow=c(2,1))
#' plotMethylationProfile(lowResProfileCG, autoscale = FALSE,
#'                        labels = LETTERS[1],
#'                        title="CG methylation on Chromosome 3",
#'                        col=c("#D55E00","#E69F00"),  pch = c(1,0),
#'                        lty = c(4,1))
#' plotMethylationProfile(lowResProfileCHG, autoscale = FALSE,
#'                        labels = LETTERS[2],
#'                        title="CHG methylation on Chromosome 3",
#'                        col=c("#0072B2", "#56B4E9"),  pch = c(16,2),
#'                        lty = c(3,2))
#' }
#'
#' @author Nicolae Radu Zabet
#' @export
plotMethylationProfile <- function(methylationProfiles,
                                   autoscale = FALSE,
                                   labels=NULL,
                                   title = "",
                                   col = NULL,
                                   pch = c(1,0,16,2,15,17),
                                   lty = c(4,1,3,2,6,5)) {
  .validateMethylationProfile(methylationProfiles)



  defaultConditionsNames <- paste("condition ", (1:length(methylationProfiles)), sep="")
  if(is.null(names(methylationProfiles))){
    names(methylationProfiles) <- defaultConditionsNames
  }

  if(!.isColor(col, minLength = length(methylationProfiles) )){
    if(length(methylationProfiles) <= 6){
      col <- c("#D55E00","#E69F00", "#0072B2", "#56B4E9", "#F0E442", "#009E73")
    } else if(length(methylationProfiles) <= 8){
      col <- c("#000000", "#E69F00", "#56B4E9", "#009E73", "#F0E442", "#0072B2", "#D55E00", "#CC79A7")
    } else{
      col <- rainbow(length(methylationProfiles))
    }
  }

  .stopIfNotAll(c(!is.null(autoscale),
                  is.logical(autoscale),
                  length(autoscale) == 1),
                " autoscale is a logical value")
  if(is.null(title)){
    title=""
  } else if(is.character(title[1]) > 1){
    title[1] = as.character(title[1])
  }

  .stopIfNotAll(c(!is.null(pch),
                  all(as.integer(pch) == pch),
                  all(pch >= 0),
                  length(pch) >= length(methylationProfiles)),
                paste(" pch is a vector of positive integers of size ",length(methylationProfiles), sep=""))

  .stopIfNotAll(c(!is.null(lty),
                  all(as.integer(lty) == lty),
                  all(lty >= 0),
                  length(lty) >= length(methylationProfiles)),
                paste(" lty is a vector of positive integers of size ",length(methylationProfiles), sep=""))


  ymax <- 1
  if(autoscale){
    ymax <- 0
    for(i in 1:length(methylationProfiles)){
      if(ymax < max(methylationProfiles[[i]]$Proportion)){
        ymax = max(methylationProfiles[[i]]$Proportion)
      }
    }
  }

  options(scipen=10)
  pos <- (start(methylationProfiles[[1]])+end(methylationProfiles[[1]]))/2
  plot(pos, methylationProfiles[[1]]$Proportion, type = "o", ylim = c(0,ymax),
       xlab="genomic coordinate", ylab="methylation", col=col[1], yaxt="n",
       pch = pch[1], lty = lty[1], main = title[1])
  if(length(methylationProfiles) > 1){
    for(i in 2:length(methylationProfiles)){
      lines(pos, methylationProfiles[[i]]$Proportion, type = "o", col=col[i], lty = lty[i], pch = pch[i])
    }
  }

  legend("topright", legend = names(methylationProfiles),
         lty = lty[1:length(methylationProfiles)],
         col = col[1:length(methylationProfiles)],
         pch = pch[1:length(methylationProfiles)], bty="n")
  if(!is.null(labels)){
    mtext(labels[1], line = 0.7, adj = 0, cex=1.4);
  }
  if(autoscale){
    axis(2,c(0,signif(ymax/2, 1),signif(ymax, 1)));
  } else{
    axis(2,c(0,.5,1));
  }
  invisible(NULL)
}

#' This function plots the low resolution profiles for all bisulfite sequencing
#' data.
#'
#' @title Plot methylation profile from data
#' @param methylationData1 the methylation data in condition 1
#' (see \code{\link{methylationDataList}}).
#' @param methylationData2 the methylation data in condition 2
#' (see \code{\link{methylationDataList}}). This is optional.
#' @param regions a \code{\link{GRanges}} object with the regions where to plot
#' the profiles.
#' @param conditionsNames the names of the two conditions. This will be used to
#' plot the legend.
#' @param context a \code{vector} with all contexts in which the DMRs are
#' computed (\code{"CG"}, \code{"CHG"} or \code{"CHH"}).
#' @param windowSize a \code{numeric} value indicating the size of the window in
#' which methylation is averaged.
#' @param autoscale a \code{logical} value indicating whether the values are
#' autoscalled for each context or not.
#' @param labels a \code{vector} of \code{character} used to add a subfigure
#' character to the plot. If \code{NULL} nothing is added.
#' @param col a \code{character} vector with the colours. It needs to contain a
#' minimum of 2 colours per condition. If not or if \code{NULL}, the defalut
#' colours will be used.
#' @param pch the R symbols used to plot the data It needs to contain a minimum
#' of 2 symbols per condition. If not or if \code{NULL}, the defalut symbols
#' will be used.
#' @param lty the line types used to plot the data. It needs to contain a
#' minimum of 2 line types per condition. If not or if \code{NULL}, the defalut
#' line types will be used.
#' @param contextPerRow a \code{logical} value indicating if the each row
#' represents an individual context. If \code{FALSE}, each column will represent
#' an individual context.
#' @return Invisibly returns \code{NULL}
#' @seealso \code{\link{plotMethylationProfile}},
#' \code{\link{computeMethylationProfile}} and \code{\link{methylationDataList}}
#' @examples
#'
#'
#' # load the methylation data
#' data(methylationDataList)
#'
#' #plot the low resolution profile at 10 Kb resolution
#' par(mar=c(4, 4, 3, 1)+0.1)
#' plotMethylationProfileFromData(methylationDataList[["WT"]],
#'                                methylationDataList[["met1-3"]],
#'                                conditionsNames=c("WT", "met1-3"),
#'                                windowSize = 20000, autoscale = TRUE,
#'                                context = c("CHG"))
#'
#' \dontrun{
#' #plot the low resolution profile at 5 Kb resolution
#' par(mar=c(4, 4, 3, 1)+0.1)
#' plotMethylationProfileFromData(methylationDataList[["WT"]],
#'                                methylationDataList[["met1-3"]],
#'                                conditionsNames=c("WT", "met1-3"),
#'                                windowSize = 5000, autoscale = TRUE,
#'                                context = c("CG", "CHG", "CHH"),
#'                                labels = LETTERS)
#' }
#'
#' @author Nicolae Radu Zabet
#' @export
plotMethylationProfileFromData <- function(methylationData1,
                                           methylationData2 = NULL,
                                           regions = NULL,
                                           conditionsNames = NULL,
                                           context = "CG",
                                           windowSize = NULL,
                                           autoscale = FALSE,
                                           labels=NULL,
                                           col=NULL,
                                           pch = c(1,0,16,2,15,17),
                                           lty = c(4,1,3,2,6,5),
                                           contextPerRow = TRUE) {
  .validateContext(context)
  .validateMethylationData(methylationData1, variableName="methylationData1")
  numberOfConditions <- 1
  if(!is.null(methylationData2)){
    .validateMethylationData(methylationData2, variableName="methylationData2")
    numberOfConditions <- 2
  }

  if(is.null(conditionsNames) | length(conditionsNames) < numberOfConditions){
    conditionsNames <- paste("condition ",(1:numberOfConditions),sep="")
  }


  number_of_symbols <-numberOfConditions*length(context)

  if(!.isColor(col, minLength = number_of_symbols )){
    if(number_of_symbols <= 6){
      col <- c("#D55E00","#E69F00", "#0072B2", "#56B4E9", "#F0E442", "#009E73")
    } else if(number_of_symbols <= 8){
      col <- c("#000000", "#E69F00", "#56B4E9", "#009E73", "#F0E442", "#0072B2", "#D55E00", "#CC79A7")
    } else{
      col <- rainbow(number_of_symbols)
    }
  }

  .stopIfNotAll(c(!is.null(pch),
                  all(as.integer(pch) == pch),
                  all(pch >= 0),
                  length(pch) >= number_of_symbols),
                paste(" pch is a vector of positive integers of size ",number_of_symbols, sep=""))

  .stopIfNotAll(c(!is.null(lty),
                  all(as.integer(lty) == lty),
                  all(lty >= 0),
                  length(lty) >= number_of_symbols),
                paste(" lty is a vector of positive integers of size ",number_of_symbols, sep=""))



  #compute all genomic regions for which data is available
  recomputeRegions <- FALSE
  if(is.null(regions)){
    recomputeRegions <- TRUE
  }
  if(!recomputeRegions & (!is(regions, "GRanges") | length(regions) < 1)){
    recomputeRegions <- TRUE
  }


  if(recomputeRegions){
    cat("Recompute regions... \n")
    regions <- GRanges()
    regions <- union(regions, getWholeChromosomes(methylationData1))
    if(numberOfConditions > 1){
      regions <- union(regions, getWholeChromosomes(methylationData2))
    }
  }
  if(is.null(windowSize)){
    windowSize = floor(min(width(regions))/500)
  }


  index <- 1;
  cat("Computing low resolution profiles... \n")


  par(mar=c(4, 4, 3, 1)+0.1)
  if(contextPerRow){
    par(mfrow=c(length(context),length(regions)))
    for(i in 1:length(context)){
      for(j in 1:length(regions)){
        buffer_profile <- GRangesList(computeMethylationProfile(methylationData1, regions[j], windowSize = windowSize, context = context[i]))
        if(numberOfConditions > 1){
          buffer_profile <- c(buffer_profile, GRangesList(computeMethylationProfile(methylationData2, regions[j], windowSize = windowSize, context = context[i])))
        }
        names(buffer_profile) <- conditionsNames

        symbols_ind <- (numberOfConditions*(i-1) + 1):(numberOfConditions*i)

        plotMethylationProfile(buffer_profile, autoscale = autoscale, labels = labels[index],
                               title=paste(context[i], " methylation on chromosome ", as.character(seqnames(regions)[j]),sep=""), col=col[symbols_ind],
                               pch = pch[symbols_ind], lty = lty[symbols_ind])
        index <- index + 1
      }
    }
  } else{
    par(mfrow=c(length(regions),length(context)))
    for(j in 1:length(regions)){
      for(i in 1:length(context)){
        buffer_profile <- GRangesList(computeMethylationProfile(methylationData1, regions[j], windowSize = windowSize, context = context[i]))
        if(numberOfConditions > 1){
          buffer_profile <- c(buffer_profile, GRangesList(computeMethylationProfile(methylationData2, regions[j], windowSize = windowSize, context = context[i])))
        }
        names(buffer_profile) <- conditionsNames

        symbols_ind <- (numberOfConditions*(i-1) + 1):(numberOfConditions*i)
        plotMethylationProfile(buffer_profile, autoscale = autoscale, labels = labels[index],
                               title=paste(context[i], " methylation on chromosome ", as.character(seqnames(regions)[j]),sep=""), col=col[symbols_ind],
                               pch = pch[symbols_ind], lty = lty[symbols_ind])
        index <- index + 1
      }
    }
  }
  invisible(NULL)
}




.countOverlaps <- function(region, subRegions){
  return(sum(countOverlaps(subRegions, region)))
}


.measureOverlaps <- function(region, subRegions){
  return(sum(width(intersect(subRegions, region))))
}



#' This function computes the distribution of a subset of regions
#' (\code{\link{GRanges}} object)  over a large region (\code{\link{GRanges}}
#' object)
#'
#' @title Compute Overlaps Profile
#' @param subRegions a \code{\link{GRanges}} object with the sub regions; e.g.
#' can be the DMRs.
#' @param largeRegion a \code{\link{GRanges}} object with the region where to
#' compute the overlaps; e.g. a chromosome
#' @param windowSize The \code{largeRegion} is partitioned into equal sized
#' tiles of width \code{windowSize}.
#' @param binary a value indicating whether to count 1 for each overlap or to
#' compute the width of the overlap
#' @param cores the number of cores used to compute the DMRs.
#' @return a \code{\link{GRanges}} object with equal sized tiles of the regions.
#' The object has one metadata file \code{score} which represents: the number of
#' subRegions overlapping with the tile, in the case of \code{binary = TRUE},
#' and the width of the subRegions overlapping with the tile , in the case of
#' \code{binary = FALSE}.
#' @seealso \code{\link{plotOverlapProfile}}, \code{\link{filterDMRs}},
#' \code{\link{computeDMRs}} and \code{\link{mergeDMRsIteratively}}
#' @examples
#' # load the methylation data
#' data(methylationDataList)
#'
#' # load the DMRs in CG context
#' data(DMRsNoiseFilterCG)
#'
#' # the coordinates of the area to be plotted
#' largeRegion <- GRanges(seqnames = Rle("Chr3"), ranges = IRanges(1,1E5))
#'
#' # compute overlaps distribution
#' hotspots <- computeOverlapProfile(DMRsNoiseFilterCG, largeRegion,
#'            windowSize = 10000, binary = FALSE)
#' @author Nicolae Radu Zabet
#' @export
computeOverlapProfile <- function(subRegions,
                                  largeRegion,
                                  windowSize = floor(width(largeRegion)/500),
                                  binary=TRUE,
                                  cores = 1){
  .stopIfNotAll(c(!is.null(subRegions),
                  is(subRegions, "GRanges")),
                paste(" subRegions neads to be a GRanges object.",sep=""))

  .stopIfNotAll(c(!is.null(largeRegion),
                  is(largeRegion, "GRanges"),
                  length(largeRegion) == 1),
                paste(" largeRegion neads to be a GRanges object with one location (e.g. entire chromosome).",sep=""))

  .stopIfNotAll(c(!is.null(binary),
                  is.logical(binary),
                  length(binary) == 1),
                paste(" binary neads to be a logical variable.",sep=""))

  .stopIfNotAll(c(.isInteger(windowSize, positive=TRUE)),
                " the window size used to compute the methylation profile is an integer higher or equal to 0")

  .stopIfNotAll(c(.isInteger(cores, positive=TRUE)),
                " the number of cores to used when computing the DMRs needs to be an integer  hirger or equal to 1.")

  seqname <- seqnames(largeRegion)
  minPos <- start(largeRegion)
  maxPos <- end(largeRegion)


  cat("Calculating overlaps for ", .printGenomicRanges(largeRegion), " using a window of ", windowSize, " bp \n")
  seqs = seq(minPos, maxPos - windowSize, windowSize);

  ranges <- GRanges(seqname, IRanges(seqs, seqs+windowSize-1))
  ranges$score <- 0


  rangesList <- tile(ranges, cores)
  if(binary){
    if(cores > 1){
      scores <- parallel::mclapply(rangesList, .countOverlaps, subRegions = subRegions, mc.cores = cores)
    } else {
      scores <- lapply(rangesList, .countOverlaps, subRegions = subRegions)
    }
  } else{
    if(cores > 1){
      scores <- parallel::mclapply(rangesList, .measureOverlaps, subRegions = subRegions, mc.cores = cores)
    } else {
      scores <- lapply(rangesList, .measureOverlaps, subRegions = subRegions)
    }
  }
  ranges$score <- unlist(scores)

  return(ranges)
}



#' This function plots the distribution of a set of subregions on a large
#' region.
#'
#' @title Plot overlap profile
#' @param overlapsProfiles1 a  \code{\link{GRanges}} object with the overlaps
#' profile; see \code{\link{computeOverlapProfile}}.
#' @param overlapsProfiles2 a  \code{\link{GRanges}} object with the overlaps
#' profile; see \code{\link{computeOverlapProfile}}. This is optional. For
#' example, one can be use \code{overlapsProfiles1} to display hypomethylated
#' regions and \code{overlapsProfiles2} the hypermethylated regions.
#' @param names a \code{vector} of \code{character} to add labels for the two
#' overlapsProfiles. This is an optinal parameter.
#' @param labels a \code{vector} of \code{character} used to add a subfigure
#' character to the plot. If \code{NULL} nothing is added.
#' @param col a \code{character} vector with the colours. It needs to contain 2
#' colours. If not or if \code{NULL}, the defalut colours will be used.
#' @param title the title of the plot.
#' @param logscale a \code{logical} value indicating if the colours are on
#' logscale or not.
#' @param maxValue a maximum value in a region. Used for the colour scheme.
#' @return Invisibly returns \code{NULL}.
#' @seealso \code{\link{computeOverlapProfile}}, \code{\link{filterDMRs}},
#' \code{\link{computeDMRs}} and \code{\link{mergeDMRsIteratively}}
#' @examples
#'
#' # load the methylation data
#' data(methylationDataList)
#'
#' # load the DMRs in CG context
#' data(DMRsNoiseFilterCG)
#'
#' # the coordinates of the area to be plotted
#' largeRegion <- GRanges(seqnames = Rle("Chr3"), ranges = IRanges(1,1E5))
#'
#' # compute overlaps distribution
#' hotspotsHypo <- computeOverlapProfile(DMRsNoiseFilterCG, largeRegion,
#'                  windowSize = 10000, binary = FALSE)
#'
#' plotOverlapProfile(GRangesList("Chr3"=hotspotsHypo),
#'                    names = c("hypomethylated"), title = "CG methylation")
#'
#' \dontrun{
#'
#' largeRegion <- GRanges(seqnames = Rle("Chr3"), ranges = IRanges(1,1E6))
#'
#' hotspotsHypo <- computeOverlapProfile(
#'                DMRsNoiseFilterCG[(DMRsNoiseFilterCG$regionType == "loss")],
#'                largeRegion, windowSize=2000, binary=TRUE, cores=1)
#'
#' hotspotsHyper <- computeOverlapProfile(
#'                DMRsNoiseFilterCG[(DMRsNoiseFilterCG$regionType == "gain")],
#'                largeRegion, windowSize=2000, binary=TRUE, cores=1)
#'
#' plotOverlapProfile(GRangesList("Chr3"=hotspotsHypo),
#'                    GRangesList("Chr3"=hotspotsHyper),
#'                    names=c("loss", "gain"), title="CG methylation")
#' }
#'
#' @author Nicolae Radu Zabet
#' @export
plotOverlapProfile <- function(overlapsProfiles1, overlapsProfiles2=NULL, names=NULL,
                               labels=NULL, col=NULL, title="", logscale=FALSE, maxValue=NULL) {

  .stopIfNotAll(c(!is.null(overlapsProfiles1),
                  is(overlapsProfiles1, "GRangesList"),
                  length(overlapsProfiles1) > 0),
                " overlapsProfiles1 needs to be a GRangesList object")
  unlisted_overlapsProfiles1 <- unlist(overlapsProfiles1, use.names=FALSE)
  score <- mcols(unlisted_overlapsProfiles1)$score
  if(is.null(score)) stop("elements of the overlapsProfiles1 are incorrect")
  maxScore <- max(score)
  minScore <- min(score)

  numberOfConditions <- 1

  if(!is.null(overlapsProfiles2)){
    .stopIfNotAll(c(is(overlapsProfiles2, "GRangesList"),
                    length(overlapsProfiles2) == length(overlapsProfiles1)),
                  " overlapsProfiles2 needs to be a GRangesList object with the same number of elements as overlapsProfiles1")
    unlisted_overlapsProfiles2 <- unlist(overlapsProfiles2, use.names=FALSE)
    score <- mcols(unlisted_overlapsProfiles2)$score
    if(is.null(score)) stop("elements of the overlapsProfiles2 are incorrect")
    maxScore <- max(maxScore, score)
    minScore <- min(minScore, score)

    numberOfConditions <- 2
  }




  .stopIfNotAll(c(!is.null(logscale),
                  is.logical(logscale),
                  length(logscale) == 1),
                paste(" logscale neads to be a logical variable.",sep=""))


  if(!.isColor(col) | length(col) < numberOfConditions){
    col <- c("#0072B2", "#D55E00")
  }

  if(!is.null(maxValue)){
    if(is.numeric(maxValue)){
      if(round(maxValue) > 0){
        maxScore <- round(maxValue)
      }
    }
  }

  if(logscale){
    minScore <- log10(1)
    maxScore <- log10(maxScore)
  }

  minScore <- round(minScore)
  maxScore <- round(maxScore)


  cols1 <- c(colorRampPalette(c("white",col[1]))(1 + maxScore - minScore))
  cols2 <- c(colorRampPalette(c("white",col[2]))(1 + maxScore - minScore))

  numberOfProfiles<-length(overlapsProfiles1)

  xmin <- min(unlist(lapply(overlapsProfiles1, start)))
  xmax <- max(unlist(lapply(overlapsProfiles1, end)))

  par(mar=c(4, 4, 3, 1)+0.1)
  plot(c(xmin,xmax),c(0, 1*numberOfProfiles), type="n",xlab="position (bp)", ylab="", yaxt="n", bty="n", main=title)

  for(i in 1:length(overlapsProfiles1)){
    if(!is.null(overlapsProfiles2)){
      #plot two conditions
      for(j in 1:length(overlapsProfiles1[[i]])){
        colId <- overlapsProfiles1[[i]][j]$score
        if(colId > 0){
          if(logscale){
            colId <- log10(colId)
          }
          rect(start(overlapsProfiles1[[i]][j]), (length(overlapsProfiles1) - i + 0.4), end(overlapsProfiles1[[i]][j]), (length(overlapsProfiles1) - i + 0.7), col=cols1[round(colId)+1], border = NA)
        }
      }

      for(j in 1:length(overlapsProfiles2[[i]])){
        colId <- overlapsProfiles2[[i]][j]$score
        if(colId > 0){
          if(logscale){
            colId <- log10(colId)
          }
          rect(start(overlapsProfiles2[[i]][j]), (length(overlapsProfiles2) - i + 0.1), end(overlapsProfiles2[[i]][j]), (length(overlapsProfiles2) - i + 0.4), col=cols2[round(colId)+1], border = NA)
        }
      }

      if(!is.null(names)){
        if(all(is.character(names)) & length(names) >= numberOfConditions){
          at_pos <- c(0.55,0.25)
          for(j in 1:numberOfConditions){
            mtext(names[j], side = 2, line=-0.5, at=(length(overlapsProfiles1) - i + at_pos[j]), col="black", las=2)
          }
        }
      }
    } else{
      #plot one conditions
      for(j in 1:length(overlapsProfiles1[[i]])){
        colId <- overlapsProfiles1[[i]][j]$score
        if(colId > 0){
          if(logscale){
            colId <- log10(colId)
          }
          rect(start(overlapsProfiles1[[i]][j]), (length(overlapsProfiles1) - i + 0.1), end(overlapsProfiles1[[i]][j]), (length(overlapsProfiles1) - i + 0.7), col=cols1[round(colId)+1], border = NA)
        }
      }
      if(!is.null(names)){
        if(all(is.character(names)) & length(names) >= numberOfConditions){
          at_pos <- c(0.4)
          mtext(names[1], side = 2, line=-0.5, at=(length(overlapsProfiles1) - i + at_pos[1]), col="black")
        }
      }
    }


    rect(xmin, (length(overlapsProfiles1) - i + 0.1), max(end(overlapsProfiles1[[i]])), (length(overlapsProfiles1) - i + 0.7), col=NA, border="#999999")
    text(0, (length(overlapsProfiles1) - i + 0.80), names(overlapsProfiles1)[i], pos=4, col="black")
  }

  if(!is.null(labels)){
    mtext(labels[1], line = 0.7, adj = 0, cex=1.4);
  }
  invisible(NULL)

}

#' This function enumerate cytosine positions in a BSgenome reference
#'
#' @title Select Cytosine Positions
#' @description
#' Constructs a \code{\link[GenomicRanges]{GRanges}} of all cytosine positions
#' in the specified BSgenome (or BSgenome package name), optionally filtering
#' by methylation context (\code{"CG"}, \code{"CHG"}, \code{"CHH"}), by
#' chromosome, and by genomic region.
#'
#' @param genome
#'   A \code{BSgenome} object or the name (character) of a BSgenome package
#'   to use as the reference genome. If a package name is given, it will be loaded
#'   automatically (default: \code{BSgenome.Hsapiens.UCSC.hg38}).
#'   \strong{Note:} When running on an HPC system, please ensure that the required
#'   \code{BSgenome} package is already installed and loaded in advance.
#' @param context
#'   A character vector of one or more methylation contexts to include:
#'   \code{"CG"}, \code{"CHG"}, and/or \code{"CHH"}.  Defaults to all three.
#' @param chr
#'   An optional character vector of chromosome names to restrict the
#'   enumeration.  If \code{NULL}, all sequences in \code{genome} are used.
#' @param region
#'   An optional \code{\link[GenomicRanges]{GRanges}} object specifying
#'   subregions to keep.  Requires \code{chr} to be non-\code{NULL}.
#'
#' @return
#' A \code{\link[GenomicRanges]{GRanges}} object with one 1-bp range per
#' cytosine, and two metadata columns:
#' \describe{
#'   \item{context}{A factor indicating the context (\code{"CG"},
#'     \code{"CHG"}, or \code{"CHH"}).}
#'   \item{trinucleotide_context}{Character or factor giving the
#'     surrounding trinucleotide sequence (or \code{NA} for \code{"CG"}).}
#' }
#' @seealso \code{\link{readONTbam}}, \code{\link{ontSampleGRangesList}}
#' @examples
#' \dontrun{
#' library(BSgenome.Hsapiens.UCSC.hg38)
#'
#' # Only CpG sites on chr1 and chr2:
#' gr_chr1_2 <- selectCytosine(context="CG", chr=c("chr1","chr2"))
#'
#' # CHH sites in a specific region on chr3:
#' my_region <- GRanges("chr3", IRanges(1e6, 1e6 + 1e5))
#' gr_region <- selectCytosine(context="CHH", chr="chr3", region=my_region)
#' }
#'
#' @author Nicolae Radu Zabet and Young Jun Kim
#' @import GenomicRanges Biostrings
#' @export
selectCytosine <- function(genome   = BSgenome.Hsapiens.UCSC.hg38,
                           context  = c("CG","CHG","CHH"),
                           chr      = NULL,
                           region   = NULL) {
  ##Parameters checking
  cat("Parameters checking ...\n")

  .validateContext(context)

  .validateGenome(genome)

  .validateChromosome(chr, genome)

  ## select the cytosine based on the reference genome
  cat("Select the Cytosine Positions in the specified BSgenome ...\n")
  ## --- literal 2-/3-mer definitions
  pat_dict <- list(
    CG  = "CG",
    CHG = c("CAG","CCG","CTG"),
    CHH = c("CAA","CAC","CAT",
            "CCA","CCC","CCT",
            "CTA","CTC","CTT")
  )
  if (!all(context %in% names(pat_dict))) {
    stop("Unknown context: ", paste(setdiff(context, names(pat_dict)), collapse=", "))
  }

  chroms <- if (is.null(chr)) seqnames(genome) else intersect(chr, seqnames(genome))
  if (!length(chroms)) stop("No valid chromosomes selected.")

  per_chr <- lapply(chroms, function(chrN) {
    seqN <- genome[[chrN]]
    seql <- seqlengths(genome)[chrN]
    ctx_hits <- lapply(intersect(context, names(pat_dict)), function(ctx) {
      patterns <- pat_dict[[ctx]]

      # Forward strand matches
      pos_fwd <- unique(unlist(lapply(patterns, function(pat) start(matchPattern(pat, seqN))), use.names=FALSE))
      pos_fwd <- pos_fwd[pos_fwd >= 1 & pos_fwd <= (seql - ifelse(ctx=="CG",1,2))]
      gr_fwd <- if (length(pos_fwd)) {
        gr <- GRanges(chrN, IRanges(start=pos_fwd, width=1L), strand="+")
        tri_fwd <- if (ctx=="CG") rep(NA_character_, length(pos_fwd))
        else as.character(getSeq(genome, chrN, pos_fwd, width=3L))
        mcols(gr) <- DataFrame(context=factor(ctx, levels=names(pat_dict)),
                               trinucleotide_context=tri_fwd)
        gr
      } else GRanges()

      # Reverse strand for CG/CHG
      gr_rev <- GRanges()
      if (ctx %in% c("CG","CHG") && length(gr_fwd)) {
        shift <- if (ctx=="CG") 1L else 2L
        gr <- GRanges(seqnames(gr_fwd), IRanges(start(gr_fwd)+shift, width=1L), strand="-")
        tri_rev <- if (ctx=="CG") rep(NA_character_, length(gr_fwd))
        else as.character(reverseComplement(DNAStringSet(mcols(gr_fwd)$trinucleotide_context)))
        mcols(gr) <- DataFrame(context=mcols(gr_fwd)$context,
                               trinucleotide_context=tri_rev)
        if (ctx %in% "CHG"){
          valid <- mcols(gr)$trinucleotide_context %in% pat_dict[["CHG"]]
          gr<- gr[valid]
        }
        gr_rev <- gr
      }

      # Reverse strand for CHH
      gr_rev_chh <- GRanges()
      if (ctx=="CHH") {
        patterns_rev <- as.character(reverseComplement(DNAStringSet(patterns)))
        pos_rev <- unique(unlist(lapply(patterns_rev, function(pat) start(matchPattern(pat, seqN))), use.names=FALSE))
        pos_rev <- pos_rev[pos_rev >=1 & pos_rev <= (seql-2)]
        if (length(pos_rev)) {
          gr2 <- GRanges(chrN, IRanges(start=pos_rev+2, width=1L), strand="-")
          tri_fwd <- as.character(getSeq(genome, chrN, pos_rev, width=3L))
          tri_rev <- as.character(reverseComplement(DNAStringSet(tri_fwd)))
          mcols(gr2) <- DataFrame(context=factor(ctx, levels=names(pat_dict)),
                                  trinucleotide_context=tri_rev)
          gr_rev_chh <- gr2
        }
      }

      c(gr_fwd, gr_rev, gr_rev_chh)
    })
    do.call(c, ctx_hits)
  })

  # Harmonize seqlevels
  per_chr <- lapply(per_chr, function(gr) {
    seqlevels(gr)  <- chroms
    seqlengths(gr) <- seqlengths(genome)[chroms]
    gr
  })

  # Flatten and sort
  ref_gr <- sort(do.call(c, per_chr))
  o <- order(as.character(seqnames(ref_gr)), start(ref_gr))
  ref_gr <- ref_gr[o]

  # If region specified, subset final GRanges
  if (!is.null(region)) {
    # Ensure seqlevels compatibility
    seqlevels(region, pruning.mode="coarse") <- seqlevels(ref_gr)
    ref_gr <- subsetByOverlaps(ref_gr, region, ignore.strand=FALSE)
  }

  ref_gr
}

#' This function read and annotate ONT MM/ML tags against a cytosine reference
#'
#' @title  Load ONT BAM, decode MM/ML, and count modified vs. unmodified reads
#' @description
#' \code{readONTbam()} takes an indexed Nanopore BAM file with MM/ML tags,
#' decodes each read’s per-C modification probabilities, and overlays
#' them on a \code{GRanges} of candidate cytosine sites.
#' It returns a copy of \code{ref_gr} augmented with:
#' \itemize{
#'   \item \code{ONT_Cm} — comma-delimited read‐indices called “modified”
#'   \item \code{ONT_C}  — comma-delimited read‐indices covering but _not_ modified
#'   \item \code{readsM} — count of modified reads at each site
#'   \item \code{readsN} — total same-strand coverage at each site
#' }
#'
#' You can either supply your own \code{ref_gr} (e.g.\ from \code{selectCytosine()})
#' or leave it \code{NULL} and pass \code{context}, \code{chr}, \code{region}
#' to build \code{ref_gr} on the fly.
#'
#' @param bamfile      Path to an indexed ONT BAM with MM/ML tags.
#' @param ref_gr       A \code{GRanges} of genomic cytosine positions to annotate.
#'                     If \code{NULL}, will be created via \code{selectCytosine()}
#'                     using \code{context}, \code{chr}, \code{region}.
#' @param modif        Character vector of MM codes to treat as “modified”
#'                     (e.g. \code{"C+m?"}, \code{"C+h?"}, \code{"C+m."}).
#' @param prob_thresh  Numeric in \[0,1\] — minimum ML probability to call a read
#'                     “modified.”
#' @param genome       A \code{BSgenome} object such as \code{BSgenome.Hsapiens.UCSC.hg38}.
#'                     This is used to extract sequence context and must be loaded in advance.
#'                     \strong{Note:} When running on an HPC system,
#'                     please ensure that the required
#'                     \code{BSgenome} package is already installed and loaded in advance.
#' @param context      Sequence context for \code{selectCytosine()}
#'                     (e.g. \code{"CG"}, \code{"CHG"}, \code{"CHH"}).
#' @param chr          Chromosome names to restrict \code{selectCytosine()}.
#' @param region       A \code{GRanges} to further subset \code{selectCytosine()}.
#' @param synonymous   Logical (default: FALSE). If TRUE, include modified calls that match the specified context sequence (e.g. CGG),
#'                     even if the site was previously excluded due to deletion or mismatch.
#'                     For example, if a deletion occurs at position 234523, but the surrounding context still forms CGG,
#'                     then the modified C at 234523 will be retained (Nmod=1).
#' @param parallel     Logical. If TRUE, automatically detect your system condition and
#'                     decoding will use parallel threads via BiocParallel::.
#'                     If FALSE (default), decoding is done serially.
#' @param BPPARAM      A \code{BiocParallelParam} object controlling parallel execution.
#'                     This value will automatically set when parallel is \code{TRUE}, also able to set as manually.
#' @param cores Integer number of workers (must not exceed BPPARAM$workers).
#'    This value will automatically set as the maximum number of system workers,
#'    also able to set as manually.
#'
#' @return A \code{GRanges} of the same length as \code{ref_gr}, with four
#'   additional metadata columns:
#'   \describe{
#'     \item{ONT_Cm}{comma-delimited read‐indices called modified}
#'     \item{ONT_C}{comma-delimited read‐indices covering but unmodified}
#'     \item{readsM}{integer count of modified reads per site}
#'     \item{readsN}{integer count of same‐strand reads covering each site}
#'   }
#' @seealso \code{\link{selectCytosine}}, \code{\link{computeDMRs}},
#' \code{\link{computePMDs}}, \code{\link{computeCoMethylatedPositions}},
#' \code{\link{computeCoMethylatedRegions}}, \code{\link{filterVMRsONT}},
#' \code{\link{ontSampleGRangesList}}, \code{\link{scanBamChr1Random5}}
#' @examples
#' \dontrun{
#' library(DMRcaller)
#' library(BSgenome.Hsapiens.UCSC.hg38)
#'
#' # set the bam file directory
#' bam_path <- system.file("extdata", "scanBamChr1Random5.bam", package="DMRcaller")
#'
#' # read ONTbam file (chromosome 1 only) in CG context with BSgenome.Hsapiens.UCSC.hg38
#' ONTSampleGRanges <- readONTbam(bamfile = bam_path, ref_gr = NULL, modif = "C+m?",
#'                          prob_thresh = 0.50,genome = BSgenome.Hsapiens.UCSC.hg38,
#'                          context = "CG", chr = "chr1", region = NULL,
#'                          synonymous = FALSE, parallel = FALSE, BPPARAM = NULL)
#' }
#'
#' @author Nicolae Radu Zabet and Young Jun Kim
#' @import GenomicAlignments
#' @import Rsamtools
#' @importFrom GenomicRanges GRanges GRangesList findOverlaps mcols
#' @importFrom IRanges IRanges IntegerList
#' @importFrom Biostrings reverseComplement matchPattern getSeq
#' @importFrom BiocParallel SerialParam bplapply
#' @importFrom cigarillo project_sequences 
#' @importFrom cigarillo cigars_as_ranges_along_query
#' @importFrom cigarillo cigars_as_ranges_along_ref 
#' @importFrom cigarillo cigar_extent_along_query 
#' @importFrom cigarillo cigar_extent_along_ref
#' @importFrom cigarillo explode_cigar_ops 
#' @importFrom cigarillo explode_cigar_oplens
#' @export
readONTbam <- function(bamfile,
                       ref_gr      = NULL,
                       modif       = "C+m?",
                       prob_thresh = 0.50,
                       genome      = BSgenome.Hsapiens.UCSC.hg38,
                       context     = "CG",
                       chr         = NULL,
                       region      = NULL,
                       synonymous  = FALSE,
                       parallel    = FALSE,
                       BPPARAM     = NULL,
                       cores       = NULL) {
  ### PUT THE include_diff = FALSE, include_nocall = FALSE) after including parameter

  ##Parameters checking
  cat("Parameters checking ...\n")

  .validateContext(context)

  .validateBamfile(bamfile)

  .validateGenome(genome)

  .validateChromosome(chr, genome)

  .validateModif(modif, bamfile, chr, genome)

  if (is.null(ref_gr)){
    ref_gr = selectCytosine(genome, context, chr, region)
  }
  region <- .validateGRanges(region, ref_gr)

  # generate the BPPARAM value if set as parallel 
  if (parallel == TRUE){
    BPPARAM <- suppressWarnings(.validateBPPARAM(BPPARAM, progressbar = TRUE)) 
  }else{
    # Force serial execution
    BPPARAM <- BiocParallel::SerialParam(progressbar = TRUE)
  }
  # If cores argument is specified
  if (!is.null(cores)) {
    .stopIfNotAll(.isInteger(cores, positive = TRUE), 
                  "the number of cores used when computing the DMRs needs to be an integer higher or equal to 1.")
    
    # Check if user requested more cores than available
    if (cores > BPPARAM$workers) {
      warning(paste0("The number of requested cores (", cores, 
                     ") exceeds the available system cores (", BPPARAM$workers, 
                     "). Automatically setting cores to the maximum available (", 
                     BPPARAM$workers, ")."))
      cores <- BPPARAM$workers
    } else {
      message(paste0("Using user-specified core count: ", cores))
    }
    
    # Apply the final core number
    BPPARAM$workers <- cores
  } else {
    cores <- BPPARAM$workers
  }
  cat("Current parallel setting, BPPARAM: ", capture.output(BPPARAM),sep = "\n")

  .stopIfNotAll(c(!is.null(prob_thresh),
                  is.numeric(prob_thresh),
                  prob_thresh > 0,
                  prob_thresh < 1),
                " the p-value threshold needs to be in the interval (0,1)")

  .stopIfNotAll(c(!is.null(synonymous),
                  is.logical(synonymous)),
                " the synonymous needs to be logical")
  if (synonymous == TRUE){
    .stopIfNotAll(c(!is.null(synonymous),
                    is.logical(synonymous),
                    context %in% c("CHG", "CHH")),
                  " the context should choose the CHG or CHH context when synonymous is TRUE \n set the synonymous value as FALSE when context is CG")
  }


   # @@@@time t1
  t1 <- proc.time()

  ## 1) preprocess the scanned bam list with filtering
  # existing indexing + ScanBamParam code
  cat("[readONTbam] Generating index file for BAM ...\n")
  if (!file.exists(paste0(bamfile,".bai")))
    Rsamtools::indexBam(bamfile)
  cat("[readONTbam] Loading BAM ...\n")
  if (!is.null(chr) && length(chr) > 0) {
    which_region <- GenomicRanges::GRanges(chr, IRanges::IRanges(1, seqlengths(genome)[chr]))
    param <- Rsamtools::ScanBamParam(
      what = c("qname", "rname", "pos", "cigar", "flag", "seq", "mapq"),
      tag  = c("MM", "ML"),
      which = which_region
    )
  } else {
    param <- Rsamtools::ScanBamParam(
      what = c("qname", "rname", "pos", "cigar", "flag", "seq", "mapq"),
      tag  = c("MM", "ML")
    )
  }
  sb <- Rsamtools::scanBam(bamfile, param = param)[[1]]
  # @@@@@time t2
  t2 <- proc.time()
  cat(sprintf("Step 1 (scanBam): %.2f sec\n", (t2-t1)[["elapsed"]]))

  # Filter out
  cat("[readONTbam] Filtering invalid or empty reads ...\n")
  valid_cigar <- !is.na(sb$cigar)
  valid_seq   <- nchar(sb$seq) > 0
  # handle primary/secondary/supplementary filtering
  is_secondary      <- bitwAnd(sb$flag, 0x100) != 0
  is_supplementary  <- bitwAnd(sb$flag, 0x800) != 0
  has_tag           <- !vapply(sb$tag$ML, function(x) all(is.na(x)), logical(1))
  has_hardclip <- grepl("H", sb$cigar)

  # primary: not secondary and not supplementary
  is_primary <- !is_secondary & !is_supplementary

  # keep secondary/supplementary
  valid_secondary_supplementary <- is_primary | ((is_secondary | is_supplementary) & has_tag & !has_hardclip & valid_seq)

  # keep only those passing both filters
  if (!is.null(chr)){
    keep <- valid_cigar & sb$rname %in% chr & valid_secondary_supplementary
  } else {
    keep <- valid_cigar & sb$rname & !is.na(sb$rname) & valid_secondary_supplementary
  }
  keep[is.na(keep)] <- FALSE

  # subset all vectors in sb
  for (nm in c("qname","rname","pos","cigar","flag","seq","mapq")) {
    sb[[nm]] <- sb[[nm]][keep]
  }
  # subset all BAM tags
  if (!is.null(sb$tag)) {
    sb$tag <- lapply(sb$tag, `[`, keep)
  }

  # Decode the strand from the FLAG bit.
  # In SAM/BAM, bit 0x10 (decimal 16) is set if the read maps to the reverse strand.
  strands <- ifelse(bitwAnd(sb$flag, 0x10) != 0, "-", "+")
  #
  # # reconstruct aligned blocks per read
  # cat("[readONTbam] Building coverage blocks per read ...\n")
  # cigar_str <- sb$cigar
  # blocks_ir <- cigars_as_ranges_along_ref(
  #   cigar_str,
  #   pos = sb$pos,
  #   ops = c("M", "=", "X")
  # )
  # blocks_grl <- GRangesList(
  #   mapply(function(rn, ir, st) {
  #     GRanges(seqnames = rn,
  #             ranges   = ir,
  #             strand   = st)
  #   },
  #   rn = sb$rname,
  #   ir = blocks_ir,
  #   st = strands,
  #   SIMPLIFY = FALSE)
  # )

  # After sb is loaded and filtered
  aln <- GAlignments(seqnames=sb$rname, pos=sb$pos, cigar=sb$cigar, seq=sb$seq, strand = strands)
  ref_seq_list <- getSeq(genome, seqnames(aln), start(aln), end(aln))
  aligned_seq_list <- cigarillo::project_sequences(sb$seq, cigars=sb$cigar, from="query", to="reference")

  # build a lookup: read name -> its index ; easy to navigating less memory
  read_order <- seq_along(sb$qname)
  names(read_order) <- sb$qname
  # @@@@ time t3
  t3 <- proc.time()
  cat(sprintf("Step 2 (filter): %.2f sec\n", (t3-t2)[["elapsed"]]))

  # --------------
  #  Helper: decode one read’s MM/ML using C‐counting
  # --------------
  # *1) Helper: extract the number of soft‐clipped bases from a CIGAR string
  # Vectorized function to extract soft‐ and hard‐clip counts from a CIGAR string

  .getClipping <- function(cigar) {
    # Leading clips: soft (S) plus hard (H)
    lead_S <- if (grepl("^\\d+S", cigar)) as.integer(sub("^(\\d+)S.*", "\\1", cigar)) else 0L
    lead_H <- if (grepl("^\\d+H", cigar)) as.integer(sub("^(\\d+)H.*", "\\1", cigar)) else 0L
    lead   <- lead_S + lead_H

    # Trailing clips: soft (S) plus hard (H)
    trail_S <- if (grepl("\\d+S$", cigar)) as.integer(sub(".*?(\\d+)S$", "\\1", cigar)) else 0L
    trail_H <- if (grepl("\\d+H$", cigar)) as.integer(sub(".*?(\\d+)H$", "\\1", cigar)) else 0L
    trail   <- trail_S + trail_H

    c(lead, trail)
  }
  # *2) Helper: parse insertions and deletions from a CIGAR string
  .getIndels <- function(cig) {
    # Insertions in read‐space
    ins_ranges <- cigars_as_ranges_along_query(cig, ops="I")
    ins_pos    <- unlist(start(ins_ranges))
    ins_len    <- unlist(width(ins_ranges))

    # Deletions: pos in read‐space, len in reference‐space
    del_ranges_q <- cigars_as_ranges_along_query(cig, ops="D")
    del_ranges_r <- cigars_as_ranges_along_ref(cig, ops="D")
    del_pos      <- unlist(start(del_ranges_q))
    del_len      <- unlist(width(del_ranges_r))

    list(
      insertion = list(pos = ins_pos, len = ins_len),
      deletion  = list(pos = del_pos, len = del_len)
    )
  }

  # *3) Helper: find all cytosine positions in the read sequence
  .getCytPos <- function(read_seq, strand) {
    # If the read is reverse‐mapped, reverse‐complement it first
    seq_to_search <- if (strand == "-") reverseComplement(read_seq) else read_seq

    # Use matchPattern() to find every “C” and return their 1-based starts
    start(matchPattern("C", seq_to_search))
  }

  # Find the reference aligned insertion position using the cigar string
  .getInsertionPos <- function(cigar){
    # Explode the CIGAR string into operations and lengths
    ops  <- explode_cigar_ops(cigar)[[1]]
    lens <- explode_cigar_oplens(cigar)[[1]]

    read_pos <- 1
    ref_pos  <- 1
    insertion_ref_sites <- c()

    for (i in seq_along(ops)) {
      op  <- ops[i]
      len <- lens[i]

      if (op == "M" || op == "=" || op == "X") {
        read_pos <- read_pos + len
        ref_pos  <- ref_pos + len
      } else if (op == "I") {
        insertion_ref_sites <- c(insertion_ref_sites, rep(ref_pos, len))
        read_pos <- read_pos + len
      } else if (op == "D" || op == "N") {
        ref_pos  <- ref_pos + len
      } else if (op == "S" || op == "H" || op == "P") {
        read_pos <- read_pos + len
      }
    }

    return(unique(insertion_ref_sites))
  }

  # *4) Main worker: decode MM/ML, apply soft‐clip and indel adjustments,
  #    and return a DataFrame of genomic positions + probabilities + modType

  .computeModifPos <- function(mm_tag, ml_tag, cigar, strand, start,
                               cyt_pos, indel_info, ref_seq, aligned_seq,
                               context, synonymous) {

    # 0) bail out if no MM or ML tag
    if (is.na(mm_tag) || length(mm_tag)==0 ||
        is.null(ml_tag) || length(ml_tag)==0) {
      return(DataFrame(pos=integer(0),
                       prob=numeric(0),
                       modType=character(0)))
    }

    # 1) directly grab the ML vector of 0–255 probabilities
    ml_vals <- as.integer(ml_tag)
    ml_idx  <- 1L

    # 2) figure out read‐length and soft‐clipped bases
    read_len   <- cigar_extent_along_query(cigar)
    read_ref_len <- cigar_extent_along_ref(cigar)
    sc         <- .getClipping(cigar)    # c(lead_clip, trail_clip)
    lead_clip  <- sc[1]
    trail_clip <- sc[2]

    # 3) pull out this read’s insertions/deletions
    ins         <- indel_info$insertion
    del         <- indel_info$deletion
    insert_pos  <- ins$pos;    insert_len  <- ins$len
    delete_pos  <- del$pos;    delete_len  <- del$len

    # filtering insertion position
    ins_align_pos <- .getInsertionPos(cigar) - 1

    # 4) prepare outputs
    out_pos  <- integer()
    out_prob <- numeric()
    out_modified_base_code  <- character()
    out_mod <- integer()
    out_insert <- integer()
    out_delete <- integer()
    out_diff <- integer()
    out_nocall <-integer()


    # 5) split your “C+m?” / “C+h?” entries
    entries <- strsplit(as.character(mm_tag), ";", fixed=TRUE)[[1]]
    for (ent in entries) {
      if (ent == "") next  # skip empty

      parts    <- strsplit(ent, ",", fixed=TRUE)[[1]]
      code     <- parts[1]                     # e.g. "C+m?"
      offsets  <- as.integer(parts[-1]) + 1L   # 1‐based step counts
      c_counts <- cumsum(offsets)              # which nth‐C in the read

      # 6) grab the ML probabilities for *those* offsets
      probs    <- ml_vals[ ml_idx + seq_along(offsets) - 1L ] / 255
      ml_idx   <- ml_idx + length(offsets)

      # 7) throw away any c_counts beyond the # of Cs in the read
      valid <- c_counts <= length(cyt_pos)

      if (!any(valid)) next

      # 8) apply your lead_clip window
      hit_cyt_pos <- cyt_pos[c_counts[valid]]
      hit_probs <- probs[valid]
      hit_code <- rep(code, sum(valid))
      n_hits <- length(hit_cyt_pos)
      if (n_hits == 0) next

      # 9) for each one that survived, map read‐space→genome

      for (k in seq_along(hit_cyt_pos)) {
        # the read-space C position for this k
        pr <- hit_probs[k]
        rp <- hit_cyt_pos[k]
        cd <- hit_code[k]
        if (strand == "+") {
          # **collect any C that was part of an insertion**
          insert <- as.integer(any(rp >= insert_pos & rp < insert_pos + insert_len))
          # delete <- as.integer(any(rp == delete_pos))
          delete <- as.integer(any(rp %in% delete_pos))

          insert_off <- sum(insert_len[insert_pos < rp])
          delete_off <- sum(delete_len[delete_pos < rp])

          if (delete == 1){
            del_idx <- which(rp >= delete_pos & rp < delete_pos + 1)[1]
            del_len <- if (!is.na(del_idx)) delete_len[del_idx] else 0L
            gp <- start - lead_clip + (rp - insert_off + delete_off) - 1L + del_len
          } else {
            gp <- start - lead_clip + (rp - insert_off + delete_off) - 1L
          }
        } else {
          # reverse strand
          rp_rev <- read_len - rp + 1L

          insert <- as.integer(any(rp_rev >= insert_pos & rp_rev <= insert_pos + insert_len))
          # delete <- as.integer(any(rp_rev == delete_pos))
          delete <- as.integer(any(rp_rev %in% delete_pos))

          insert_off <- sum(insert_len[insert_pos < rp_rev])
          delete_off <- sum(delete_len[delete_pos < rp_rev])

          if (delete == 1) {
            del_idx <- which(rp_rev >= delete_pos & rp_rev < delete_pos + 1)[1]
            del_len <- if (!is.na(del_idx)) delete_len[del_idx] else 0L
            within_deletion <- any(rp_rev >= delete_pos & rp_rev < delete_pos + delete_len)
            adjust <- if (within_deletion) -1L else 0L
            gp <- start - lead_clip + (rp_rev - insert_off + delete_off) - 1L + del_len  + adjust
          } else {
            gp <- start - lead_clip + (rp_rev - insert_off + delete_off) - 1L
          }
        }
        if (gp >= start + read_ref_len | gp < start) next

        # Fetch ref/read base at this position
        ref_idx <- gp - start + 1
        if (ref_idx < 1 || ref_idx > length(ref_seq)) next
        ref_base <- ref_seq[ref_idx]
        aligned_base <- aligned_seq[ref_idx]

        # Diff flag: ref!=read (strict SNV filtering)
        if (strand == "+"){
          diff <- as.integer(ref_base != aligned_base | as.character(aligned_base) != "C")
        }else if(strand == "-"){
          diff <- as.integer(ref_base != aligned_base | as.character(aligned_base) != "G")
        }

        # filtering only if insertion is exactly *within* the C site (not adjacent)
        if (ref_idx %in% ins_align_pos & strand == "+") {
          insert = 1
        } else if ((ref_idx - 1) %in% ins_align_pos & strand == "-") {
          insert = 1
        }

        # Nocall flag: unmatched context (e.g. CG, CHG)
        nocall <- 0L
        if (context %in% "CG") {
          if (strand == "+") {
            ref_idx2 <- ref_idx + 1
            if (ref_idx2 >= 1 && ref_idx2 <= length(ref_seq)) {
              ref_next <- ref_seq[ref_idx2]
              aligned_next <- aligned_seq[ref_idx2]
              nocall <- as.integer(ref_next != aligned_next)
            }
          } else if (strand == "-") {
            ref_idx2 <- ref_idx - 1
            if (ref_idx2 >= 1 && ref_idx2 <= length(ref_seq)) {
              ref_prev <- ref_seq[ref_idx2]
              aligned_prev <- aligned_seq[ref_idx2]
              nocall <- as.integer(ref_prev != aligned_prev)
            }
          }

        }
        if (context %in% "CHG") {
          if (strand == "+") {
            if (ref_idx + 2 <= length(ref_seq)) {
              ref_c  <- as.character(ref_seq[ref_idx])
              ref_h  <- as.character(ref_seq[ref_idx + 1])
              ref_g  <- as.character(ref_seq[ref_idx + 2])

              aln_c  <- as.character(aligned_seq[ref_idx])
              aln_h  <- as.character(aligned_seq[ref_idx + 1])
              aln_g  <- as.character(aligned_seq[ref_idx + 2])

              nocall <- 1L
              if (synonymous) {
                if (ref_c == "C" && aln_c == "C" &&
                    ref_g == "G" && aln_g == "G" &&
                    ref_h %in% c("A", "T", "C") && aln_h %in% c("A", "T", "C")) {
                  nocall <- 0L  # context preserved → allow
                }
              } else {
                # strict match
                if (ref_c == aln_c && ref_h == aln_h && ref_g == aln_g) {
                  nocall <- 0L
                }
              }
            }
          } else if (strand == "-") {
            if (ref_idx - 2 >= 1) {
              ref_g  <- as.character(ref_seq[ref_idx])
              ref_h  <- as.character(ref_seq[ref_idx - 1])
              ref_c  <- as.character(ref_seq[ref_idx - 2])

              aln_g  <- as.character(aligned_seq[ref_idx])
              aln_h  <- as.character(aligned_seq[ref_idx - 1])
              aln_c  <- as.character(aligned_seq[ref_idx - 2])

              nocall <- 1L
              if (synonymous) {
                if (ref_g == "G" && aln_g == "G" &&
                    ref_c == "C" && aln_c == "C" &&
                    ref_h %in% c("A", "T", "G") && aln_h %in% c("A", "T", "G")) {
                  nocall <- 0L
                }
              } else {
                if (ref_g == aln_g && ref_h == aln_h && ref_c == aln_c) {
                  nocall <- 0L
                }
              }
            }
          }
        }
        if (context %in% "CHH"){
          if (strand == "+") {
            if (ref_idx + 2 <= length(ref_seq)) {
              ref_c  <- as.character(ref_seq[ref_idx])
              ref_h1  <- as.character(ref_seq[ref_idx + 1])
              ref_h2  <- as.character(ref_seq[ref_idx + 2])

              aln_c  <- as.character(aligned_seq[ref_idx])
              aln_h1  <- as.character(aligned_seq[ref_idx + 1])
              aln_h2  <- as.character(aligned_seq[ref_idx + 2])

              nocall <- 1L
              if (synonymous) {
                if (ref_c == "C" && aln_c == "C" &&
                    ref_h1 %in% c("A", "T", "C") && aln_h1 %in% c("A", "T", "C") &&
                    ref_h2 %in% c("A", "T", "C") && aln_h2 %in% c("A", "T", "C")) {
                  nocall <- 0L  # context preserved → allow
                }
              } else {
                # strict match
                if (ref_c == aln_c && ref_h1 == aln_h1 && ref_h2 == aln_h2) {
                  nocall <- 0L
                }
              }
            }
          } else if (strand == "-") {
            if (ref_idx - 2 >= 1) {
              ref_g  <- as.character(ref_seq[ref_idx])
              ref_h1  <- as.character(ref_seq[ref_idx - 1])
              ref_h2  <- as.character(ref_seq[ref_idx - 2])

              aln_g  <- as.character(aligned_seq[ref_idx])
              aln_h1  <- as.character(aligned_seq[ref_idx - 1])
              aln_h2  <- as.character(aligned_seq[ref_idx - 2])

              nocall <- 1L
              if (synonymous) {
                if (ref_g == "G" && aln_g == "G" &&
                    ref_h1 %in% c("A", "T", "G") && aln_h1 %in% c("A", "T", "G") &&
                    ref_h2 %in% c("A", "T", "G") && aln_h2 %in% c("A", "T", "G")) {
                  nocall <- 0L
                }
              } else {
                if (ref_g == aln_g && ref_h1 == aln_h1 && ref_h2 == aln_h2) {
                  nocall <- 0L
                }
              }
            }
          }
        }

        # mod flag
        mod <- ifelse(insert == 0 & diff == 0 & nocall == 0, 1, 0)

        out_pos  <- c(out_pos,  gp)
        out_prob <- c(out_prob, pr)
        out_modified_base_code  <- c(out_modified_base_code, cd)
        out_mod <- c(out_mod, mod)
        out_insert <- c(out_insert, insert)
        out_delete <- c(out_delete, delete)
        out_diff <- c(out_diff, diff)
        out_nocall <- c(out_nocall, nocall)

        # # DEBUG
        # cat(sprintf("k=%d | start=%d | rp=%d | insert=%d | delete=%d |insert_off=%d | delete_off=%d | gp=%d | ref_base=%s | aln_base=%s\n  | del_len=%d",
        #             k, start, rp, insert, delete, insert_off, delete_off, gp, ref_base, aligned_base, del_len))
      }

    }
    DataFrame(pos=out_pos, prob=out_prob, modType=out_modified_base_code,
              Nmod=out_mod, Ninsert=out_insert, Ndelete=out_delete, Ndiff=out_diff, Nnocall=out_nocall)
  }

  # *5) Helper: Run .computeModifPos using the scanned bam list and return the modified
  # position and proportion values of the sequence as GRanges
  .getCovSequence <- function(i) {
    if (i %in% progress_points) {
      pct <- round(i / N * 100)
      cat(sprintf("[readONTbam] Decoding MM/ML: %d%% (%d/%d reads)\n", pct, i, N))
    }
    # run .computeModifPos and generate the data_frame
    df <- .computeModifPos(
      mm_tag      = sb$tag$MM[[i]],
      ml_tag      = sb$tag$ML[[i]],
      cigar       = sb$cigar[[i]],
      strand      = strands[[i]],
      start       = sb$pos[[i]],
      cyt_pos     = cyt_pos_list[[i]],
      indel_info  = indel_info_list[[i]],
      ref_seq     = ref_seq_list[[i]],
      aligned_seq = aligned_seq_list[[i]],
      context     = context,
      synonymous  = synonymous
    )

    # exclude the rows if they didn't hold the Nmod == 1
    df <- df[df$Nmod == 1, ]
    if (nrow(df) == 0) return(GRanges())

    # wrap up the results as the GRanges with using the index of sequence
    rr <- GRanges(
      seqnames = sb$rname[[i]],
      ranges   = IRanges(start = df$pos, width = 1),
      strand   = strands[[i]]
    )
    mcols(rr)$prob    <- df$prob
    mcols(rr)$modType <- df$modType
    mcols(rr)$read    <- read_order[[i]]

    # ## DEBUG
    # cat("current position i:", i, "\n")
    return(rr)
  }
  # --------------
  # *preparation of the decoding the modified information by
  # creating cytosine position and indel information list
  cat("[readONTbam] Generating cytosine position list and indel_info ...\n")
  cyt_pos_list <- mapply(.getCytPos,
                         read_seq = sb$seq,
                         strand   = strands,
                         SIMPLIFY = FALSE)

  indel_info_list   <- mapply(.getIndels,
                         cig      = sb$cigar,
                         SIMPLIFY = FALSE)

  # --------------
  ## 2) decode every read
  cat("[readONTbam] Start Decoding MM/ML ...\n")
  N <- length(sb$qname)
  progress_points <- round(seq(0, N, length.out = 11)[-1])  # 10%, 20%, ..., 100%

  per_read <- bplapply(seq_len(N), .getCovSequence, BPPARAM = BPPARAM)

  valid_gr <- per_read[sapply(per_read, function(x) is(x, "GRanges") && length(x) > 0)]

  if (length(valid_gr) == 0) {
    all_calls <- GRanges()
  } else {
    all_calls <- suppressWarnings(do.call(c, valid_gr))
  }
  rm(per_read)


  # @@@@ time t4
  t4 <- proc.time()
  cat(sprintf("Step 3 (decode Bam file): %.2f sec\n", (t4-t3)[["elapsed"]]))

  ## 3) initialize ONT_Cm and ONT_C as simple character columns
  mcols(ref_gr)$ONT_Cm <- CharacterList(vector("list", length(ref_gr)))
  mcols(ref_gr)$ONT_C  <- CharacterList(vector("list", length(ref_gr)))

  ref_gr$readsM <- integer(length(ref_gr))
  ref_gr$readsN <- integer(length(ref_gr))


  cat("[readONTbam] Calculating overlaps with reference cytosines ...\n")
  ov_cov <- findOverlaps(ref_gr, all_calls, ignore.strand = FALSE)
  df_cov <- DataFrame(
    qhit   = queryHits(ov_cov),
    read   = all_calls$read[subjectHits(ov_cov)],
    rstrand= as.character(strand(all_calls))[subjectHits(ov_cov)],
    modType = mcols(all_calls)$modType[subjectHits(ov_cov)],
    prob    = mcols(all_calls)$prob[subjectHits(ov_cov)]
  )

  # subsetting the df_cov as the modif preset
  df_cov   <- df_cov[df_cov$modType %in% modif, ]

  # @@@@ time t5
  t5 <- proc.time()
  cat(sprintf("Step 4 (coverage): %.2f sec\n", (t5-t4)[["elapsed"]]))

  ## 4)  save the modified and read‐indices of modified (readsN, ONT_C) in ref_gr
  cat("[readONTbam] Annotating high-confidence modified reads ...\n")

  df_mod   <- df_cov[df_cov$modType %in% modif & df_cov$prob >= prob_thresh, ]
  mod_list <- lapply(split(df_mod$read, df_mod$qhit), unique)
  mod_char_list <- lapply(mod_list, function(x) paste(x, collapse = ","))
  # collapse into semicolon-separated strings
  idx_mod <- as.integer(names(mod_list))

  ref_gr$ONT_Cm[idx_mod] <- CharacterList(mod_char_list)
  ref_gr$readsM[idx_mod] <- lengths(mod_list)

  # @@@@ time t6
  t6 <- proc.time()
  cat(sprintf("Step 5 (ONT_Cm & readsM meta-column): %.2f sec\n", (t6-t5)[["elapsed"]]))

  # 5) save the coverage and read‐indices of unmodified (readsN, ONT_C) in ref_gr
  cat("[readONTbam] Annotating unmodified (covered) reads & Calculating coverage per position...\n")

  df_unmod   <- df_cov[df_cov$modType %in% modif & df_cov$prob < prob_thresh, ]

  unmod_list <- lapply(split(df_unmod$read, df_unmod$qhit), unique)
  cov_list   <- lapply(split(df_cov$read, df_cov$qhit), unique)
  unmod_char_list <- lapply(unmod_list, function(x) paste(x, collapse = ","))
  # collapse into semicolon-separated strings
  idx_unmod <- as.integer(names(unmod_list))
  idx_cov   <- as.integer(names(cov_list))

  ref_gr$ONT_C[idx_unmod] <- CharacterList(unmod_char_list)
  ref_gr$readsN[idx_cov] <- lengths(cov_list)

  # @@@@ time t6
  t7 <- proc.time()
  cat(sprintf("Step 6 (ONT_C & readsN meta-column): %.2f sec\n", (t7-t6)[["elapsed"]]))
  cat("[readONTbam] Done. Returning annotated GRanges.\n")
  cat(sprintf("Total: %.2f sec\n", (t7-t1)[["elapsed"]]))

  return(ref_gr)

}

#' This function asses an expression and if this is FALSE stops the execution
#' and prints a message
#'
#' @title .stopIfNotAll
#' @param expr an array of logical expressions
#' @param errorMsg the error message to be printed
#'
#' @author Radu Zabet
.stopIfNotAll <- function(exprs, errorMsg) {
  for(expr in exprs){
    if (! expr)
      stop(errorMsg, call. = FALSE)
  }
}

#' This function checks is a variable is an integer number
#'
#' @title .is.integer
#' @param x the variable
#' @param positive a logical variable indicating whether to check if the number
#' is a positive integer
#' @return a logical number whether the variable x is an integer or not
#'
#' @author Radu Zabet
.isInteger <- function(x, positive = FALSE){
  isInteger <- TRUE
  if (is.null(x)){
    isInteger <- FALSE
  } else{
    if (!is.numeric(x)){
      isInteger <- FALSE
    } else{
      if(x%%1!=0){
        isInteger <- FALSE
      } else{
        if(positive & x < 0){
          isInteger <- FALSE
        }
      }
    }
  }
  return(isInteger)
}


#' This function that prints a GenomicRanges object
#'
#' @title Print GenomicRanges
#' @param gr the GenomicRanges object
#' @return a vector of type \code{character} with template chr:start..end
#'
#' @author Radu Zabet
.printGenomicRanges <- function(gr){

  .stopIfNotAll(c(!is.null(gr),
                  is(gr, "GRanges")),
                " gr is a GenomicRanges object");
  result=c();
  for(index in 1:length(gr)){
    result <- c(result,
                paste(seqnames(gr)[index],":",start(gr)[index],"..",end(gr)[index],sep=""))
  }
  return(result)
}

#' This function checks whether the argument is a vector containing colours
#'
#' @title Is color
#' @param x the vector to be validated
#' @param minLength the  minimum length of the vector. If NULL the minimum
#' length is 1
#' @return a \code{logical} value indicating whether \code{x} is a vector
#' containing only colors
#'
#' @author Radu Zabet
.isColor <- function(x, minLength=NULL){
  isColor <- TRUE
  if(is.null(x)){
    isColor <- FALSE
  }

  if(is.null(minLength)){
    minLength = 1
  }

  if(isColor & length(x) < minLength){
    isColor <- FALSE
  }

  if(isColor){
    for(i in 1:length(x)){
      if(!(x[i]%in% grDevices::colors()) & length(grep("^#[0-9A-Fa-f]{6}$", x[i])) < 1){
        isColor <- FALSE
      }
    }
  }
  return(isColor)
}


#' Returns a \code{\link{GRanges}} object spanning from the first cytocine until
#' the last one on each chromosome
#'
#' @title Get whole chromosomes from methylation data
#' @param methylationData the methylation data stored as a \code{\link{GRanges}}
#' object with four metadata columns (see \code{\link{methylationDataList}}).
#' @return a \code{\link{GRanges}} object will all chromosomes.
#'
#' @examples
#' # load the methylation data
#' data(methylationDataList)
#'
#' # get all chromosomes
#' chromosomes <- getWholeChromosomes(methylationDataList[["WT"]])
#'
#' @author Nicolae Radu Zabet
#' @export
getWholeChromosomes <- function(methylationData){
  max <- c()
  min <- c()
  seqnames <- c()
  for(chr in levels(seqnames(methylationData))){
    indexes=which(as.character(seqnames(methylationData)) == chr)
    if(length(indexes) > 0){
      seqnames <- c(seqnames, chr)
      max <- c(max, max(start(ranges(methylationData))[indexes]))
      min <- c(min, min(start(ranges(methylationData))[indexes]))
    }
  }

  regions <- GRanges(seqnames = Rle(seqnames), ranges   = IRanges(min,max))

  return(regions)
}

#' Checks whether the passed parameter has the correct format for methylation data
#'
#' @title Validate methylation data
#' @param methylationData the methylation data stored as a \code{\link{GRanges}}
#' object containing all the replicates.
#'
#' @author Alessandro Pio Greco and Nicolae Radu Zabet
.validateMethylationData <- function(methylationData, variableName="methylationData",
                                     manageDuplicates = "mean"){
  .stopIfNotAll(c(!is.null(methylationData),
                  is(methylationData, "GRanges")),
                " methylationData needs to be a GRanges object")
  .stopIfNotAll(c(ncol(mcols(methylationData)[grepl("reads",
                  names(mcols(methylationData)))])%%2 == 0,
                  length(methylationData) > 0,
                  any(grepl("context", names(mcols(methylationData)))) == TRUE,
                  any(grepl("trinucleotide_context", names(mcols(methylationData)))) == TRUE),
                paste(" ",variableName," the object does not contain the correct metadata columns", sep=" "))
  if(any(duplicated(methylationData)) == TRUE){
    indexesDuplicated <- which(ranges(methylationData) ==
                               ranges(methylationData)[duplicated(methylationData)])
    checkMetadataEqual <- all(mcols(methylationData[indexesDuplicated[1:(length(indexesDuplicated)/2)]]) ==
                                mcols(methylationData[indexesDuplicated[((length(indexesDuplicated)/2)+1):
                                                              length(indexesDuplicated)]]))
    if(all(checkMetadataEqual) == TRUE){
      if(manageDuplicates == "mean"){
      cat("Cytosines that were duplicated and had the different metadata columns were merged by meaning readings \n")


    } else if(manageDuplicates == "sum"){


    } else if(manageDuplicates == "discard"){
      cat("Cytosines that were duplicated (",indexesDuplicated, ") and had the same metadata columns were discarded \n", sep = " ")
      methylationData <- unique(methylationData)
    }


    } else{
      stop(" context or trinucleotide context on duplicated cytosines (", indexesDuplicated, ") are not equal")
    }

  }
}


#' Checks whether the passed parameter has the correct format for methylation data
#'
#' @title Validate methylation data
#' @param methylationData the methylation data stored as a \code{\link{GRanges}}
#' object with six metadata columns (see \code{\link{methylationData}}).
#'
#' @author Radu Zabet
.validateMethylationDataList <- function(methylationDataList){
  .stopIfNotAll(c(!is.null(methylationDataList),
                is(methylationDataList, "GRangesList"),
                length(methylationDataList) > 0),
                " methylationDataList needs to be a GRangesList object")

  for(i in 1:length(methylationDataList)){
    .stopIfNotAll(c(.validateMethylationData(methylationDataList[[i]])),
                 paste(" element ", i," of the methylationDataList is incorrect", sep=""))
  }

}

#' Checks whether the passed parameter is the context
#'
#' @title Validate context
#' @param context the context in which the DMRs are computed (\code{"CG"},
#' \code{"CHG"} or \code{"CHH"})
#' @param length the expected length of the vector. If NULL any length is
#' allowed
#'
#' @author Radu Zabet
.validateContext <- function(context, length=NULL){
  .stopIfNotAll(c(!is.null(context), all(is.character(context)),
                  all(context %in% c("CG","CHG","CHH"))),
                " context can be only CG,CHG or CHH")
  if(!is.null(length)){
    .stopIfNotAll(c(is.numeric(length), length(context) == length),
                  paste(" context needs to contain exactly ", length," elements", sep=""))
  }
}



#' Checks whether the passed parameter is the statistical test
#'
#' @title Validate statistial test
#' @param test the statistical test used to call DMRs (\code{"fisher"} for
#' Fisher's exact test or \code{"score"} for Score test).
#'
#' @author Radu Zabet
.validateStatisticalTest <- function(test){
  .stopIfNotAll(c(!is.null(test), is.character(test), length(test) == 1, test %in% c("fisher","score")),
                " test needs to be one of the following \"fisher\" for Fisher's exact test or \"score\" for Score test")
}


#' Checks whether the passed parameter is a \code{\link{GRanges}} object
#'
#' @title Validate GRanges
#' @param regions a \code{\link{GRanges}} object. If \code{NULL} and
#' \code{generateGenomeWide} is true it uses the \code{methylationData} to
#' compute the regions and returns this \code{\link{GRanges}} object
#' @param methylationData the methylation data stored as a \code{\link{GRanges}}
#' object with six metadata columns (see \code{\link{methylationData}}).
#' @param length the expected length of the vector. If \code{NULL} any length is
#' allowed.
#' @param generateGenomeWide logical value to indicate whether to compute the
#' regions that span over all the \code{methylationData}
#' @return a \code{\link{GRanges}} object representing the regions
#'
#' @author Radu Zabet
.validateGRanges <- function(regions, methylationData, length=NULL, generateGenomeWide=TRUE, variableName="regions", minLength=0){

  if(is.null(regions) & generateGenomeWide){
    if(is(methylationData, "GRangesList") & length(methylationData) > 0){
      regions <- NULL
      for(i in 1:length(methylationData)){
        if(is.null(regions)){
          regions <- getWholeChromosomes(methylationData[[i]])
        } else{
          regions <- union(regions, getWholeChromosomes(methylationData[[i]]))
        }
      }
    } else if(is(methylationData, "GRanges")){
      regions <- getWholeChromosomes(methylationData)
    }
  }

  .stopIfNotAll(c(!is.null(regions),
                  is(regions, "GRanges")),
                paste(" ",variableName," neads to be a GRanges object. If NULL, the DMRs are computed genome-wide.",sep=""))

  if(!is.null(length)){
    .stopIfNotAll(c(is.numeric(length), length(regions) == length),
                  paste(" ",variableName," needs to contain exactly ", length," elements", sep=""))
  }
  if(!is.null(minLength)){
    .stopIfNotAll(c(is.numeric(minLength), length(regions) >= minLength),
                  paste(" ",variableName," needs to contain more than ", minLength," elements", sep=""))
  }
  return(regions)
}

#' Checks whether the passed parameter is a \code{GRangesList} object that
#' represents the methylation profile
#'
#' @title Validate methylation profile
#' @param methylationProfile a \code{GRangesList} object
#' (see \code{\link{computeMethylationProfile}}).
#'
#' @author Radu Zabet
.validateMethylationProfile <- function(methylationProfile){
  .stopIfNotAll(c(!is.null(methylationProfile),
                is(methylationProfile, "GRangesList")),
                " methylationProfile needs to be a GRangesList")


  for(i in 1:length(methylationProfile)){
    .stopIfNotAll(c(!is.null(methylationProfile[[i]]),
                  is(methylationProfile[[i]], "GRanges")),
                  paste(" element ",i," of the methylationProfile is not a GRanges object", sep=""))
    .stopIfNotAll(c(ncol(mcols(methylationProfile[[i]])) == 5,
                    length(methylationProfile[[i]]) > 0),
                  paste(" element ",i," of the methylationProfile is not a GRanges object with five metadata columns (see computeMethylationProfile function).", sep=""))
  }

}

#' Checks whether the passed parameter is the statistical test for detecting Co-methylation
#'
#' @title Validate statistial test for Co-methylation
#' @param test the statistical test used to call Co-methylation (\code{"fisher"} for
#' Fisher's exact test, \code{"score"} for Score test or \code{"permutation"} for Permutation test).
#' @keywords internal
#' @author Radu Zabet and Young Jun Kim
.validateCoMethylationStatTest <- function(test){
  .stopIfNotAll(c(!is.null(test), is.character(test), length(test) == 1, test %in% c("fisher","score","permutation","binom")),
                " Test needs to be one of the following \"fisher\" for Fisher's exact test, \"score\" for Score test or \"permutation\" for Permutation test" )
}

#' Checks whether the passed parameter is the alternative for detecting Co-methylation
#'
#' @title Validate alternative value for Co-methylation
#' @param test the alternative valuet used to Fisher exact's test must be one of
#' \code{"two.sided"}, \code{"greater"} or \code{"less"}. You can specify just
#' the initial letter.
#' @keywords internal
#' @author Radu Zabet and Young Jun Kim
.validateCoMethylationAlternative <- function(alternative){
  .stopIfNotAll(c(!is.null(alternative), is.character(alternative), length(alternative) == 1, alternative %in% c("two.sided","greater","less")),
                " Alternative needs to be one of the following \"two.sided\", \"greater\" or \"less\" for Fisher's exact test" )
}

#' Checks whether the passed parameter is the modified context
#'
#' @title Validate modified context
#' @param test the  modified context used to call methylation information from bamfile must be one of
#' sequence context for \code{selectCytosine()} (e.g. \code{"CG"}, \code{"CHG"}, \code{"CHH"})
#' @keywords internal
#' @author Radu Zabet and Young Jun Kim
.validateContext <- function(context){
  .stopIfNotAll(c(!is.null(context), is.character(context), context %in% c("CG","CHG","CHH")),
                " Modified context needs to be one of the following \"CG\", \"CHG\" or \"CHH\"" )
}


#' @title Validate BAM Filename
#' @description
#' Checks that `bamfile` exists and has a ".bam" extension.
#'
#' @param bamfile Character scalar. Path to a BAM file.
#' @return Invisibly `TRUE` if file exists and extension is ".bam"; otherwise errors.
#' @keywords internal
#' @author Radu Zabet and Young Jun Kim
.validateBamfile <- function(bamfile){
  # stop if bamfile directory is not include the .bam file format
  .stopIfNotAll(c(!is.null(bamfile),is.character(bamfile), file.exists(bamfile), endsWith(tolower(bamfile), ".bam")),
                "bamfile directory name should need .bam extension")
}

#' @title Validate BSgenome Object
#' @description
#' Ensures `genome` inherits from BSgenome and that its package is installed.
#'
#' @param genome A BSgenome object (e.g. BSgenome.Hsapiens.UCSC.hg38).
#' @return Invisibly `TRUE` if valid; otherwise throws an error.
#' @keywords internal
#' @author Radu Zabet and Young Jun Kim
.validateGenome <- function(genome) {
  # 1) must be a BSgenome object
  if (!inherits(genome, "BSgenome")) {
    stop(
      "`genome` must be supplied as a BSgenome object, not as a string literal;\n",
      "please call e.g. `.validateGenome(BSgenome.Hsapiens.UCSC.hg38)` without quotes.",
      call. = FALSE
    )
  }

  ## 2) extract its package name
  pkg <- genome@pkgname

  ## 3) make sure BSgenome machinery is there
  if (!requireNamespace("BSgenome", quietly=TRUE)) {
    stop("please install the BSgenome package before specifying a genome", call. = FALSE)
  }

  ## 4) get the official list of BSgenome packages
  avail <- tryCatch(
    {suppressMessages(suppressPackageStartupMessages(BSgenome::available.genomes()))},
    error = function(e) NULL
  )

  ## 5) fallback: any installed BSgenome.* packages
  if (is.null(avail) || !is.character(avail)) {
    avail <- grep("^BSgenome\\.", rownames(utils::installed.packages()), value=TRUE)
  }

  ## 6) error if this genome isn't known
  if (!pkg %in% avail) {
    stop(
      sprintf(
        "Loaded BSgenome object has package name '%s',\n  but that package is not among the installed/available BSgenome.* packages.\n  Known packages: %s%s",
        pkg,
        paste(head(avail, 3), collapse=", "),
        if (length(avail)>3) ", _" else ""
      ),
      call. = FALSE
    )
  }
  invisible(TRUE)
}

#' @title Validate Chromosome Names
#' @description
#' Checks that `chr` entries are present in the supplied BSgenome.
#'
#' @param chr Character vector of chromosome names.
#' @param genome A validated BSgenome object.
#' @return Invisibly `TRUE` if all `chr` are in `seqnames(genome)`; otherwise errors.
#' @keywords internal
#' @author Radu Zabet and Young Jun Kim
.validateChromosome <- function(chr, genome){
  # check the Chromosome is included in the genome
  .stopIfNotAll(c(is.character(chr), chr %in% seqnames(genome)),
                " Chromosome should select same code as the BSgenome provided" )
}

#' @title Validate Modification Tag in BAM
#' @description
#' Scans the BAM for MM/ML tags in the specified region (or whole file)
#' and checks that `modif` is one of the observed modification codes.
#'
#' @param modif Character scalar. The MM tag code to validate (e.g. "C+m?").
#' @param bamfile Character path to BAM (must exist plus .bai index).
#' @param chr Optional character vector of chromosome(s) to restrict scan.
#' @param genome A BSgenome, used to determine seqlengths if `chr` is set.
#' @return Invisibly `TRUE` if `modif` is found among the MM tags; otherwise errors.
#' @author Radu Zabet and Young Jun Kim
.validateModif <- function(modif, bamfile, chr, genome){
  # stop if modif is not included in the tags in the bam
  if (!file.exists(paste0(bamfile,".bai")))
    Rsamtools::indexBam(bamfile)
  if (!is.null(chr) && length(chr) > 0) {
    which_region <- GenomicRanges::GRanges(seqnames = chr,
      ranges = IRanges::IRanges(1, seqlengths(genome)[chr]))
    param <- Rsamtools::ScanBamParam(tag = c("MM","ML"), which = which_region)
  } else {
    param <- Rsamtools::ScanBamParam(tag = c("MM","ML"))
  }

  bam <- Rsamtools::scanBam(bamfile, param = param)[[1]]
  bam_tag <- unique(bam$tag$MM)
  bam_tag <- bam_tag[!is.na(bam_tag)]

  if (length(bam_tag) == 0) {
    stop("No MM tag found in the BAM file for the selected region/chromosome.")
  }

  # flatten all tags
  bam_tag_split <- unlist(strsplit(as.character(bam_tag), ";", fixed = TRUE))
  bam_tag_split <- bam_tag_split[nzchar(bam_tag_split)]  # remove empty strings


  code_list <- character()
  for (i in seq_along(bam_tag_split)) {
    parts <- strsplit(bam_tag_split[[i]], ",", fixed = TRUE)[[1]]
    code_list <- c(code_list, parts[1])
  }

  codes <- unique(code_list)
  .stopIfNotAll(c(!is.null(modif), is.character(modif), length(modif) == 1, modif %in% codes),
                c("Modified context must be one of: ", paste(shQuote(codes), collapse=", " )))
}

#' @title Validate or Auto-Select a BiocParallelParam
#' @description
#' If `BPPARAM` is `NULL`, picks `SnowParam` or `MulticoreParam` based on env vars
#' or OS; otherwise checks class and fork compatibility.
#'
#' @param BPPARAM A \code{BiocParallelParam} or `NULL`.
#' @param progressbar Logical; whether to show a progress bar.
#' @return A valid `BiocParallelParam` object.
#' @keywords internal
#' @author Radu Zabet and Young Jun Kim
.validateBPPARAM <- function(BPPARAM = NULL, progressbar = FALSE) {
  if (is.null(BPPARAM)) {
    BPPARAM <- .chooseBPPARAM(progressbar = progressbar)
    return(BPPARAM)
  }
  # must inherit from BiocParallelParam
  if (!inherits(BPPARAM, "BiocParallelParam")) {
    stop("`BPPARAM` must be a BiocParallelParam object, got: ",
         paste(class(BPPARAM), collapse = "/"))
  }
  # no fork on Windows
  if (.Platform$OS.type == "windows" &&
      methods::is(BPPARAM, "MulticoreParam")) {
    stop("MulticoreParam() (fork) not supported on Windows; please use SnowParam().")
  }
  invisible(BPPARAM)
}

#' @title Auto-choose BiocParallelParam Backend
#' @description
#' 1) If `NSLOTS` or `SLURM_CPUS_ON_NODE` are set, uses `SnowParam()` with that many workers.
#' 2) On Windows, `SnowParam(detectCores())`.
#' 3) Otherwise on Unix, `MulticoreParam(detectCores())`.
#'
#' @param workers Optional integer to override detected cores.
#' @param progressbar Logical; show progress bar.
#' @param cluster_type Type passed to `SnowParam()`, defaults to "SOCK".
#' @param ... Further args passed to `SnowParam` or `MulticoreParam`.
#' @return A `SnowParam` or `MulticoreParam` configured with the desired workers.
#' @keywords internal
#' @author Radu Zabet and Young Jun Kim
.chooseBPPARAM <- function(workers = NULL, progressbar = FALSE, cluster_type = "SOCK", ...) {
  # 1) HPC scheduler slots
  nslots <- as.integer(Sys.getenv("NSLOTS", unset = NA))
  if (!is.na(nslots) && nslots > 1) {
    cat("Detected SGE/PBS (NSLOTS=", nslots, ") -> using SnowParam()")
    return(SnowParam(workers = nslots, type = cluster_type,
                     progressbar = progressbar, ...))
  }
  # SLURM
  cpus <- as.integer(Sys.getenv("SLURM_CPUS_ON_NODE", unset = NA))
  if (!is.na(cpus) && cpus > 1) {
    cat("Detected SLURM (SLURM_CPUS_ON_NODE=", cpus, ") -> using SnowParam()")
    return(SnowParam(workers = cpus, type = cluster_type,
                     progressbar = progressbar, ...))
  }

  # 2) local fallback
  os <- .Platform$OS.type
  ncore <- if (is.null(workers)) parallel::detectCores(logical = FALSE) else workers
  if (os == "windows") {
    cat("Detected Windows -> using SnowParam() with ", ncore, " workers")
    return(SnowParam(workers = ncore, progressbar = progressbar, ...))
  }
  # 3) Unix
  cat("Local Unix -> using MulticoreParam() with forking and ", ncore, " workers")
  MulticoreParam(workers = ncore, progressbar = progressbar, ...)
}

#' This function takes as input a CX report file produced by Bismark
#' and returns a \code{\link{GRanges}} object with four metadata columns
#' The file represents the bisulfite sequencing methylation data.
#'
#' @title Read Bismark
#' @param file The filename (including path) of the methylation
#' (CX report generated by Bismark) to be read.
#' @return the methylation data stored as a \code{\link{GRanges}}
#' object with four metadata columns (see \code{\link{methylationDataList}}).
#' @examples
#'
#' # load methylation data object
#' data(methylationDataList)
#'
#' # save the one datasets into a file
#' saveBismark(methylationDataList[["WT"]], "chr3test_a_thaliana_wt.CX_report")
#'
#' # load the data
#' methylationDataWT <- readBismark("chr3test_a_thaliana_wt.CX_report")
#'
#' #check that the loading worked
#' all(methylationDataWT == methylationDataList[["WT"]])
#'
#' @author Nicolae Radu Zabet and Jonathan Michael Foonlan Tsang
#'
#' @export
readBismark <- function(file) {
  .stopIfNotAll(c(!is.null(file),
                  is.character(file),
                  file.exists(file)),
                " file does not exist")

  cat("Reading file: ",file,"\n",sep="")
  cx <- scan(
    file = file,
    what = list(character(), integer(), character(), integer(), integer(), character(), character()),
    sep  = "\t",
    skip = 0);
  chrs <- unique(cx[[1]])
  chrs <- chrs[order(chrs)]
  dat <- data.frame(
    chr = factor(cx[[1]], levels = chrs),
    pos = as.integer(cx[[2]]),
    strand = factor(cx[[3]], levels =c("+","-")),
    context = factor(cx[[6]], levels = c("CG","CHG","CHH")),
    trinucleotide_context = cx[[7]],
    readsM = as.integer(cx[[4]]),
    readsN = as.integer((cx[[4]]+cx[[5]])));
  #dat <- dat[with(dat, order(chr, pos, strand)), ];
  dat <- GRanges(seqnames = dat$chr,
                 ranges   = IRanges(start = dat$pos, end = dat$pos),
                 strand   = dat$strand,
                 context = dat$context,
                 readsM = dat$readsM,
                 readsN = dat$readsN,
                 trinucleotide_context = dat$trinucleotide_context);
  cat("Finished reading file: ",file,"\n",sep="")
  return(dat)
}

#' This function takes as input a vector of CX report file produced by Bismark
#' and returns a \code{\link{GRanges}} object with four metadata columns
#' (see \code{\link{methylationDataList}}). The file represents the pooled
#' bisulfite sequencing data.
#'
#' @title Read Bismark pool
#' @param files The filenames (including path) of the methylation
#' (CX report generated with Bismark) to be read
#' @return the methylation data stored as a \code{\link{GRanges}}
#' object with four metadata columns (see \code{\link{methylationDataList}}).
#' @examples
#'
#' # load methylation data object
#' data(methylationDataList)
#'
#' # save the two datasets
#' saveBismark(methylationDataList[["WT"]],
#'            "chr3test_a_thaliana_wt.CX_report")
#' saveBismark(methylationDataList[["met1-3"]],
#'            "chr3test_a_thaliana_met13.CX_report")
#'
#' # reload the two datasets and pool them
#' filenames <- c("chr3test_a_thaliana_wt.CX_report",
#'                "chr3test_a_thaliana_met13.CX_report")
#' methylationDataPool <- readBismarkPool(filenames)
#'
#' @author Nicolae Radu Zabet and Jonathan Michael Foonlan Tsang
#'
#' @export
readBismarkPool <- function(files) {
  .stopIfNotAll(c(!is.null(files),
                  length(files) > 0),
                " files is a vector containing filenames")

  #read the data
  data <- GRangesList()
  for(i in 1:length(files)){
    data <- c(data, GRangesList(readBismark(files[i])))
  }

  #pool the data
  pooledData <- poolMethylationDatasets(data)

  return(pooledData)
}

#' This function takes as input a \code{\link{GRanges}} object generated with
#' \code{\link{readBismark}} and saves the output to a file using
#' Bismark CX report format.
#'
#' @title Save Bismark
#' @param methylationData the methylation data stored as a \code{\link{GRanges}}
#' object with four metadata columns (see \code{\link{methylationDataList}}).
#' @param filename the filename where the data will be saved.
#' @return Invisibly returns \code{NULL}
#' @examples
#'
#' # load methylation data object
#' data(methylationDataList)
#'
#' # save one dataset to a file
#' saveBismark(methylationDataList[["WT"]], "chr3test_a_thaliana_wt.CX_report")
#'
#' @author Nicolae Radu Zabet
#'
#' @export
saveBismark <- function(methylationData, filename){
  .stopIfNotAll(c(!is.null(filename),
                  is.character(filename),
                  length(filename)==1),
                " filename should specify the filename, where the methylation data will be saved")

  .validateMethylationData(methylationData)



  chrs <- as.character(seqnames(methylationData))
  start <- start(methylationData)
  strand <- as.character(strand(methylationData))

  readsM <- methylationData$readsM
  readsU <- methylationData$readsN - methylationData$readsM

  context <- methylationData$context
  trinucleotide_context <-  methylationData$trinucleotide_context

  dataFile <- data.frame("chr"=chrs, "position"=start, "strand"=strand, "count methylated"=readsM, "count unmethylated"=readsU, "context"=context, "trinucleotide context"=trinucleotide_context)
  write.table(dataFile, file=filename, quote=FALSE, row.names = FALSE, col.names = FALSE, sep  = "\t")
  invisible(NULL)
}






#' This function takes as input two \code{\link{GRanges}} objects of CX reports
#' (Bismark) for two conditions and  returns a \code{\link{GRanges}} object with
#' five metadata columns
#'
#' @title Join methylation data
#' @param cx1 a \code{\link{GRanges}} object of CX reports (Bismark) with three
#' metadata columns the cytosine context (CG, CHG or CHH), the number of
#' methylated reads and the total number of reads at that position in condition
#' 1
#' @param cx2 a \code{\link{GRanges}} object of CX reports (Bismark) with three
#' metadata columns the cytosine context (CG, CHG or CHH), the number of
#' methylated reads and the total number of reads at that position in condition
#' 2
#' @return a the methylation data stored as a \code{\link{GRanges}}
#' object with six metadata columns (see \code{\link{methylationDataList}}).
.joinMethylationData <- function(cx1, cx2, sample1_name = "Sample1", sample2_name = "Sample2"){
  overlaps <- findOverlaps(cx1, cx2)
  indexes  <- which(!duplicated(queryHits(overlaps)))
  i1 <- queryHits(overlaps)[indexes]
  i2 <- subjectHits(overlaps)[indexes]

  methylData <- GRanges(
    seqnames = seqnames(cx1)[i1],
    ranges   = ranges(cx1)[i1],
    strand   = strand(cx1)[i1],
    context  = cx1$context[i1],
    trinucleotide_context = cx1$trinucleotide_context[i1],
    readsM1 = cx1$readsM[i1], readsN1 = cx1$readsN[i1],
    readsM2 = cx2$readsM[i2], readsN2 = cx2$readsN[i2]
  )

  hasCm <- "ONT_Cm" %in% names(mcols(cx1)) || "ONT_Cm" %in% names(mcols(cx2))
  hasC  <- "ONT_C"  %in% names(mcols(cx1)) || "ONT_C"  %in% names(mcols(cx2))

  if (hasCm) {
    ont1_cm <- as.list(cx1$ONT_Cm[i1])
    ont2_cm <- as.list(cx2$ONT_Cm[i2])
    mcols(methylData)$ONT_Cm <- CharacterList(
      mapply(function(a, b) {
        a <- a[!is.na(a) & a != ""]
        b <- b[!is.na(b) & b != ""]
        c(if(length(a)) paste0(sample1_name, "_", a),
                 if(length(b)) paste0(sample2_name, "_", b))
      }, ont1_cm, ont2_cm, SIMPLIFY = FALSE)
    )
  }
  if (hasC) {
    ont1_c <- as.list(cx1$ONT_C[i1])
    ont2_c <- as.list(cx2$ONT_C[i2])
    mcols(methylData)$ONT_C <- CharacterList(
      mapply(function(a, b) {
        a <- a[!is.na(a) & a != ""]
        b <- b[!is.na(b) & b != ""]
        c(if(length(a)) paste0(sample1_name, "_", a),
                 if(length(b)) paste0(sample2_name, "_", b))
      }, ont1_c, ont2_c, SIMPLIFY = FALSE)
    )
  }
  return(methylData)
}

#' This function pools together multiple methylation datasets.
#'
#' @title Pool methylation data
#' @param methylationDataList a \code{\link{GRangesList}} object where each
#' element of the list is a \code{\link{GRanges}} object with the methylation
#' data in the corresponding condition (see \code{\link{methylationDataList}}).
#' @return the methylation data stored as a \code{\link{GRanges}}
#' object with four metadata columns (see \code{\link{methylationDataList}}).
#' If the Granges are from ONT datasets, its have six metedata columns include as
#' ONT_Cm and ONT_C (see \code{\link{readONTbam}}).
#' @examples
#' # load methylation data object
#' data(methylationDataList)
#'
#' # pools the two datasets together
#' pooledMethylationData <- poolMethylationDatasets(methylationDataList)
#'
#' @author Nicolae Radu Zabet updated by Young Jun Kim
#'
#' @export
poolMethylationDatasets <- function(methylationDataList){
  .validateMethylationDataList(methylationDataList)

  n <- length(methylationDataList)
  sample_names <- names(methylationDataList)
  if (is.null(sample_names)) sample_names <- paste0("Sample", seq_len(n))

  pooledMethylationData <- methylationDataList[[1]]
  if (n > 1) {
    for (i in 2:n) {
      cat("joining two datasets ...\n")
      buffer <- .joinMethylationData(pooledMethylationData, methylationDataList[[i]],
                                     sample1_name = sample_names[i-1],
                                     sample2_name = sample_names[i])
      if(!is.null(buffer$ONT_Cm)){
        pooledMethylationData <- GRanges(seqnames = seqnames(buffer),
                                         ranges   = ranges(buffer),
                                         strand   = strand(buffer),
                                         context  = buffer$context,
                                         trinucleotide_context = buffer$trinucleotide_context,
                                         ONT_Cm   = buffer$ONT_Cm,
                                         ONT_C    = buffer$ONT_C,
                                         readsM  = (buffer$readsM1 + buffer$readsM2),
                                         readsN  = (buffer$readsN1 + buffer$readsN2))
      } else{
        pooledMethylationData <- GRanges(seqnames = seqnames(buffer),
                                         ranges   = ranges(buffer),
                                         strand   = strand(buffer),
                                         context  = buffer$context,
                                         trinucleotide_context = buffer$trinucleotide_context,
                                         readsM  = (buffer$readsM1 + buffer$readsM2),
                                         readsN  = (buffer$readsN1 + buffer$readsN2))
      }
      cat("the sum was performed ...\n")
    }
  }
  return(pooledMethylationData)
}

#' This function pools together two methylation datasets.
#'
#' @title Pool two methylation datasets
#' @param methylationData1 a \code{\link{GRanges}} object with the methylation
#' data (see \code{\link{methylationDataList}}).
#' @param methylationData2 a \code{\link{GRanges}} object with the methylation
#' data (see \code{\link{methylationDataList}}).
#' @param sample1_name the label used for sample 1.
#' @param sample2_name the label used for sample 2.
#' @return the methylation data stored as a \code{\link{GRanges}}
#' object with four metadata columns (see \code{\link{methylationDataList}}).
#' If the Granges are from ONT datasets, its have six metedata columns include as
#' ONT_Cm and ONT_C (see \code{\link{readONTbam}}).
#'
#' @examples
#' # load methylation data object
#' data(methylationDataList)
#'
#' # save the two datasets together
#' pooledMethylationData <- poolTwoMethylationDatasets(methylationDataList[[1]],
#'                          methylationDataList[[2]])
#'
#' @author Nicolae Radu Zabet updated by Young Jun Kim
#'
#' @export
poolTwoMethylationDatasets <- function(methylationData1, methylationData2,
                                       sample1_name = NULL, sample2_name = NULL) {
  .validateMethylationData(methylationData1)
  .validateMethylationData(methylationData2)
  if (is.null(sample1_name)) sample1_name <- deparse(substitute(methylationData1))
  if (is.null(sample2_name)) sample2_name <- deparse(substitute(methylationData2))

  cat("joining two datasets ...\n")
  buffer <- .joinMethylationData(methylationData1, methylationData2,
                                 sample1_name = sample1_name, sample2_name = sample2_name)
  if(!is.null(buffer$ONT_Cm)){
    pooledMethylationData <- GRanges(seqnames = seqnames(buffer),
                                     ranges   = ranges(buffer),
                                     strand   = strand(buffer),
                                     context  = buffer$context,
                                     trinucleotide_context = buffer$trinucleotide_context,
                                     ONT_Cm   = buffer$ONT_Cm,
                                     ONT_C    = buffer$ONT_C,
                                     readsM  = (buffer$readsM1 + buffer$readsM2),
                                     readsN  = (buffer$readsN1 + buffer$readsN2))
  } else {
    pooledMethylationData <- GRanges(seqnames = seqnames(buffer),
                                    ranges   = ranges(buffer),
                                    strand   = strand(buffer),
                                    context  = buffer$context,
                                    trinucleotide_context = buffer$trinucleotide_context,
                                    readsM  = (buffer$readsM1 + buffer$readsM2),
                                    readsN  = (buffer$readsN1 + buffer$readsN2))
  }
  cat("the sum was performed ...\n")
  return(pooledMethylationData)
}

#' This function draws the genetic elements over a plot
#'
#' @title Plot genetic elements
#' @param gff a \code{\link{GRanges}} object with all elements usually imported
#' from a GFF3 file
#' @param region the location on the genome where to draw the genetic elements
#' @param col the colors used to draw the exons and the transposable elemnets
#'
#' @author Radu Zabet
.plotGeneticElements <- function(gff, region, col){

  seqname <- seqnames(region)
  minPos <- start(region)
  maxPos <- end(region)


  # Select the genes that lie in the region of interest.
  gff <- gff[queryHits(findOverlaps(gff, region))]
  # Chop off the ends of anything sticking out...
  start(gff) <- pmax(start(gff), minPos)
  end(gff)   <- pmin(end(gff), maxPos)

  genes <- gff[gff$type == 'gene']
  genesPos <- genes[strand(genes) == '+' | strand(genes) == '*']
  genesNeg <- genes[strand(genes) == '-' | strand(genes) == '*']
  exons <- gff[gff$type == 'exon']
  exons <- exons[overlapsAny(exons, genes)];
  exonsPos <- exons[strand(exons) == '+' | strand(exons) == '*']
  exonsNeg <- exons[strand(exons) == '-' | strand(exons) == '*']

  transposons <- gff[gff$type == 'transposable_element'];
  transposonsPos <- transposons[strand(transposons) == '+' | strand(transposons) == '*'];
  transposonsNeg <- transposons[strand(transposons) == '-' | strand(transposons) == '*'];

  negativeStrandPosition <- -0.175
  positiveStrandPosition <- -0.075


  text(maxPos + (maxPos-minPos)/100, positiveStrandPosition, '+');
  text(maxPos + (maxPos-minPos)/100, negativeStrandPosition, '-');
  lines(c(minPos, maxPos),c(-0.14,-0.14), lty=1, lwd=0.75, col="black")

  if(length(genesPos)>0) {
    segments(start(genesPos), positiveStrandPosition, end(genesPos), positiveStrandPosition);
    text(start(genesPos), -0.115, genesPos$ID, pos=4, cex=0.5);
  }
  if(length(genesNeg)>0) {
    segments(start(genesNeg), -0.175, end(genesNeg), negativeStrandPosition);
    text(start(genesNeg), -0.23, genesNeg$ID, pos=4, cex=0.5);
  }
  if(length(exonsPos)>0) {
    rect(start(exonsPos), -0.05, end(exonsPos), -0.09, col=col[1], border = NA);
  }
  if(length(exonsNeg)>0) {
    rect(start(exonsNeg), -0.16, end(exonsNeg), -0.2, col=col[1], border = NA);
  }
  if(length(transposonsPos)>0) {
    rect(start(transposonsPos), -0.05, end(transposonsPos), -0.09, col=col[2], border = col[2], density=30, angle=30);
    text(start(transposonsPos), -0.115, transposonsPos$ID, pos=4, cex=0.5);
  }
  if(length(transposonsNeg)>0) {
    rect(start(transposonsNeg), -0.16, end(transposonsNeg), -0.2, col=col[2], border = col[2], density=30, angle=30);
    text(start(transposonsNeg), -0.23, transposonsNeg$ID, pos=4, cex=0.5);
  }

}


#' This function plots the methylation profile at one locus for the bisulfite
#' sequencing data.The points on the graph represent methylation proportion of
#' individual cytosines, their colour which sample they belong to and the
#' intesity of the the colour how many reads that particular cytosine had. This
#' means that darker colors indicate stronger evidence that the corresponding
#' cytosine has the corresponding methylation proportion, while lighter colors
#' indicate a weaker evidence. The solid lines represent the smoothed profiles
#' and the intensity of the line the coverage at the corresponding position
#' (darker colors indicate more reads while lighter ones less reads). The boxes
#' on top represent the DMRs, where a filled box will represent a DMR which
#' gained methylation while a box with a pattern represent a DMR that lost
#' methylation. The DMRs need to have a metadafield \code{"regionType"} which
#' can be either \code{"gain"} (where there is more methylation in condition 2
#' compared to condition 1) or \code{"loss"} (where there is less methylation in
#' condition 2 compared to condition 1). In case this metadafield is missing all
#' DMRs are drawn using a filled box. Finally, we also allow annotation of the
#' DNA sequence. We represent by a black boxes all the exons, which are joined
#' by a horizontal black line, thus, marking the full body of the gene. With
#' grey boxes we mark the transposable elements. Both for genes and transposable
#' elements we plot them over a mid line if they are on the positive strand and
#' under the mid line if they are on the negative strand.
#'
#' @title Plot local methylation profile
#' @param methylationData1 the methylation data in condition 1
#' (see \code{\link{methylationDataList}}).
#' @param methylationData2 the methylation data in condition 2
#' (see \code{\link{methylationDataList}}).
#' @param region a \code{\link{GRanges}} object with the region where to plot
#' the high resolution profile.
#' @param DMRs a \code{\link{GRangesList}} object or a list with the list of
#' DMRs (see \code{\link{computeDMRs}} or \code{\link{filterDMRs}}.
#' @param conditionsNames the names of the two conditions. This will be used to
#' plot the legend.
#' @param gff a \code{\link{GRanges}} object with all elements usually imported
#' from a GFF3 file. The gff file needs to have an metafield \code{"type"}. Only
#' the elements of type  \code{"gene"}, \code{"exon"} and
#' \code{"transposable_element"} are plotted. Genes are represented as
#' horizontal black lines, exons as a black rectangle and transposable elements
#' as a grey rectangle. The elements are plotted on the corresponding strand
#' (\code{+} or \code{-}).
#' @param windowSize the size of the triangle base used to smooth the average
#' methylation profile.
#' @param context the context in which the DMRs are computed (\code{"CG"},
#' \code{"CHG"} or \code{"CHH"}).
#' @param labels a \code{vector} of \code{character} used to add a subfigure
#' characters to the plot. If \code{NULL} nothing is added.
#' @param col a \code{character} vector with the colors. It needs to contain a
#' minimum of \code{4 length(DMRs)} colors. If not or if \code{NULL}, the
#' defalut colors will be used.
#' @param main a \code{character} with the title of the plot
#' @param plotMeanLines a \code{logical} value indicating whether to plot the
#' mean lines or not.
#' @param plotPoints a \code{logical} value indicating whether to plot the
#' points or not.
#' @return Invisibly returns \code{NULL}
#' @examples
#'
#' # load the methylation data
#' data(methylationDataList)
#' # load the gene annotation data
#' data(GEs)
#'
#' #select the genes
#' genes <- GEs[which(GEs$type == "gene")]
#'
#' # the coordinates of the area to be plotted
#' chr3Reg <- GRanges(seqnames = Rle("Chr3"), ranges = IRanges(510000,530000))
#'
#' # load the DMRs in CG context
#' data(DMRsNoiseFilterCG)
#'
#' DMRsCGlist <- list("noise filter"=DMRsNoiseFilterCG)
#'
#'
#' # plot the CG methylation
#' par(mar=c(4, 4, 3, 1)+0.1)
#' par(mfrow=c(1,1))
#' plotLocalMethylationProfile(methylationDataList[["WT"]],
#'                            methylationDataList[["met1-3"]], chr3Reg,
#'                            DMRsCGlist, c("WT", "met1-3"), GEs,
#'                            windowSize=100, main="CG methylation")
#'
#' @author Nicolae Radu Zabet
#' @export
plotLocalMethylationProfile <- function(methylationData1,
                                        methylationData2,
                                        region,
                                        DMRs = NULL,
                                        conditionsNames = NULL,
                                        gff = NULL,
                                        windowSize = 150,
                                        context = "CG",
                                        labels=NULL,
                                        col=NULL,
                                        main="",
                                        plotMeanLines=TRUE,
                                        plotPoints=TRUE) {

  .validateMethylationData(methylationData1, variableName="methylationData1")
  .validateMethylationData(methylationData2, variableName="methylationData1")

  .stopIfNotAll(c(!is.null(plotMeanLines),
                  is.logical(plotMeanLines)),
                " plotMeanLines is a logical parameter and can be only TRUE or FALSE.")
  .stopIfNotAll(c(!is.null(plotPoints),
                  is.logical(plotPoints)),
                " plotPoints is a logical parameter and can be only TRUE or FALSE.")

  numberOfConditions <- 2

  if(is.null(conditionsNames) | length(conditionsNames) < numberOfConditions | !all(is.character(conditionsNames))){
    conditionsNames <- paste("condition ",(1:numberOfConditions),sep="")
  }

  .validateGRanges(region, length=1, generateGenomeWide=FALSE)

  .stopIfNotAll(c(.isInteger(windowSize, positive=TRUE)),
                " the window size used by the interpolation method is an integer higher or equal to 0")
  .validateContext(context)



  numberOfDMRs <- 0
  if(!is.null(DMRs)){
    numberOfDMRs <- length(DMRs)
  }

  if(!is.null(labels) & (length(labels) < 1 | !is.character(labels))){
    labels <- LETTERS[1:length(labels)]
  }

  if(!.isColor(col, minLength = (4+numberOfDMRs))){
    col <- c("#000000", "#E69F00", "#56B4E9", "#009E73", "#F0E442", "#0072B2", "#D55E00", "#CC79A7", "#999999")

    cond1Color <- col[7]
    cond2Color <- col[6]
    geneColor <- col[1]
    TEColor <- col[9]
    DMRsColor <- col[c(4, 5, 8, 2, 3)]

  } else{
    cond1Color <- col[1]
    cond2Color <- col[2]
    geneColor <- col[3]
    TEColor <- col[4]
    DMRsColor <- col[5:length(col)]
  }

  seqname <- seqnames(region)
  minPos <- start(region)
  maxPos <- end(region)

  methylationData <- .joinMethylationData(methylationData1, methylationData2)

  hits <- findOverlaps(methylationData, region)

  localMethylationData <- methylationData[queryHits(hits)];
  contextMethylationData <- localMethylationData[localMethylationData$context%in%context];


  ramp1 <- colorRampPalette(c("white", cond1Color))
  colramp1 <- ramp1(100)

  ramp2 <- colorRampPalette(c("white", cond2Color))
  colramp2 <- ramp2(100)



  # plot the points
  proportion1 <- rep(0, length(contextMethylationData))
  index <- which(contextMethylationData$readsM1 >= 0 & contextMethylationData$readsN1 > 0)
  proportion1[index] <- contextMethylationData$readsM1[index] / contextMethylationData$readsN1[index]

  proportion2 <- rep(0, length(contextMethylationData))
  index <- which(contextMethylationData$readsM2 >= 0 & contextMethylationData$readsN2 > 0)
  proportion2[index] <- contextMethylationData$readsM2[index] / contextMethylationData$readsN2[index]

  maxColor <- max(c(contextMethylationData$readsN1[!is.na(contextMethylationData$readsN1)],
                    contextMethylationData$readsN2[!is.na(contextMethylationData$readsN2)]))



  plot(start(contextMethylationData), proportion1,
       col=colramp1[round(99 * log(contextMethylationData$readsN1) / log(maxColor))+1],
       pch = 16, cex=0.7,
       xlim = c(minPos, maxPos),
       ylim = c(-.2, 1.2 + numberOfDMRs*.1),
       xlab = paste("genomic coordinate on chromosome ",seqname, sep="") ,
       ylab = "methylation proportion",
       yaxt = "n",
       main = main,
       mar  = c(4,4,2,2)+.1,
       type = "n")
  axis(2,c(0,.5,1));


  if(plotPoints){
    points(start(contextMethylationData), proportion1,
           col=colramp1[round(99 * log(contextMethylationData$readsN1) / log(maxColor))+1],
           pch = 16, cex=0.7)
    points(start(contextMethylationData), proportion2,
           col=colramp2[round(99 * log(contextMethylationData$readsN2) / log(maxColor))+1],
           pch = 15, cex=0.7)
    if(!plotMeanLines){
      legend("topright", bty="n", col=c(cond1Color, cond2Color), legend=conditionsNames, pch=c(15,16), horiz=TRUE)
    }
  }

  #axis

  windowSizeHalf <- floor((windowSize - 1)/2)



  if(plotMeanLines){
    positions <- start(region) : end(region)


    movingAverageMethylReads1 <- .movingAverage(start(region),
                                                end(region),
                                                start(contextMethylationData),
                                                contextMethylationData$readsM1,
                                                windowSizeHalf = windowSizeHalf)
    movingAverageTotalReads1 <- .movingAverage(start(region),
                                               end(region),
                                               start(contextMethylationData),
                                               contextMethylationData$readsN1,
                                               windowSizeHalf = windowSizeHalf)
    movingAverageProportion1 <- movingAverageMethylReads1 / movingAverageTotalReads1

    movingAverageMethylReads2 <- .movingAverage(start(region),
                                                end(region),
                                                start(contextMethylationData),
                                                contextMethylationData$readsM2,
                                                windowSizeHalf = windowSizeHalf)
    movingAverageTotalReads2 <- .movingAverage(start(region),
                                               end(region),
                                               start(contextMethylationData),
                                               contextMethylationData$readsN2,
                                               windowSizeHalf = windowSizeHalf)
    movingAverageProportion2 <- movingAverageMethylReads2 / movingAverageTotalReads2

    maxColor <- max(c(movingAverageTotalReads1[!is.na(movingAverageTotalReads1)],
                      movingAverageTotalReads2[!is.na(movingAverageTotalReads2)]))


    #lines(positions, movingAverageProportion1, lty=1, lwd=2, col=cond1Color)
    #lines(positions, movingAverageProportion2, lty=1, lwd=2, col=cond2Color)
    #compute coordinates for mean lines condition 1
    bufferIndex <- !is.na(movingAverageProportion1) & !is.na(movingAverageTotalReads1)
    if(any(bufferIndex)){
      minID <- min(which(bufferIndex))
      maxID <- max(which(bufferIndex))
      colorId <- round(99 * log(round(movingAverageTotalReads1)) / log(maxColor))+1
      colorId <- round((colorId[minID:(maxID-1)] +  colorId[(minID+1):(maxID)])/2)
      colorId[is.na(colorId)] <- 1
      segments(positions[minID:(maxID-1)],
               movingAverageProportion1[minID:(maxID-1)],
               positions[(minID+1):(maxID)],
               movingAverageProportion1[(minID+1):(maxID)],
               col=colramp1[colorId],
               lty=1, lwd=2)


    }
    #compute coordinates for mean lines condition 2
    bufferIndex <- !is.na(movingAverageProportion2) & !is.na(movingAverageTotalReads2)
    if(any(bufferIndex)){
      minID <- min(which(bufferIndex))
      maxID <- max(which(bufferIndex))
      colorId <- round(99 * log(round(movingAverageTotalReads2)) / log(maxColor))+1
      colorId <- round((colorId[minID:(maxID-1)] +  colorId[(minID+1):(maxID)])/2)
      colorId[is.na(colorId)] <- 1
      segments(positions[minID:(maxID-1)],
               movingAverageProportion2[minID:(maxID-1)],
               positions[(minID+1):(maxID)],
               movingAverageProportion2[(minID+1):(maxID)],
               col=colramp2[colorId],
               lty=1, lwd=2)
    }

    legend("topright", bty="n", col=c(cond1Color, cond2Color), legend=conditionsNames, lty=1, lwd = 2, horiz=TRUE)
  }




  if(numberOfDMRs > 0) {
    for(i in 1:numberOfDMRs) {
      range <- DMRs[[i]][queryHits(findOverlaps(DMRs[[i]],region))]

      if(length(range) > 0 ){
        start(range) <- pmax(start(range), start(region))
        end(range)   <- pmin(end(range), end(region))

        if(length(which(range$regionType == "gain"))>0){
          rect(start(range)[range$regionType == "gain"], 1.025 + (length(DMRs) - i)*.1,
               end(range)[range$regionType == "gain" ], 1.075 + (length(DMRs)-i)*.1,
               col=DMRsColor[i], border=DMRsColor[i])
        }
        if(length(which(range$regionType == "loss"))>0){
          rect(start(range)[range$regionType == "loss"], 1.025 + (length(DMRs) - i)*.1,
               end(range)[range$regionType == "loss"], 1.075 + (length(DMRs)-i)*.1,
               col=DMRsColor[i], border=DMRsColor[i], density =30)
        }

        if(length(range) > 0 & (is.null(range$regionType))){
          rect(start(range), (1.025 + (length(DMRs) - i)*.1), end(range), (1.075 + (length(DMRs)-i)*.1),
               col=DMRsColor[i], border=DMRsColor[i])
        }

      }
      text(par('usr')[1], 1.050 + (length(DMRs)-i)*.1, pos=4, names(DMRs)[i])
    }
  }

  if(!is.null(gff)){
    if(is(gff, "GRanges")){
      .plotGeneticElements(gff,region, c(geneColor, TEColor))
    }
  }

  if(!is.null(labels)){
    if(length(labels) >= 1){
      mtext(labels[1], line = 0.7, adj = 0, cex=1.0);
    }
  }
  invisible(NULL)
}

#' This function computes the correlation of the methylation levels for
#' Cytosines located at a certain distance apart. The function returns a the
#' correlation of methylation levels at distance equal to the specified
#' thresholds.
#'
#' @title Compute methylation data spatial correlation
#' @param methylationData the methylation data stored as a \code{\link[GenomicRanges]{GRanges}}
#' object with four metadata columns (see \code{\link{methylationDataList}}).
#' @param regions a \code{\link[GenomicRanges]{GRanges}} object with the regions where to
#' compute the correlation If NULL, the correlation is computed genome-wide.
#' @param distance the distance used when computing the correlation of
#' methylation levels
#'
#' @return correlation of the methylation levels for Cytosines located at a
#' certain distance apart.
#'
#' @author Radu Zabet
.computeMethylationDataSpatialCorrelation <- function(methylationData, regions, distance){
  values1 <- c()
  values2 <- c()
  for(i in 1:length(regions)){
    localMethylationData <- methylationData[queryHits(findOverlaps(methylationData, regions[i]))]
    localMethylationData <- localMethylationData[localMethylationData$readsN > 0]
    localMethylationDataValues <- localMethylationData$readsM/localMethylationData$readsN
    localMethylationDataPositions <- start(localMethylationData)

    localMethylationDatamatch <- match(localMethylationDataPositions,
                                       (localMethylationDataPositions + distance))


    ids <- !is.na(localMethylationDatamatch)

    values1 <- c(values1, localMethylationDataValues[ids])
    values2 <- c(values2, localMethylationDataValues[localMethylationDatamatch[ids]])

  }

  return(cor(values1, values2))
}

#' This function computes the correlation of the methylation levels as a
#' function of the distances between the Cytosines. The function returns a
#' \code{vector} with the correlation of methylation levels at distance equal to
#' a \code{vector} of specified thresholds.
#'
#' @title Compute methylation data spatial correlation
#' @param methylationData the methylation data stored as a \code{\link[GenomicRanges]{GRanges}}
#' object with four metadata columns (see \code{\link{methylationDataList}}).
#' @param regions a \code{\link[GenomicRanges]{GRanges}} object with the regions where to
#' compute the correlation. If \code{NULL}, the correlation is computed
#' genome-wide.
#' @param context the context in which the correlation is computed (\code{"CG"},
#' \code{"CHG"} or \code{"CHH"}).
#' @param distances a \code{numeric} vector specifing the different values for
#' the distances when computing the correlation.
#'
#' @return a \code{vector} with the correlation of the methylation levels for
#' Cytosines located at distances specified in the \code{distances}
#' \code{vector}.
#'
#' @seealso \code{\link{plotMethylationDataSpatialCorrelation}},
#' \code{\link{methylationDataList}}
#' @examples
#'
#'
#' # load the methylation data
#' data(methylationDataList)
#'
#' # compute spatial correlation in CG context
#' distances <- c(1,5,10,15)
#' correlation_CG_wt <- computeMethylationDataSpatialCorrelation(methylationDataList[["WT"]],
#'                  context="CG", distances=distances)
#'
#'
#'
#' @author Nicolae Radu Zabet
#' @export
computeMethylationDataSpatialCorrelation <- function(methylationData,
                                                      regions = NULL,
                                                      context = "CG",
                                                      distances=NULL) {

  .validateMethylationData(methylationData)

  .stopIfNotAll(c(all(sapply(distances,.isInteger, positive=TRUE)),
                  (length(distances) > 1)),
                " the distances is a vector of integers with the distance between the positions .")

  regions <- .validateGRanges(regions, methylationData)

  .validateContext(context)

  contextMethylationData <- methylationData[methylationData$context%in%context];
  rm(methylationData)
  localMethylationData <- contextMethylationData[queryHits(findOverlaps(contextMethylationData, regions))]
  rm(contextMethylationData)

  correlation <- c()

  for(k in distances){
    cat("Computing methylation levels correlation for distances of ",k," bp\n")
    correlation <- c(correlation,
                     .computeMethylationDataSpatialCorrelation(localMethylationData,
                                                                regions,
                                                                k))
  }
  names(correlation) <- distances

  return(correlation)
}



#' This function plots the correlation of methylation levels for Cytosines
#' located at a certain distance apart.
#'
#' @title Plot methylation data spatial correlation
#' @param methylationData1 the methylation data in condition 1
#' (see \code{\link{methylationDataList}}).
#' @param methylationData2 the methylation data in condition 2
#' (see \code{\link{methylationDataList}}). This is optional.
#' @param distances a \code{numeric} vector specifing the different values for
#' the distances when computing the correlation.
#' @param regions a \code{\link[GenomicRanges]{GRanges}} object with the regions where to
#' compute the correlation. If \code{NULL}, the coverage is computed genome-wide.
#' @param conditionsNames a vector of character with the names of the conditions
#' for \code{methylationData1} and \code{methylationData2}.
#' @param context the context in which the DMRs are computed (\code{"CG"},
#' \code{"CHG"} or \code{"CHH"}).
#' @param labels a \code{vector} of \code{character} used to add a subfigure
#' character to the plot. If \code{NULL} nothing is added.
#' @param col a \code{character} vector with the colors. It needs to contain a
#' minimum of 2 colors per condition. If not or if \code{NULL}, the defalut
#' colors will be used.
#' @param pch the R symbols used to plot the data. It needs to contain a minimum
#' of 2 symbols per condition. If not or if \code{NULL}, the defalut symbols
#' will be used.
#' @param lty the line types used to plot the data. It needs to contain a
#' minimum of 2 line types per condition. If not or if \code{NULL}, the defalut
#' line types will be used.
#' @param contextPerRow a \code{logical} value indicating if the each row
#' represents an individual context. If \code{FALSE}, each column will represent
#' an individual context.
#' @param log a \code{character} indicating if any of the axes will be displayed
#' on log scale. This argument will be passed to \code{\link{plot}} function.
#' @details This function plots the proportion of cytosines in a specific
#' context that have at least a certain number of reads (x-axis)
#' @seealso \code{\link{computeMethylationDataSpatialCorrelation}},
#' \code{\link{methylationDataList}}
#' @return Invisibly returns \code{NULL}
#' @examples
#'
#' 
#' # load the methylation data
#' data(methylationDataList)
#'
#' # plot the spatial correlation in CG context
#' par(mar=c(4, 4, 3, 1)+0.1)
#' plotMethylationDataSpatialCorrelation(methylationDataList[["WT"]],
#'                            methylationDataList[["met1-3"]],
#'                            distances = c(1,5,10,15), regions = NULL,
#'                            conditionsNames = c("WT","met1-3"),
#'                            context = c("CG"),
#'                            labels = LETTERS, col = NULL,
#'                            pch = c(1,0,16,2,15,17), lty = c(4,1,3,2,6,5),
#'                            contextPerRow = FALSE)
#' \dontrun{
#' # plot the spatial correlation in all three contexts
#' plotMethylationDataSpatialCorrelation(methylationDataList[["WT"]],
#'                            methylationDataList[["met1-3"]],
#'                            distances = c(1,5,10,15,20,50,100,150,200,500,1000),
#'                            regions = NULL, conditionsNames = c("WT","met1-3"),
#'                            context = c("CG", "CHG", "CHH"),
#'                            labels = LETTERS, col = NULL,
#'                            pch = c(1,0,16,2,15,17), lty = c(4,1,3,2,6,5),
#'                            contextPerRow = FALSE, log="x")
#' }
#'
#' @author Nicolae Radu Zabet
#' @export
plotMethylationDataSpatialCorrelation <- function(methylationData1,
                                                   methylationData2 = NULL,
                                                   distances,
                                                   regions = NULL,
                                                   conditionsNames = NULL,
                                                   context = "CG",
                                                   labels=NULL,
                                                   col=NULL,
                                                   pch = c(1,0,16,2,15,17),
                                                   lty = c(4,1,3,2,6,5),
                                                   contextPerRow = FALSE,
                                                   log="") {
  .validateContext(context)
  .validateMethylationData(methylationData1, variableName="methylationData1")
  numberOfConditions <- 1
  if(!is.null(methylationData2)){
    .validateMethylationData(methylationData2, variableName="methylationData2")
    numberOfConditions <- 2
  }

  if(is.null(conditionsNames) | length(conditionsNames) < numberOfConditions){
    conditionsNames <- paste("condition ",(1:numberOfConditions),sep="")
  }

  number_of_symbols <- numberOfConditions*length(context)

  if(!.isColor(col, minLength = number_of_symbols )){
    if(number_of_symbols <= 6){
      col <- c("#D55E00","#E69F00", "#0072B2", "#56B4E9", "#F0E442", "#009E73")
    } else if(number_of_symbols <= 8){
      col <- c("#000000", "#E69F00", "#56B4E9", "#009E73", "#F0E442", "#0072B2", "#D55E00", "#CC79A7")
    } else{
      col <- rainbow(number_of_symbols)
    }
  }

  .stopIfNotAll(c(!is.null(pch),
                  all(as.integer(pch) == pch),
                  all(pch >= 0),
                  length(pch) >= number_of_symbols),
                paste(" pch is a vector of positive integers of size ",number_of_symbols, sep=""))

  .stopIfNotAll(c(!is.null(lty),
                  all(as.integer(lty) == lty),
                  all(lty >= 0),
                  length(lty) >= number_of_symbols),
                paste(" lty is a vector of positive integers of size ",number_of_symbols, sep=""))


  if(!is.null(labels) & (length(labels) < length(coverage) | !is.character(labels))){
    labels <- LETTERS[1:length(coverage)]
  }



  par(mar=c(4, 4, 3, 1)+0.1)
  if(contextPerRow){
    par(mfrow = c(length(context),1))
  } else{
    par(mfrow = c(1,length(context)))
  }

  for(i in 1:length(context)){
    buffer_correlation <- matrix(0, nrow = length(distances), ncol=numberOfConditions)
    buffer_correlation[,1] <- computeMethylationDataSpatialCorrelation(methylationData1,
                                                                     regions = regions,
                                                                     context = context[i],
                                                                     distances=distances)

    if(numberOfConditions > 1){
      buffer_correlation[,2] <- computeMethylationDataSpatialCorrelation(methylationData2,
                                                                       regions = regions,
                                                                       context = context[i],
                                                                       distances=distances)
    }
    colnames(buffer_correlation) <- conditionsNames

    symbols_ind <- (numberOfConditions*(i-1) + 1):(numberOfConditions*i)

    plot(distances, buffer_correlation[, 1], type = "o", ylim = c(0,1), main = paste("Correlation of methylation levels in ",context[i]," context",sep=""),
         xlab="distance between cytosines (bp)", ylab="correlation", col=col[symbols_ind[1]], xaxt="n", yaxt="n",
         lty = lty[symbols_ind[1]], pch = pch[symbols_ind[1]], log=log)

    if(numberOfConditions > 1){
      lines(distances, buffer_correlation[, 2], type = "o", col=col[symbols_ind[2]], lty = lty[symbols_ind[2]], pch = pch[symbols_ind[2]])
    }

    axis(1, at=distances,labels=distances, las=1)
    axis(2, at=seq(0,1,0.1),labels=seq(0,1,0.1), las=1)


    legend("topright", legend = colnames(buffer_correlation), lty = lty[symbols_ind], col = col[symbols_ind], pch = pch[symbols_ind], bty="n")


    mtext(labels[i], line = 0.7, adj = 0, cex=1.0);

  }
  invisible(NULL)
}

#' This function computes the number of cytosines with more reads than a
#' threshold in the methylation data
#' @title Count methylation data coverage
#' @param methylationData the methylation data stored as a \code{\link{GRanges}}
#' object with four metadata columns (see \code{\link{methylationDataList}}).
#' @param regions a \code{\link{GRanges}} object with the regions where to
#' compute the coverage. If NULL, the coverage is computed genome-wide.
#' @param threshold the threshold used when counting cytosines that have at
#' least a certain number of reads
#'
#' @return a the number of cytosines with more reads than the threshold in the
#' specified region
#'
#' @author Radu Zabet
.countMethylationDataCoverage <- function(methylationData, regions, threshold){
  count <- 0
  for (index in 1:length(regions)) {
    currentRegion <- regions[index]
    localMethylationData <- methylationData[queryHits(findOverlaps(methylationData, currentRegion))]
    count <- count + length(which(localMethylationData$readsN >= threshold))
  }
  return(count)
}


#' This function computes the coverage for bisulfite sequencing data. It
#' returns a \code{vector} with the proportion (or raw count) of cytosines that
#' have the number of reads higher or equal than a \code{vector} of specified
#' thresholds.
#'
#' @title Compute methylation data coverage
#' @param methylationData the methylation data stored as a \code{\link{GRanges}}
#' object with four metadata columns (see \code{\link{methylationDataList}}).
#' @param regions a \code{\link{GRanges}} object with the regions where to
#' compute the coverage. If \code{NULL}, the coverage is computed genome-wide.
#' @param context the context in which the DMRs are computed (\code{"CG"},
#' \code{"CHG"} or \code{"CHH"}).
#' @param breaks a \code{numeric} vector specifing the different values for the
#' thresholds when computing the coverage.
#' @param proportion a \code{logical} value indicating whether to compute the
#' proportion (\code{TRUE}) or raw counts (\code{FALSE}).
#'
#' @return a \code{vector} with the proportion (or raw count) of cytosines that
#' have the number of reads higher or equal than the threshold values specified
#' in the \code{breaks} \code{vector}.
#' @seealso \code{\link{plotMethylationDataCoverage}},
#' \code{\link{methylationDataList}}
#' @examples
#'
#' # load the methylation data
#' data(methylationDataList)
#'
#' # compute coverage in CG context
#' breaks <- c(1,5,10,15)
#' coverage_CG_wt <- computeMethylationDataCoverage(methylationDataList[["WT"]],
#'                  context="CG", breaks=breaks)
#'
#'
#' @author Nicolae Radu Zabet and Jonathan Michael Foonlan Tsang
#' @export
computeMethylationDataCoverage <- function(methylationData,
                                           regions = NULL,
                                           context = "CG",
                                           breaks=NULL,
                                           proportion = TRUE) {

  .validateMethylationData(methylationData)

  .stopIfNotAll(c(all(sapply(breaks,.isInteger, positive=TRUE)),
                  (length(breaks) > 1)),
                " the breaks is a vector of integers with the number of reads used
                as thresholds when computing the coverage.")

  regions <- .validateGRanges(regions, methylationData)

  .validateContext(context)

  contextMethylationData <- methylationData[methylationData$context%in%context];
  rm(methylationData)

  counts <- c()
  for(i in breaks){
    counts <- c(counts,
                .countMethylationDataCoverage(contextMethylationData,regions,i))
  }


  if(proportion){
    counts <- counts/.countTotalNumberOfCytosines(contextMethylationData, regions)
  }

  return(counts)
}



#' This function plots the coverage for the bisulfite sequencing data.
#'
#' @title Plot methylation data coverage
#' @param methylationData1 the methylation data in condition 1
#' (see \code{\link{methylationDataList}}).
#' @param methylationData2 the methylation data in condition 2
#' (see \code{\link{methylationDataList}}). This is optional.
#' @param breaks a \code{numeric} vector specifing the different values for the
#' thresholds when computing the coverage.
#' @param regions a \code{\link{GRanges}} object with the regions where to
#' compute the coverage. If \code{NULL}, the coverage is computed genome-wide.
#' @param conditionsNames a vector of character with the names of the conditions
#' for \code{methylationData1} and \code{methylationData2}.
#' @param context the context in which the DMRs are computed (\code{"CG"},
#' \code{"CHG"} or \code{"CHH"}).
#' @param proportion a \code{logical} value indicating whether proportion or
#' counts will be plotted.
#' @param labels a \code{vector} of \code{character} used to add a subfigure
#' character to the plot. If \code{NULL} nothing is added.
#' @param col a \code{character} vector with the colors. It needs to contain a
#' minimum of 2 colors per condition. If not or if \code{NULL}, the defalut
#' colors will be used.
#' @param pch the R symbols used to plot the data. It needs to contain a minimum
#' of 2 symbols per condition. If not or if \code{NULL}, the defalut symbols
#' will be used.
#' @param lty the line types used to plot the data. It needs to contain a
#' minimum of 2 line types per condition. If not or if \code{NULL}, the defalut
#' line types will be used.
#' @param contextPerRow a \code{logical} value indicating if the each row
#' represents an individual context. If \code{FALSE}, each column will represent
#' an individual context.
#' @details This function plots the proportion of cytosines in a specific
#' context that have at least a certain number of reads (x-axis)
#' @seealso \code{\link{computeMethylationDataCoverage}},
#' \code{\link{methylationDataList}}
#' @return Invisibly returns \code{NULL}
#' @examples
#'
#' # load the methylation data
#' data(methylationDataList)
#'
#' # plot the coverage in CG context
#' par(mar=c(4, 4, 3, 1)+0.1)
#' plotMethylationDataCoverage(methylationDataList[["WT"]],
#'                            methylationDataList[["met1-3"]],
#'                            breaks = c(1,5,10,15), regions = NULL,
#'                            conditionsNames = c("WT","met1-3"),
#'                            context = c("CG"), proportion = TRUE,
#'                            labels = LETTERS, col = NULL,
#'                            pch = c(1,0,16,2,15,17), lty = c(4,1,3,2,6,5),
#'                            contextPerRow = FALSE)
#'
#' \dontrun{
#' # plot the coverage in all three contexts
#' plotMethylationDataCoverage(methylationDataList[["WT"]],
#'                            methylationDataList[["met1-3"]],
#'                            breaks = 1:15, regions = NULL,
#'                            conditionsNames = c("WT","met1-3"),
#'                            context = c("CG", "CHG", "CHH"),
#'                            proportion = TRUE, labels = LETTERS, col = NULL,
#'                            pch = c(1,0,16,2,15,17), lty = c(4,1,3,2,6,5),
#'                            contextPerRow = FALSE)
#' }
#'
#' @author Nicolae Radu Zabet
#' @export
plotMethylationDataCoverage <- function(methylationData1,
                                        methylationData2 = NULL,
                                        breaks,
                                        regions = NULL,
                                        conditionsNames = NULL,
                                        context = "CG",
                                        proportion = TRUE,
                                        labels=NULL,
                                        col=NULL,
                                        pch = c(1,0,16,2,15,17),
                                        lty = c(4,1,3,2,6,5),
                                        contextPerRow = FALSE) {
  .validateContext(context)
  .validateMethylationData(methylationData1, variableName="methylationData1")
  numberOfConditions <- 1
  if(!is.null(methylationData2)){
    .validateMethylationData(methylationData2, variableName="methylationData2")
    numberOfConditions <- 2
  }

  if(is.null(conditionsNames) | length(conditionsNames) < numberOfConditions){
    conditionsNames <- paste("condition ",(1:numberOfConditions),sep="")
  }

  number_of_symbols <- numberOfConditions*length(context)

  if(!.isColor(col, minLength = number_of_symbols )){
    if(number_of_symbols <= 6){
      col <- c("#D55E00","#E69F00", "#0072B2", "#56B4E9", "#F0E442", "#009E73")
    } else if(number_of_symbols <= 8){
      col <- c("#000000", "#E69F00", "#56B4E9", "#009E73", "#F0E442", "#0072B2", "#D55E00", "#CC79A7")
    } else{
      col <- rainbow(number_of_symbols)
    }
  }

  .stopIfNotAll(c(!is.null(pch),
                  all(as.integer(pch) == pch),
                  all(pch >= 0),
                  length(pch) >= number_of_symbols),
                paste(" pch is a vector of positive integers of size ",number_of_symbols, sep=""))

  .stopIfNotAll(c(!is.null(lty),
                  all(as.integer(lty) == lty),
                  all(lty >= 0),
                  length(lty) >= number_of_symbols),
                paste(" lty is a vector of positive integers of size ",number_of_symbols, sep=""))


  if(!is.null(labels) & (length(labels) < length(coverage) | !is.character(labels))){
    labels <- LETTERS[1:length(coverage)]
  }



  par(mar=c(4, 4, 3, 1)+0.1)
  if(contextPerRow){
    par(mfrow = c(length(context),1))
  } else{
    par(mfrow = c(1,length(context)))
  }

  for(i in 1:length(context)){
    buffer_coverage <- matrix(0, nrow = length(breaks), ncol=numberOfConditions)
    buffer_coverage[,1] <- computeMethylationDataCoverage(methylationData1,
                                                          regions = regions,
                                                          context = context[i],
                                                          breaks=breaks,
                                                          proportion = proportion)

    if(numberOfConditions > 1){
      buffer_coverage[,2] <- computeMethylationDataCoverage(methylationData2,
                                                            regions = regions,
                                                            context = context[i],
                                                            breaks=breaks,
                                                            proportion = proportion)
    }
    colnames(buffer_coverage) <- conditionsNames

    symbols_ind <- (numberOfConditions*(i-1) + 1):(numberOfConditions*i)

    plot(breaks, buffer_coverage[, 1], type = "o", ylim = c(0,1), main = paste("Coverage in ",context[i]," context",sep=""),
         xlab="minimum number of reads", ylab="coverage", col=col[symbols_ind[1]], xaxt="n", yaxt="n",
         lty = lty[symbols_ind[1]], pch = pch[symbols_ind[1]])

    if(numberOfConditions > 1){
      lines(breaks, buffer_coverage[, 2], type = "o", col=col[symbols_ind[2]], lty = lty[symbols_ind[2]], pch = pch[symbols_ind[2]])
    }

    axis(1, at=breaks,labels=breaks, las=1)
    axis(2, at=seq(0,1,0.1),labels=seq(0,1,0.1), las=1)


    legend("topright", legend = colnames(buffer_coverage), lty = lty[symbols_ind], col = col[symbols_ind], pch = pch[symbols_ind], bty="n")


    mtext(labels[i], line = 0.7, adj = 0, cex=1.0);

  }
  invisible(NULL)
}

#' This function joins together data that come from biological replicates to
#' perform analysis
#'
#' @title Joins together two GRange objects in a single containing all the
#' replicates
#'
#' @param methylationData1 the methylation data stored as a \code{\link[GenomicRanges]{GRanges}}
#' object with four metadata columns (see \code{\link{methylationDataList}}).
#'
#' @param methylationData2 the methylation data stored as a \code{\link[GenomicRanges]{GRanges}}
#' object with four metadata columns (see \code{\link{methylationDataList}}).
#'
#' @param usecomplete Boolean, determine wheter, when the two dataset differ for
#' number of cytosines, if the smaller dataset should be added with zero reads
#' to match the bigger dataset.
#'
#' @return returns a \code{\link[GenomicRanges]{GRanges}} object containing multiple metadata
#' columns with the reads from each object passed as parameter
#'
#' @examples
#'
#'
#' # load the methylation data
#' data(methylationDataList)
#'
#' # Joins the wildtype and the mutant in a single object
#' joined_data <- joinReplicates(methylationDataList[["WT"]],
#'                               methylationDataList[["met1-3"]], FALSE)
#'
#'
#' @author Alessandro Pio Greco and Nicolae Radu Zabet
#'
#' @export
joinReplicates <- function(methylationData1, methylationData2, usecomplete = FALSE){
  cat("Validating objects \n")
  .validateMethylationData(methylationData1, "methylationData1")
  .validateMethylationData(methylationData2, "methylationData2")
  cat("Finding overlaps \n")
  # Checking the elements of the objects
  overlaps <- findOverlaps(methylationData1, methylationData2)
  # checking that the two objects are the same length
  if(queryLength(overlaps) != subjectLength(overlaps) & usecomplete == FALSE){
    warning("The number of elements in each dataset is different,
            non replicated data will be dropped")
  } else if(queryLength(overlaps) != subjectLength(overlaps) & usecomplete == TRUE){
    if(queryLength(overlaps) > subjectLength(overlaps)){
      cx_2 <- methylationData1[which(methylationData1 %in% methylationData1[queryHits(overlaps)] == FALSE)]
      mcols(cx_2)[which(grepl("reads", names(mcols(cx_2))))] <- 0
      methylationData2 <- c(methylationData2, cx_2)
      overlaps <- findOverlaps(methylationData1, methylationData2)
    } else{
      cx_1 <- methylationData2[which(methylationData2 %in% methylationData2[queryHits(overlaps)] == FALSE)]
      mcols(cx_1)[which(grepl("reads", names(mcols(cx_1))))] <- 0
      methylationData1 <- c(methylationData1, cx_1)
      overlaps <- findOverlaps(methylationData1, methylationData2)
    }

  }
  flag_multiple <- NULL
  # Read if one of the two objects presents "union columns" already
  # if not elements are just joined together
  cat("Joining objects \n")
  if(length(grep("readsM", names(mcols(methylationData1)))) == 1 &
     length(grep("readsM", names(mcols(methylationData2)))) == 1){

    result <- GRanges(seqnames = seqnames(methylationData1[queryHits(overlaps)]),
                      ranges = ranges(methylationData1[queryHits(overlaps)]),
                      strand = strand(methylationData1[queryHits(overlaps)]),
                      context = methylationData1$context[queryHits(overlaps)],
                      readsM1 = methylationData1$readsM[queryHits(overlaps)],
                      readsN1 = methylationData1$readsN[queryHits(overlaps)],
                      readsM2 = methylationData2$readsM[subjectHits(overlaps)],
                      readsN2 = methylationData2$readsN[subjectHits(overlaps)],
                      trinucleotide_context = methylationData1$trinucleotide_context[
                        queryHits(overlaps)])
  } else{
    if(length(grep("readsM", names(mcols(methylationData1)))) > 1 &
       length(grep("readsM", names(mcols(methylationData2)))) > 1){
      flag_multiple <- 1
    }

    if(!is.null(flag_multiple)){
      result <- methylationData1
      # bind together the reads (M & N)
      mcols(result) <- cbind(mcols(result), mcols(methylationData2)[subjectHits(overlaps),
                                                                    which(grepl("reads",names(mcols(methylationData2))))])
      #generates col names
      new_col_names_M <- paste0("readsM", 1:
                                  length(grep("readsM", names(mcols(result)))))
      new_col_names_N <- paste0("readsN", 1:
                                  length(grep("readsN", names(mcols(result)))))

      names(mcols(result))[grep("readsM", names(mcols(result)))] <- new_col_names_M
      names(mcols(result))[grep("readsN", names(mcols(result)))] <- new_col_names_N
      # reordering mcols indexes are 1 to (trinuc -1), (trinuc +1) to the total
      # length, trinuc
      mcols(result) <- mcols(result)[c((1:(which(names(mcols(result)) ==
                                                   "trinucleotide_context")-1)), (which(names(mcols(
                                                     result)) == "trinucleotide_context")+1):length(
                                                       names(mcols(result))), which(names(mcols(result)) ==
                                                                                      "trinucleotide_context"))]
    } else{
      result <- methylationData1
      # columns are selected according to overlaps
      mcols(result) <- cbind(mcols(result), data.frame("readsM" = mcols(methylationData2)[
        subjectHits(overlaps), which(grepl("readsM", names(mcols(methylationData2))))],
        "readsN" = mcols(methylationData2)[subjectHits(overlaps),which(grepl("readsN",
                                                                             names(mcols(methylationData2))))]))
      # this selection allows to be generic and work all the time
      new_col_names_M <- paste0("readsM", 1:
                                  length(grep("readsM", names(mcols(result)))))
      new_col_names_N <- paste0("readsN", 1:
                                  length(grep("readsN", names(mcols(result)))))

      names(mcols(result))[grep("readsM", names(mcols(result)))] <- new_col_names_M
      names(mcols(result))[grep("readsN", names(mcols(result)))] <- new_col_names_N
      # reordering mcols indexes are 1 to (trinuc -1), (trinuc +1) to the total
      # length, trinuc
      mcols(result) <- mcols(result)[c((1:(which(names(mcols(result)) ==
                                                   "trinucleotide_context")-1)), (which(names(mcols(
                                                     result)) == "trinucleotide_context")+1):length(
                                                       names(mcols(result))), which(names(mcols(result)) ==
                                                                                      "trinucleotide_context"))]
    }

  }
  return(result)
  }

#' Checks whether the passed parameter is the statistical test
#'
#' @title Validate statistial test
#' @param test the statistical test used to call DMRs (\code{"betareg"} for
#' Beta regression.
#'
#' @author Alessandro Pio Greco and Nicolae Radu Zabet
.validateStatisticalTestReplicates <- function(test){
  .stopIfNotAll(c(!is.null(test), is.character(test), length(test) == 1, test %in% "betareg"),
                " test needs to be one of the following \"betareg\" for beta regression")
}

#' Checks whether the passed parameter is a valid condition for the data
#'
#' @title Validate the condition of the experiment
#' @param condition The vector containing the two conditions for the experiment.
#'
#' @author Alessandro Pio Greco and Nicolae Radu Zabet
.validateConditionReplicates <- function(condition, methylationData){
  .stopIfNotAll(c(!is.null(condition), is.character(condition), length(condition) ==
                    length(grep("readsM", names(mcols(methylationData))))),
                " condition need to be a vector of characters and the length needs to much the number of replicates")
  .stopIfNotAll(length(levels(as.factor(condition))) == 2,
                " only two levels are allowed for the condition")
}

#' This function computes the differentially methylated regions between
#' replicates with two conditions.
#'
#' @title Compute DMRs
#' @param methylationData the methylation data containing all the conditions
#' for all the replicates.
#' @param condition a vector of strings indicating the conditions for each
#' sample in \code{methylationData}. Two different values are allowed
#' (for the two conditions).
#' @param regions a \code{\link[GenomicRanges]{GRanges}} object with the regions where to
#' compute the DMRs. If \code{NULL}, the DMRs are computed genome-wide.
#' @param context the context in which the DMRs are computed (\code{"CG"},
#' \code{"CHG"} or \code{"CHH"}).
#' @param method the method used to compute the DMRs \code{"neighbourhood"}
#' or \code{"bins"}). The \code{"neighbourhood"} method computates
#' differentially methylated cytosines. Finally, the \code{"bins"}
#' method partiones the genome into equal sized tilling bins and performs the
#' statistical test between the two conditions in each bin. For all three
#' methods, the cytosines or bins are then merged into DMRs without affecting
#' the inital parameters used when calling the differentiall methylated
#' cytosines/bins (p-value, difference in methylation levels, minimum number of
#' reads per cytosine).
#' @param binSize the size of the tiling bins in nucleotides. This parameter is
#' required only if the selected method is \code{"bins"}.
#' @param test the statistical test used to call DMRs (\code{"betareg"} for
#' Beta regression).
#' @param pseudocountM numerical value to be added to the methylated reads
#' before modelling beta regression.
#' @param pseudocountN numerical value to be added to the total reads
#' before modelling beta regression.
#' @param pValueThreshold DMRs with p-values (when performing the statistical
#' test; see \code{test}) higher or equal than \code{pValueThreshold} are
#' discarded. Note that we adjust the p-values using the Benjamini and
#' Hochberg's method to control the false discovery rate.
#' @param minCytosinesCount DMRs with less cytosines in the specified context
#' than \code{minCytosinesCount} will be discarded.
#' @param minProportionDifference DMRs where the difference in methylation
#' proportion between the two conditions is lower than
#' \code{minProportionDifference} are discarded.
#' @param minGap DMRs separated by a gap of at least \code{minGap} are not
#' merged. Note that only DMRs where the change in methylation is in the same
#' direction are joined.
#' @param minSize DMRs with a size smaller than \code{minSize} are discarded.
#' @param minReadsPerCytosine  DMRs with the average number of reads lower than
#' \code{minReadsPerCytosine} are discarded.
#' @param parallel Logical; run in parallel if \code{TRUE}.
#' @param BPPARAM A \code{BiocParallelParam} object controlling parallel execution.
#'    This value will automatically set when parallel is \code{TRUE}, also able to set as manually.
#' @param cores Integer number of workers (must not exceed BPPARAM$workers).
#'    This value will automatically set as the maximum number of system workers,
#'    also able to set as manually.
#' @return the DMRs stored as a \code{\link[GenomicRanges]{GRanges}} object with the following
#' metadata columns:
#' \describe{
#'  \item{direction}{a number indicating whether the region lost (-1)  or gain
#'  (+1) methylation in condition 2 compared to condition 1.}
#'  \item{context}{the context in which the DMRs was computed (\code{"CG"},
#'  \code{"CHG"} or \code{"CHH"}).}
#'  \item{sumReadsM1}{the number of methylated reads in condition 1.}
#'  \item{sumReadsN1}{the total number of reads in condition 1.}
#'  \item{proportion1}{the proportion methylated reads in condition 1.}
#'  \item{sumReadsM2}{the number of methylated reads in condition 2.}
#'  \item{sumReadsN2}{the total number reads in condition 2.}
#'  \item{proportion2}{the proportion methylated reads in condition 2.}
#'  \item{cytosinesCount}{the number of cytosines in the DMR.}
#'  \item{regionType}{a string indicating whether the region lost (\code{"loss"})
#'  or gained (\code{"gain"}) methylation in condition 2 compared to condition 1.}
#'  \item{pValue}{the p-value (adjusted to control the false discovery rate with
#'  the Benjamini and Hochberg's method) of the statistical test when the DMR was
#'  called.}
#' }
#'
#' @examples
#'
#' \dontrun{
#' # starting with data joined using joinReplicates
#' data("syntheticDataReplicates")
#'
#' # compute the DMRs in CG context with neighbourhood method
#'
#' # creating condition vector
#' condition <- c("a", "a", "b", "b")
#'
#' # computing DMRs using the neighbourhood method
#' DMRsReplicatesNeighbourhood <- computeDMRsReplicates(methylationData = syntheticDataReplicates,
#'                                                      condition = condition,
#'                                                      regions = NULL,
#'                                                      context = "CHH",
#'                                                      method = "neighbourhood",
#'                                                      test = "betareg",
#'                                                      pseudocountM = 1,
#'                                                      pseudocountN = 2,
#'                                                      pValueThreshold = 0.01,
#'                                                      minCytosinesCount = 4,
#'                                                      minProportionDifference = 0.4,
#'                                                      minGap = 200,
#'                                                      minSize = 50,
#'                                                      minReadsPerCytosine = 4,
#'                                                      cores = 1)
#' }
#' @import betareg
#' @author Alessandro Pio Greco and Nicolae Radu Zabet
#' @export
computeDMRsReplicates <- function(methylationData,
                                  condition = NULL,
                                  regions = NULL,
                                  context = "CG",
                                  method= "neighbourhood",
                                  binSize = 100,
                                  test = "betareg",
                                  pseudocountM = 1,
                                  pseudocountN = 2,
                                  pValueThreshold = 0.01,
                                  minCytosinesCount = 4,
                                  minProportionDifference = 0.4,
                                  minGap = 200,
                                  minSize = 50,
                                  minReadsPerCytosine = 4,
                                  parallel = FALSE,
                                  BPPARAM = NULL,
                                  cores = NULL) {
  #Parameters checking
  cat("Parameters checking ...\n")

  # generate the BPPARAM value if set as parallel
  if (parallel == TRUE){
    BPPARAM <- suppressWarnings(.validateBPPARAM(BPPARAM, progressbar = TRUE))
  }else{
    # Force serial execution
    BPPARAM <- BiocParallel::SerialParam(progressbar = TRUE)
  }
  # If cores argument is specified
  if (!is.null(cores)) {
    .stopIfNotAll(.isInteger(cores, positive = TRUE), 
                  "the number of cores used when computing the DMRs needs to be an integer higher or equal to 1.")
    
    # Check if user requested more cores than available
    if (cores > BPPARAM$workers) {
      warning(paste0("The number of requested cores (", cores, 
                     ") exceeds the available system cores (", BPPARAM$workers, 
                     "). Automatically setting cores to the maximum available (", 
                     BPPARAM$workers, ")."))
      cores <- BPPARAM$workers
    } else {
      message(paste0("Using user-specified core count: ", cores))
    }
    
    # Apply the final core number
    BPPARAM$workers <- cores
  } else {
    cores <- BPPARAM$workers
  }
  cat("Current parallel setting, BPPARAM: ", capture.output(BPPARAM),sep = "\n")

  .validateMethylationData(methylationData, variableName="methylationData")

  .validateConditionReplicates(condition, methylationData)

  regions <- .validateGRanges(regions, methylationData)

  .validateContext(context)

  .stopIfNotAll(c(!is.null(method),
                  all(is.character(method)),
                  length(method) == 1,
                  all(method %in% c("neighbourhood","bins"))),
                " method can be only neighbourhood or bins")


  if(method == "bins"){
    .stopIfNotAll(c(.isInteger(binSize, positive=TRUE)),
                  " the bin size used by the method is an integer higher than 0")

  }

  .validateStatisticalTestReplicates(test)

  .stopIfNotAll(c(!is.null(pValueThreshold),
                  is.numeric(pValueThreshold),
                  pValueThreshold > 0,
                  pValueThreshold < 1),
                " the p-value threshold needs to be in the interval (0,1)")

  .stopIfNotAll(c(.isInteger(minCytosinesCount, positive=TRUE)),
                " the minCytosinesCount is an integer higher or equal to 0")

  .stopIfNotAll(c(!is.null(minProportionDifference),
                  is.numeric(minProportionDifference),
                  minProportionDifference > 0,
                  minProportionDifference < 1),
                " the minimum difference in methylation needs to be in the interval (0,1)")

  .stopIfNotAll(c(.isInteger(minGap, positive=TRUE)),
                " the minimum gap between DMRs is an integer higher or equal to 0")

  .stopIfNotAll(c(.isInteger(minSize, positive=TRUE)),
                " the minimum size of a DMR is an integer higher or equal to 0")

  .stopIfNotAll(c(.isInteger(minReadsPerCytosine, positive=TRUE)),
                " the minimum average number of reads in a DMR is an integer higher or equal to 0")

  computedDMRs <- GRanges()

  if(method == "neighbourhood"){
    computedDMRs <- .computeDMRsReplicatesNeighbourhood(methylationData = methylationData,
                                                        condition = condition,
                                                        regions = regions,
                                                        context = context,
                                                        pseudocountM = pseudocountM,
                                                        pseudocountN = pseudocountN,
                                                        pValueThreshold = pValueThreshold,
                                                        minCytosinesCount = minCytosinesCount,
                                                        minProportionDifference =minProportionDifference,
                                                        minGap = minGap,
                                                        minSize = minSize,
                                                        minReadsPerCytosine = minReadsPerCytosine,
                                                        BPPARAM = BPPARAM,
                                                        cores = cores)
  } else if(method == "bins"){
    computedDMRs <- .computeDMRsReplicatesBins(methylationData = methylationData,
                                               condition = condition,
                                               regions = regions,
                                               context = context,
                                               binSize = binSize,
                                               pseudocountM = pseudocountM,
                                               pseudocountN = pseudocountN,
                                               pValueThreshold = pValueThreshold,
                                               minCytosinesCount = minCytosinesCount,
                                               minProportionDifference =minProportionDifference,
                                               minGap = minGap,
                                               minSize = minSize,
                                               minReadsPerCytosine = minReadsPerCytosine,
                                               BPPARAM = BPPARAM,
                                               cores = cores)

  } else{
    cat("Unknown method: ",method," \n")
  }

  return(computedDMRs)
}

#' This function computes the differentially methylated regions between replicates
#' using the neighbourhood method.
.computeDMRsReplicatesNeighbourhood <- function(methylationData,
                                                condition,
                                                regions = NULL,
                                                context = "CG",
                                                pseudocountM = 1,
                                                pseudocountN = 2,
                                                pValueThreshold = 0.01,
                                                minCytosinesCount = 4,
                                                minProportionDifference = 0.4,
                                                minGap = 200,
                                                minSize = 50,
                                                minReadsPerCytosine = 4,
                                                BPPARAM = BPPARAM,
                                                cores = cores){
  condition <- as.factor(condition)

  regions <- reduce(regions)
  regionsList <- .splitGRangesEqualy(regions, cores)

  # extract the methylation data in the correct context
  cat("Extract methylation in the corresponding context \n")
  localContextMethylationData <- methylationData[methylationData$context%in%context]
  localContextMethylationData <- localContextMethylationData[queryHits(findOverlaps(
    localContextMethylationData, regions))]


  # create dataframe row wise with proportions
  m <- grep("readsM", names(mcols(localContextMethylationData)))
  n <- grep("readsN", names(mcols(localContextMethylationData)))

  # inner loop function for BiocParallel::bplapply
  .computeDMPsReplicatesNeighbourhoodLoop = function(i){
    computedDMPs <- GRanges()
    for(index in 1:length(regionsList[[i]])){
      cat("Computing DMRs \n")
      currentRegion <- regionsList[[i]][index]

      overlapsCs <- findOverlaps(localContextMethylationData, currentRegion)

      if(length(overlapsCs) > 0){
        localMethylationData <- localContextMethylationData[queryHits(overlapsCs)]

        proportions <-proportions <- (as.matrix(mcols(localMethylationData)[,m]) + pseudocountM) /
          (as.matrix(mcols(localMethylationData)[,n]) + pseudocountN)

        DMPs <- GRanges()
        if(length(localMethylationData) > 0){
          DMPs <- localMethylationData
          DMPs$pValue <- .computeAdjuestedPValuesReplicates(proportions, condition, cores=1)
          DMPs <- DMPs[!is.na(DMPs$pValue)]
          if(length(DMPs) > 0){
            DMPs$sumReadsM1 <- apply(mcols(DMPs)[m[which(condition == unique(condition)[1])]],1,sum)
            DMPs$sumReadsN1 <- apply(mcols(DMPs)[n[which(condition == unique(condition)[1])]],1,sum)
            DMPs$proportion1 <- DMPs$sumReadsM1 / DMPs$sumReadsN1
            DMPs$sumReadsM2 <- apply(mcols(DMPs)[m[which(condition == unique(condition)[2])]],1,sum)
            DMPs$sumReadsN2 <- apply(mcols(DMPs)[n[which(condition == unique(condition)[2])]],1,sum)
            DMPs$proportion2 <- DMPs$sumReadsM2 / DMPs$sumReadsN2
            DMPs$cytosinesCount <- 1
            DMPs$direction <- sign(DMPs$proportion2 - DMPs$proportion1)
          }
        }
        if(length(DMPs) > 0){
          bufferIndex <- DMPs$pValue < pValueThreshold &
            abs(DMPs$proportion2 - DMPs$proportion1) >= minProportionDifference &
            DMPs$sumReadsN1 >=minReadsPerCytosine &
            DMPs$sumReadsN2 >=minReadsPerCytosine
          DMPs <- DMPs[bufferIndex]
          strand(DMPs) <- "*"
        }

        if(is.null(DMPs)){
          DMPs <- GRanges()
        }


        # append current DMRs to the global list of DMRs
        if(length(computedDMPs) == 0){
          computedDMPs <- DMPs
        } else{
          computedDMPs <- c(computedDMPs,DMPs)
        }
      }
    }
    return(computedDMPs)
  }


  # compute the DMRs
  if(cores > 1){
    cat("Compute the DMRs using ", cores, "cores\n")
    computedDMPs <- BiocParallel::bplapply(1:length(regionsList), .computeDMPsReplicatesNeighbourhoodLoop, BPPARAM = BPPARAM)
  } else {
    computedDMPs <- lapply(1:length(regionsList), .computeDMPsReplicatesNeighbourhoodLoop)
  }

 computedDMRs <- GRanges()
  computedDMRs <- subset(computedDMPs, !sapply(computedDMPs, is.null))
  if(length(computedDMRs) > 0){
    computedDMRs <- unlist(GRangesList(computedDMRs))
    if(length(computedDMRs) > 0){
      cat("Merge adjacent DMRs\n")
      computedDMRs <- computedDMRs[order(computedDMRs)]
      computedDMRs$pValue <- .computeaAjustedPValuesInDMRsReplicates(localContextMethylationData ,computedDMRs, condition, cores, m, n, pseudocountM, pseudocountN)
      cat("Merge DMRs iteratively\n")
      # Get rid of small gaps between DMRs.
      if(minGap > 0){
        computedDMRs <- .smartMergeDMRsReplicates(computedDMRs,
                                                  minGap = minGap,
                                                  respectSigns = TRUE,
                                                  methylationData = localContextMethylationData,
                                                  minProportionDifference=minProportionDifference,
                                                  minReadsPerCytosine = minReadsPerCytosine,
                                                  pseudocountM = pseudocountM,
                                                  pseudocountN = pseudocountN,
                                                  pValueThreshold=pValueThreshold,
                                                  condition = condition,
                                                  m = m,
                                                  n = n,
                                                  cores = cores,
                                                  BPPARAM = BPPARAM)
      }


      cat("Filter DMRs \n")
      if(length(computedDMRs) > 0){
        #remove small DMRs
        computedDMRs <- computedDMRs[width(computedDMRs) >= minSize]
        if(length(computedDMRs) > 0){
          #remove DMRs with few cytosines
          computedDMRs <- computedDMRs[computedDMRs$cytosinesCount >= minCytosinesCount]
          #recompute the adjusted p-values
          if(length(computedDMRs) > 0){
            computedDMRs$pValue <- .computeaAjustedPValuesInDMRsReplicates(localContextMethylationData ,computedDMRs, condition, cores, m, n, pseudocountM, pseudocountN)
            computedDMRs$regionType <- rep("loss", length(computedDMRs))
            computedDMRs$regionType[which(computedDMRs$proportion1 < computedDMRs$proportion2)] <- "gain"

          }
        }
      }
    }
  }

  return(computedDMRs)
}

#' This function computes the differentially methylated regions between replicates
#' using the bins method.
.computeDMRsReplicatesBins <- function(methylationData,
                                       condition = condition,
                                       regions = NULL,
                                       context = "CG",
                                       binSize = 100,
                                       pseudocountM = 1,
                                       pseudocountN = 2,
                                       pValueThreshold = 0.01,
                                       minCytosinesCount = 4,
                                       minProportionDifference = 0.4,
                                       minGap = 200,
                                       minSize = 50,
                                       minReadsPerCytosine = 4,
                                       BPPARAM = BPPARAM,
                                       cores = cores) {

  condition <- as.factor(condition)

  regions <- reduce(regions)

  # extract the methylation data in the correct context
  cat("Extract methylation in the corresponding context \n")
  localContextMethylationData <- methylationData[methylationData$context%in%context]
  localContextMethylationData <- localContextMethylationData[queryHits(findOverlaps(
    localContextMethylationData, regions))]

  regionsList <- .splitGRangesEqualy(regions, cores)

  m <- grep("readsM", names(mcols(localContextMethylationData)))
  n <- grep("readsN", names(mcols(localContextMethylationData)))

  # inner loop function for BiocParallel::bplapply
  .computeDMRsReplicatesBinsLoop = function(i){
    computedDMRs <- GRanges()
    for(index in 1:length(regionsList[[i]])){
      currentRegion <- regionsList[[i]][index]

      cat("Computing DMRs at ",.printGenomicRanges(currentRegion),"\n")

      seqs <- seq(start(currentRegion), (end(currentRegion)-binSize), by = binSize);

      bins <- GRanges(seqnames(currentRegion), IRanges(seqs, (seqs+binSize-1)))

      overlapsBins <- findOverlaps(localContextMethylationData, currentRegion)

      if(length(overlapsBins) > 0){
        localMethylationData <- localContextMethylationData[queryHits(overlapsBins)]

        cat("Count inside each bin...\n")
        #bins <- .analyseReadsInsideRegions(localMethylationData, bins, context, cores)
        bins <- .analyseReadsInsideBinsReplicates(localMethylationData, bins, currentRegion, condition, pseudocountM, pseudocountN)

        cat("Filter the bins...\n")
        # Get rid of the bins with fewer than minCytosinesCount cytosines inside.
        bins  <- bins[bins$cytosinesCount >= minCytosinesCount]


        # Get rid of the bins with fewer than minReadsPerCytosine reads per cytosine.
        bins  <- bins[(bins$sumReadsN1/bins$cytosinesCount >= minReadsPerCytosine) &
                        (bins$sumReadsN2/bins$cytosinesCount >= minReadsPerCytosine)]

        # Get rid of the bins with small difference in proportion of methylation
        bins  <- bins[(abs(bins$proportion1 - bins$proportion2) >= minProportionDifference)]

        proportions <- as.matrix(mcols(bins)[grep("proportionsR", names(mcols(bins)))])

        if(nrow(proportions) > 0){
          cat("Identifying DMRs...\n")
          pValue <- .computeAdjuestedPValuesReplicates(proportions, condition, cores)
          mcols(bins)[grep("proportionsR", names(mcols(bins)))] <- NULL
          bins <- bins[!is.na(pValue) & pValue < pValueThreshold ]
          if(length(bins) > 0){
            bins$context <- rep(paste(context, collapse = "_"), length(bins))
            bins$direction <- rep(NA, length(bins))
            bins$direction <- sign(bins$proportion2 - bins$proportion1)
            # Select the crude list of DMRs
            DMRs <- bins[!is.na(bins$direction) & (bins$direction == 1 | bins$direction == -1)]
          }
        } else{
          DMRs <- bins
        }

        if(is.null(DMRs)){
          DMRs <- GRanges()
        }

        # append current DMRs to the global list of DMRs
        if(length(computedDMRs) == 0){
          computedDMRs <- DMRs
        } else{
          computedDMRs <- c(computedDMRs,DMRs)
        }
      }
    }
    return(computedDMRs)
  }

  # compute the DMRs
  if(cores > 1){
    cat("Compute the DMRs using ", cores, "cores\n")
    computedDMRs <- BiocParallel::bplapply(1:length(regionsList), .computeDMRsReplicatesBinsLoop, BPPARAM = BPPARAM)
  } else {
    computedDMRs <- lapply(1:length(regionsList), .computeDMRsReplicatesBinsLoop)
  }

  if(length(computedDMRs) > 0){
    computedDMRs <- subset(computedDMRs, !sapply(computedDMRs, is.null))
    if(length(computedDMRs) > 0){
      computedDMRs <- unlist(GRangesList(computedDMRs))
    } else{
      computedDMRs <- GRanges()
    }


    if(length(computedDMRs) > 0){

      cat("Merge adjacent DMRs\n")
      computedDMRs <- computedDMRs[order(computedDMRs)]
      computedDMRs$pValue <- .computeaAjustedPValuesInDMRsReplicates(localContextMethylationData ,computedDMRs, condition, cores, m, n, pseudocountM, pseudocountN)
      cat("Merge DMRs iteratively\n")
      # Get rid of small gaps between DMRs.
      if(minGap > 0){
        computedDMRs <- .smartMergeDMRsReplicates(computedDMRs,
                                                  minGap = minGap,
                                                  respectSigns = TRUE,
                                                  methylationData = localContextMethylationData,
                                                  minProportionDifference=minProportionDifference,
                                                  minReadsPerCytosine = minReadsPerCytosine,
                                                  pseudocountM = pseudocountM,
                                                  pseudocountN = pseudocountN,
                                                  pValueThreshold=pValueThreshold,
                                                  condition = condition,
                                                  m = m,
                                                  n = n,
                                                  cores = cores,
                                                  BPPARAM = BPPARAM)
      }

      cat("Filter DMRs \n")
      if(length(computedDMRs) > 0){
        #remove small DMRs
        computedDMRs <- computedDMRs[width(computedDMRs) >= minSize]

        if(length(computedDMRs) > 0){
          #remove DMRs with few cytosines
          computedDMRs <- computedDMRs[computedDMRs$cytosinesCount >= minCytosinesCount]
          #recompute the adjusted p-values
          if(length(computedDMRs) > 0){
            computedDMRs$pValue <- .computeaAjustedPValuesInDMRsReplicates(localContextMethylationData ,computedDMRs, condition, cores, m, n, pseudocountM, pseudocountN)
            computedDMRs$regionType <- rep("loss", length(computedDMRs))
            computedDMRs$regionType[which(computedDMRs$proportion1 < computedDMRs$proportion2)] <- "gain"
          }
        }
      }
    }
  } else{
    computedDMRs <- GRanges()
  }
  ## remove proportions R1,2,3,4..
  cols_to_remove <- grep("^proportionsR[1-4]+$", names(mcols(computedDMRs)), value = TRUE)
  if (length(cols_to_remove) > 0) {
    mcols(computedDMRs)[, cols_to_remove] <- NULL
  }
  return(computedDMRs)
}

#' This function extracts the p-value for each cytosine from the betareg object
.convertResult <- function(x){
  if(any(is.na(x))){
    result <- NA
  } else{
    result <- summary(x)[[1]][[1]][,4][2]
  }
  return(result)
}

# formula for eliminating extreme values (0,1) from the proportion matrix
.convertProportions <- function(x){
  result <- (x *(length(x) - 1) + 0.5)/length(x)
  return(result)
}

#' This function computes the p-values of the beta regression test
#'
#' @title Beta regression
#' @param y vector containing the proportions (readsM/readsN) for a single cytosine.
#' @param condition vector containing the conditions of the experiment.
#' @return The p-values of the beta regression test.
#'
#' @author Alessandro Pio Greco and Nicolae Radu Zabet
.computeBetaRegSingle <- function(y, condition){
  if(!.suitableForBetaReg(y)){
    result <- NA
  } else{
    result <- betareg(.convertProportions(y) ~ condition, na.action = na.pass)
  }
  return(result)
}

#https://stats.stackexchange.com/questions/220868/questionable-beta-regression-results
#https://stats.stackexchange.com/questions/89999/how-to-replicate-statas-robust-binomial-glm-for-proportion-data-in-r/205040#205040
.suitableForBetaReg <- function(y){
  if(any(is.na(y))){
    return(FALSE)
  } else if(length(unique(y)) < 4 | any(y==0) | any(y==1)){
    return(FALSE)
  }
  return(TRUE)
}

#' This function applies the .computeBetaRegSingle to the entire dataset
.computeBetaReg <- function(y, condition, cores){
  if(is.null(nrow(y))){
    result <- .computeBetaRegSingle(y, condition)
    result <- .convertResult(result)
  } else {
    suitableForBetaReg <- apply(y, 1, .suitableForBetaReg)
    result <- rep(NA, nrow(y))
    if(sum(suitableForBetaReg) >= 1){
      ySuitable <- matrix(y[suitableForBetaReg,], ncol=length(condition))
      #buffer <- mclapply(1:nrow(ySuitable), FUN = function(x) .computeBetaRegSingle(ySuitable[x,], condition), mc.cores = cores)
      buffer <- apply(ySuitable, 1, .computeBetaRegSingle, condition=condition)
      buffer <- vapply(buffer, FUN=.convertResult, FUN.VALUE = 0.0)
      #buffer <- unlist(buffer)
      result[suitableForBetaReg] <- buffer
    }
  }
  return(result)
}

#' This function computes the adjusted p-values (using Benjamini & Hochberg
#' method)
#'
#'
#' @title Compute adjusted p-values
#' @param conditions matrix containing the proportions (readsM/readsN) for all the cytosines.
#' @param condition vector containing the conditions of the experiment.
#' @param cores the number of cores to be used to fit the model.
#' @return The adjusted p-values of the statistical test.
#' @author  Nicolae Radu Zabet and Alessandro Pio Greco
.computeAdjuestedPValuesReplicates <- function(proportions, condition, cores){

  pValue <- .computeBetaReg(proportions, condition, cores)
  # convert p-values to FDR
  adjPValue <- rep(NA, times=length(pValue))

  adjPValue[which(!is.na(pValue))] <- stats::p.adjust(pValue[which(!is.na(pValue))], method="fdr")

  return(adjPValue)
}

#' This function computes the adjusted p-values (using Benjamini & Hochberg
#' method)
#'
#'
#' @title Compute adjusted p-values
#' @param conditions matrix containing the proportions (readsM/readsN) for all the cytosines.
#' @param condition vector containing the conditions of the experiment.
#' @param cores the number of cores to be used to fit the model.
#' @return The adjusted p-values of the statistical test.
#' @author  Nicolae Radu Zabet and Alessandro Pio Greco
.computeaAjustedPValuesInDMRsReplicates <- function(methylationData,
                                                    DMRs,
                                                    condition,
                                                    cores,
                                                    indexM,
                                                    indexN,
                                                    pseudocountM,
                                                    pseudocountN){

  M <- sapply(1:length(DMRs),
              function(x){.computeProportionsInDMRs(DMRs[x],
                                                    methylationData = methylationData,
                                                    col_indexes = indexM)})
  N <- sapply(1:length(DMRs),
              function(x){.computeProportionsInDMRs(DMRs[x],
                                                    methylationData = methylationData,
                                                    col_indexes = indexN)})

  proportions <- (M + pseudocountM) / (N + pseudocountN)
  proportions <- t(proportions)


  pValue <- .computeBetaReg(proportions, condition, cores)


  # convert p-values to FDR
  adjPValue <- rep(NA, times=length(pValue))

  adjPValue[which(!is.na(pValue))] <- stats::p.adjust(pValue[which(!is.na(pValue))], method="fdr")

  return(adjPValue)
}

#' This function recalculate proportions between methylated and total reads in a new region.
.computeProportionsInDMRs <- function(DMRs, methylationData,col_indexes){
  regions <- methylationData[queryHits(findOverlaps(methylationData, DMRs))]
  computedDMRsproportions <- apply(mcols(regions)[col_indexes], 2, sum)
  return(computedDMRsproportions)
}

#' Performs the analysis in equal width regions of an \code{\link[GenomicRanges]{GRanges}}
#' object
#'
#' @title Analyse reads inside regions
#' @param methylationData a \code{\link[GenomicRanges]{GRanges}} object with five metadata
#' columns; see \code{\link{methylationDataList}}
#' @param currentRegion a \code{\link[GenomicRanges]{GRanges}} object with the identified regions
#' @param condition The vector containing the two conditions for the experiment.
#' @param pseudocountM numerical value to be added to the methylated reads
#' before modelling beta regression.
#' @param pseudocountN numerical value to be added to the total reads
#' before modelling beta regression.
#' @return a \code{\link[GenomicRanges]{GRanges}} object with eaual sized tiles of the regions.
#' The object consists of the following metadata
#' \describe{
#'  \item{sumReadsM1}{the number of methylated reads in condition 1}
#'  \item{sumReadsN1}{the total number of reads in condition 1}
#'  \item{proportion1}{the proportion of methylated reads in condition 1}
#'  \item{sumReadsM2}{the number of methylated reads in condition 2}
#'  \item{sumReadsN2}{the total number of reads in condition 2}
#'  \item{proportion2}{the proportion of methylated reads in condition 2}
#'  \item{cytosinesCount}{the number of cytosines in the correct context}
#' }
#'
#' @author Nicolae Radu Zabet and Alessandro Pio Greco
.analyseReadsInsideBinsReplicates <- function(methylationData, bins, currentRegion,
                                              condition, pseudocountM, pseudocountN){


  binSize <- min(unique(width(bins)))
  #Rcpp
  m <- grep("readsM", names(mcols(methylationData)))
  n <- grep("readsN", names(mcols(methylationData)))
  readsM <- matrix(0, ncol = length(m), nrow=length(bins))
  for(i in 1:length(m)){
    test <- .movingSum(start(currentRegion), end(currentRegion), start(methylationData), mcols(methylationData)[[m[i]]], windowSize = binSize)
    readsM[,i] <- test[seq(1,length(test)-binSize, by=binSize)]
  }

  readsN <- matrix(0, ncol = length(n), nrow=length(bins))
  for(i in 1:length(n)){
    test2 <- .movingSum(start(currentRegion), end(currentRegion), start(methylationData), mcols(methylationData)[[n[i]]], windowSize = binSize)
    readsN[,i] <- test2[seq(1,length(test)-binSize, by=binSize)]
  }

  proportions <- (readsM + pseudocountM)/ (readsN + pseudocountN)
  proportions <- as.data.frame(proportions)
  names_prop <- paste0("proportionsR", 1:ncol(proportions))
  colnames(proportions) <- names_prop

  m1 <- m[which(condition == unique(condition)[1])]
  n1 <- n[which(condition == unique(condition)[1])]
  m2 <- m[which(condition == unique(condition)[2])]
  n2 <- n[which(condition == unique(condition)[2])]

  # sumReadsM1 <- sum(readsM[,which(condition == unique(condition)[1])])
  # sumReadsN1 <- sum(readsN[,which(condition == unique(condition)[1])])
  # proportion1 <- (sumReadsM1 + pseudocountM)/ (sumReadsN1 + pseudocountN)
  #
  # sumReadsM2 <- sum(readsM[,which(condition == unique(condition)[2])])
  # sumReadsN2 <- sum(readsN[,which(condition == unique(condition)[2])])
  # proportion2 <- (sumReadsM2 + pseudocountM)/ (sumReadsN2 + pseudocountN)

  readsM1 <- readsM[,which(condition == unique(condition)[1])]
  # readsM1 <- matrix(0, ncol = length(m1), nrow=length(bins))
  # for(i in 1:length(m1)){
  #   test <- .movingSum(start(currentRegion), end(currentRegion), start(methylationData), mcols(methylationData)[[m1[i]]], windowSize = binSize)
  #   readsM1[,i] <- test[seq(1,length(test)-binSize, by=binSize)]
  # }
  # sumReadsM1 <-  apply(readsM1,1,sum)
  if (is.null(dim(readsM1))) {
    sumReadsM1 <- sum(readsM1)
  } else {
    sumReadsM1 <- rowSums(readsM1)
  }


  readsN1 <- readsN[,which(condition == unique(condition)[1])]
  # readsN1 <- matrix(0, ncol = length(n1), nrow=length(bins))
  # for(i in 1:length(n1)){
  #   test <- .movingSum(start(currentRegion), end(currentRegion), start(methylationData), mcols(methylationData)[[n1[i]]], windowSize = binSize)
  #   readsN1[,i] <- test[seq(1,length(test)-binSize, by=binSize)]
  # }
  # sumReadsN1 <- apply(readsN1,1,sum)
  if (is.null(dim(readsN1))) {
    sumReadsN1 <- sum(readsN1)
  } else {
    sumReadsN1 <- rowSums(readsN1)
  }
  proportion1 <- (sumReadsM1 + pseudocountM)/ (sumReadsN1 + pseudocountN)


  readsM2 <- readsM[,which(condition == unique(condition)[2])]
  # readsM2 <- matrix(0, ncol = length(m2), nrow=length(bins))
  # for(i in 1:length(m2)){
  #   test <- .movingSum(start(currentRegion), end(currentRegion), start(methylationData), mcols(methylationData)[[m2[i]]], windowSize = binSize)
  #   readsM2[,i] <- test[seq(1,length(test)-binSize, by=binSize)]
  # }
  # sumReadsM2 <- apply(readsM2,1,sum)
  if (is.null(dim(readsM2))) {
    sumReadsM2 <- sum(readsM2)
  } else {
    sumReadsM2 <- rowSums(readsM2)
  }

  readsN2 <- readsN[,which(condition == unique(condition)[2])]
  # readsN2 <- matrix(0, ncol = length(n2), nrow=length(bins))
  # for(i in 1:length(n2)){
  #   test <- .movingSum(start(currentRegion), end(currentRegion), start(methylationData), mcols(methylationData)[[n2[i]]], windowSize = binSize)
  #   readsN2[,i] <- test[seq(1,length(test)-binSize, by=binSize)]
  # }
  # sumReadsN2 <- apply(readsN2,1,sum)
  if (is.null(dim(readsN2))) {
    sumReadsN2 <- sum(readsN2)
  } else {
    sumReadsN2 <- rowSums(readsN2)
  }
  proportion2 <- (sumReadsM2 + pseudocountM)/ (sumReadsN2 + pseudocountN)

  cytosines <- .movingSum(start(currentRegion), end(currentRegion), start(methylationData), rep(1, length(start(methylationData))), windowSize = binSize)
  cytosinesCount <- cytosines[seq(1,length(cytosines)-binSize, by=binSize)]


  bins$sumReadsM1 <- sumReadsM1
  bins$sumReadsN1 <- sumReadsN1
  bins$proportion1 <- proportion1
  bins$sumReadsM2 <- sumReadsM2
  bins$sumReadsN2 <- sumReadsN2
  bins$proportion2 <- proportion2
  bins$cytosinesCount <- cytosinesCount
  mcols(bins) <- cbind(mcols(bins), proportions)
  return(bins)

}

#' Performs the analysis in all regions in a \code{\link[GenomicRanges]{GRanges}} object
#'
#' @title Analyse reads inside regions
#' @param methylationData a \code{\link[GenomicRanges]{GRanges}} object with five metadata
#' columns see \code{\link{methylationDataList}}
#' @param regions a \code{\link[GenomicRanges]{GRanges}} object with the identified regions
#' @param condition The vector containing the two conditions for the experiment.
#' @param m indexes of methylated reads for creation of the proportions matrix
#' @param n indexes of total reads for creation of the proportions matrix
#' @return a \code{\link[GenomicRanges]{GRanges}} object with eaual sized tiles of the regions.
#' The object consists of the following metadata
#' \describe{
#'  \item{sumReadsM1}{the number of methylated reads in condition 1}
#'  \item{sumReadsN1}{the total number of reads in condition 1}
#'  \item{proportion1}{the proportion of methylated reads in condition 1}
#'  \item{sumReadsM2}{the number of methylated reads in condition 2}
#'  \item{sumReadsN2}{the total number of reads in condition 2}
#'  \item{proportion2}{the proportion of methylated reads in condition 2}
#'  \item{cytosinesCount}{the number of cytosines in the correct context}
#' }
#'
#' @author Nicolae Radu Zabet and Alessandro Pio Greco
.analyseReadsInsideRegionsReplicates <- function(methylationData, regions, condition, m, n){
  overlaps <- findOverlaps(methylationData, regions, ignore.strand = TRUE)
  methylationDataContextList <- S4Vectors::splitAsList(methylationData[queryHits(overlaps)],  subjectHits(overlaps))
  regionsIndexes <- as.integer(names(methylationDataContextList))
  # methylationDataContextList <-
  # regions <- methylationData[queryHits(findOverlaps(methylationData, DMRs))]

  regions$sumReadsM1 <- rep(0, times=length(regions))
  regions$sumReadsN1 <- rep(0, times=length(regions))
  regions$proportion1 <- rep(0, times=length(regions))
  regions$sumReadsM2 <- rep(0, times=length(regions))
  regions$sumReadsN2 <- rep(0, times=length(regions))
  regions$proportion2 <- rep(0, times=length(regions))
  regions$cytosinesCount <- rep(0, times=length(regions))

  if(length(regionsIndexes) > 0){
    regions$sumReadsM1[regionsIndexes] <- sum(apply(mcols(unlist(methylationDataContextList))[m[which(condition == unique(condition)[1])]],1,sum))
    regions$sumReadsM2 <- sum(apply(mcols(unlist(methylationDataContextList))[m[which(condition == unique(condition)[2])]],1,sum))
    regions$sumReadsN1 <- sum(apply(mcols(unlist(methylationDataContextList))[n[which(condition == unique(condition)[1])]],1,sum))
    regions$sumReadsN2 <- sum(apply(mcols(unlist(methylationDataContextList))[n[which(condition == unique(condition)[2])]],1,sum))
    regions$cytosinesCount[regionsIndexes] <- sapply(methylationDataContextList,length)

    valid <- regions$cytosinesCount[regionsIndexes] > 0
    regions$proportion1[regionsIndexes[valid]] <- regions$sumReadsM1[regionsIndexes[valid]]/regions$sumReadsN1[regionsIndexes[valid]]
    regions$proportion2[regionsIndexes[valid]] <- regions$sumReadsM2[regionsIndexes[valid]]/regions$sumReadsN2[regionsIndexes[valid]]
  }
  return(regions)
}



.joinDMRsReplicates <- function(DMRs,
                                minGap = minGap,
                                respectSigns = TRUE,
                                methylationData = methylationData,
                                minProportionDifference=0.4,
                                minReadsPerCytosine = 4,
                                pseudocountM = pseudocountM,
                                pseudocountN = pseudocountN,
                                pValueThreshold=0.01,
                                condition = condition,
                                m = m,
                                n = n,
                                cores = 1
){

  #are within the requeired distance
  if(length(reduce(DMRs, drop.empty.ranges=TRUE, min.gapwidth=minGap,
                   ignore.strand=TRUE)) == 1){
    #all have the same direction
    if(!respectSigns | length(unique(DMRs$direction)) == 1){
      direction <- unique(DMRs$direction)
      if(length(unique(strand(DMRs))) == 1){
        localDMR <- DMRs[1]
        end(localDMR) <- max(end(DMRs))
        start(localDMR) <- min(start(DMRs))
        localDMR <- .analyseReadsInsideRegionsReplicates(methylationData, localDMR, condition, m, n)
        localDMR$pValue <- .computeaAjustedPValuesInDMRsReplicates(methylationData, localDMR, condition, cores, m, n, pseudocountM, pseudocountN)
        if(!is.na(localDMR$pValue)){
          if(abs(localDMR$proportion1 - localDMR$proportion2) >= minProportionDifference &
             localDMR$pValue <= pValueThreshold &
             (localDMR$sumReadsN1 / localDMR$cytosinesCount) >= minReadsPerCytosine &
             (localDMR$sumReadsN2 / localDMR$cytosinesCount) >= minReadsPerCytosine){
            DMRs <- localDMR
          }
        }
      }
    }
  }
  return(DMRs)
}

.getLongestDMRsReplicates <- function(DMRs,
                                      minGap = minGap,
                                      respectSigns = TRUE,
                                      methylationData = methylationData,
                                      minProportionDifference=0.4,
                                      minReadsPerCytosine = 4,
                                      pseudocountM = pseudocountM,
                                      pseudocountN = pseudocountN,
                                      pValueThreshold=0.01,
                                      condition = condition,
                                      m = m,
                                      n = n,
                                      cores = 1){

  newDMRs <-.joinDMRsReplicates(DMRs,
                                minGap = minGap,
                                respectSigns = respectSigns,
                                methylationData = methylationData,
                                minProportionDifference=minProportionDifference,
                                minReadsPerCytosine = minReadsPerCytosine,
                                pseudocountM = pseudocountM,
                                pseudocountN = pseudocountN,
                                pValueThreshold=pValueThreshold,
                                condition = condition,
                                m = m,
                                n = n,
                                cores = 1)
  if(length(newDMRs) == 1){
    result <- newDMRs
  } else{
    result <- .mergeDMRsIterativelyReplicates(DMRs,
                                              minGap = minGap,
                                              respectSigns = respectSigns,
                                              methylationData = methylationData,
                                              minProportionDifference=minProportionDifference,
                                              minReadsPerCytosine = minReadsPerCytosine,
                                              pseudocountM = pseudocountM,
                                              pseudocountN = pseudocountN,
                                              pValueThreshold=pValueThreshold,
                                              condition = condition,
                                              m = m,
                                              n = n,
                                              cores = 1)

  }
  return(result)
}

#' This takes a list of DMRs and attempts to merge DMRs while keeping the new
#' DMRs statistically significant
.smartMergeDMRsReplicates <- function(DMRs,
                                      minGap,
                                      respectSigns = TRUE,
                                      methylationData,
                                      minProportionDifference=0.4,
                                      minReadsPerCytosine = 4,
                                      pseudocountM = 1,
                                      pseudocountN = 2,
                                      pValueThreshold=0.01,
                                      condition = condition,
                                      m = m,
                                      n = n,
                                      cores = 1,
                                      BPPARAM = BPPARAM){



  overlaps <- countOverlaps(DMRs, DMRs, maxgap = minGap, ignore.strand = TRUE)
  notToJoin <- DMRs[overlaps == 1]

  DMRsToJoin <- DMRs[overlaps > 1]


  if(length(DMRsToJoin) > 0){
    overlaps <- findOverlaps(DMRsToJoin,
                             reduce(DMRsToJoin, min.gapwidth = minGap,
                                    ignore.strand=TRUE),
                             maxgap = minGap, ignore.strand = TRUE)
    DMRsList <- S4Vectors::splitAsList(DMRsToJoin[queryHits(overlaps)],
                                     subjectHits(overlaps))



    if(cores > 1){
      bufferDMRs <- BiocParallel::bplapply(1:length(DMRsList), function(i){ .getLongestDMRsReplicates(DMRsList[[i]],
                                                                                                  minGap = minGap,
                                                                                                  respectSigns = respectSigns,
                                                                                                  methylationData = methylationData,
                                                                                                  minProportionDifference=minProportionDifference,
                                                                                                  minReadsPerCytosine = minReadsPerCytosine,
                                                                                                  pseudocountM = pseudocountM,
                                                                                                  pseudocountN = pseudocountN,
                                                                                                  pValueThreshold=pValueThreshold,
                                                                                                  condition = condition,
                                                                                                  m = m,
                                                                                                  n = n,
                                                                                                  cores = 1)},
                                       BPPARAM = BPPARAM)

      bufferDMRs <- unlist(GRangesList(bufferDMRs))
    } else{
      bufferDMRs <- GRanges()
      for(i in 1:length(DMRsList)){
        bufferDMRs <- c(bufferDMRs, .getLongestDMRsReplicates(DMRsList[[i]],
                                                              minGap = minGap,
                                                              respectSigns = respectSigns,
                                                              methylationData = methylationData,
                                                              minProportionDifference=minProportionDifference,
                                                              minReadsPerCytosine = minReadsPerCytosine,
                                                              pseudocountM = pseudocountM,
                                                              pseudocountN = pseudocountN,
                                                              pValueThreshold=pValueThreshold,
                                                              condition = condition,
                                                              m = m,
                                                              n = n,
                                                              cores = 1))
      }
    }

    joinedDMRs <- c(bufferDMRs, notToJoin)
  } else{
    joinedDMRs <- notToJoin
  }

  joinedDMRs <- joinedDMRs[order(joinedDMRs)]

  return(joinedDMRs)
}

.mergeDMRsIterativelyReplicates <- function(DMRs,
                                            minGap,
                                            respectSigns = TRUE,
                                            methylationData,
                                            minProportionDifference=0.4,
                                            minReadsPerCytosine = 4,
                                            pseudocountM = pseudocountM,
                                            pseudocountN = pseudocountN,
                                            pValueThreshold=0.01,
                                            condition = condition,
                                            m = m,
                                            n = n,
                                            cores = 1){

  overlaps <- countOverlaps(DMRs, DMRs, maxgap = minGap, ignore.strand = TRUE)
  notToJoin <- DMRs[overlaps == 1]
  DMRs <- DMRs[overlaps > 1]

  joinedAny <- TRUE
  iteration <- 1
  while(joinedAny){
    joinedAny <- FALSE
    bufferDMRs <-GRanges()
    index <- 1
    localIndex <- index
    joinedCount <- 0
    while(localIndex < length(DMRs)){
      localDMRs <- DMRs[index]
      canJoin <- TRUE
      while(canJoin){
        localIndex <- localIndex + 1
        newDMRs <-.joinDMRsReplicates(c(localDMRs, DMRs[localIndex]),
                                      minGap = minGap,
                                      respectSigns = respectSigns,
                                      methylationData = methylationData,
                                      minProportionDifference=minProportionDifference,
                                      minReadsPerCytosine = minReadsPerCytosine,
                                      pseudocountM = pseudocountM,
                                      pseudocountN = pseudocountN,
                                      pValueThreshold=pValueThreshold,
                                      condition = condition,
                                      m = m,
                                      n = n,
                                      cores = 1)

        if(length(newDMRs) == 1){
          joinedCount <- joinedCount + 1
          localDMRs <- newDMRs
          joinedAny <- TRUE
          if(localIndex == length(DMRs)){
            canJoin <- FALSE
            bufferDMRs <- c(bufferDMRs, newDMRs)
          }
        } else{
          bufferDMRs <- c(bufferDMRs, localDMRs)
          canJoin <- FALSE
          index <- localIndex
          if(localIndex == length(DMRs)){
            bufferDMRs <- c(bufferDMRs, DMRs[localIndex])
          }
        }
      }
    }
    DMRs <- bufferDMRs
    iteration <- iteration + 1
  }

  DMRs <- c(DMRs, notToJoin)

  DMRs <- DMRs[order(DMRs)]

  return(DMRs)
}

---
title: "Overview of the DMRcaller package"
author: "Nicolae Radu Zabet\\thanks{e-mail: \\texttt{r.zabet@qmul.ac.uk}, Blizard
  Institute, Barts and The London School of Medicine and Dentistry, Queen Mary University
  of London, UK}"
date: "`r Sys.Date()`"
output:
  bookdown::html_document2:
    toc: true
    toc_depth: 3
    number_sections: true
    toc_float: true
    fig_caption: true
  bookdown::pdf_document2:
    toc: true
    toc_depth: 3
    number_sections: true
    fig_caption: true
  BiocStyle::html_document:
bibliography: DMRcaller.bib
csl: apa.csl
link-citations: true
vignette: >
  %\VignetteIndexEntry{Overview of the DMRcaller package}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
local_caption <- paste(
  "Local methylation profile.",
  "The points on the graph represent methylation proportion of individual cytosines,",
  "their colour (red or blue) which sample they belong to and the intensity of the colour how many reads that particular cytosine had.",
  "This means that darker colours indicate stronger evidence that the corresponding cytosine has the corresponding methylation proportion,",
  "while lighter colours indicate a weaker evidence.",
  "The solid lines represent the smoothed profiles and the intensity of the colour the coverage at the corresponding position (darker colours indicate more reads while lighter ones less reads).",
  "The boxes on top represent the DMRs, where a filled box will represent a DMR which gained methylation while a box with a pattern represent a DMR that lost methylation.",
  "The DMRs need to have a metadata column `regionType` which can be either `gain` (where there is more methylation in condition 2 compared to condition 1) or `loss` (where there is less methylation in condition 2 compared to condition 1).",
  "In case this metadata column is missing all DMRs are drawn using filled boxes.",
  "Finally we also allow annotation of the DNA sequence.",
  "We represent by black boxes all the exons, which are joined by a horizontal black line, thus, marking the full body of the gene.",
  "With grey boxes we mark the transposable elements.",
  "Both for genes and transposable elements we plot them over a mid line if they are on the positive strand and under the mid line if they are on the negative strand."
)

knitr::opts_chunk$set(
  echo = TRUE,
  message = FALSE,
  warning = FALSE,
  fig.width = 7,
  fig.height = 5)
```

\setlength\parindent{12pt}

# Introduction

DNA methylation is an epigenetic modification of the DNA where a methyl group is added to the cytosine nucleotides. This modification is heritable, able to control the gene regulation and, in general, is associated with transcriptional gene silencing. While in mammals the DNA is predominantly methylated in CG context, in plants non-CG methylation (CHG and CHH, where H can be any of the A, C or T nucleotides) is also present and is important for the epigenetic regulation of transcription. Sequencing of bisulfite converted DNA has become the method of choice to determine genome wide methylation distribution. The *DMRcaller* package computes the set of Differentially Methylated Regions (DMRs) between two samples. *DMRcaller* will compute the differentially methylated regions from Whole Genome Bisulfite Sequencing (WGBS) or Reduced Representation Bisulfite Sequencing (RRBS) data. There are several tools able to call DMRs, but most work has been done in mammalian systems and, thus, they were designed to primarily call CG methylation.

# Methods

The package computes the DMRs using the CX report files generated by Bismark [@krueger_2011], which contain the number of methylated and unmmethylated reads for each cytosine in the genome. The coverage at each position on the genome is not homogeneous and this makes it difficult to compute the differentially methylated cytosines. Here, we implemented three methods:

-   **noise_filter** where we use a kernel [@hebestreit_2013] to smooth the number of methylated reads and the total number of reads (the *DMRcaller* package provides four kernels: `"uniform"`, `"triangular"`, `"gaussian"` and `"epanechnicov"`)

-   **neighbourhood** where individual cytosines in a specific context are considered in the analysis without any smoothing

-   **bins** where the genome is split into equal bins where all the reads are pooled together

The DMRs are then computed by performing a statistical test between the number of methylated reads and the total number of reads in the two conditions for each position, cytosine or bin. In particular, we implemented two statistical tests: $(i)$ Fisher's exact test and $(ii)$ the Score test. The former (Fisher's exact test) uses the `fisher.test` in the *stats* package.

The Score test is a statistical test of a simple null hypothesis that a parameter of interest is equal to some particular value. In our case, we are interested if the methylation levels in the two samples are equal or different. Given that $m_1$ is the number of methylated reads in condition 1, $m_2$ is the number of methylated reads in condition 2, $n_1$ is the total number of reads in condition 1 and $n_2$ is the total number of reads in condition 2, the Z-score of the Score test is

$$ Z = \frac{(p_1 - p_2)\,\nu}{\sqrt{p(1-p)}} $$

where $p_1=m_1/n_1$, $p_2 = m_2/n_2$,

$$ p =  \frac{m_1 + m_2}{n_1 + n_2}\quad \textrm{and}\quad \nu = \sqrt{\frac{n_1n_2}{n_1 + n_2}} $$

We then convert the Z-score to the p-value assuming a normal distribution and a two sided test.

```{r pvalue, echo=TRUE, message=FALSE, eval=FALSE}
pValue <- 2*pnorm(-abs(zScore))
```

Finally, for both statistical tests (Fisher's exact test and Score test), we adjust the p-values for multiple testing using Benjamini and Hochberg's method [@benjamini_1995] to control the false discovery

```{r pvalue_adjust, echo=TRUE, message=FALSE, eval=FALSE}
pValue <- p.adjust(pValue, method="fdr")
```

The algorithm performs the statistical test for each position, cytosine or bin and then marks as DMRs all positions/cytosines/bins that satisfy the following three conditions:

-   the difference in methylation levels between the two conditions is statistically significant according to the statistical test;

-   the difference in methylation proportion between the two conditions is higher than a threshold value;

-   the mean number of reads per cytosine is higher than a threshold.

To group adjacent DMRs, we run an iterative process, where neighbouring DMRs (within a certain distance of each other) are joined only if these three conditions are still met after joining the DMRs.

Finally, we filter the DMRs as follow

-   Remove DMRs whose lengths are less than a minimum size.

-   Remove DMRs with fewer cytosines than a threshold value.

For a set of potential DMRs (e.g. genes, transposable elements or CpG islands) the user can call the function `filterDMRs` where all reads in a set of provided regions are pooled together and then the algorithm performs the statistical test for each region.

# Description

## Data

### Bisulfite sequencing data

Bismark [@krueger_2011] is a popular tool for methylation call on WGBS or RRBS data. *DMRcaller* takes as inputs the CX report files generated by Bismark and stores this data in a `GRanges` object. In the package, we included two CX report files that contain the methylation calls of WT and `met1-3` `Arabidopsis thaliana` [@stroud_2013]. `MET1` gene encodes for the main DNA methyltransferase in `Arabidopsis thaliana` and the `met1-3` mutation results in a genome-wide loss of DNA methylation (mainly in CG context). Due to running time, we restricted the data and analysis to the first $1\ Mb$ of the third chromosome of `A. thaliana`.

```{r load_presaved, message=FALSE,cache=FALSE}
library(DMRcaller)

#load presaved data
data(methylationDataList)
```

To load a different dataset, one can use `readBismark` function, which takes as input the filename of the CX report file to be loaded.

```{r load, echo=TRUE, message=FALSE, cache=FALSE, eval=FALSE}
# specify the Bismark CX report files
saveBismark(methylationDataList[["WT"]],
            "chr3test_a_thaliana_wt.CX_report")
saveBismark(methylationDataList[["met1-3"]],
            "chr3test_a_thaliana_met13.CX_report")

# load the data
methylationDataWT <- readBismark("chr3test_a_thaliana_wt.CX_report")
methylationDataMet13 <- readBismark("chr3test_a_thaliana_met13.CX_report")
methylationDataList <- GRangesList("WT" = methylationDataWT,
                                   "met1-3" = methylationDataMet13)
```

`methylationDataList` is a `GRangesList` object , where the `GRanges` elements contain four metadata columns

-   **context** - the context of the Cytosine (CG, CHG or CHH)

-   **readsM** - the number of methylated reads

-   **readsN** - the total number of reads

-   **trinucleotide_context** - the specific context of the cytosine (H is replaced by the actual nucleotide)

If the data consists of two or more replicates, these can be pooled together using the function `poolMethylationDatasets` or `poolTwoMethylationDatasets` (in the case of pooling only two datasets). The latter function (`poolTwoMethylationDatasets`) is useful when the datasets are large and creating a `GRangesList` object is not possible (e.g. the `GRanges` objects are two large).

```{r pool_data, echo=TRUE, message=FALSE, cache=FALSE, eval=FALSE}
# load the data
methylationDataAll <- poolMethylationDatasets(methylationDataList)

# In the case of 2 elements, this is equivalent to
methylationDataAll <- poolTwoMethylationDatasets(methylationDataList[[1]],
                                                 methylationDataList[[2]])
```

Alternatively, one can use `ReadBismarkPool` to directly read a list of CX report files and pool them together.

```{r read_pool_data, echo=TRUE, message=FALSE, cache=FALSE, eval=FALSE}
# load the data
methylationDataAll <- readBismarkPool(c(file_wt, file_met13))
```

### Oxford Nanopore methylation data

Oxford Nanopore Technologies (ONT) provides direct detection of modified bases (e.g., `5mC`) during sequencing by modelling electrical current disruptions at DNA pores. The methylation information is embedded as `MM` (modified base) and `ML` (modification probability) tags in aligned `BAM` files, which are typically produced after basecalling and alignment from raw signal files (`.pod5`).

To analyse ONT methylation data with *DMRcaller*, users should first:

1.  Get raw signal data (`.pod5`) from ONT sequencing device or public resources (e.g., [the 1000 Genomes ONT dataset](https://s3.amazonaws.com/1000g-ont/index.html?prefix=)).

2.  Basecall the signals using [*dorado*](https://github.com/nanoporetech/dorado), a GPU-accelerated basecaller by ONT.

3.  Align the reads and embed `MM`/`ML` tags using [*dorado*](https://github.com/nanoporetech/dorado) with methylation-aware models.

4.  Decode the tags and summarise methylation calls with `readONTbam()`.

#### Oxford Nanopore cytosine reference selection

To analyse Oxford Nanopore (ONT) methylation data, one must first define the genomic positions of interest, i.e., the candidate cytosines in a specific methylation context (`CG`, `CHG`, or `CHH`). The `selectCytosine()` function provides this functionality by scanning a reference genome and returning a `GRanges` object of relevant cytosine positions.

```{r selectCytosine_example, echo=TRUE, message=FALSE, cache=FALSE,eval=FALSE}
library(BSgenome.Hsapiens.UCSC.hg38)

# Select cytosines in CG context on chromosome 1
gr_cpg <- selectCytosine(
  genome = BSgenome.Hsapiens.UCSC.hg38,
  context = "CG",
  chr     = "chr1"
)
gr_cpg
```

The resulting `GRanges` object contains one genomic range per cytosine that matches the specified context. For `CG` context, each range spans two bases (e.g., `CpG`). The following metadata columns are attached:

-   **context** - A factor indicating the context (`"CG"`, `"CHG"`, or `"CHH"`).

-   **trinucleotide_context** - Character or factor giving the surrounding trinucleotide sequence (or `NA` for `"CG"`).

This function is particularly useful when preparing to quantify methylation signals from ONT `BAM` files using the `readONTbam()` function, as it provides the input cytosine sites to be queried in the aligned sequencing data.

#### Reading Oxford Nanopore BAM files

After alignment and basecalling, you can extract methylation calls using the `readONTbam()` function. This function decodes `MM`/`ML` tags and calculates per-site read-level methylation counts across the genome

```{r read_ONT, echo=TRUE,message=FALSE,cache=FALSE,eval=FALSE}
# Path to ONT BAM (with MM/ML tags)
bamfile <- system.file("extdata",
                       "scanBamChr1Random5.bam",
                       package="DMRcaller")

# Decode methylation and compute read-level counts
OntChr1CG <- readONTbam(
  bamfile    = bamfile,
  ref_gr     = gr_cpg,
  genome     = BSgenome.Hsapiens.UCSC.hg38,
  modif      = "C+m?",      
  chr        = "chr1",
  region     = FALSE,
  prob_thresh= 0.5,
  parallel   = FALSE,
)

# View results
OntChr1CG
```

**Notes on important parameters**

-   `genome`: The genome argument is required if `region = FALSE`, and must be a `BSgenome` object from the Bioconductor package *BSgenome*. This allows the function to extract cytosine coordinates and sequence contexts from the full reference genome. For example:

```{r BSgenome_available, echo=TRUE, cache=FALSE, message=FALSE, eval=FALSE}
head(BSgenome::available.genomes())

```

-   `modif`: The modif parameter specifies which base modifications to extract based on the `BAM` file’s `MM` tag. The syntax follows the [SAMtags specification](chrome-extension://efaidnbmnnnibpcajpcglclefindmkaj/https://samtools.github.io/hts-specs/SAMtags.pdf) and is usually one of "`C+m?`", "`C+m.`", "`C+h?`" or "`C+h.`"

For example, in this package we provide methylation data derived from the `GM18501` and `GM18876` B-Lymphocyte cell lines from the 1000 Genomes ONT dataset [@gustafson_2024]:

-   **Raw data**: [GM18501](https://s3.amazonaws.com/1000g-ont/index.html?prefix=pod5_data/GM18501_R9/), [GM18876](https://s3.amazonaws.com/1000g-ont/index.html?prefix=pod5_data/GM18876_R9/)
-   **Basecalling**: performed using `dorado v0.9.6` with the model `dna_r10.4.1_e8.2_400bps_hac@v5.2.0`

The resulting BAM files contain `MM` and `ML` tags for `5mC` detection. These are processed with `readONTbam()` to compute per-cytosine methylation statistics. The example data included in the package is accessible as:

```{r load_ontSampleGRangesList, message=FALSE, cache=FALSE}
# load presaved ont GRanges data
data(ontSampleGRangesList)
```

This is a `GRangesList` object with one `GRanges` per sample (`GM18501` and `GM18876`), covering $1.5–2 Mbp$ of human chromosome 1. Each `GRanges` object contains the following metadata columns:

-   **context** - the context of the Cytosine (`CG`, `CHG` or `CHH`)

-   **readsM** - the number of methylated reads

-   **readsN** - the total number of reads

-   **trinucleotide_context** - the specific context of the cytosine (`H` is replaced by the actual nucleotide)

-   **ONT_Cm** - comma-delimited read-indices called modified

-   **ONT_C** - comma-delimited read-indices covering but unmodified

This function allows for fine-grained downstream analysis, such as identifying differentially methylated regions (DMRs), variably methylated domains (VMDs), or co-methylation structures across samples.

## Low resolution profiles

The *DMRcaller* package also offers the possibility to visualise context specific global changes in the methylation profile. To achieve this, the user can call `plotMethylationProfileFromData` function, which computes the mean methylation proportion in tiling bins of fixed size; see Figure \@ref(fig:lowResProfile).

```{r lowResProfile, fig.height=4, fig.width=12, fig.cap="Low resolution profile in CG context for WT and met1-3", out.width='95%',echo=TRUE, fig.align='center', message=FALSE, cache=FALSE}
plotMethylationProfileFromData(methylationDataList[["WT"]],
                               methylationDataList[["met1-3"]],
                               conditionsNames = c("WT","met1-3"),
                               windowSize = 10000,
                               autoscale = FALSE,
                               context = c("CG"))
```

Alternatively, for a finer control, the user can use `computeMethylationProfile` function to compute the methylation profile at certain locations on the genome. This function returns a `Ranges` object with four metadata columns

-   **sumReadsM** - the number of methylated reads

-   **sumReadsN** - the total number of reads

-   **Proportion** - the proportion of methylated reads

-   **context** - the context

One or more of these `GRanges` objects can be put in a `GRangesList` object which is then passed as a parameter to the `plotMethylationProfile` function.

```{r profile_fine, echo=TRUE, message=FALSE, cache=FALSE,  fig.width=5,fig.height=5, fig.align='center', eval=FALSE}
regions <- GRanges(seqnames = Rle("Chr3"), ranges = IRanges(1,1E6))

# compute low resolution profile in 10 Kb windows
profileCGWT <- computeMethylationProfile(methylationDataList[["WT"]],
                                         regions,
                                         windowSize = 10000,
                                         context = "CG")

profileCGMet13 <- computeMethylationProfile(methylationDataList[["met1-3"]],
                                            regions,
                                            windowSize = 10000,
                                            context = "CG")

profilesCG <- GRangesList("WT" = profileCGWT, "met1-3" = profileCGMet13)

#plot the low resolution profile
par(mar=c(4, 4, 3, 1)+0.1)
par(mfrow=c(1,1))
plotMethylationProfile(profilesCG,
                       autoscale = FALSE,
                       labels = NULL,
                       title="CG methylation on Chromosome 3",
                       col=c("#D55E00","#E69F00"),
                       pch = c(1,0),
                       lty = c(4,1))
```

## Coverage of the bisulfite sequencing data

The number of reads from the bisulfite sequencing can differ significantly between different locations on the genome in the sense that cytosines in the same context (including neighbouring cytosines) can display large variability in the coverage. To plot the coverage of the bisulfite sequencing datasets, one can use `plotMethylationDataCoverage` function which takes as input one or two datasets and the vector with the thresholds used to compute the proportion of cytosines with at least that many reads; see Figure \@ref(fig:coverage).

```{r coverage, fig.width=6,fig.height=5,out.width='60%', fig.cap="Coverage. For example, this figure shows that in WT only $30%$ of the cytosines in CHH context have at least 10 reads", fig.align='center', message=FALSE,cache=FALSE}
# plot the coverage in the two contexts
par(mar=c(4, 4, 3, 1)+0.1)
plotMethylationDataCoverage(methylationDataList[["WT"]],
                            methylationDataList[["met1-3"]],
                            breaks = c(1,5,10,15),
                            regions = NULL,
                            conditionsNames=c("WT","met1-3"),
                            context = c("CHH"),
                            proportion = TRUE,
                            labels=LETTERS,
                            contextPerRow = FALSE)
```

Alternatively, the *DMRcaller* also provides the `computeMethylationDataCoverage` function which returns a numeric vector with the number or proportion of cytosines in a specific context that have at least a certain number of reads specified by the input vector `breaks`.

```{r coverage_compute, message=FALSE,cache=FALSE, eval=FALSE}
coverageCGWT <- computeMethylationDataCoverage(methylationDataList[["WT"]],
                                               context="CG",
                                               breaks = c(1,5,10,15))
```

## Spatial correlation of methylation levels

Methylation levels are often spatially correlated and some methods to detect DMRs assume this spatial correlation. Nevertheless, different tissues, samples and even methylation context will display different levels of correlation. *DMRcaller* implements `plotMethylationDataSpatialCorrelation` function that plots the correlation of methylation levels as function of distance between cytosines. This function takes as input one or two datasets and the vector with the distances between cytosines; see Figure \@ref(fig:correlationPlot).

```{r correlationPlot, fig.width=6,fig.height=5, fig.cap="Spatial correlation of methylation levels", fig.align='center', out.width='60%', message=FALSE,cache=FALSE}
# compute the spatial correlation of methylation levels
plotMethylationDataSpatialCorrelation(methylationDataList[["WT"]],
                            distances = c(1,100,10000), regions = NULL,
                            conditionsNames = c("WT"),
                            context = c("CG"),
                            labels = LETTERS, col = NULL,
                            pch = c(1,0,16,2,15,17), lty = c(4,1,3,2,6,5),
                            contextPerRow = FALSE,
                            log = "x")
```

Alternatively, the *DMRcaller* also provides the `computeMethylationDataSpatialCorrelation` function which returns a numeric vector with the correlation of methylation levels of cytosines separated by certain distances in a specific context.

```{r correlation_compute, message=FALSE,cache=FALSE, eval=FALSE}
correlation_CG_wt <- computeMethylationDataSpatialCorrelation(methylationDataList[["WT"]],
                                                              context="CG",
                                                              distances=c(1,10,100,1000,10000))
```

## Calling DMRs

*DMRcaller* package provides `computeDMRs` function to call DMRs. The output of this function is a `GRanges` with 11 metadata columns.

-   **direction** - a numeric value indicating whether the methylation was lost in the second condition compared to the first one ($-1$) or gained ($+1$)

-   **context** - the context of the cytosine (CG, CHG or CHH)

-   **sumReadsM1** - the number of methylated reads in the DMR in condition 1

-   **sumReadsN1** - the total number of reads in the DMR in condition 1

-   **proportion1** - the proportion of methylated reads in the DMR in condition 1

-   **sumReadsM2** - the number of methylated reads in the DMR in condition 2

-   **sumReadsN2** - the total number of reads in the DMR in condition 2

-   **proportion2** - the proportion of methylated reads in the DMR in condition 2

-   **cytosinesCount** - the number of cytosines in the DMR

-   **pValue** - the adjusted p-value of the statistical test

-   **regionType** - a character string indicating whether the methylation was lost in the second condition compared to the first one (*"loss"*) or gained (*"gain"*)

For predifined regions (e.g. genes, transposons or CpG islands) the user can call `filterDMRs` function to extract the list of regions that are differentially methylated. The output of this function is again a GRanges with the same 11 metadata columns.

Below we present examples of calling both functions.

```{r compute_DMRs_CG_noise_filter, message=FALSE,cache=FALSE}
chr_local <- GRanges(seqnames = Rle("Chr3"), ranges = IRanges(5E5,6E5))

# compute the DMRs in CG context with noise_filter method
DMRsNoiseFilterCG <- computeDMRs(methylationDataList[["WT"]],
                      methylationDataList[["met1-3"]],
                      regions = chr_local,
                      context = "CG",
                      method = "noise_filter",
                      windowSize = 100,
                      kernelFunction = "triangular",
                      test = "score",
                      pValueThreshold = 0.01,
                      minCytosinesCount = 4,
                      minProportionDifference = 0.4,
                      minGap = 0,
                      minSize = 50,
                      minReadsPerCytosine = 4,
                      cores = 1)

print(DMRsNoiseFilterCG)
```

```{r compute_DMRs_CG_neighbourhood, message=FALSE,cache=FALSE}
# compute the DMRs in CG context with neighbourhood method
DMRsNeighbourhoodCG <- computeDMRs(methylationDataList[["WT"]],
                                   methylationDataList[["met1-3"]],
                                   regions = chr_local,
                                   context = "CG",
                                   method = "neighbourhood",
                                   test = "score",
                                   pValueThreshold = 0.01,
                                   minCytosinesCount = 4,
                                   minProportionDifference = 0.4,
                                   minGap = 200,
                                   minSize = 1,
                                   minReadsPerCytosine = 4,
                                   cores = 1)

print(DMRsNeighbourhoodCG)
```

```{r compute_DMRs_CG_bins, message=FALSE,cache=FALSE}
# compute the DMRs in CG context with bins method
DMRsBinsCG <- computeDMRs(methylationDataList[["WT"]],
                          methylationDataList[["met1-3"]],
                          regions = chr_local,
                          context = "CG",
                          method = "bins",
                          binSize = 100,
                          test = "score",
                          pValueThreshold = 0.01,
                          minCytosinesCount = 4,
                          minProportionDifference = 0.4,
                          minGap = 200,
                          minSize = 50,
                          minReadsPerCytosine = 4,
                          cores = 1)

print(DMRsBinsCG)
```

```{r compute_DMRs_CG_GE, message=FALSE,cache=FALSE}
# load the gene annotation data
data(GEs)

#select the genes
genes <- GEs[which(GEs$type == "gene")]

# compute the DMRs in CG context over genes
DMRsGenesCG <- filterDMRs(methylationDataList[["WT"]],
                          methylationDataList[["met1-3"]],
                          potentialDMRs = genes[overlapsAny(genes, chr_local)],
                          context = "CG",
                          test = "score",
                          pValueThreshold = 0.01,
                          minCytosinesCount = 4,
                          minProportionDifference = 0.4,
                          minReadsPerCytosine = 3,
                          cores = 1)
print(DMRsGenesCG)
```

## Merge DMRs

Finally, for merging adjacent DMRs, *DMRcaller* provides the function `mergeDMRsIteratively` which can be used as follows:

```{r merge_DMRs_CG_noise_filter, message=FALSE,cache=FALSE}
DMRsNoiseFilterCGMerged <- mergeDMRsIteratively(DMRsNoiseFilterCG,
                                                minGap = 200,
                                                respectSigns = TRUE,
                                                methylationDataList[["WT"]],
                                                methylationDataList[["met1-3"]],
                                                context = "CG",
                                                minProportionDifference = 0.4,
                                                minReadsPerCytosine = 4,
                                                pValueThreshold = 0.01,
                                                test="score")
print(DMRsNoiseFilterCGMerged)
```

Note that two neighbouring DMRs will be merged if all the conditions below are met

-   they are within a distance from each other smaller than minGap

-   the difference in methylation levels between the two conditions is statistically significant according to the statistical test when the two DMRs are joined

-   the difference in methylation proportion between the two conditions is higher than a threshold value when the two DMRs are joined

-   the number of reads per cytosine is higher than a threshold when the two DMRs are joined

## Calling DMRs using biological replicates

The package also contains a set of functions for the analysis of multiple biological replicates.

The synthetic dataset is made by 300 different cytosines, extracted from those present in the *A. thaliana* dataset. The value for **readsN** are created using the function `rnorm`, while the values for `readsM` are generated using the function `rbinom`. The probabilities used are 0.1 in the external region and 0.8 in the central region. In this way a DMR should be detected in the central region of the synthetic dataset.

The difference in proportion is plotted in figure \@ref(fig:differenceMethylationReplicates).

```{r differenceMethylationReplicates, fig.width=12,fig.height=5,out.width='95%', fig.cap="Methylation proportions in the synthetic dataset.", fig.align='center',message=FALSE,cache=FALSE}
# loading synthetic data
data("syntheticDataReplicates")

# create vector with colours for plotting
cbbPalette <- c("#000000", "#E69F00", "#56B4E9", "#009E73", "#F0E442",
                "#0072B2", "#D55E00", "#CC79A7")

# plotting the difference in proportions
plot(start(syntheticDataReplicates),
     syntheticDataReplicates$readsM1/syntheticDataReplicates$readsN1,
     ylim=c(0,1), col=cbbPalette[2], xlab="Position in Chr3 (bp)",
     ylab="Methylation proportion")
points(start(syntheticDataReplicates),
       syntheticDataReplicates$readsM2/syntheticDataReplicates$readsN2,
       col=cbbPalette[7], pch=4)
points(start(syntheticDataReplicates),
       syntheticDataReplicates$readsM3/syntheticDataReplicates$readsN3,
       col=cbbPalette[3], pch=2)
points(start(syntheticDataReplicates),
       syntheticDataReplicates$readsM4/syntheticDataReplicates$readsN4,
       col=cbbPalette[6], pch=3)
legend(x = "topleft", legend=c("Treatment 1", "Treatment 2", "Control 1",
                               "Control 2"), pch=c(1,4,2,3),
       col=cbbPalette[c(2,7,3,6)], bty="n", cex=1.0)
```

The DMRs are computed using the function `computeDMRsReplicates`, which uses beta regression [@ferrari_2004] to detect differential methylation.

```{r computingBiologicalReplicates, message=FALSE,cache=FALSE, eval=TRUE}
if (requireNamespace("GenomeInfoDb", quietly = TRUE)) {
# starting with data joined using joinReplicates

# creating condition vector
condition <- c("a", "a", "b", "b")

# computing DMRs using the neighbourhood method
DMRsReplicatesBins <- computeDMRsReplicates(methylationData = syntheticDataReplicates,
                                            condition = condition,
                                            regions = NULL,
                                            context = "CG",
                                            method = "bins",
                                            binSize = 100,
                                            test = "betareg",
                                            pseudocountM = 1,
                                            pseudocountN = 2,
                                            pValueThreshold = 0.01,
                                            minCytosinesCount = 4,
                                            minProportionDifference = 0.4,
                                            minGap = 0,
                                            minSize = 50,
                                            minReadsPerCytosine = 4,
                                            cores = 1)

print(DMRsReplicatesBins)
}
```

## Extract methylation data in regions

`analyseReadsInsideRegionsForCondition` function can extract additional information in a set of genomic regions (including DMRs) from any `methylationData` object. For example, to establish a link between the CG and CHH methylation, one might want to extract the number of methylated reads and the total number of reads in CHH context inside DMRs called in CG context.

```{r analyse_reads_inside_regions_for_condition, message=FALSE,cache=FALSE}
#retrive the number of reads in CHH context in WT in CG DMRs
DMRsNoiseFilterCGreadsCHH <- analyseReadsInsideRegionsForCondition(
                              DMRsNoiseFilterCGMerged,
                              methylationDataList[["WT"]], context = "CHH",
                              label = "WT")
print(DMRsNoiseFilterCGreadsCHH)
```

## Calling partially methylated domains (PMDs)

The *DMRcaller* package provides the `computePMDs` function to identify **partially methylated domains (PMDs)** from both WGBS and Oxford Nanopore methylation data. PMDs are large genomic regions with intermediate methylation levels (e.g., 40–60%) and can be identified using three available methods: "`noise_filter`", "`neighbourhood`", and "`bins`"

The output of this function is a GRanges object with five metadata columns:

-   context – the context of the cytosine (CG, CHG or CHH)

-   sumReadsM – the number of methylated reads in the PMD

-   sumReadsN – the total number of reads in the PMD

-   proportion – the average methylation proportion in the PMD (between minMethylation and maxMethylation)

-   cytosinesCount – the number of cytosines considered in the PMD

To compute PMDs in a specific region of interest, define a GRanges object. If not specified, PMDs are computed genome-wide.

```{r compute_PMDs_CG_noise_filter, message=FALSE, cache=FALSE,eval=FALSE}
# load the ONT methylation data
data(ontSampleGRangesList)

# define the region of interest on chr1
chr1_ranges <- GRanges(seqnames = Rle("chr1"), ranges = IRanges(1E6+5E5,2E6))

# compute the PMDs in CG context using the noise_filter method
PMDsNoiseFilterCG <- computePMDs(ontSampleGRangesList[["GM18501"]],
                                 regions = chr1_ranges,
                                 context = "CG",
                                 method = "noise_filter",
                                 windowSize = 100,
                                 kernelFunction = "triangular", 
                                 lambda = 0.5,
                                 minCytosinesCount = 4, 
                                 minMethylation = 0.4, 
                                 maxMethylation = 0.6, 
                                 minGap = 200, 
                                 minSize = 50, 
                                 minReadsPerCytosine = 4, 
                                 cores = 1,
                                 parallel = FALSE)

PMDsNoiseFilterCG
```

```{r compute_PMDs_CG_neighbourhood, message=FALSE, cache=FALSE, eval=FALSE}
# compute the PMDs using the neighbourhood method
PMDsNeighbourhoodCG <- computePMDs(ontSampleGRangesList[["GM18501"]],
                                   regions = chr1_ranges,
                                   context = "CG",
                                   method = "neighbourhood",
                                   minCytosinesCount = 4, 
                                   minMethylation = 0.4, 
                                   maxMethylation = 0.6, 
                                   minGap = 200, 
                                   minSize = 50, 
                                   minReadsPerCytosine = 4, 
                                   cores = 1,
                                   parallel = FALSE)

PMDsNeighbourhoodCG
```

```{r compute_PMDs_CG_bins, message=FALSE, cache=FALSE, eval=FALSE}
# compute the PMDs using the bins method
PMDsBinsCG <- computePMDs(ontSampleGRangesList[["GM18501"]],
                          regions = chr1_ranges,
                          context = "CG",
                          method = "bins",
                          binSize = 100,
                          minCytosinesCount = 4, 
                          minMethylation = 0.4, 
                          maxMethylation = 0.6, 
                          minGap = 200, 
                          minSize = 50, 
                          minReadsPerCytosine = 4, 
                          cores = 1,
                          parallel = FALSE)

PMDsBinsCG
```

## Calling variably methylated domains (VMDs) using ONT data

The *DMRcaller* package provides the `computeVMDs` function to identify **variably methylated domains (VMDs)** — genomic regions showing high or low variability in methylation levels across reads at each cytosine. This is particularly relevant for identifying regulatory instability or stochastic epigenetic variation using Oxford Nanopore sequencing data.

The function segments the genome into *fixed-size bins*, evaluates the *per-read methylation variability* within each bin, and filters bins based on their **weighted standard deviation (SD)** using user-defined or data-driven thresholds.

The output is a GRanges object with the following metadata columns:

-   **context** – the context in which the VMDs were computed ("`CG`", "`CHG`", or "`CHH`")
-   **sumReadsM** – total number of methylated reads across the region
-   **sumReadsN** – total number of reads across the region
-   **proportion** – average methylation level across cytosines
-   **cytosinesCount** – number of cytosines in the region
-   **mean** – mean of per-read methylation proportions
-   **sd** – standard deviation of per-read methylation proportions
-   **w_mean** – weighted mean (based on read count per cytosine)
-   **w_sd** – weighted standard deviation

To compute VMDs in a specific region or genome-wide, you must specify bin size and a method for selecting variance thresholds.

```{r compute_VMDs_CG, message=FALSE, cache=FALSE, eval=FALSE}
# load the ONT methylation data
data(ontSampleGRangesList)

# define the region of interest on chr1
chr1_ranges <- GRanges(seqnames = Rle("chr1"), ranges = IRanges(1E6+5E5, 1E6+6E5))

# compute the VMDs in CG context using the "EDE.high" variance threshold
VMDsBinsCG <- computeVMDs(ontSampleGRangesList[["GM18501"]],
                          regions = chr1_ranges,
                          context = "CG",
                          binSize = 100,
                          minCytosinesCount = 4,
                          sdCutoffMethod = "per.high",  # elbow-detected high variance
                          percentage = 0.05,            # only used if sdCutoffMethod = "per.high" or "per.low"
                          minGap = 200,
                          minSize = 50,
                          minReadsPerCytosine = 4,
                          parallel = FALSE,
                          cores = 1)

VMDsBinsCG
```

The `sdCutoffMethod` determines how to select high or low variance regions:

-   "`per.high`" or "`per.low`" filter to top/bottom bins based on a quantile cutoff
-   "`EDE.high`" or "`EDE.low`" apply an elbow detection algorithm by the `inflection::ede` from package *inflection* to automatically determine the cutoff point in the SD distribution.

## Detecting Variably Methylated Regions (VMRs) using ONT data

The function `filterVMRsONT()` is designed to identify **Variably Methylated Regions (VMRs)** between two ONT methylation datasets. These are genomic regions (e.g., genes, transposable elements, or CpG islands) that exhibit statistically significant differences in **methylation variability** or **methylation levels** between two conditions.

This function applies the following ONT-specific statistical filters:

-   **Wilcoxon rank-sum test** on per-read methylation proportions
-   **F-test** on variance of per-read methylation proportions
-   Optional: fold-change cutoffs on variance ratio, confidence interval filtering, and coverage/methylation thresholds

We use the example ONT methylation data and transcript annotation to identify VMRs across genes within a region of chromosome 1.

```{r filter_VMRs, message=FALSE, cache=FALSE, eval=FALSE}
# load the ONT methylation data 
data(ontSampleGRangesList)
# load the gene annotation data
data(GEs_hg38)

# select the transcript
transcript <- GEs_hg38[which(GEs_hg38$type == "transcript")]

# the regions where to compute the PMDs
regions <- GRanges(seqnames = Rle("chr1"), ranges = IRanges(1E6+5E5,2E6))
transcript <- transcript[overlapsAny(transcript, regions)]

# filter genes that are differntially methylated in the two conditions
VMRsGenesCG <- filterVMRsONT(ontSampleGRangesList[["GM18501"]], 
               ontSampleGRangesList[["GM18876"]], potentialVMRs = transcript, 
               context = "CG", pValueThreshold = 0.01, 
               minCytosinesCount = 4, minProportionDifference = 0.01, 
               minReadsPerCytosine = 3, ciExcludesOne = TRUE,
               varRatioFc = NULL, parallel = TRUE) # parallel recommended
```

The function returns a `GRanges` object containing the same genomic intervals as the input `potentialVMRs`, enriched with the following metadata columns:

-   **sumReadsM1** / **sumReadsN1**: Total methylated / total reads in condition 1
-   **sumReadsM2** / **sumReadsN2**: Total methylated / total reads in condition 2
-   **proportion1** / **proportion2**: Methylation proportions
-   **variance1** / **variance2**: Variance of per-read methylation levels
-   **wilcox_pvalue**: FDR-adjusted p-value from *Wilcoxon test*
-   **f_pvalue**: FDR-adjusted p-value from *F-test*
-   **var_ratio**: Ratio of variances (condition1 / condition2)
-   **is_DMR**: TRUE if *Wilcoxon test* passed significance threshold
-   **is_VMR**: TRUE if *F-test* passed significance threshold
-   **regionType**: "`gain`" or "`loss`" of methylation
-   **direction**: `+1` (gain) or `-1` (loss)

Only regions satisfying all filtering criteria (e.g., coverage, cytosine count, variance CI exclusion, proportion difference) are retained in the output. For strict variance detection, set `ciExcludesOne = TRUE` and optionally apply `varRatioFc = 2` for 2-fold variance filtering.

## Calling Co-Methylated Positions (CMPs) using ONT data 

The *DMRcaller* package provides the `computeCoMethylatedPositions` function to detect **pairwise co-methylation** patterns between CpG, CpHpG or CpHpH sites within user-defined genomic regions, such as **PMDs**, **genes**, or **repeat elements**. This is especially useful when exploring coordinated methylation regulation or identifying methylation-linked chromatin interactions using *long-read Nanopore sequencing data*.

For each region, the function computes **strand-specific pairwise associations** between methylation positions that fall within a specified genomic distance range. It constructs a 2×2 contingency table summarising methylation co-occurrence across reads and performs a statistical test to evaluate co-dependency.

The output is a list of `GInteractions` objects (one per input region) containing significant pairs and their co-methylation statistics.

Each significant interaction includes:

-   **C1_C2** – number of reads methylated at both cytosines
-   **C1_only** – number of reads methylated only at the first cytosine
-   **C2_only** – number of reads methylated only at the second cytosine
-   **neither** – number of reads methylated at neither cytosines
-   **strand** – DNA strand of the pair ("`+`" or "`-`")
-   **genomic_position** – region ID in UCSC/IGV format (e.g. "`chr1:1522971-1523970`")
-   **p.value** – FDR-adjusted p-value from Fisher’s exact or permutation test

```{r compute_comethylation_fisher, message=FALSE, cache=FALSE, eval=TRUE}
# load ONT methylation and PMD data
data("ont_gr_GM18870_chr1_PMD_bins_1k")
data("ont_gr_GM18870_chr1_sorted_bins_1k")

# compute co-methylation using Fisher's exact test
coMethylationFisher <- computeCoMethylatedPositions(
  ont_gr_GM18870_chr1_sorted_bins_1k,
  regions = ont_gr_GM18870_chr1_PMD_bins_1k[1:4],
  minDistance = 150,
  maxDistance = 1000,
  minCoverage = 1,
  pValueThreshold = 0.01,
  test = "fisher",
  parallel = FALSE)

coMethylationFisher
```

**Note on region size and multiple testing correction**

The `computeCoMethylatedPositions` function applies `FDR` correction globally across **all cytosine pairs** tested within the input regions. As such, the number and range of regions provided can influence the adjusted p-values. For example, providing smaller or fewer regions results in fewer total comparisons, which can lead to more conservative (higher) adjusted p-values for the same cytosine pairs.

Therefore, when designing region inputs for co-methylation analysis, it’s important to note that adjusted significance depends not only on the pairwise test but also on the size and number of regions analysed.

## Calling Co-Methylated Regions (CMRs) using ONT Data

The *DMRcaller* package also provides the `computeCoMethylatedRegions` function to detect **correlation between methylation levels** of different genomic regions, such as **PMDs**, **genes**, or **regulatory domains**, using *long-read Nanopore methylation data*. This approach identifies region pairs that are **coordinated in their methylation profiles across single reads**. For example, two PMDs showing a high positive correlation may reflect **shared regulatory influence**, spatial proximity, or joint epigenetic repression.

For each candidate region pair within the user-defined **distance range** (e.g. 500–50,000.bp), the function computes:

-   Per-read methylation **proportions** across each region
-   **Correlation coefficients** (Pearson, Spearman, or Kendall) between these proportions
-   FDR-adjusted **p-values** to identify significantly co-methylated region pairs

The function returns a `GInteractions` object, where:

-   Each row represents a **significantly correlated region pair**
-   Anchors (`anchor1` and `anchor2`) correspond to the original input `regions`
-   Metadata columns include:
    -   `correlation` – correlation coefficient (e.g. Pearson's r)
    -   `coverage` – number of shared reads used for correlation
    -   `p.value` – FDR-adjusted p-value
    -   `genomic_position` – string representation of both region coordinates

```{r compute_cmrs_pearson, message=FALSE, cache=FALSE, eval=TRUE}
# load ONT methylation and PMD data
data("ont_gr_GM18870_chr1_sorted_bins_1k")
data("ont_gr_GM18870_chr1_PMD_bins_1k")

# compute highly correlated PMD region pairs using Pearson method
coMethRegions <- computeCoMethylatedRegions(
  methylationData = ont_gr_GM18870_chr1_sorted_bins_1k,
  regions = ont_gr_GM18870_chr1_PMD_bins_1k[1:5],
  minDistance = 500,
  maxDistance = 50000,
  minCoverage = 4,
  pValueThreshold = 0.05,
  correlation_test = "pearson",
  minCorrelation = -0.5,
  maxCorrelation = 0.5,
  parallel = FALSE
)

coMethRegions
```

## Plotting the distribution of DMRs, PMDs, VMDs or VMRs

Sometimes, it is useful to obtain the distribution of the DMRs over the chromosomes. The *DMRcaller* provides the `computeOverlapProfile` function, which computes this distribution. The `GRanges` object generated by this function can then be added to a `GRangesList` object, which can be plotted using `plotOverlapProfile` function; see Figure \@ref(fig:DMRdensity). Additionally, the `plotOverlapProfile` function allows the user to specify two `GRangesList`, thus, allowing the plotting of distributions of hypo or hyper methylated DMRs separately.

```{r DMRdensity, fig.width=15,fig.height=2.5,out.width='95%', fig.cap= "Distribution of DMRs. Darker colour indicates higher density, while lighter colour lower density.", fig.align='center', message=FALSE,cache=FALSE}
# compute the distribution of DMRs
hotspots <- computeOverlapProfile(DMRsNoiseFilterCGMerged, chr_local,
                                  windowSize=5000, binary=TRUE)
# plot the distribution of DMRs
plotOverlapProfile(GRangesList("Chr3"=hotspots))
```

## Plotting profiles with DMRs, PMDs, VMDs or VMRs

Finally, *DMRcaller* package also provides a function to plot methylation profiles at a specific location on the genome. To plot the methylation profile the user needs to call the `plotLocalMethylationProfile` function; see Figure \@ref(fig:localMethylationProfile).

```{r localMethylationProfile, fig.width=15,fig.height=5,out.width='95%',  fig.cap=local_caption, fig.align='center',message=FALSE,cache=FALSE}
# select a 20 Kb location on the Chr3
chr3Reg <- GRanges(seqnames = Rle("Chr3"), ranges = IRanges(510000,530000))

# create a list with all DMRs
DMRsCGList <- list("noise filter" = DMRsNoiseFilterCGMerged,
                   "neighbourhood" = DMRsNeighbourhoodCG,
                   "bins" = DMRsBinsCG,
                   "genes" = DMRsGenesCG)
# plot the local profile
par(cex=0.9)
par(mar=c(4, 4, 3, 1)+0.1)
plotLocalMethylationProfile(methylationDataList[["WT"]],
                            methylationDataList[["met1-3"]],
                            chr3Reg,
                            DMRsCGList,
                            conditionsNames = c("WT", "met1-3"),
                            GEs,
                            windowSize = 300,
                            main="CG methylation")
```

# Parallel computation

Computing the DMRs can be computationally intensive. For example, in the case of *A. thaliana* (with a genome of $\approx 130\ Mb$), it can take several hours to compute the DMRs depending on the method used and on the number of DMRs. To speed up computations, *DMRcaller* supports parallel computing of DMRs using the package *BiocParallel*, which able to run parallel works in any types of processor (including Windows, Mac, Linux and HPC conditions).

```{r DMRinParallel, message=FALSE,cache=FALSE}
# Parallel; compute the DMRs in CG context with noise_filter method
DMRsNoiseFilterCG_3cores <- computeDMRs(methylationDataList[["WT"]],
                      methylationDataList[["met1-3"]],
                      regions = chr_local,
                      context = "CG",
                      method = "noise_filter",
                      windowSize = 100,
                      kernelFunction = "triangular",
                      test = "score",
                      pValueThreshold = 0.01,
                      minCytosinesCount = 4,
                      minProportionDifference = 0.4,
                      minGap = 0,
                      minSize = 50,
                      minReadsPerCytosine = 4,
                      parallel = TRUE, # default: FALSE
                      BPPARAM = NULL, # default: NULL
                      cores = 3 # default: 1
                      ) 

print(DMRsNoiseFilterCG_3cores)
```

The five functions used for computing and filtering the DMRs, PMDs and VMRs (`readONTbam`,  `computeDMRs`, `filterDMRs`, `mergeDMRsIteratively`, `analyseReadsInsideRegionsForCondition`, `computeDMRsReplicates`,`computePMDs`, `filterPMDs`, `mergePMDsIteratively`, `analyseReadsInsideRegionsForConditionPMD`, `filterVMRsONT`, `computeCoMethylatedPositions` and `computeCoMethylatedRegions`) accept three parameters

-   `parallel` - run in parallel if `TRUE`.; *default setting*: `FALSE`

-   `BPPARAM` - A `BiocParallelParam` object controlling parallel execution. This value will automatically set when parallel is `TRUE`, also able to set as manually.; *default setting*: `NULL`

-   `cores` - which specifies the number of cores that can be used when performing the corresponding computations (must not exceed `BPPARAM$workers`). This value will automatically set as the maximum number of system workers, also able to set as manually.; *default setting*: `1`

When using 10 cores, it can take between 10 and 30 minutes to compute the DMRs in *A. thaliana* depending on the selected parameters.

# Session information

```{r session_info, echo=TRUE}
sessionInfo()
```

# References

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/filterVMRsONT.R
\name{filterVMRsONT}
\alias{filterVMRsONT}
\title{Filter VMRs for ONT Data}
\usage{
filterVMRsONT(
  methylationData1,
  methylationData2,
  potentialVMRs,
  context = "CG",
  pValueThreshold = 0.01,
  minCytosinesCount = 4,
  minProportionDifference = 0.4,
  minReadsPerCytosine = 3,
  ciExcludesOne = TRUE,
  varRatioFc = NULL,
  parallel = FALSE,
  BPPARAM = NULL,
  cores = NULL
)
}
\arguments{
\item{methylationData1}{A \code{GRanges} of methylation calls for condition 1
(see \code{\link{ontSampleGRangesList}}).}

\item{methylationData2}{A \code{GRanges} of methylation calls for condition 2.}

\item{potentialVMRs}{A \code{GRanges} of candidate VMR regions (genes, TEs,
CpG islands, etc.).}

\item{context}{Character string specifying cytosine context ("CG", "CHG",
or "CHH").}

\item{pValueThreshold}{Numeric p-value threshold (0<value<1) for both
Wilcoxon and F-tests after FDR adjustment.}

\item{minCytosinesCount}{Integer minimum number of cytosines per region.}

\item{minProportionDifference}{Numeric minimum methylation difference
between conditions (0<value<1).}

\item{minReadsPerCytosine}{Integer minimum average coverage per cytosine.}

\item{ciExcludesOne}{Logical; if \code{TRUE}, filter out regions whose
F-test 95\% confidence interval spans 1 (i.e., no significant variance change).}

\item{varRatioFc}{Optional; numeric fold-change cutoff on variance ratio
(e.g., 2 for twofold variance difference). Regions with variance ratio
outside \code{[1/varRatioFc, varRatioFc]} are kept when set.}

\item{parallel}{Logical; run in parallel if \code{TRUE}.}

\item{BPPARAM}{A \code{BiocParallelParam} object controlling parallel execution.
This value will automatically set when parallel is \code{TRUE}, also able to set as manually.}

\item{cores}{Integer number of workers (must not exceed \code{BPPARAM$workers}).
This value will automatically set as the maximum number of system workers,
also able to set as manually.}
}
\value{
A \code{GRanges} with the same ranges as \code{regions}, plus these metadata:
\describe{
  \item{sumReadsM1}{total methylated reads in condition 1}
  \item{sumReadsN1}{total reads in condition 1}
  \item{proportion1}{methylation proportion (sumReadsM1/sumReadsN1)}
  \item{variance1}{variance of per-read methylation proportions in condition 1}
  \item{sumReadsM2}{total methylated reads in condition 2}
  \item{sumReadsN2}{total reads in condition 2}
  \item{proportion2}{methylation proportion (sumReadsM2/sumReadsN2)}
  \item{variance2}{variance of per-read methylation proportions in condition 2}
  \item{cytosinesCount}{number of cytosines observed in each region}
  \item{wilcox_pvalue}{FDR adjusted p-value from Wilcoxon rank-sum test comparing per-read proportions}
  \item{f_pvalue}{FDR adjusted p-value from F-test comparing variances of per-read proportions}
  \item{var_ratio}{Ratio of variances (variance1 / variance2)}
  \item{wilcox_result}{Full \code{htest} object returned by \code{wilcox.test}}
  \item{F_test_result}{Full \code{htest} object returned by \code{var.test}}
  \item{direction}{a number indicating whether the region lost (-1)  or gain
    (+1) methylation in condition 2 compared to condition 1.}
  \item{regionType}{a string indicating whether the region lost (\code{"loss"})
    or gained (\code{"gain"}) methylation in condition 2 compared to condition 1.}
  \item{is_DMR}{logical; \code{TRUE} if region passed the \code{wilcox.test}}
  \item{is_VMR}{logical; \code{TRUE} if region passed the \code{var.test}}
}
}
\description{
Filter VMRs with ONT-specific variance tests and CI filters
}
\details{
This function verifies whether a set of potential VMRs (e.g., genes,
transposons, CpG islands) are differentially methylated or not in ONT data,
adding per-read Wilcoxon and F-tests on per-site proportions, confidence interval
filtering, and optional variance-fold change cutoffs.


For each potential VMR, per-site methylation proportions are aggregated per read,
then a two-sample Wilcoxon rank-sum test compares means (\code{wilcox_pvalue}), and
an F-test compares variances (\code{f_pvalue}). You may further filter by requiring
the 95% confidence interval from the F-test to exclude 1 (\code{ciExcludesOne}) and/or
apply a fold-change cutoff on the variance ratio (\code{varRatioFc}).
}
\examples{
\dontrun{
# load the ONT methylation data
data(ontSampleGRangesList)
# load the gene annotation data
data(GEs_hg38)

# select the transcript
transcript <- GEs_hg38[which(GEs_hg38$type == "transcript")]

# the regions where to compute the PMDs
regions <- GRanges(seqnames = Rle("chr1"), ranges = IRanges(1E6+5E5,2E6))
transcript <- transcript[overlapsAny(transcript, regions)]

# filter genes that are differntially methylated in the two conditions
VMRsGenesCG <- filterVMRsONT(ontSampleGRangesList[["GM18501"]],
               ontSampleGRangesList[["GM18876"]], potentialVMRs = transcript,
               context = "CG", pValueThreshold = 0.01,
               minCytosinesCount = 4, minProportionDifference = 0.01,
               minReadsPerCytosine = 3, ciExcludesOne = TRUE,
               varRatioFc = NULL, parallel = TRUE) # parallel recommended
}
}
\seealso{
\code{\link{readONTbam}},
\code{\link{computePMDs}}, \code{\link{computeCoMethylatedPositions}},
\code{\link{ontSampleGRangesList}}, \code{\link{GEs_hg38}}
}
\author{
Nicolae Radu Zabet and Young Jun Kim
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Replicates.R
\name{.smartMergeDMRsReplicates}
\alias{.smartMergeDMRsReplicates}
\title{This takes a list of DMRs and attempts to merge DMRs while keeping the new
DMRs statistically significant}
\usage{
.smartMergeDMRsReplicates(
  DMRs,
  minGap,
  respectSigns = TRUE,
  methylationData,
  minProportionDifference = 0.4,
  minReadsPerCytosine = 4,
  pseudocountM = 1,
  pseudocountN = 2,
  pValueThreshold = 0.01,
  condition = condition,
  m = m,
  n = n,
  cores = 1,
  BPPARAM = BPPARAM
)
}
\description{
This takes a list of DMRs and attempts to merge DMRs while keeping the new
DMRs statistically significant
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/regions.R
\name{.countCytosinesInside}
\alias{.countCytosinesInside}
\title{Count cytosines inside}
\usage{
.countCytosinesInside(methylationData, regions)
}
\arguments{
\item{methylationData}{a \code{\link[GenomicRanges]{GRanges}} object with five metadata 
columns; see \code{\link{methylationDataList}}}

\item{regions}{a \code{\link[GenomicRanges]{GRanges}} with the computed regions}
}
\value{
a \code{vector} with the number of cytosines in each DMR
}
\description{
This function counts the number of cytosines in each DMR
}
\author{
Jonathan Michael Foonlan Tsang
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/computeDMRs.R
\name{analyseReadsInsideRegionsForCondition}
\alias{analyseReadsInsideRegionsForCondition}
\title{Analyse reads inside regions for condition}
\usage{
analyseReadsInsideRegionsForCondition(
  regions,
  methylationData,
  context,
  label = "",
  parallel = FALSE,
  BPPARAM = NULL,
  cores = NULL
)
}
\arguments{
\item{regions}{a \code{\link{GRanges}} object with a list of regions on the 
genome; e.g. could be a list of DMRs}

\item{methylationData}{the methylation data in one condition
(see \code{\link{methylationDataList}}).}

\item{context}{the context in which to extract the reads (\code{"CG"}, 
\code{"CHG"} or \code{"CHH"}).}

\item{label}{a string to be added to the columns to identify the condition}

\item{parallel}{Logical; run in parallel if \code{TRUE}.}

\item{BPPARAM}{A \code{BiocParallelParam} object controlling parallel execution.
This value will automatically set when parallel is \code{TRUE}, also able to set as manually.}

\item{cores}{Integer number of workers (must not exceed BPPARAM$workers).
This value will automatically set as the maximum number of system workers,
also able to set as manually.}
}
\value{
a \code{\link{GRanges}} object with additional four metadata columns
\describe{
 \item{sumReadsM}{the number of methylated reads}
 \item{sumReadsN}{the total number of reads} 
 \item{proportion}{the proportion methylated reads} 
 \item{cytosinesCount}{the number of cytosines in the regions} 
}
}
\description{
This function extracts from the methylation data the total number of reads, 
the number of methylated reads and the number of cytosines in the specific 
context from a region (e.g. DMRs)
}
\examples{

# load the methylation data
data(methylationDataList)
 
#load the DMRs in CG context. These DMRs were computed with minGap = 200.
data(DMRsNoiseFilterCG)

#retrive the number of reads in CHH context in WT
DMRsNoiseFilterCGreadsCHH <- analyseReadsInsideRegionsForCondition(
                             DMRsNoiseFilterCG[1:10], 
                             methylationDataList[["WT"]], context = "CHH", 
                             label = "WT")


}
\seealso{
\code{\link{readONTbam}}, \code{\link{filterDMRs}}, \code{\link{computeDMRs}}, 
\code{\link{DMRsNoiseFilterCG}}, and \code{\link{mergeDMRsIteratively}}
}
\author{
Nicolae Radu Zabet
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/computeCoMethylatedRegions.R
\name{computeCoMethylatedRegions}
\alias{computeCoMethylatedRegions}
\title{Compute pairwise co-methylation statistics between regions}
\usage{
computeCoMethylatedRegions(
  methylationData,
  regions,
  minDistance = 500,
  maxDistance = 50000,
  minCoverage = 4,
  pValueThreshold = 0.05,
  correlation_test = "pearson",
  minCorrelation = -0.5,
  maxCorrelation = 0.5,
  parallel = FALSE,
  BPPARAM = NULL,
  cores = NULL
)
}
\arguments{
\item{methylationData}{A \code{GRanges} object containing cytosine sites, annotated with
ONT methylation metadata (columns \code{ONT_Cm}, \code{ONT_C}, etc.).}

\item{regions}{A \code{GRanges} object defining genomic regions (e.g., PMDs, Enhancer binding sites) to
evaluate for CMRs.}

\item{minDistance}{Minimum genomic distance (in base pairs) between two regions to be considered (default: 500).}

\item{maxDistance}{Maximum genomic distance (in base pairs) between two regions (default: 50,000).}

\item{minCoverage}{Minimum number of shared reads (per region pair) required to compute correlation (default: 4).}

\item{pValueThreshold}{Significance threshold for FDR-adjusted p-values (default: 0.05).}

\item{correlation_test}{Statistical method to compute correlation; must be one of \code{"pearson"}, \code{"spearman"}, or \code{"kendall"} (default: \code{"pearson"}).}

\item{minCorrelation}{Minimum allowed correlation value for a significant result (must be in between -1 and 0; default: -0.5).}

\item{maxCorrelation}{Maximum allowed correlation value for a significant result (must be in between 0 and 1; default: 0.5).}

\item{parallel}{Logical; run in parallel if \code{TRUE}.}

\item{BPPARAM}{A \code{BiocParallelParam} object controlling parallel execution.
This value will automatically set when parallel is \code{TRUE}, also able to set as manually.}

\item{cores}{Integer number of workers (must not exceed BPPARAM$workers).
This value will automatically set as the maximum number of system workers,
also able to set as manually.}
}
\value{
A \code{GInteractions} object, where each row represents a significantly correlated pair of genomic regions
from the input \code{regions}. The anchors of each interaction correspond to original regions,
and their genomic coordinates are retained in the \code{anchor1} and \code{anchor2} slots.

Additionally, a \code{genomic_position} meta-column is included to indicate the original coordinate ranges
(in UCSC/IGV format) for each interaction, aiding downstream interpretation or visualisation.

Each interaction is annotated with:
\describe{
  \item{correlation}{Correlation coefficient between the two regions}
  \item{coverage}{Number of shared reads used for correlation}
  \item{p.value}{FDR-adjusted p-value for the correlation test}
}
}
\description{
\code{computeCoMethylatedRegions()} calculates pairwise correlation statistics for methylation levels
across defined genomic regions (e.g., PMDs, Enhancer binding sites). For each region pair within the
specified distance range, the function computes per-read methylation proportions and
performs correlation testing (Pearson, Spearman, or Kendall). Pairs with strong correlations
(beyond user-defined thresholds) and significant p-values (FDR-adjusted) are returned.
}
\details{
Compute Co-Methylated Regions (CMRs)


The function first identifies all region pairs within the user-defined distance range. For each pair,
it calculates methylation proportions per read across both regions, extracts common reads,
and tests correlation using the selected method. FDR correction is applied globally across all region pairs.
}
\examples{
# Load methylation data and PMD regions
data("ont_gr_GM18870_chr1_sorted_bins_1k")
data("ont_gr_GM18870_chr1_PMD_bins_1k")

# Compute highly correlated regions (Pearson)
coMethylationRegion_pearson <- computeCoMethylatedRegions(ont_gr_GM18870_chr1_sorted_bins_1k,
                                                          ont_gr_GM18870_chr1_PMD_bins_1k[1:5],
                                                          minDistance = 500,
                                                          maxDistance = 50000,
                                                          minCoverage = 4,
                                                          pValueThreshold = 0.05,
                                                          correlation_test = "pearson",
                                                          minCorrelation = -0.5,
                                                          maxCorrelation = 0.5,
                                                          parallel = FALSE,
                                                          BPPARAM  = NULL)

}
\seealso{
\code{\link{readONTbam}}, \code{\link{computePMDs}},
\code{\link{ontSampleGRangesList}}
}
\author{
Nicolae Radu Zabet and Young Jun Kim
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/loadData.R
\name{saveBismark}
\alias{saveBismark}
\title{Save Bismark}
\usage{
saveBismark(methylationData, filename)
}
\arguments{
\item{methylationData}{the methylation data stored as a \code{\link{GRanges}}
object with four metadata columns (see \code{\link{methylationDataList}}).}

\item{filename}{the filename where the data will be saved.}
}
\value{
Invisibly returns \code{NULL}
}
\description{
This function takes as input a \code{\link{GRanges}} object generated with
\code{\link{readBismark}} and saves the output to a file using
Bismark CX report format.
}
\examples{

# load methylation data object
data(methylationDataList)

# save one dataset to a file
saveBismark(methylationDataList[["WT"]], "chr3test_a_thaliana_wt.CX_report")

}
\author{
Nicolae Radu Zabet
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/computePMDs.R
\name{computePMDs}
\alias{computePMDs}
\title{Compute PMDs}
\usage{
computePMDs(
  methylationData,
  regions = NULL,
  context = "CG",
  method = "noise_filter",
  windowSize = 100,
  kernelFunction = "triangular",
  lambda = 0.5,
  binSize = 100,
  minCytosinesCount = 4,
  minMethylation = 0.4,
  maxMethylation = 0.6,
  minGap = 200,
  minSize = 50,
  minReadsPerCytosine = 4,
  parallel = FALSE,
  BPPARAM = NULL,
  cores = NULL
)
}
\arguments{
\item{methylationData}{the methylation data in condition
(see \code{\link{ontSampleGRangesList}}).}

\item{regions}{a \code{\link[GenomicRanges]{GRanges}} object with the regions where to 
compute the PMDs. If \code{NULL}, the PMDs are computed genome-wide.}

\item{context}{the context in which the PMDs are computed (\code{"CG"}, 
\code{"CHG"} or \code{"CHH"}).}

\item{method}{Character string specifying the algorithm for PMD detection. 
If \code{"noise_filter"}, a sliding window of size \code{windowSize} is applied 
with the specified \code{kernelFunction} (and \code{lambda} for a Gaussian kernel) 
to smooth methylation counts before calling and merging PMDs. 
If \code{"neighbourhood"}, individual partially methylated cytosines are 
identified first and then merged into PMDs. 
If \code{"bins"}, the genome is partitioned into fixed bins of size \code{binSize}, 
partially methylation is sorted per bin, and significant bins are merged.}

\item{windowSize}{the size of the triangle base measured in nucleotides. 
This parameter is required only if the selected method is 
\code{"noise_filter"}.}

\item{kernelFunction}{a \code{character} indicating which kernel function to 
be used. Can be one of \code{"uniform"}, \code{"triangular"}, 
\code{"gaussian"} or \code{"epanechnicov"}. This is required only if the 
selected method is \code{"noise_filter"}.}

\item{lambda}{numeric value required for the Gaussian filter 
(\code{K(x) = exp(-lambda*x^2)}). This is required only if the selected 
method is \code{"noise_filter"} and the selected kernel function is 
\code{"gaussian"}.}

\item{binSize}{the size of the tiling bins in nucleotides. This parameter is 
required only if the selected method is \code{"bins"}.}

\item{minCytosinesCount}{PMDs with less cytosines in the specified context 
than \code{minCytosinesCount} will be discarded.}

\item{minMethylation}{Numeric [0,1]; minimum mean methylation proportion.}

\item{maxMethylation}{Numeric [0,1]; maximum mean methylation proportion.}

\item{minGap}{PMDs separated by a gap of at least \code{minGap} are not 
merged. Note that only PMDs where the change in methylation is in the same 
direction are joined.}

\item{minSize}{PMDs with a size smaller than \code{minSize} are discarded.}

\item{minReadsPerCytosine}{PMDs with the average number of reads lower than 
\code{minReadsPerCytosine} are discarded.}

\item{parallel}{Logical; run in parallel if \code{TRUE}.}

\item{BPPARAM}{A \code{BiocParallelParam} object controlling parallel execution.
This value will automatically set when parallel is \code{TRUE}, also able to set as manually.}

\item{cores}{Integer number of workers (must not exceed BPPARAM$workers).
This value will automatically set as the maximum number of system workers,
also able to set as manually.}
}
\value{
the PMDs stored as a \code{\link[GenomicRanges]{GRanges}} object with the following 
metadata columns:
\describe{
 \item{context}{the context in which the PMDs was computed (\code{"CG"}, 
 \code{"CHG"} or \code{"CHH"}).}
 \item{sumReadsM}{the number of methylated reads.}
 \item{sumReadsN}{the total number of reads.} 
 \item{proportion}{the proportion methylated reads filtered between
 \code{minMethylation} and \code{maxMethylation}.} 
 \item{cytosinesCount}{the number of cytosines in the PMDs.} 
}
}
\description{
This function computes the partially methylated domains between pre-set
min and max proportion values.
}
\examples{
# load the ONT methylation data 
data(ontSampleGRangesList)

# the regions where to compute the PMDs
chr1_ranges <- GRanges(seqnames = Rle("chr1"), ranges = IRanges(1E6+5E5,2E6))

# compute the PMDs in CG context with noise_filter method
PMDsNoiseFilterCG <- computePMDs(ontSampleGRangesList[["GM18501"]],
                                 regions = chr1_ranges,
                                 context = "CG",
                                 windowSize = 100,
                                 method = "noise_filter",
                                 kernelFunction = "triangular", 
                                 lambda = 0.5,
                                 minCytosinesCount = 4, 
                                 minMethylation = 0.4, 
                                 maxMethylation = 0.6, 
                                 minGap = 200, 
                                 minSize = 50, 
                                 minReadsPerCytosine = 4, 
                                 cores = 1,
                                 parallel = FALSE)
 
\dontrun{                                
# compute the PMDs in CG context with neighbourhood method
PMDsNeighbourhoodCG <- computePMDs(ontSampleGRangesList[["GM18501"]],
                                   regions = chr1_ranges,
                                   context = "CG",
                                   method = "neighbourhood"
                                   minCytosinesCount = 4, 
                                   minMethylation = 0.4, 
                                   maxMethylation = 0.6, 
                                   minGap = 200, 
                                   minSize = 50, 
                                   minReadsPerCytosine = 4, 
                                   cores = 1,
                                   parallel = FALSE)
                                   
# compute the PMDs in CG context with bins method
PMDsBinsCG <- computePMDs(ontSampleGRangesList[["GM18501"]],
                          regions = chr1_ranges,
                          context = "CG",
                          method = "bins",
                          binSize = 100,
                          minCytosinesCount = 4, 
                          minMethylation = 0.4, 
                          maxMethylation = 0.6, 
                          minGap = 200, 
                          minSize = 50, 
                          minReadsPerCytosine = 4, 
                          cores = 1,
                          parallel = FALSE)
}
}
\seealso{
\code{\link{readONTbam}}, \code{\link{filterPMDs}}, \code{\link{mergePMDsIteratively}}, 
\code{\link{analyseReadsInsideRegionsForConditionPMD}} and 
\code{\link{PMDsNoiseFilterCG}}
}
\author{
Nicolae Radu Zabet, Jonathan Michael Foonlan Tsang and Young Jun Kim
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/loadData.R
\name{poolMethylationDatasets}
\alias{poolMethylationDatasets}
\title{Pool methylation data}
\usage{
poolMethylationDatasets(methylationDataList)
}
\arguments{
\item{methylationDataList}{a \code{\link{GRangesList}} object where each
element of the list is a \code{\link{GRanges}} object with the methylation
data in the corresponding condition (see \code{\link{methylationDataList}}).}
}
\value{
the methylation data stored as a \code{\link{GRanges}}
object with four metadata columns (see \code{\link{methylationDataList}}).
If the Granges are from ONT datasets, its have six metedata columns include as
ONT_Cm and ONT_C (see \code{\link{readONTbam}}).
}
\description{
This function pools together multiple methylation datasets.
}
\examples{
# load methylation data object
data(methylationDataList)

# pools the two datasets together
pooledMethylationData <- poolMethylationDatasets(methylationDataList)

}
\author{
Nicolae Radu Zabet updated by Young Jun Kim
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/coverage.R
\name{computeMethylationDataCoverage}
\alias{computeMethylationDataCoverage}
\title{Compute methylation data coverage}
\usage{
computeMethylationDataCoverage(
  methylationData,
  regions = NULL,
  context = "CG",
  breaks = NULL,
  proportion = TRUE
)
}
\arguments{
\item{methylationData}{the methylation data stored as a \code{\link{GRanges}}
object with four metadata columns (see \code{\link{methylationDataList}}).}

\item{regions}{a \code{\link{GRanges}} object with the regions where to
compute the coverage. If \code{NULL}, the coverage is computed genome-wide.}

\item{context}{the context in which the DMRs are computed (\code{"CG"},
\code{"CHG"} or \code{"CHH"}).}

\item{breaks}{a \code{numeric} vector specifing the different values for the
thresholds when computing the coverage.}

\item{proportion}{a \code{logical} value indicating whether to compute the
proportion (\code{TRUE}) or raw counts (\code{FALSE}).}
}
\value{
a \code{vector} with the proportion (or raw count) of cytosines that
have the number of reads higher or equal than the threshold values specified
in the \code{breaks} \code{vector}.
}
\description{
This function computes the coverage for bisulfite sequencing data. It
returns a \code{vector} with the proportion (or raw count) of cytosines that
have the number of reads higher or equal than a \code{vector} of specified
thresholds.
}
\examples{

# load the methylation data
data(methylationDataList)

# compute coverage in CG context
breaks <- c(1,5,10,15)
coverage_CG_wt <- computeMethylationDataCoverage(methylationDataList[["WT"]],
                 context="CG", breaks=breaks)


}
\seealso{
\code{\link{plotMethylationDataCoverage}},
\code{\link{methylationDataList}}
}
\author{
Nicolae Radu Zabet and Jonathan Michael Foonlan Tsang
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/computeVMDs.R
\name{computeVMDs}
\alias{computeVMDs}
\title{Compute VMDs}
\usage{
computeVMDs(
  methylationData,
  regions = NULL,
  context = "CG",
  binSize = 100,
  minCytosinesCount = 4,
  sdCutoffMethod = "per.high",
  percentage = 0.05,
  minGap = 200,
  minSize = 50,
  minReadsPerCytosine = 4,
  parallel = FALSE,
  BPPARAM = NULL,
  cores = NULL
)
}
\arguments{
\item{methylationData}{the methylation data in condition
(see \code{\link{ontSampleGRangesList}}).}

\item{regions}{a \code{\link{GRanges}} object with the regions where to 
compute the VMDs. If \code{NULL}, the VMDs are computed genome-wide.}

\item{context}{the context in which the VMDs are computed (\code{"CG"}, 
\code{"CHG"} or \code{"CHH"}).}

\item{binSize}{the size of the tiling bins in nucleotides. This parameter is 
required only if the selected method is \code{"bins"}.}

\item{minCytosinesCount}{VMDs with less cytosines in the specified context 
than \code{minCytosinesCount} will be discarded.}

\item{sdCutoffMethod}{Character string specifying how to determine the cutoff
for filtering VMDs based on their methylation variance (standard deviation).
Available options are:
\describe{
  \item{\code{"per.high"}}{Selects the top \code{percentage} of regions with the highest variance (standard deviation).}
  \item{\code{"per.low"}}{Selects the bottom \code{percentage} of regions with the lowest variance.}
  \item{\code{"EDE.high"}}{Uses the elbow point (inflection/knee) from the descendingly sorted variance values to determine a data-driven high-variance cutoff. Retains regions with SD above this elbow point.}
  \item{\code{"EDE.low"}}{Uses the elbow point from the ascendingly sorted variance values to define a low-variance cutoff. Retains regions with SD below this point.}
}
This allows either quantile-based filtering or automatic detection of variance thresholds based on distribution shape.}

\item{percentage}{Numeric cutoff used when \code{sdCutoffMethod} is set to
\code{"per.high"} or \code{"per.low"}. Represents the quantile threshold:
for example, \code{percentage = 0.05} keeps the top 5\% or bottom 5\% of
bins based on standard deviation, depending on the selected method.}

\item{minGap}{VMDs separated by a gap of at least \code{minGap} are not 
merged. Note that only VMDs where the change in methylation is in the same 
direction are joined.}

\item{minSize}{VMDs with a size smaller than \code{minSize} are discarded.}

\item{minReadsPerCytosine}{VMDs with the average number of reads lower than 
\code{minReadsPerCytosine} are discarded.}

\item{parallel}{Logical; run in parallel if \code{TRUE}.}

\item{BPPARAM}{A \code{BiocParallelParam} object controlling parallel execution.
This value will automatically set when parallel is \code{TRUE}, also able to set as manually.}

\item{cores}{Integer number of workers (must not exceed BPPARAM$workers).
This value will automatically set as the maximum number of system workers,
also able to set as manually.}
}
\value{
the VMDs stored as a \code{\link{GRanges}} object with the following 
metadata columns:
\describe{
 \item{context}{the context in which the VMDs was computed (\code{"CG"}, 
 \code{"CHG"} or \code{"CHH"}).}
 \item{sumReadsM}{the number of methylated reads.}
 \item{sumReadsN}{the total number of reads.} 
 \item{proportion}{the proportion from total methylated reads.} 
 \item{cytosinesCount}{the number of cytosines in the VMDs.}
 \item{mean}{mean value comparing per‐read proportions}
 \item{sd}{standard deviation comparing per‐read proportions}
 \item{w_mean}{weighted mean value comparing per‐read proportions}
 \item{w_sd}{weighted standard deviation comparing per‐read proportions} 
}
}
\description{
This function computes the variance methylated domains between pre-set
min and max proportion values.
}
\examples{
\dontrun{
# load the ONT methylation data 
data(ontSampleGRangesList)

# the regions where to compute the VMDs
chr1_ranges <- GRanges(seqnames = Rle("chr1"), ranges = IRanges(1E6+5E5,1E6+6E5))

# compute the VMDs in CG context with bins method
VMDsBinsCG <- computeVMDs(ontSampleGRangesList[["GM18501"]],
                          regions = NULL,
                          context = "CG",
                          binSize = 100,
                          minCytosinesCount = 4,
                          sdCutoffMethod = "EDE.high",
                          percentage = 0.05,
                          minGap = 200,
                          minSize = 50,
                          minReadsPerCytosine = 4,
                          parallel = FALSE,
                          BPPARAM = NULL,
                          cores = 1)
}
}
\seealso{
\code{\link{readONTbam}}, \code{\link{filterVMDs}}
and \code{\link{analyseReadsInsideRegionsForCondition}}
}
\author{
Nicolae Radu Zabet, Jonathan Michael Foonlan Tsang and Young Jun Kim
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generalFunctions.R
\name{.validateGenome}
\alias{.validateGenome}
\title{Validate BSgenome Object}
\usage{
.validateGenome(genome)
}
\arguments{
\item{genome}{A BSgenome object (e.g. BSgenome.Hsapiens.UCSC.hg38).}
}
\value{
Invisibly `TRUE` if valid; otherwise throws an error.
}
\description{
Ensures `genome` inherits from BSgenome and that its package is installed.
}
\author{
Radu Zabet and Young Jun Kim
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/DMRcaller.R
\docType{package}
\name{DMRcaller}
\alias{DMRcaller-package}
\alias{DMRcaller}
\title{Call Differentially Methylated Regions (DMRs) between two samples}
\description{
Supports both Bisulfite Sequencing (Bismark CX reports) and Oxford Nanopore
Sequencing (MM/ML tags) for per-site methylation calling. Identifies
differentially methylated regions between two samples in CG and non-CG
contexts.
}
\details{
For bisulfite data, the input is Bismark CX report files and the output
is a list of DMRs stored as a \code{GRanges} object.
\itemize{
  \item \code{readsM} — count of modified reads per site
  \item \code{readsN} — total same-strand coverage per site
}

For Nanopore data, the input is an indexed ONT BAM with calling
\code{\link{readONTbam}} function in this package and the output is a
\code{GRanges} augmented with metadata columns:
\itemize{
  \item \code{ONT_Cm} — comma-delimited read-indices called “modified”
  \item \code{ONT_C}  — comma-delimited read-indices covering but not modified
  \item \code{readsM} — count of modified reads per site
  \item \code{readsN} — total same-strand coverage per site
}

The most important functions in the \pkg{DMRcaller} are:
\describe{
 \item{\code{\link{readBismark}}}{reads the Bismark CX report files in a
       \code{\link[GenomicRanges]{GRanges}} object.}
 \item{\code{\link{readBismarkPool}}}{Reads multiple CX report files and
       pools them together.}
 \item{\code{\link{saveBismark}}}{saves the methylation data stored in a
       \code{\link[GenomicRanges]{GRanges}} object into a Bismark CX report file.}
 \item{\code{\link{selectCytosine}}}{Enumerates cytosine positions in a
       BSgenome reference, optionally filtering by methylation context
       (CG/CHG/CHH), chromosome and genomic region.}
 \item{\code{\link{readONTbam}}}{Loads an Oxford Nanopore BAM
       (with MM/ML tags), decodes per‐C modification probabilities and
       counts modified vs. unmodified reads per site.}
 \item{\code{\link{poolMethylationDatasets}}}{pools together multiple
       methylation datasets.}
 \item{\code{\link{poolTwoMethylationDatasets}}}{pools together two
       methylation datasets.}
 \item{\code{\link{computeMethylationDataCoverage}}}{Computes the coverage
       for the bisulfite sequencing data.}
 \item{\code{\link{plotMethylationDataCoverage}}}{Plots the coverage for the
       bisulfite sequencing data.}
 \item{\code{\link{computeMethylationDataSpatialCorrelation}}}{Computes the
       correlation between methylation levels  as a function of the distances
       between the Cytosines.}
 \item{\code{\link{plotMethylationDataSpatialCorrelation}}}{Plots the
       correlation of methylation levels for Cytosines located at a certain
       distance apart.}
 \item{\code{\link{computeMethylationProfile}}}{Computes the low resolution
       profiles for the bisulfite sequencing data at certain locations.}
 \item{\code{\link{plotMethylationProfile}}}{Plots the low resolution
       profiles for the bisulfite sequencing data at certain locations.}
 \item{\code{\link{plotMethylationProfileFromData}}}{Plots the low resolution
       profiles for the loaded bisulfite sequencing data.}
 \item{\code{\link{computeDMRs}}}{Computes the differentially methylated
       regions between two conditions.}
 \item{\code{\link{filterDMRs}}}{Filters a list of (potential) differentially
       methylated regions.}
 \item{\code{\link{mergeDMRsIteratively}}}{Merge DMRs iteratively.}
 \item{\code{\link{analyseReadsInsideRegionsForCondition}}}{Analyse reads
       inside regions for condition.}
 \item{\code{\link{plotLocalMethylationProfile}}}{Plots the methylation
       profile at one locus for the bisulfite sequencing data.}
 \item{\code{\link{computeOverlapProfile}}}{Computes the distribution of a
       set of subregions on a large region.}
 \item{\code{\link{plotOverlapProfile}}}{Plots the distribution of a set of
       subregions on a large region.}
 \item{\code{\link{getWholeChromosomes}}}{Computes the GRanges objects with
       each chromosome as an element from the methylationData.}
 \item{\code{\link{joinReplicates}}}{Merges two GRanges objects with single
       reads columns. It is necessary to start the analysis of DMRs with
       biological replicates.}
 \item{\code{\link{computeDMRsReplicates}}}{Computes the differentially
       methylated regions between two conditions with multiple
       biological replicates.}
 \item{\code{\link{selectCytosine}}}{Enumerates cytosine positions in a
       BSgenome reference.}
 \item{\code{\link{readONTbam}}}{Loads an ONT BAM (MM/ML tags),
       decodes per‐C modification probabilities,
       and counts modified vs. unmodified reads per site.}
 \item{\code{\link{computePMDs}}}{Partitions the genome into PMDs via
       three methods ("noise_filter", "neighbourhood", "bins").}
 \item{\code{\link{filterPMDs}}}{Filters a set of PMDs by methylation level
       and read depth.}
 \item{\code{\link{mergePMDsIteratively}}}{Merge PMDs while preserving
       statistical significance.}
 \item{\code{\link{analyseReadsInsideRegionsForConditionPMD}}}{Counts reads
       in each PMD for one condition.}
 \item{\code{\link{computeVMDs}}}{Computes the variance methylated domains 
       between pre-set min and max proportion values}
 \item{\code{\link{filterVMRsONT}}}{Filters VMRs with ONT-specific variance 
       tests and CI filters}
 \item{\code{\link{computeCoMethylatedPositions}}}{Computes pairwise 
       co‐methylation between Cytosine sites within regions.}
 \item{\code{\link{computeCoMethylatedRegions}}}{Computes pairwise 
       co-methylation statistics between regions.}
}
}
\examples{
\dontrun{
# load the methylation data
data(methylationDataList)
library(BSgenome.Hsapiens.UCSC.hg38)

# All cytosines in hg38:
gr_all <- selectCytosine()

# Only CpG sites on chr1 and chr2:
gr_chr1_2 <- selectCytosine(context="CG", chr=c("chr1","chr2"))

# CHH sites in a specific region on chr3:
my_region <- GRanges("chr3", IRanges(1e6, 1e6 + 1e5))
gr_region <- selectCytosine(context="CHH", chr="chr3", region=my_region)

# set the bam file directory
bam_path <- system.file("extdata", "scanBamChr1Random5.bam", package="DMRcaller")

# read ONTbam file (chromosome 1 only) in CG context with BSgenome.Hsapiens.UCSC.hg38
ONTSampleGRanges <- readONTbam(bamfile = bam_path, ref_gr = NULL, modif = "C+m?",
                         prob_thresh = 0.50,genome = BSgenome.Hsapiens.UCSC.hg38,
                         context = "CG", chr = "chr1", region = NULL,
                         synonymous = FALSE, parallel = FALSE, BPPARAM = NULL)

# plot the low resolution profile at 5 Kb resolution
par(mar=c(4, 4, 3, 1)+0.1)
plotMethylationProfileFromData(methylationDataList[["WT"]],
                               methylationDataList[["met1-3"]],
                               conditionsNames=c("WT", "met1-3"),
                               windowSize = 5000, autoscale = TRUE,
                               context = c("CG", "CHG", "CHH"),
                               labels = LETTERS)

# compute low resolution profile in 10 Kb windows in CG context
lowResProfileWTCG <- computeMethylationProfile(methylationDataList[["WT"]],
                     region, windowSize = 10000, context = "CG")

lowResProfileMet13CG <- computeMethylationProfile(
                     methylationDataList[["met1-3"]], region,
                     windowSize = 10000, context = "CG")

lowResProfileCG <- GRangesList("WT" = lowResProfileWTCG,
                   "met1-3" = lowResProfileMet13CG)

# compute low resolution profile in 10 Kb windows in CHG context
lowResProfileWTCHG <- computeMethylationProfile(methylationDataList[["WT"]],
                     region, windowSize = 10000, context = "CHG")

lowResProfileMet13CHG <- computeMethylationProfile(
                     methylationDataList[["met1-3"]], region,
                     windowSize = 10000, context = "CHG")

lowResProfileCHG <- GRangesList("WT" = lowResProfileWTCHG,
                   "met1-3" = lowResProfileMet13CHG)

# plot the low resolution profile
par(mar=c(4, 4, 3, 1)+0.1)
par(mfrow=c(2,1))
plotMethylationProfile(lowResProfileCG, autoscale = FALSE,
                       labels = LETTERS[1],
                       title="CG methylation on Chromosome 3",
                       col=c("#D55E00","#E69F00"),  pch = c(1,0),
                       lty = c(4,1))
plotMethylationProfile(lowResProfileCHG, autoscale = FALSE,
                       labels = LETTERS[2],
                       title="CHG methylation on Chromosome 3",
                       col=c("#0072B2", "#56B4E9"),  pch = c(16,2),
                       lty = c(3,2))

# plot the coverage in all three contexts
plotMethylationDataCoverage(methylationDataList[["WT"]],
                           methylationDataList[["met1-3"]],
                           breaks = 1:15, regions = NULL,
                           conditionsNames = c("WT","met1-3"),
                           context = c("CG", "CHG", "CHH"),
                           proportion = TRUE, labels = LETTERS, col = NULL,
                           pch = c(1,0,16,2,15,17), lty = c(4,1,3,2,6,5),
                           contextPerRow = FALSE)

# plot the correlation of methylation levels as a function of distance
plotMethylationDataSpatialCorrelation(methylationDataList[["WT"]],
                           distances = c(1,5,10,15), regions = NULL,
                           conditionsNames = c("WT","met1-3"),
                           context = c("CG"),
                           labels = LETTERS, col = NULL,
                           pch = c(1,0,16,2,15,17), lty = c(4,1,3,2,6,5),
                           contextPerRow = FALSE)

# the regions where to compute the DMRs
regions <- GRanges(seqnames = Rle("Chr3"), ranges = IRanges(1,1E6))

# compute the DMRs in CG context with noise_filter method
DMRsNoiseFilterCG <- computeDMRs(methylationDataList[["WT"]],
                     methylationDataList[["met1-3"]], regions = regions,
                     context = "CG", method = "noise_filter",
                     windowSize = 100, kernelFunction = "triangular",
                     test = "score", pValueThreshold = 0.01,
                     minCytosinesCount = 4, minProportionDifference = 0.4,
                     minGap = 200, minSize = 50, minReadsPerCytosine = 4,
                     cores = 1)

# compute the DMRs in CG context with neighbourhood method
DMRsNeighbourhoodCG <- computeDMRs(methylationDataList[["WT"]],
                       methylationDataList[["met1-3"]], regions = regions,
                       context = "CG", method = "neighbourhood",
                       test = "score", pValueThreshold = 0.01,
                       minCytosinesCount = 4, minProportionDifference = 0.4,
                       minGap = 200, minSize = 50, minReadsPerCytosine = 4,
                       cores = 1)

# compute the DMRs in CG context with bins method
DMRsBinsCG <- computeDMRs(methylationDataList[["WT"]],
               methylationDataList[["met1-3"]], regions = regions,
               context = "CG", method = "bins", binSize = 100,
               test = "score", pValueThreshold = 0.01, minCytosinesCount = 4,
               minProportionDifference = 0.4, minGap = 200, minSize = 50,
               minReadsPerCytosine = 4, cores = 1)

# load the gene annotation data
data(GEs)

# select the genes
genes <- GEs[which(GEs$type == "gene")]

# the regions where to compute the DMRs
genes <- genes[overlapsAny(genes, regions)]

# filter genes that are differntially methylated in the two conditions
DMRsGenesCG <- filterDMRs(methylationDataList[["WT"]],
               methylationDataList[["met1-3"]], potentialDMRs = genes,
               context = "CG", test = "score", pValueThreshold = 0.01,
               minCytosinesCount = 4, minProportionDifference = 0.4,
               minReadsPerCytosine = 3, cores = 1)

# merge the DMRs
DMRsNoiseFilterCGLarger <- mergeDMRsIteratively(DMRsNoiseFilterCG,
                           minGap = 500, respectSigns = TRUE,
                           methylationDataList[["WT"]],
                           methylationDataList[["met1-3"]],
                           context = "CG", minProportionDifference=0.4,
                           minReadsPerCytosine = 1, pValueThreshold=0.01,
                           test="score",alternative = "two.sided")

# select the genes
genes <- GEs[which(GEs$type == "gene")]

# the coordinates of the area to be plotted
chr3Reg <- GRanges(seqnames = Rle("Chr3"), ranges = IRanges(510000,530000))

# load the DMRs in CG context
data(DMRsNoiseFilterCG)

DMRsCGlist <- list("noise filter"=DMRsNoiseFilterCG,
                   "neighbourhood"=DMRsNeighbourhoodCG,
                   "bins"=DMRsBinsCG,
                   "genes"=DMRsGenesCG)


# plot the CG methylation
par(mar=c(4, 4, 3, 1)+0.1)
par(mfrow=c(1,1))
plotLocalMethylationProfile(methylationDataList[["WT"]],
                           methylationDataList[["met1-3"]], chr3Reg,
                           DMRsCGlist, c("WT", "met1-3"), GEs,
                           windowSize=100, main="CG methylation")


hotspotsHypo <- computeOverlapProfile(
               DMRsNoiseFilterCG[(DMRsNoiseFilterCG$regionType == "loss")],
               region, windowSize=2000, binary=TRUE, cores=1)

hotspotsHyper <- computeOverlapProfile(
               DMRsNoiseFilterCG[(DMRsNoiseFilterCG$regionType == "gain")],
               region, windowSize=2000, binary=TRUE, cores=1)

plotOverlapProfile(GRangesList("Chr3"=hotspotsHypo),
                   GRangesList("Chr3"=hotspotsHyper),
                   names=c("loss", "gain"), title="CG methylation")
# loading synthetic data
data("syntheticDataReplicates")

# creating condition vector
condition <- c("a", "a", "b", "b")

# computing DMRs using the neighbourhood method
DMRsReplicatesNeighbourhood <- computeDMRsReplicates(methylationData = methylationData,
                                                     condition = condition,
                                                     regions = NULL,
                                                     context = "CHH",
                                                     method = "neighbourhood",
                                                     test = "betareg",
                                                     pseudocountM = 1,
                                                     pseudocountN = 2,
                                                     pValueThreshold = 0.01,
                                                     minCytosinesCount = 4,
                                                     minProportionDifference = 0.4,
                                                     minGap = 200,
                                                     minSize = 50,
                                                     minReadsPerCytosine = 4,
                                                     cores = 1)
# load the ONT methylation data
data(ontSampleGRangesList)

# the regions where to compute the PMDs
chr1_ranges <- GRanges(seqnames = Rle("chr1"), ranges = IRanges(1E6+5E5,2E6))

# compute the PMDs in CG context with noise_filter method
PMDsNoiseFilterCG <- computePMDs(ontSampleGRangesList[["GM18501"]],
                                 regions = chr1_ranges,
                                 context = "CG",
                                 windowSize = 100,
                                 method = "noise_filter",
                                 kernelFunction = "triangular",
                                 lambda = 0.5,
                                 minCytosinesCount = 4,
                                 minMethylation = 0.4,
                                 maxMethylation = 0.6,
                                 minGap = 200,
                                 minSize = 50,
                                 minReadsPerCytosine = 4,
                                 cores = 1,
                                 parallel = FALSE)

# compute the PMDs in CG context with neighbourhood method
PMDsNeighbourhoodCG <- computePMDs(ontSampleGRangesList[["GM18501"]],
                                   regions = chr1_ranges,
                                   context = "CG",
                                   method = "neighbourhood"
                                   minCytosinesCount = 4,
                                   minMethylation = 0.4,
                                   maxMethylation = 0.6,
                                   minGap = 200,
                                   minSize = 50,
                                   minReadsPerCytosine = 4,
                                   cores = 1,
                                   parallel = FALSE)

# compute the PMDs in CG context with bins method
PMDsBinsCG <- computePMDs(ontSampleGRangesList[["GM18501"]],
                          regions = chr1_ranges,
                          context = "CG",
                          method = "bins",
                          binSize = 100,
                          minCytosinesCount = 4,
                          minMethylation = 0.4,
                          maxMethylation = 0.6,
                          minGap = 200,
                          minSize = 50,
                          minReadsPerCytosine = 4,
                          cores = 1,
                          parallel = FALSE)

# load the gene annotation data
data(GEs_hg38)

# select the transcript
transcript <- GEs_hg38[which(GEs_hg38$type == "transcript")]

# the regions where to compute the PMDs
regions <- GRanges(seqnames = Rle("chr1"), ranges = IRanges(1E6+5E5,2E6))
transcript <- transcript[overlapsAny(transcript, regions)]

# filter genes that are partially methylated in the two conditions
PMDsGenesCG <- filterPMDs(ontSampleGRangesList[["GM18501"]],
               potentialPMDs = transcript,
               context = "CG", minMethylation = 0.4, maxMethylation = 0.6,
               minCytosinesCount = 4, minReadsPerCytosine = 3, cores = 1)

# load the PMDs in CG context they were computed with minGap = 200
data(PMDsNoiseFilterCG)


# merge the PMDs
PMDsNoiseFilterCGLarger <- mergePMDsIteratively(PMDsNoiseFilterCG[1:100],
                           minGap = 500, respectSigns = TRUE,
                           ontSampleGRangesList[["GM18501"]], context = "CG",
                           minReadsPerCytosine = 4, minMethylation = 0.4,
                           maxMethylation = 0.6, cores = 1)

# set genomic coordinates where to compute PMDs
chr1_ranges <- GRanges(seqnames = Rle("chr1"), ranges = IRanges(1E6+5E5,2E6))

# compute PMDs and remove gaps smaller than 200 bp
PMDsNoiseFilterCG200 <- computePMDs(ontSampleGRangesList[["GM18501"]],
                       regions = chr1_ranges, context = "CG", method = "noise_filter",
                       windowSize = 100, kernelFunction = "triangular",
                       minCytosinesCount = 1, minMethylation = 0.4,
                       maxMethylation = 0.6, minGap = 0, minSize = 200,
                       minReadsPerCytosine = 1, cores = 1)
PMDsNoiseFilterCG0 <- computePMDs(ontSampleGRangesList[["GM18501"]],
                       regions = chr1_ranges, context = "CG", method = "noise_filter",
                       windowSize = 100, kernelFunction = "triangular",
                       minCytosinesCount = 1, minMethylation = 0.4,
                       maxMethylation = 0.6, minGap = 0, minSize = 0,
                       minReadsPerCytosine = 1, cores = 1)
PMDsNoiseFilterCG0Merged200 <- mergePMDsIteratively(PMDsNoiseFilterCG0,
                             minGap = 200, respectSigns = TRUE,
                             ontSampleGRangesList[["GM18501"]], context = "CG",
                             minReadsPerCytosine = 4, minMethylation = 0.4,
                             maxMethylation = 0.6, cores = 1)

#check that all newley computed PMDs are identical
print(all(PMDsNoiseFilterCG200 == PMDsNoiseFilterCG0Merged200))

#retrive the number of reads in CG context in GM18501
PMDsNoiseFilterCGreadsCG <- analyseReadsInsideRegionsForConditionPMD(
                             PMDsNoiseFilterCG[1:10],
                             ontSampleGRangesList[["GM18501"]], context = "CG",
                             label = "GM18501")

# load the PMD data
data(PMDsBinsCG)

# compute the co-methylations with Fisher's exact test
coMetylationFisher <- computeCoMethylatedPositions(
  ontSampleGRangesList[[1]],
  regions = PMDsBinsCG,
  minDistance = 150,
  maxDistance = 1000,
  minCoverage = 4,
  pValueThreshold = 0.01,
  test = "fisher",
  parallel = FALSE)

# compute the co-methylations with Permuation test
coMetylationPermutation <- computeCoMethylatedPositions(
  ontSampleGRangesList[[1]],
  regions = PMDsBinsCG,
  minDistance = 150,
  maxDistance = 1000,
  minCoverage = 4,
  pValueThreshold = 0.01,
  test = "permutation",
  parallel = FALSE) # highly recommended to set as TRUE

# select the transcript
transcript <- GEs_hg38[which(GEs_hg38$type == "transcript")]

# the regions where to compute the PMDs
regions <- GRanges(seqnames = Rle("chr1"), ranges = IRanges(1E6+5E5,2E6))
transcript <- transcript[overlapsAny(transcript, regions)]

# filter genes that are differntially methylated in the two conditions
VMRsGenesCG <- filterVMRsONT(ontSampleGRangesList[["GM18501"]],
               ontSampleGRangesList[["GM18876"]], potentialVMRs = transcript,
               context = "CG", pValueThreshold = 0.01,
               minCytosinesCount = 4, minProportionDifference = 0.01,
               minReadsPerCytosine = 3, ciExcludesOne = TRUE,
               varRatioFc = NULL, parallel = TRUE) # parallel recommended

}


}
\seealso{
See \code{vignette("rd", package = "DMRcaller")} for an overview
of the package.
}
\author{
Nicolae Radu Zabet \email{r.zabet@qmul.ac.uk},
Jonathan Michael Foonlan Tsang \email{jmft2@cam.ac.uk},
Alessandro Pio Greco \email{apgrec@essex.ac.uk},
Young Jun Kim \email{qc25039@qmul.ac.uk}

Maintainer: Nicolae Radu Zabet \email{r.zabet@qmul.ac.uk}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/regions.R
\name{.mergeDMRsIteratively}
\alias{.mergeDMRsIteratively}
\title{This takes a list of DMRs and attempts to merge DMRs while keeping the new 
DMRs statistically significant}
\usage{
.mergeDMRsIteratively(
  DMRs,
  minGap,
  respectSigns = TRUE,
  methylationData,
  minProportionDifference = 0.4,
  minReadsPerCytosine = 4,
  pValueThreshold = 0.01,
  test = "fisher",
  alternative = "two.sided"
)
}
\description{
This takes a list of DMRs and attempts to merge DMRs while keeping the new 
DMRs statistically significant
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/DMRcaller.R
\docType{data}
\name{GEs}
\alias{GEs}
\title{The genetic elements data}
\format{
A \code{GRanges} object
}
\source{
The object was created by calling \code{import.gff3} function
from \code{rtracklayer} package for
\url{ftp://ftp.arabidopsis.org/Maps/gbrowse_data/TAIR10/TAIR10_GFF3_genes_transposons.gff}
}
\description{
A \code{GRanges} object containing the annotation of the Arabidopsis thaliana
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/DMRcaller.R
\docType{data}
\name{syntheticDataReplicates}
\alias{syntheticDataReplicates}
\title{Simulated data for biological replicates}
\format{
A \code{\link[GenomicRanges]{GRanges}} object containing multiple metadata
columns with the reads from each object passed as parameter
}
\source{
The object was created by calling \code{\link{joinReplicates}}
 function.
}
\description{
A \code{GRanges} object containing simulated date for methylation in four
samples. The conditions assciated witch each sample are a, a, b and b.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/profile.R
\name{plotMethylationProfileFromData}
\alias{plotMethylationProfileFromData}
\title{Plot methylation profile from data}
\usage{
plotMethylationProfileFromData(
  methylationData1,
  methylationData2 = NULL,
  regions = NULL,
  conditionsNames = NULL,
  context = "CG",
  windowSize = NULL,
  autoscale = FALSE,
  labels = NULL,
  col = NULL,
  pch = c(1, 0, 16, 2, 15, 17),
  lty = c(4, 1, 3, 2, 6, 5),
  contextPerRow = TRUE
)
}
\arguments{
\item{methylationData1}{the methylation data in condition 1
(see \code{\link{methylationDataList}}).}

\item{methylationData2}{the methylation data in condition 2
(see \code{\link{methylationDataList}}). This is optional.}

\item{regions}{a \code{\link{GRanges}} object with the regions where to plot
the profiles.}

\item{conditionsNames}{the names of the two conditions. This will be used to
plot the legend.}

\item{context}{a \code{vector} with all contexts in which the DMRs are
computed (\code{"CG"}, \code{"CHG"} or \code{"CHH"}).}

\item{windowSize}{a \code{numeric} value indicating the size of the window in
which methylation is averaged.}

\item{autoscale}{a \code{logical} value indicating whether the values are
autoscalled for each context or not.}

\item{labels}{a \code{vector} of \code{character} used to add a subfigure
character to the plot. If \code{NULL} nothing is added.}

\item{col}{a \code{character} vector with the colours. It needs to contain a
minimum of 2 colours per condition. If not or if \code{NULL}, the defalut
colours will be used.}

\item{pch}{the R symbols used to plot the data It needs to contain a minimum
of 2 symbols per condition. If not or if \code{NULL}, the defalut symbols
will be used.}

\item{lty}{the line types used to plot the data. It needs to contain a
minimum of 2 line types per condition. If not or if \code{NULL}, the defalut
line types will be used.}

\item{contextPerRow}{a \code{logical} value indicating if the each row
represents an individual context. If \code{FALSE}, each column will represent
an individual context.}
}
\value{
Invisibly returns \code{NULL}
}
\description{
This function plots the low resolution profiles for all bisulfite sequencing
data.
}
\examples{


# load the methylation data
data(methylationDataList)

#plot the low resolution profile at 10 Kb resolution
par(mar=c(4, 4, 3, 1)+0.1)
plotMethylationProfileFromData(methylationDataList[["WT"]],
                               methylationDataList[["met1-3"]],
                               conditionsNames=c("WT", "met1-3"),
                               windowSize = 20000, autoscale = TRUE,
                               context = c("CHG"))

\dontrun{
#plot the low resolution profile at 5 Kb resolution
par(mar=c(4, 4, 3, 1)+0.1)
plotMethylationProfileFromData(methylationDataList[["WT"]],
                               methylationDataList[["met1-3"]],
                               conditionsNames=c("WT", "met1-3"),
                               windowSize = 5000, autoscale = TRUE,
                               context = c("CG", "CHG", "CHH"),
                               labels = LETTERS)
}

}
\seealso{
\code{\link{plotMethylationProfile}},
\code{\link{computeMethylationProfile}} and \code{\link{methylationDataList}}
}
\author{
Nicolae Radu Zabet
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/statisticalTests.R
\name{.computeAdjuestedPValues}
\alias{.computeAdjuestedPValues}
\title{Compute adjusted p-values}
\usage{
.computeAdjuestedPValues(test, m1, n1, m2, n2, alternative = "two.sided")
}
\arguments{
\item{test}{a \code{character} indicating the statistical test used.
(\code{"fisher"} for Fisher's exact test or \code{"score"} for Score test).}

\item{m1}{the number of methylated reads in condition 1.}

\item{n1}{the total number of methylated reads in condition 1.}

\item{m2}{the number of methylated reads in condition 2.}

\item{n2}{the total number of methylated reads in condition 2.}

\item{alternative}{indicates the alternative hypothesis and must be one of
\code{"two.sided"}, \code{"greater"} or \code{"less"}. You can specify just
the initial letter. Only used in the 2 by 2 case. This is used only for
Fisher's test}
}
\value{
The adjusted p-values of the statistical test.
}
\description{
This function computes the adjusted p-values (using Benjamini & Hochberg
method)
}
\author{
Radu Zabet
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/coverage.R
\name{plotMethylationDataCoverage}
\alias{plotMethylationDataCoverage}
\title{Plot methylation data coverage}
\usage{
plotMethylationDataCoverage(
  methylationData1,
  methylationData2 = NULL,
  breaks,
  regions = NULL,
  conditionsNames = NULL,
  context = "CG",
  proportion = TRUE,
  labels = NULL,
  col = NULL,
  pch = c(1, 0, 16, 2, 15, 17),
  lty = c(4, 1, 3, 2, 6, 5),
  contextPerRow = FALSE
)
}
\arguments{
\item{methylationData1}{the methylation data in condition 1
(see \code{\link{methylationDataList}}).}

\item{methylationData2}{the methylation data in condition 2
(see \code{\link{methylationDataList}}). This is optional.}

\item{breaks}{a \code{numeric} vector specifing the different values for the
thresholds when computing the coverage.}

\item{regions}{a \code{\link{GRanges}} object with the regions where to
compute the coverage. If \code{NULL}, the coverage is computed genome-wide.}

\item{conditionsNames}{a vector of character with the names of the conditions
for \code{methylationData1} and \code{methylationData2}.}

\item{context}{the context in which the DMRs are computed (\code{"CG"},
\code{"CHG"} or \code{"CHH"}).}

\item{proportion}{a \code{logical} value indicating whether proportion or
counts will be plotted.}

\item{labels}{a \code{vector} of \code{character} used to add a subfigure
character to the plot. If \code{NULL} nothing is added.}

\item{col}{a \code{character} vector with the colors. It needs to contain a
minimum of 2 colors per condition. If not or if \code{NULL}, the defalut
colors will be used.}

\item{pch}{the R symbols used to plot the data. It needs to contain a minimum
of 2 symbols per condition. If not or if \code{NULL}, the defalut symbols
will be used.}

\item{lty}{the line types used to plot the data. It needs to contain a
minimum of 2 line types per condition. If not or if \code{NULL}, the defalut
line types will be used.}

\item{contextPerRow}{a \code{logical} value indicating if the each row
represents an individual context. If \code{FALSE}, each column will represent
an individual context.}
}
\value{
Invisibly returns \code{NULL}
}
\description{
This function plots the coverage for the bisulfite sequencing data.
}
\details{
This function plots the proportion of cytosines in a specific
context that have at least a certain number of reads (x-axis)
}
\examples{

# load the methylation data
data(methylationDataList)

# plot the coverage in CG context
par(mar=c(4, 4, 3, 1)+0.1)
plotMethylationDataCoverage(methylationDataList[["WT"]],
                           methylationDataList[["met1-3"]],
                           breaks = c(1,5,10,15), regions = NULL,
                           conditionsNames = c("WT","met1-3"),
                           context = c("CG"), proportion = TRUE,
                           labels = LETTERS, col = NULL,
                           pch = c(1,0,16,2,15,17), lty = c(4,1,3,2,6,5),
                           contextPerRow = FALSE)

\dontrun{
# plot the coverage in all three contexts
plotMethylationDataCoverage(methylationDataList[["WT"]],
                           methylationDataList[["met1-3"]],
                           breaks = 1:15, regions = NULL,
                           conditionsNames = c("WT","met1-3"),
                           context = c("CG", "CHG", "CHH"),
                           proportion = TRUE, labels = LETTERS, col = NULL,
                           pch = c(1,0,16,2,15,17), lty = c(4,1,3,2,6,5),
                           contextPerRow = FALSE)
}

}
\seealso{
\code{\link{computeMethylationDataCoverage}},
\code{\link{methylationDataList}}
}
\author{
Nicolae Radu Zabet
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Replicates.R
\name{.validateConditionReplicates}
\alias{.validateConditionReplicates}
\title{Validate the condition of the experiment}
\usage{
.validateConditionReplicates(condition, methylationData)
}
\arguments{
\item{condition}{The vector containing the two conditions for the experiment.}
}
\description{
Checks whether the passed parameter is a valid condition for the data
}
\author{
Alessandro Pio Greco and Nicolae Radu Zabet
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/computePMDs.R
\name{analyseReadsInsideRegionsForConditionPMD}
\alias{analyseReadsInsideRegionsForConditionPMD}
\title{Analyse reads inside regions for condition}
\usage{
analyseReadsInsideRegionsForConditionPMD(
  regions,
  methylationData,
  context,
  label = "",
  parallel = FALSE,
  BPPARAM = NULL,
  cores = NULL
)
}
\arguments{
\item{regions}{a \code{\link[GenomicRanges]{GRanges}} object with a list of regions on the 
genome; e.g. could be a list of PMDs}

\item{methylationData}{the methylation data in one condition
(see \code{\link{ontSampleGRangesList}}).}

\item{context}{the context in which to extract the reads (\code{"CG"}, 
\code{"CHG"} or \code{"CHH"}).}

\item{label}{a string to be added to the columns to identify the condition}

\item{parallel}{Logical; run in parallel if \code{TRUE}.}

\item{BPPARAM}{A \code{BiocParallelParam} object controlling parallel execution.
This value will automatically set when parallel is \code{TRUE}, also able to set as manually.}

\item{cores}{Integer number of workers (must not exceed BPPARAM$workers).
This value will automatically set as the maximum number of system workers,
also able to set as manually.}
}
\value{
a \code{\link[GenomicRanges]{GRanges}} object with additional four metadata columns
\describe{
 \item{sumReadsM}{the number of methylated reads}
 \item{sumReadsN}{the total number of reads} 
 \item{proportion}{the proportion methylated reads} 
 \item{cytosinesCount}{the number of cytosines in the regions} 
}
}
\description{
This function extracts from the methylation data the total number of reads, 
the number of methylated reads and the number of cytosines in the specific 
context from a region (e.g. PMDs)
}
\examples{

# load the ONT methylation data
data(ontSampleGRangesList)
 
#load the PMDs in CG context. These PMDs were computed with minGap = 200.
data(PMDsNoiseFilterCG)

#retrive the number of reads in CG context in GM18501
PMDsNoiseFilterCGreadsCG <- analyseReadsInsideRegionsForConditionPMD(
                             PMDsNoiseFilterCG[1:10], 
                             ontSampleGRangesList[["GM18501"]], context = "CG", 
                             label = "GM18501")


}
\seealso{
\code{\link{filterPMDs}}, \code{\link{computePMDs}}, 
\code{\link{PMDsNoiseFilterCG}}, and \code{\link{mergePMDsIteratively}}
}
\author{
Nicolae Radu Zabet and Young Jun Kim
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Replicates.R
\name{.computeProportionsInDMRs}
\alias{.computeProportionsInDMRs}
\title{This function recalculate proportions between methylated and total reads in a new region.}
\usage{
.computeProportionsInDMRs(DMRs, methylationData, col_indexes)
}
\description{
This function recalculate proportions between methylated and total reads in a new region.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/computeDMRs.R
\name{computeDMRs}
\alias{computeDMRs}
\title{Compute DMRs}
\usage{
computeDMRs(
  methylationData1,
  methylationData2,
  regions = NULL,
  context = "CG",
  method = "noise_filter",
  windowSize = 100,
  kernelFunction = "triangular",
  lambda = 0.5,
  binSize = 100,
  test = "fisher",
  pValueThreshold = 0.01,
  minCytosinesCount = 4,
  minProportionDifference = 0.4,
  minGap = 200,
  minSize = 50,
  minReadsPerCytosine = 4,
  parallel = FALSE,
  BPPARAM = NULL,
  cores = NULL
)
}
\arguments{
\item{methylationData1}{the methylation data in condition 1
(see \code{\link{methylationDataList}}).}

\item{methylationData2}{the methylation data in condition 2
(see \code{\link{methylationDataList}}).}

\item{regions}{a \code{\link{GRanges}} object with the regions where to 
compute the DMRs. If \code{NULL}, the DMRs are computed genome-wide.}

\item{context}{the context in which the DMRs are computed (\code{"CG"}, 
\code{"CHG"} or \code{"CHH"}).}

\item{method}{the method used to compute the DMRs (\code{"noise_filter"}, 
\code{"neighbourhood"} or \code{"bins"}). The \code{"noise_filter"} method 
uses a triangular kernel to smooth the number of reads and then performs a 
statistical test to determine which regions dispay different levels of 
methylation in the two conditions. The \code{"neighbourhood"} method 
computates differentially methylated cytosines. Finally, the \code{"bins"} 
method partiones the genome into equal sized tilling bins and performs the 
statistical test between the two conditions in each bin. For all three 
methods, the cytosines or bins are then merged into DMRs without affecting 
the inital parameters used when calling the differentiall methylated 
cytosines/bins (p-value, difference in methylation levels, minimum number of 
reads per cytosine).}

\item{windowSize}{the size of the triangle base measured in nucleotides. 
This parameter is required only if the selected method is 
\code{"noise_filter"}.}

\item{kernelFunction}{a \code{character} indicating which kernel function to 
be used. Can be one of \code{"uniform"}, \code{"triangular"}, 
\code{"gaussian"} or \code{"epanechnicov"}. This is required only if the 
selected method is \code{"noise_filter"}.}

\item{lambda}{numeric value required for the Gaussian filter 
(\code{K(x) = exp(-lambda*x^2)}). This is required only if the selected 
method is \code{"noise_filter"} and the selected kernel function is 
\code{"gaussian"}.}

\item{binSize}{the size of the tiling bins in nucleotides. This parameter is 
required only if the selected method is \code{"bins"}.}

\item{test}{the statistical test used to call DMRs (\code{"fisher"} for 
Fisher's exact test or \code{"score"} for Score test).}

\item{pValueThreshold}{DMRs with p-values (when performing the statistical 
test; see \code{test}) higher or equal than \code{pValueThreshold} are 
discarded. Note that we adjust the p-values using the Benjamini and 
Hochberg's method to control the false discovery rate.}

\item{minCytosinesCount}{DMRs with less cytosines in the specified context 
than \code{minCytosinesCount} will be discarded.}

\item{minProportionDifference}{DMRs where the difference in methylation 
proportion between the two conditions is lower than 
\code{minProportionDifference} are discarded.}

\item{minGap}{DMRs separated by a gap of at least \code{minGap} are not 
merged. Note that only DMRs where the change in methylation is in the same 
direction are joined.}

\item{minSize}{DMRs with a size smaller than \code{minSize} are discarded.}

\item{minReadsPerCytosine}{DMRs with the average number of reads lower than 
\code{minReadsPerCytosine} are discarded.}

\item{parallel}{Logical; run in parallel if \code{TRUE}.}

\item{BPPARAM}{A \code{BiocParallelParam} object controlling parallel execution.
This value will automatically set when parallel is \code{TRUE}, also able to set as manually.}

\item{cores}{Integer number of workers (must not exceed BPPARAM$workers).
This value will automatically set as the maximum number of system workers,
also able to set as manually.}
}
\value{
the DMRs stored as a \code{\link{GRanges}} object with the following 
metadata columns:
\describe{
 \item{direction}{a number indicating whether the region lost (-1)  or gain 
 (+1) methylation in condition 2 compared to condition 1.}
 \item{context}{the context in which the DMRs was computed (\code{"CG"}, 
 \code{"CHG"} or \code{"CHH"}).}
 \item{sumReadsM1}{the number of methylated reads in condition 1.}
 \item{sumReadsN1}{the total number of reads in condition 1.} 
 \item{proportion1}{the proportion methylated reads in condition 1.} 
 \item{sumReadsM2}{the number of methylated reads in condition 2.} 
 \item{sumReadsN2}{the total number reads in condition 2.}
 \item{proportion2}{the proportion methylated reads in condition 2.} 
 \item{cytosinesCount}{the number of cytosines in the DMR.} 
 \item{regionType}{a string indicating whether the region lost (\code{"loss"})  
 or gained (\code{"gain"}) methylation in condition 2 compared to condition 1.} 
 \item{pValue}{the p-value (adjusted to control the false discovery rate with 
 the Benjamini and Hochberg's method) of the statistical test when the DMR was 
 called.} 
}
}
\description{
This function computes the differentially methylated regions between two 
conditions.
}
\examples{

# load the methylation data
data(methylationDataList)

# the regions where to compute the DMRs
regions <- GRanges(seqnames = Rle("Chr3"), ranges = IRanges(1,1E5))

# compute the DMRs in CG context with noise_filter method
DMRsNoiseFilterCG <- computeDMRs(methylationDataList[["WT"]], 
                     methylationDataList[["met1-3"]], regions = regions, 
                     context = "CG", method = "noise_filter", 
                     windowSize = 100, kernelFunction = "triangular",  
                     test = "score", pValueThreshold = 0.01, 
                     minCytosinesCount = 4, minProportionDifference = 0.4, 
                     minGap = 200, minSize = 50, minReadsPerCytosine = 4, 
                     cores = 1)

\dontrun{
# compute the DMRs in CG context with neighbourhood method
DMRsNeighbourhoodCG <- computeDMRs(methylationDataList[["WT"]], 
                       methylationDataList[["met1-3"]], regions = regions, 
                       context = "CG", method = "neighbourhood", 
                       test = "score", pValueThreshold = 0.01, 
                       minCytosinesCount = 4, minProportionDifference = 0.4, 
                       minGap = 200, minSize = 50, minReadsPerCytosine = 4, 
                       cores = 1)

# compute the DMRs in CG context with bins method
DMRsBinsCG <- computeDMRs(methylationDataList[["WT"]], 
               methylationDataList[["met1-3"]], regions = regions, 
               context = "CG", method = "bins", binSize = 100, 
               test = "score", pValueThreshold = 0.01, minCytosinesCount = 4, 
               minProportionDifference = 0.4, minGap = 200, minSize = 50, 
               minReadsPerCytosine = 4, cores = 1)

}
}
\seealso{
\code{\link{filterDMRs}}, \code{\link{mergeDMRsIteratively}}, 
\code{\link{analyseReadsInsideRegionsForCondition}} and 
\code{\link{DMRsNoiseFilterCG}}
}
\author{
Nicolae Radu Zabet and Jonathan Michael Foonlan Tsang
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/DMRcaller.R
\docType{data}
\name{GEs_hg38}
\alias{GEs_hg38}
\title{The genetic elements data of GRCh38 Genome Reference}
\format{
A \code{GRanges} object
}
\source{
The object was created by loading the \code{gtf} file from
\url{https://hgdownload.soe.ucsc.edu/goldenPath/hg38/bigZips/genes/hg38.refGene.gtf.gz}.
and concatenated by range in 1.5 ~ 2 Mbp from Chromosome 1
}
\description{
A \code{GRanges} object containing the annotation of the Homo sapiens (hg38)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generalFunctions.R
\name{.isColor}
\alias{.isColor}
\title{Is color}
\usage{
.isColor(x, minLength = NULL)
}
\arguments{
\item{x}{the vector to be validated}

\item{minLength}{the  minimum length of the vector. If NULL the minimum
length is 1}
}
\value{
a \code{logical} value indicating whether \code{x} is a vector
containing only colors
}
\description{
This function checks whether the argument is a vector containing colours
}
\author{
Radu Zabet
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Replicates.R
\name{.computeDMRsReplicatesBins}
\alias{.computeDMRsReplicatesBins}
\title{This function computes the differentially methylated regions between replicates
using the bins method.}
\usage{
.computeDMRsReplicatesBins(
  methylationData,
  condition = condition,
  regions = NULL,
  context = "CG",
  binSize = 100,
  pseudocountM = 1,
  pseudocountN = 2,
  pValueThreshold = 0.01,
  minCytosinesCount = 4,
  minProportionDifference = 0.4,
  minGap = 200,
  minSize = 50,
  minReadsPerCytosine = 4,
  BPPARAM = BPPARAM,
  cores = cores
)
}
\description{
This function computes the differentially methylated regions between replicates
using the bins method.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/computeDMRs.R
\name{filterDMRs}
\alias{filterDMRs}
\title{Filter DMRs}
\usage{
filterDMRs(
  methylationData1,
  methylationData2,
  potentialDMRs,
  context = "CG",
  test = "fisher",
  pValueThreshold = 0.01,
  minCytosinesCount = 4,
  minProportionDifference = 0.4,
  minReadsPerCytosine = 3,
  parallel = FALSE,
  BPPARAM = NULL,
  cores = NULL
)
}
\arguments{
\item{methylationData1}{the methylation data in condition 1
(see \code{\link{methylationDataList}}).}

\item{methylationData2}{the methylation data in condition 2
(see \code{\link{methylationDataList}}).}

\item{potentialDMRs}{a \code{\link{GRanges}} object with potential DMRs 
where to compute the DMRs. This can be a a list of gene and/or transposable 
elements coordinates.}

\item{context}{the context in which the DMRs are computed (\code{"CG"}, 
\code{"CHG"} or \code{"CHH"}).}

\item{test}{the statistical test used to call DMRs (\code{"fisher"} for 
Fisher's exact test or \code{"score"} for Score test).}

\item{pValueThreshold}{DMRs with p-values (when performing the statistical 
test; see \code{test}) higher or equal than \code{pValueThreshold} are 
discarded. Note that we adjust the p-values using the Benjamini and 
Hochberg's method to control the false discovery rate.}

\item{minCytosinesCount}{DMRs with less cytosines in the specified context 
than \code{minCytosinesCount} will be discarded.}

\item{minProportionDifference}{DMRs where the difference in methylation 
proportion between the two conditions is lower than 
\code{minProportionDifference} are discarded.}

\item{minReadsPerCytosine}{DMRs with the average number of reads lower than 
\code{minReadsPerCytosine} are discarded.}

\item{parallel}{Logical; run in parallel if \code{TRUE}.}

\item{BPPARAM}{A \code{BiocParallelParam} object controlling parallel execution.
This value will automatically set when parallel is \code{TRUE}, also able to set as manually.}

\item{cores}{Integer number of workers (must not exceed BPPARAM$workers).
This value will automatically set as the maximum number of system workers,
also able to set as manually.}
}
\value{
a \code{\link{GRanges}} object with 11 metadata columns that contain 
the DMRs; see \code{\link{computeDMRs}}.
}
\description{
This function verifies whether a set of pottential DMRs (e.g. genes, 
transposons, CpG islands) are differentially methylated or not.
}
\examples{
# load the methylation data
data(methylationDataList)
# load the gene annotation data
data(GEs)

#select the genes
genes <- GEs[which(GEs$type == "gene")]

# the regions where to compute the DMRs
regions <- GRanges(seqnames = Rle("Chr3"), ranges = IRanges(1,1E5))
genes <- genes[overlapsAny(genes, regions)]

# filter genes that are differntially methylated in the two conditions
DMRsGenesCG <- filterDMRs(methylationDataList[["WT"]], 
               methylationDataList[["met1-3"]], potentialDMRs = genes, 
               context = "CG", test = "score", pValueThreshold = 0.01, 
               minCytosinesCount = 4, minProportionDifference = 0.4, 
               minReadsPerCytosine = 3, cores = 1)

}
\seealso{
\code{\link{DMRsNoiseFilterCG}}, \code{\link{computeDMRs}}, 
\code{\link{analyseReadsInsideRegionsForCondition}}  
and \code{\link{mergeDMRsIteratively}}
}
\author{
Nicolae Radu Zabet
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generalFunctions.R
\name{.chooseBPPARAM}
\alias{.chooseBPPARAM}
\title{Auto-choose BiocParallelParam Backend}
\usage{
.chooseBPPARAM(workers = NULL, progressbar = FALSE, cluster_type = "SOCK", ...)
}
\arguments{
\item{workers}{Optional integer to override detected cores.}

\item{progressbar}{Logical; show progress bar.}

\item{cluster_type}{Type passed to `SnowParam()`, defaults to "SOCK".}

\item{...}{Further args passed to `SnowParam` or `MulticoreParam`.}
}
\value{
A `SnowParam` or `MulticoreParam` configured with the desired workers.
}
\description{
1) If `NSLOTS` or `SLURM_CPUS_ON_NODE` are set, uses `SnowParam()` with that many workers.
2) On Windows, `SnowParam(detectCores())`.
3) Otherwise on Unix, `MulticoreParam(detectCores())`.
}
\author{
Radu Zabet and Young Jun Kim
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/DMRcaller.R
\docType{data}
\name{scanBamChr1Random5}
\alias{scanBamChr1Random5}
\title{The bam file from ONT nanopore .pod5 files}
\format{
A \code{.bam} object
}
\source{
The object was created by base-calling and alignment by \code{dorado, ver.0.9.6}
\url{https://github.com/nanoporetech/dorado?tab=readme-ov-file#dna-models}
to generate the \code{bam} files with \code{dna_r10.4.1_e8.2_400bps_hac@v5.2.0} as basecalling model
from GM18501 cell line \url{https://s3.amazonaws.com/1000g-ont/index.html?prefix=pod5_data/GM18501_R9/}.
To subset, call the \code{bam} file by \code{\link[Rsamtools]{scanBam}} function from
\code{\link[Rsamtools]{Rsamtools}} package and randomly select the 5 sequences.
and repackaging the subsetted \code{list} to the \code{bam} file for test running.
}
\description{
A \code{.bam} file containing the basecalling result of 5 sequences
containing with the MM, ML tag from \code{dorado}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/localProfile.R
\name{plotLocalMethylationProfile}
\alias{plotLocalMethylationProfile}
\title{Plot local methylation profile}
\usage{
plotLocalMethylationProfile(
  methylationData1,
  methylationData2,
  region,
  DMRs = NULL,
  conditionsNames = NULL,
  gff = NULL,
  windowSize = 150,
  context = "CG",
  labels = NULL,
  col = NULL,
  main = "",
  plotMeanLines = TRUE,
  plotPoints = TRUE
)
}
\arguments{
\item{methylationData1}{the methylation data in condition 1
(see \code{\link{methylationDataList}}).}

\item{methylationData2}{the methylation data in condition 2
(see \code{\link{methylationDataList}}).}

\item{region}{a \code{\link{GRanges}} object with the region where to plot
the high resolution profile.}

\item{DMRs}{a \code{\link{GRangesList}} object or a list with the list of
DMRs (see \code{\link{computeDMRs}} or \code{\link{filterDMRs}}.}

\item{conditionsNames}{the names of the two conditions. This will be used to
plot the legend.}

\item{gff}{a \code{\link{GRanges}} object with all elements usually imported
from a GFF3 file. The gff file needs to have an metafield \code{"type"}. Only
the elements of type  \code{"gene"}, \code{"exon"} and
\code{"transposable_element"} are plotted. Genes are represented as
horizontal black lines, exons as a black rectangle and transposable elements
as a grey rectangle. The elements are plotted on the corresponding strand
(\code{+} or \code{-}).}

\item{windowSize}{the size of the triangle base used to smooth the average
methylation profile.}

\item{context}{the context in which the DMRs are computed (\code{"CG"},
\code{"CHG"} or \code{"CHH"}).}

\item{labels}{a \code{vector} of \code{character} used to add a subfigure
characters to the plot. If \code{NULL} nothing is added.}

\item{col}{a \code{character} vector with the colors. It needs to contain a
minimum of \code{4 length(DMRs)} colors. If not or if \code{NULL}, the
defalut colors will be used.}

\item{main}{a \code{character} with the title of the plot}

\item{plotMeanLines}{a \code{logical} value indicating whether to plot the
mean lines or not.}

\item{plotPoints}{a \code{logical} value indicating whether to plot the
points or not.}
}
\value{
Invisibly returns \code{NULL}
}
\description{
This function plots the methylation profile at one locus for the bisulfite
sequencing data.The points on the graph represent methylation proportion of
individual cytosines, their colour which sample they belong to and the
intesity of the the colour how many reads that particular cytosine had. This
means that darker colors indicate stronger evidence that the corresponding
cytosine has the corresponding methylation proportion, while lighter colors
indicate a weaker evidence. The solid lines represent the smoothed profiles
and the intensity of the line the coverage at the corresponding position
(darker colors indicate more reads while lighter ones less reads). The boxes
on top represent the DMRs, where a filled box will represent a DMR which
gained methylation while a box with a pattern represent a DMR that lost
methylation. The DMRs need to have a metadafield \code{"regionType"} which
can be either \code{"gain"} (where there is more methylation in condition 2
compared to condition 1) or \code{"loss"} (where there is less methylation in
condition 2 compared to condition 1). In case this metadafield is missing all
DMRs are drawn using a filled box. Finally, we also allow annotation of the
DNA sequence. We represent by a black boxes all the exons, which are joined
by a horizontal black line, thus, marking the full body of the gene. With
grey boxes we mark the transposable elements. Both for genes and transposable
elements we plot them over a mid line if they are on the positive strand and
under the mid line if they are on the negative strand.
}
\examples{

# load the methylation data
data(methylationDataList)
# load the gene annotation data
data(GEs)

#select the genes
genes <- GEs[which(GEs$type == "gene")]

# the coordinates of the area to be plotted
chr3Reg <- GRanges(seqnames = Rle("Chr3"), ranges = IRanges(510000,530000))

# load the DMRs in CG context
data(DMRsNoiseFilterCG)

DMRsCGlist <- list("noise filter"=DMRsNoiseFilterCG)


# plot the CG methylation
par(mar=c(4, 4, 3, 1)+0.1)
par(mfrow=c(1,1))
plotLocalMethylationProfile(methylationDataList[["WT"]],
                           methylationDataList[["met1-3"]], chr3Reg,
                           DMRsCGlist, c("WT", "met1-3"), GEs,
                           windowSize=100, main="CG methylation")

}
\author{
Nicolae Radu Zabet
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Replicates.R
\name{.computeDMRsReplicatesNeighbourhood}
\alias{.computeDMRsReplicatesNeighbourhood}
\title{This function computes the differentially methylated regions between replicates
using the neighbourhood method.}
\usage{
.computeDMRsReplicatesNeighbourhood(
  methylationData,
  condition,
  regions = NULL,
  context = "CG",
  pseudocountM = 1,
  pseudocountN = 2,
  pValueThreshold = 0.01,
  minCytosinesCount = 4,
  minProportionDifference = 0.4,
  minGap = 200,
  minSize = 50,
  minReadsPerCytosine = 4,
  BPPARAM = BPPARAM,
  cores = cores
)
}
\description{
This function computes the differentially methylated regions between replicates
using the neighbourhood method.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/computeDMRs.R
\name{mergeDMRsIteratively}
\alias{mergeDMRsIteratively}
\title{Merge DMRs iteratively}
\usage{
mergeDMRsIteratively(
  DMRs,
  minGap,
  respectSigns = TRUE,
  methylationData1,
  methylationData2,
  context = "CG",
  minProportionDifference = 0.4,
  minReadsPerCytosine = 4,
  pValueThreshold = 0.01,
  test = "fisher",
  alternative = "two.sided",
  parallel = FALSE,
  BPPARAM = NULL,
  cores = NULL
)
}
\arguments{
\item{DMRs}{the list of DMRs as a \code{\link{GRanges}} object; 
e.g. see \code{\link{computeDMRs}}}

\item{minGap}{DMRs separated by a gap of at least \code{minGap} are not 
merged.}

\item{respectSigns}{logical value indicating whether to respect the sign when 
joining DMRs.}

\item{methylationData1}{the methylation data in condition 1
(see \code{\link{methylationDataList}}).}

\item{methylationData2}{the methylation data in condition 2
(see \code{\link{methylationDataList}}).}

\item{context}{the context in which the DMRs are computed (\code{"CG"}, 
\code{"CHG"} 
or \code{"CHH"}).}

\item{minProportionDifference}{two adjacent DMRs are merged only if the 
difference in methylation proportion of the new DMR is higher than 
\code{minProportionDifference}.}

\item{minReadsPerCytosine}{two adjacent DMRs are merged only if the number of 
reads per cytosine of the new DMR is higher than \code{minReadsPerCytosine}.}

\item{pValueThreshold}{two adjacent DMRs are merged only if the p-value of 
the new DMR (see \code{test} below) is lower than \code{pValueThreshold}. 
Note that we adjust the p-values using the Benjamini and Hochberg's method to 
control the false discovery rate.}

\item{test}{the statistical test used to call DMRs (\code{"fisher"} for 
Fisher's exact test or \code{"score"} for Score test).}

\item{alternative}{indicates the alternative hypothesis and must be one of 
\code{"two.sided"}, \code{"greater"} or \code{"less"}.}

\item{parallel}{Logical; run in parallel if \code{TRUE}.}

\item{BPPARAM}{A \code{BiocParallelParam} object controlling parallel execution.
This value will automatically set when parallel is \code{TRUE}, also able to set as manually.}

\item{cores}{Integer number of workers (must not exceed BPPARAM$workers).
This value will automatically set as the maximum number of system workers,
also able to set as manually.}
}
\value{
the reduced list of DMRs as a \code{\link{GRanges}} object; 
e.g. see \code{\link{computeDMRs}}
}
\description{
This function takes a list of DMRs and attempts to merge DMRs while keeping 
the new DMRs statistically significant.
}
\examples{
# load the methylation data
data(methylationDataList)

#load the DMRs in CG context they were computed with minGap = 200
data(DMRsNoiseFilterCG)


#merge the DMRs 
DMRsNoiseFilterCGLarger <- mergeDMRsIteratively(DMRsNoiseFilterCG[1:100], 
                           minGap = 500, respectSigns = TRUE, 
                           methylationDataList[["WT"]], 
                           methylationDataList[["met1-3"]],
                           context = "CG", minProportionDifference=0.4, 
                           minReadsPerCytosine = 1, pValueThreshold=0.01, 
                           test="score",alternative = "two.sided")


\dontrun{
#set genomic coordinates where to compute DMRs
regions <- GRanges(seqnames = Rle("Chr3"), ranges = IRanges(1,1E5))

# compute DMRs and remove gaps smaller than 200 bp
DMRsNoiseFilterCG200 <- computeDMRs(methylationDataList[["WT"]], 
                       methylationDataList[["met1-3"]], regions = regions, 
                       context = "CG", method = "noise_filter", 
                       windowSize = 100, kernelFunction = "triangular",  
                       test = "score", pValueThreshold = 0.01, 
                       minCytosinesCount = 1, minProportionDifference = 0.4, 
                       minGap = 200, minSize = 0, minReadsPerCytosine = 1, 
                       cores = 1)
                       
DMRsNoiseFilterCG0 <- computeDMRs(methylationDataList[["WT"]], 
                       methylationDataList[["met1-3"]], regions = regions, 
                       context = "CG", method = "noise_filter", 
                       windowSize = 100, kernelFunction = "triangular",  
                       test = "score", pValueThreshold = 0.01, 
                       minCytosinesCount = 1, minProportionDifference = 0.4, 
                       minGap = 0, minSize = 0, minReadsPerCytosine = 1, 
                       cores = 1)
DMRsNoiseFilterCG0Merged200 <- mergeDMRsIteratively(DMRsNoiseFilterCG0, 
                             minGap = 200, respectSigns = TRUE, 
                             methylationDataList[["WT"]], 
                             methylationDataList[["met1-3"]],
                             context = "CG", minProportionDifference=0.4, 
                             minReadsPerCytosine = 1, pValueThreshold=0.01, 
                             test="score",alternative = "two.sided")                      

#check that all newley computed DMRs are identical
print(all(DMRsNoiseFilterCG200 == DMRsNoiseFilterCG0Merged200))

}

}
\seealso{
\code{\link{filterDMRs}}, \code{\link{computeDMRs}}, 
\code{\link{analyseReadsInsideRegionsForCondition}} and 
\code{\link{DMRsNoiseFilterCG}}
}
\author{
Nicolae Radu Zabet
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/computeVMDs.R
\name{.analyseReadsInsideRegionsVMDs}
\alias{.analyseReadsInsideRegionsVMDs}
\title{Analyse reads inside regions with computing mean and standard deviation per-read proportions}
\usage{
.analyseReadsInsideRegionsVMDs(methylationData, regions)
}
\arguments{
\item{methylationData}{A \code{GRanges} of methylation calls. Must include metadata:
\describe{
  \item{readsM, readsN}{methylated and total read counts in condition}
  \item{ONT_Cm, ONT_C}{per-read methylation and coverage indices (character vectors)}
}}

\item{regions}{A \code{GRanges} of genomic intervals to summarise.}
}
\value{
A \code{GRanges} with the same ranges as \code{regions}, plus these metadata:
\describe{
  \item{sumReadsM}{total methylated reads in condition}
  \item{sumReadsN}{total reads in condition}
  \item{proportion}{methylation proportion (sumReadsM/sumReadsN)}
  \item{cytosinesCount}{number of cytosines observed in each region}
  \item{mean}{mean value comparing per‐read proportions}
  \item{sd}{standard deviation comparing per‐read proportions}
  \item{w_mean}{weighted mean value comparing per‐read proportions}
  \item{w_sd}{weighted standard deviation comparing per‐read proportions}
}
}
\description{
Performs the analysis in all regions in a \code{\link{GRanges}} object,
computing per-region mean and standard deviation of proportions.
}
\author{
Radu Zabet and Young Jun Kim
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generalFunctions.R
\name{.validateMethylationData}
\alias{.validateMethylationData}
\title{Validate methylation data}
\usage{
.validateMethylationData(
  methylationData,
  variableName = "methylationData",
  manageDuplicates = "mean"
)
}
\arguments{
\item{methylationData}{the methylation data stored as a \code{\link{GRanges}}
object containing all the replicates.}
}
\description{
Checks whether the passed parameter has the correct format for methylation data
}
\author{
Alessandro Pio Greco and Nicolae Radu Zabet
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generalFunctions.R
\name{.validateMethylationProfile}
\alias{.validateMethylationProfile}
\title{Validate methylation profile}
\usage{
.validateMethylationProfile(methylationProfile)
}
\arguments{
\item{methylationProfile}{a \code{GRangesList} object
(see \code{\link{computeMethylationProfile}}).}
}
\description{
Checks whether the passed parameter is a \code{GRangesList} object that
represents the methylation profile
}
\author{
Radu Zabet
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/computePMDs.R
\name{filterPMDs}
\alias{filterPMDs}
\title{Filter PMDs}
\usage{
filterPMDs(
  methylationData,
  potentialPMDs,
  context = "CG",
  minCytosinesCount = 4,
  minMethylation = 0.4,
  maxMethylation = 0.6,
  minReadsPerCytosine = 3,
  parallel = FALSE,
  BPPARAM = NULL,
  cores = NULL
)
}
\arguments{
\item{methylationData}{the methylation data in condition
(see \code{\link{ontSampleGRangesList}}).}

\item{potentialPMDs}{a \code{\link[GenomicRanges]{GRanges}} object with potential PMDs 
where to compute the PMDs. This can be a a list of gene and/or transposable 
elements coordinates.}

\item{context}{the context in which the PMDs are computed (\code{"CG"}, 
\code{"CHG"} or \code{"CHH"}).}

\item{minCytosinesCount}{PMDs with less cytosines in the specified context 
than \code{minCytosinesCount} will be discarded.}

\item{minMethylation}{Numeric [0,1]; minimum mean methylation proportion.}

\item{maxMethylation}{Numeric [0,1]; maximum mean methylation proportion.}

\item{minReadsPerCytosine}{PMDs with the average number of reads lower than 
\code{minReadsPerCytosine} are discarded.}

\item{parallel}{Logical; run in parallel if \code{TRUE}.}

\item{BPPARAM}{A \code{BiocParallelParam} object controlling parallel execution.
This value will automatically set when parallel is \code{TRUE}, also able to set as manually.}

\item{cores}{Integer number of workers (must not exceed BPPARAM$workers).
This value will automatically set as the maximum number of system workers,
also able to set as manually.}
}
\value{
a \code{\link[GenomicRanges]{GRanges}} object with 5 metadata columns that contain 
the PMDs; see \code{\link{computePMDs}}.
}
\description{
This function verifies whether a set of potential PMDs (e.g. genes, 
transposons, CpG islands) are partially methylated or not.
}
\examples{
# load the ONT methylation data 
data(ontSampleGRangesList)
# load the gene annotation data
data(GEs_hg38)

# select the transcript
transcript <- GEs_hg38[which(GEs_hg38$type == "transcript")]

# the regions where to compute the PMDs
regions <- GRanges(seqnames = Rle("chr1"), ranges = IRanges(1E6+5E5,2E6))
transcript <- transcript[overlapsAny(transcript, regions)]

# filter genes that are partially methylated in the two conditions
PMDsGenesCG <- filterPMDs(ontSampleGRangesList[["GM18501"]], 
               potentialPMDs = transcript, 
               context = "CG", minMethylation = 0.4, maxMethylation = 0.6,
               minCytosinesCount = 4, minReadsPerCytosine = 3, cores = 1)

}
\seealso{
\code{\link{PMDsNoiseFilterCG}}, \code{\link{computePMDs}}, 
\code{\link{analyseReadsInsideRegionsForCondition}}  
and \code{\link{mergePMDsIteratively}}
}
\author{
Nicolae Radu Zabet and Young Jun Kim
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/filterVMRsONT.R
\name{.analyseReadsInsideRegionsVMR}
\alias{.analyseReadsInsideRegionsVMR}
\title{Analyse reads inside regions (with VMR testing)}
\usage{
.analyseReadsInsideRegionsVMR(methylationData, regions)
}
\arguments{
\item{methylationData}{A \code{GRanges} of methylation calls. Must include metadata:
\describe{
  \item{readsM1, readsN1}{methylated and total read counts in condition 1}
  \item{readsM2, readsN2}{methylated and total read counts in condition 2}
  \item{ONT_Cm, ONT_C}{per-read methylation and coverage indices (character vectors)}
}}

\item{regions}{A \code{GRanges} of genomic intervals to summarise.}
}
\value{
A \code{GRanges} with the same ranges as \code{regions}, plus these metadata:
\describe{
  \item{sumReadsM1}{total methylated reads in condition 1}
  \item{sumReadsN1}{total reads in condition 1}
  \item{proportion1}{methylation proportion (sumReadsM1/sumReadsN1)}
  \item{variance1}{variance of per-read methylation proportions in condition 1}
  \item{sumReadsM2}{total methylated reads in condition 2}
  \item{sumReadsN2}{total reads in condition 2}
  \item{proportion2}{methylation proportion (sumReadsM2/sumReadsN2)}
  \item{variance2}{variance of per-read methylation proportions in condition 2}
  \item{cytosinesCount}{number of cytosines observed in each region}
  \item{wilcox_pvalue}{p-value from Wilcoxon rank-sum test comparing per-read proportions}
  \item{f_pvalue}{p-value from F-test comparing variances of per-read proportions}
  \item{var_ratio}{Ratio of variances (variance1 / variance2)}
  \item{wilcox_result}{Full \code{htest} object returned by \code{wilcox.test}}
  \item{F_test_result}{Full \code{htest} object returned by \code{var.test}}
}
}
\description{
Performs the analysis in all regions in a \code{\link[GenomicRanges]{GRanges}} object,
computing per-region summary statistics and statistical tests.
}
\author{
Radu Zabet and Young Jun Kim
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/profile.R
\name{plotMethylationProfile}
\alias{plotMethylationProfile}
\title{Plot Methylation Profile}
\usage{
plotMethylationProfile(
  methylationProfiles,
  autoscale = FALSE,
  labels = NULL,
  title = "",
  col = NULL,
  pch = c(1, 0, 16, 2, 15, 17),
  lty = c(4, 1, 3, 2, 6, 5)
)
}
\arguments{
\item{methylationProfiles}{a \code{GRangesList} object. Each
\code{\link{GRanges}} object in the list is generated by calling  the
function \code{\link{computeMethylationProfile}}.}

\item{autoscale}{a \code{logical} value indicating whether the values are
autoscalled for each context or not.}

\item{labels}{a \code{vector} of \code{character} used to add a subfigure
characters to the plot. If \code{NULL} nothing is added.}

\item{title}{the plot title.}

\item{col}{a \code{character} vector with the colours. It needs to contain a
minimum of 2 colours per context. If not or if NULL, the defalut colours will
be used.}

\item{pch}{the R symbols used to plot the data.}

\item{lty}{the line types used to plot the data.}
}
\value{
Invisibly returns \code{NULL}
}
\description{
This function plots the low resolution profiles for the bisulfite sequencing
data.
}
\examples{


# load the methylation data
data(methylationDataList)

# the region where to compute the profile
region <- GRanges(seqnames = Rle("Chr3"), ranges = IRanges(1,1E6))

# compute low resolution profile in 20 Kb windows
lowResProfileWTCG <- computeMethylationProfile(methylationDataList[["WT"]],
                     region, windowSize = 20000, context = "CG")

lowResProfilsCG <- GRangesList("WT" = lowResProfileWTCG)

#plot the low resolution profile
par(mar=c(4, 4, 3, 1)+0.1)
par(mfrow=c(1,1))
plotMethylationProfile(lowResProfilsCG, autoscale = FALSE,
                       title="CG methylation on Chromosome 3",
                       col=c("#D55E00","#E69F00"),  pch = c(1,0),
                       lty = c(4,1))

\dontrun{
# compute low resolution profile in 10 Kb windows in CG context
lowResProfileWTCG <- computeMethylationProfile(methylationDataList[["WT"]],
                     region, windowSize = 10000, context = "CG")

lowResProfileMet13CG <- computeMethylationProfile(
                     methylationDataList[["met1-3"]], region,
                     windowSize = 10000, context = "CG")

lowResProfileCG <- GRangesList("WT" = lowResProfileWTCG,
                   "met1-3" = lowResProfileMet13CG)

# compute low resolution profile in 10 Kb windows in CHG context
lowResProfileWTCHG <- computeMethylationProfile(methylationDataList[["WT"]],
                     region, windowSize = 10000, context = "CHG")

lowResProfileMet13CHG <- computeMethylationProfile(
                     methylationDataList[["met1-3"]], region,
                     windowSize = 10000, context = "CHG")

lowResProfileCHG <- GRangesList("WT" = lowResProfileWTCHG,
                   "met1-3" = lowResProfileMet13CHG)

# plot the low resolution profile
par(mar=c(4, 4, 3, 1)+0.1)
par(mfrow=c(2,1))
plotMethylationProfile(lowResProfileCG, autoscale = FALSE,
                       labels = LETTERS[1],
                       title="CG methylation on Chromosome 3",
                       col=c("#D55E00","#E69F00"),  pch = c(1,0),
                       lty = c(4,1))
plotMethylationProfile(lowResProfileCHG, autoscale = FALSE,
                       labels = LETTERS[2],
                       title="CHG methylation on Chromosome 3",
                       col=c("#0072B2", "#56B4E9"),  pch = c(16,2),
                       lty = c(3,2))
}

}
\seealso{
\code{\link{plotMethylationProfileFromData}},
\code{\link{computeMethylationProfile}} and \code{\link{methylationDataList}}
}
\author{
Nicolae Radu Zabet
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generalFunctions.R
\name{.isInteger}
\alias{.isInteger}
\title{.is.integer}
\usage{
.isInteger(x, positive = FALSE)
}
\arguments{
\item{x}{the variable}

\item{positive}{a logical variable indicating whether to check if the number
is a positive integer}
}
\value{
a logical number whether the variable x is an integer or not
}
\description{
This function checks is a variable is an integer number
}
\author{
Radu Zabet
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/regions.R
\name{.smartMergeDMRs}
\alias{.smartMergeDMRs}
\title{This takes a list of DMRs and attempts to merge DMRs while keeping the new 
DMRs statistically significant}
\usage{
.smartMergeDMRs(
  DMRs,
  minGap,
  respectSigns = TRUE,
  methylationData,
  minProportionDifference = 0.4,
  minReadsPerCytosine = 4,
  pValueThreshold = 0.01,
  test = "fisher",
  alternative = "two.sided",
  cores = 1,
  BPPARAM = BPPARAM
)
}
\description{
This takes a list of DMRs and attempts to merge DMRs while keeping the new 
DMRs statistically significant
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/DMRcaller.R
\docType{data}
\name{DMRsNoiseFilterCG}
\alias{DMRsNoiseFilterCG}
\title{The DMRs between WT and met1-3 in CG context}
\format{
The \code{\link[GenomicRanges]{GRanges}} element contain 11 metadata columns;
see \code{\link{computeDMRs}}
}
\description{
A \code{GRangesList} object containing the DMRs between  Wild Type (WT) and
met1-3 mutant (met1-3) in Arabidopsis thaliana
(see \code{\link{methylationDataList}}). The DMRs were computed on the first
1 Mbp from Chromosome 3 with noise filter method using a triangular kernel
and a windowSize of 100 bp
}
\seealso{
\code{\link{filterDMRs}}, \code{\link{computeDMRs}},
\code{\link{analyseReadsInsideRegionsForCondition}}
and \code{\link{mergeDMRsIteratively}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/regions.R
\name{.smartMergePMDs}
\alias{.smartMergePMDs}
\title{This takes a list of PMDs and attempts to merge PMDs while keeping the new 
PMDs statistically significant}
\usage{
.smartMergePMDs(
  PMDs,
  minGap,
  respectSigns = TRUE,
  methylationData,
  minReadsPerCytosine = 4,
  minMethylation = 0.4,
  maxMethylation = 0.6,
  cores = 1,
  BPPARAM = BPPARAM
)
}
\description{
This takes a list of PMDs and attempts to merge PMDs while keeping the new 
PMDs statistically significant
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spatialCorrelation.R
\name{plotMethylationDataSpatialCorrelation}
\alias{plotMethylationDataSpatialCorrelation}
\title{Plot methylation data spatial correlation}
\usage{
plotMethylationDataSpatialCorrelation(
  methylationData1,
  methylationData2 = NULL,
  distances,
  regions = NULL,
  conditionsNames = NULL,
  context = "CG",
  labels = NULL,
  col = NULL,
  pch = c(1, 0, 16, 2, 15, 17),
  lty = c(4, 1, 3, 2, 6, 5),
  contextPerRow = FALSE,
  log = ""
)
}
\arguments{
\item{methylationData1}{the methylation data in condition 1
(see \code{\link{methylationDataList}}).}

\item{methylationData2}{the methylation data in condition 2
(see \code{\link{methylationDataList}}). This is optional.}

\item{distances}{a \code{numeric} vector specifing the different values for
the distances when computing the correlation.}

\item{regions}{a \code{\link[GenomicRanges]{GRanges}} object with the regions where to
compute the correlation. If \code{NULL}, the coverage is computed genome-wide.}

\item{conditionsNames}{a vector of character with the names of the conditions
for \code{methylationData1} and \code{methylationData2}.}

\item{context}{the context in which the DMRs are computed (\code{"CG"},
\code{"CHG"} or \code{"CHH"}).}

\item{labels}{a \code{vector} of \code{character} used to add a subfigure
character to the plot. If \code{NULL} nothing is added.}

\item{col}{a \code{character} vector with the colors. It needs to contain a
minimum of 2 colors per condition. If not or if \code{NULL}, the defalut
colors will be used.}

\item{pch}{the R symbols used to plot the data. It needs to contain a minimum
of 2 symbols per condition. If not or if \code{NULL}, the defalut symbols
will be used.}

\item{lty}{the line types used to plot the data. It needs to contain a
minimum of 2 line types per condition. If not or if \code{NULL}, the defalut
line types will be used.}

\item{contextPerRow}{a \code{logical} value indicating if the each row
represents an individual context. If \code{FALSE}, each column will represent
an individual context.}

\item{log}{a \code{character} indicating if any of the axes will be displayed
on log scale. This argument will be passed to \code{\link{plot}} function.}
}
\value{
Invisibly returns \code{NULL}
}
\description{
This function plots the correlation of methylation levels for Cytosines
located at a certain distance apart.
}
\details{
This function plots the proportion of cytosines in a specific
context that have at least a certain number of reads (x-axis)
}
\examples{


# load the methylation data
data(methylationDataList)

# plot the spatial correlation in CG context
par(mar=c(4, 4, 3, 1)+0.1)
plotMethylationDataSpatialCorrelation(methylationDataList[["WT"]],
                           methylationDataList[["met1-3"]],
                           distances = c(1,5,10,15), regions = NULL,
                           conditionsNames = c("WT","met1-3"),
                           context = c("CG"),
                           labels = LETTERS, col = NULL,
                           pch = c(1,0,16,2,15,17), lty = c(4,1,3,2,6,5),
                           contextPerRow = FALSE)
\dontrun{
# plot the spatial correlation in all three contexts
plotMethylationDataSpatialCorrelation(methylationDataList[["WT"]],
                           methylationDataList[["met1-3"]],
                           distances = c(1,5,10,15,20,50,100,150,200,500,1000),
                           regions = NULL, conditionsNames = c("WT","met1-3"),
                           context = c("CG", "CHG", "CHH"),
                           labels = LETTERS, col = NULL,
                           pch = c(1,0,16,2,15,17), lty = c(4,1,3,2,6,5),
                           contextPerRow = FALSE, log="x")
}

}
\seealso{
\code{\link{computeMethylationDataSpatialCorrelation}},
\code{\link{methylationDataList}}
}
\author{
Nicolae Radu Zabet
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generalFunctions.R
\name{getWholeChromosomes}
\alias{getWholeChromosomes}
\title{Get whole chromosomes from methylation data}
\usage{
getWholeChromosomes(methylationData)
}
\arguments{
\item{methylationData}{the methylation data stored as a \code{\link{GRanges}}
object with four metadata columns (see \code{\link{methylationDataList}}).}
}
\value{
a \code{\link{GRanges}} object will all chromosomes.
}
\description{
Returns a \code{\link{GRanges}} object spanning from the first cytocine until
the last one on each chromosome
}
\examples{
# load the methylation data
data(methylationDataList)

# get all chromosomes
chromosomes <- getWholeChromosomes(methylationDataList[["WT"]])

}
\author{
Nicolae Radu Zabet
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/regions.R
\name{.countTotalNumberOfCytosines}
\alias{.countTotalNumberOfCytosines}
\title{Count total number of Cytosines}
\usage{
.countTotalNumberOfCytosines(methylationData, regions)
}
\arguments{
\item{methylationData}{a \code{\link[GenomicRanges]{GRanges}} object with five metadata 
columns: see \code{\link{methylationDataList}}}

\item{regions}{a \code{\link[GenomicRanges]{GRanges}} object with the identified regions}
}
\value{
a the total number of Cytosines
}
\description{
Counts the number of cytosines in all the regions
}
\author{
Radu Zabet
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/statisticalTests.R
\name{.fisherTest}
\alias{.fisherTest}
\title{Fisher test}
\usage{
.fisherTest(m1, n1, m2, n2, alternative = c("two.sided", "less", "greater"))
}
\arguments{
\item{m1}{the number of methylated reads in condition 1}

\item{n1}{the total number of methylated reads in condition 1}

\item{m2}{the number of methylated reads in condition 2}

\item{n2}{the total number of methylated reads in condition 2}

\item{alternative}{indicates the alternative hypothesis and must be one of
\code{"two.sided"}, \code{"greater"} or \code{"less"}. You can specify just
the initial letter. Only used in the 2 by 2 case.}
}
\value{
The p-values of Fisher's exact test.
}
\description{
This function computes the p-values of the Fisher's exact test.
}
\author{
Radu Zabet
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/loadData.R
\name{poolTwoMethylationDatasets}
\alias{poolTwoMethylationDatasets}
\title{Pool two methylation datasets}
\usage{
poolTwoMethylationDatasets(
  methylationData1,
  methylationData2,
  sample1_name = NULL,
  sample2_name = NULL
)
}
\arguments{
\item{methylationData1}{a \code{\link{GRanges}} object with the methylation
data (see \code{\link{methylationDataList}}).}

\item{methylationData2}{a \code{\link{GRanges}} object with the methylation
data (see \code{\link{methylationDataList}}).}

\item{sample1_name}{the label used for sample 1.}

\item{sample2_name}{the label used for sample 2.}
}
\value{
the methylation data stored as a \code{\link{GRanges}}
object with four metadata columns (see \code{\link{methylationDataList}}).
If the Granges are from ONT datasets, its have six metedata columns include as
ONT_Cm and ONT_C (see \code{\link{readONTbam}}).
}
\description{
This function pools together two methylation datasets.
}
\examples{
# load methylation data object
data(methylationDataList)

# save the two datasets together
pooledMethylationData <- poolTwoMethylationDatasets(methylationDataList[[1]],
                         methylationDataList[[2]])

}
\author{
Nicolae Radu Zabet updated by Young Jun Kim
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/computeVMDs.R
\name{.computeVMDsBins}
\alias{.computeVMDsBins}
\title{This function computes the variance methylated regions 
using the bins method.}
\usage{
.computeVMDsBins(
  methylationData,
  regions = NULL,
  context = "CG",
  binSize = 100,
  minCytosinesCount = 4,
  minGap = 200,
  minSize = 50,
  minReadsPerCytosine = 4,
  sdCutoffMethod = "per.high",
  percentage = 0.05,
  cores = 1,
  parallel = parallel,
  BPPARAM = BPPARAM
)
}
\description{
This function computes the variance methylated regions 
using the bins method.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spatialCorrelation.R
\name{computeMethylationDataSpatialCorrelation}
\alias{computeMethylationDataSpatialCorrelation}
\title{Compute methylation data spatial correlation}
\usage{
computeMethylationDataSpatialCorrelation(
  methylationData,
  regions = NULL,
  context = "CG",
  distances = NULL
)
}
\arguments{
\item{methylationData}{the methylation data stored as a \code{\link[GenomicRanges]{GRanges}}
object with four metadata columns (see \code{\link{methylationDataList}}).}

\item{regions}{a \code{\link[GenomicRanges]{GRanges}} object with the regions where to
compute the correlation. If \code{NULL}, the correlation is computed
genome-wide.}

\item{context}{the context in which the correlation is computed (\code{"CG"},
\code{"CHG"} or \code{"CHH"}).}

\item{distances}{a \code{numeric} vector specifing the different values for
the distances when computing the correlation.}
}
\value{
a \code{vector} with the correlation of the methylation levels for
Cytosines located at distances specified in the \code{distances}
\code{vector}.
}
\description{
This function computes the correlation of the methylation levels as a
function of the distances between the Cytosines. The function returns a
\code{vector} with the correlation of methylation levels at distance equal to
a \code{vector} of specified thresholds.
}
\examples{


# load the methylation data
data(methylationDataList)

# compute spatial correlation in CG context
distances <- c(1,5,10,15)
correlation_CG_wt <- computeMethylationDataSpatialCorrelation(methylationDataList[["WT"]],
                 context="CG", distances=distances)



}
\seealso{
\code{\link{plotMethylationDataSpatialCorrelation}},
\code{\link{methylationDataList}}
}
\author{
Nicolae Radu Zabet
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/DMRcaller.R
\docType{data}
\name{methylationDataList}
\alias{methylationDataList}
\title{The methylation data list}
\format{
The \code{GRanges} elements contain four metadata columns
\describe{
 \item{context}{the context in which the DMRs are computed (\code{"CG"},
               \code{"CHG"} or \code{"CHH"}).}
 \item{readsM}{the number of methylated reads.}
 \item{readsN}{the total number of reads.}
 \item{trinucleotide_context}{the specific context of the cytosine (H is
                             replaced by the actual nucleotide).}
}
}
\source{
Each element was created by by calling \code{\link{readBismark}}
function on the CX report files generated by Bismark
\url{http://www.bioinformatics.babraham.ac.uk/projects/bismark/}
for \url{http://www.ncbi.nlm.nih.gov/geo/query/acc.cgi?acc=GSM980986} dataset
in the case of Wild Type (WT) and
\url{http://www.ncbi.nlm.nih.gov/geo/query/acc.cgi?acc=GSM981032}
dataset in the case of met1-3 mutant (met1-3).
}
\description{
A \code{GRangesList} object containing the methylation data at each cytosine
location in the genome in  Wild Type (WT) and  met1-3 mutant (met1-3) in
Arabidopsis thaliana. The data only contains the first 1 Mbp from Chromosome 3.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/computeCoMethylatedPostions.R
\name{computeCoMethylatedPositions}
\alias{computeCoMethylatedPositions}
\title{Compute pairwise co-methylation statistics for cytosine sites within regions}
\usage{
computeCoMethylatedPositions(
  methylationData,
  regions,
  minDistance = 150,
  maxDistance = 1000,
  minCoverage = 4,
  pValueThreshold = 0.01,
  alternative = "two.sided",
  test = "fisher",
  parallel = FALSE,
  BPPARAM = NULL,
  cores = NULL
)
}
\arguments{
\item{methylationData}{A \code{GRanges} object containing cytosine sites, annotated with
per-site ONT methylation calls (columns \code{ONT_Cm}, \code{ONT_C}, \code{readsN}, etc).}

\item{regions}{A \code{GRanges} object with list including genomic context such as gene and/or transposable
elements coordinates which possibly have DMRs, VMRs or PMDs.}

\item{minDistance}{Minimum distance (in bp) between two cytosines to consider for co-methylation (default: 150).}

\item{maxDistance}{Maximum distance (in bp) between two cytosines to consider (default: 1000).}

\item{minCoverage}{Minimum read coverage required for both cytosines in a pair (default: 4).}

\item{pValueThreshold}{FDR-adjusted p-value threshold for reporting significant co-methylation (default: 0.01).}

\item{alternative}{indicates the alternative hypothesis and must be one of
\code{"two.sided"}, \code{"greater"} or \code{"less"}. You can specify just
the initial letter. Only used in the 2 by 2 case. This is used only for
Fisher's test.}

\item{test}{Statistical test to use for co-methylation (\code{"fisher"}
for Fisher's exact [default], or \code{"permutation"} for chi-squared).
NOTE: highly recommended to do parallel when use permutation test.}

\item{parallel}{Logical; run in parallel if \code{TRUE}.}

\item{BPPARAM}{A \code{BiocParallelParam} object controlling parallel execution.
This value will automatically set when parallel is \code{TRUE}, also able to set as manually.}

\item{cores}{Integer number of workers (must not exceed BPPARAM$workers).
This value will automatically set as the maximum number of system workers,
also able to set as manually.}
}
\value{
A list of length equal to \code{regions}, where each entry is a \code{GInteractions} object
  of significant cytosine pairs (by strand), annotated with:
  \describe{
    \item{C1_C2}{number of reads methylated at both cytosines}
    \item{C1_only}{number methylated at only first cytosine}
    \item{C2_only}{number methylated at only second cytosine}
    \item{neither}{number methylated at neither cytosines}
    \item{strand}{The DNA strand ("+" or "-") on which the two CpGs reside.}
    \item{genomic_position}{The original region (from \code{regions}) containing
    this cytosines pair, formatted in UCSC or IGV style, e.g. \code{"chr1:1522971-1523970"}.}
    \item{p.value}{FDR-adjusted p-value for co-methylation association}
  }
}
\description{
\code{computeCoMethylatedPositions()} calculates pairwise co-methylation between all cytosine sites
within each given region, using ONT methylation calls annotated to each site.
For each pair of cytosines within the same strand and PMD, it builds a 2x2 contingency table
reflecting the overlap state of reads (both methylated, only one methylated, or neither),
performs a statistical test (Fisher's exact by default), and reports FDR-adjusted p-values.
}
\details{
Compute Co-Methylation Positions within Regions (CMPs)


Pairwise tests are performed separately for each strand (+ and -) within each region.
FDR correction is performed for all pairs within each region and strand.
}
\examples{
\dontrun{
# load the ONT methylation data and PMD data
data(ont_gr_GM18870_chr1_PMD_bins_1k)
data(ont_gr_GM18870_chr1_sorted_bins_1k)

# compute the co-methylations with Fisher's exact test
coMetylationFisher <- computeCoMethylatedPositions(
  ont_gr_GM18870_chr1_sorted_bins_1k,
  regions = ont_gr_GM18870_chr1_PMD_bins_1k[1:4],
  minDistance = 150,
  maxDistance = 1000,
  minCoverage = 4,
  pValueThreshold = 0.01,
  test = "fisher",
  parallel = FALSE)

# compute the co-methylations with Permuation test
coMetylationPermutation <- computeCoMethylatedPositions(
  ont_gr_GM18870_chr1_sorted_bins_1k,
  regions = ont_gr_GM18870_chr1_PMD_bins_1k[1:4],
  minDistance = 150,
  maxDistance = 1000,
  minCoverage = 4,
  pValueThreshold = 0.01,
  test = "permutation",
  parallel = FALSE) # highly recommended to set as TRUE
}

}
\seealso{
\code{\link{readONTbam}}, \code{\link{computePMDs}},
\code{\link{ontSampleGRangesList}}
}
\author{
Nicolae Radu Zabet and Young Jun Kim
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/statisticalTests.R
\name{.scoreTest}
\alias{.scoreTest}
\title{Score test}
\usage{
.scoreTest(m1, n1, m2, n2)
}
\arguments{
\item{m1}{the number of methylated reads in condition 1}

\item{n1}{the total number of methylated reads in condition 1}

\item{m2}{the number of methylated reads in condition 2}

\item{n2}{the total number of methylated reads in condition 2}
}
\value{
The p-values of the Score test
}
\description{
This function computes the p-values of the Score test.
}
\author{
Radu Zabet
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generalFunctions.R
\name{.validateMethylationDataList}
\alias{.validateMethylationDataList}
\title{Validate methylation data}
\usage{
.validateMethylationDataList(methylationDataList)
}
\arguments{
\item{methylationData}{the methylation data stored as a \code{\link{GRanges}}
object with six metadata columns (see \code{\link{methylationData}}).}
}
\description{
Checks whether the passed parameter has the correct format for methylation data
}
\author{
Radu Zabet
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generalFunctions.R
\name{.validateCoMethylationAlternative}
\alias{.validateCoMethylationAlternative}
\title{Validate alternative value for Co-methylation}
\usage{
.validateCoMethylationAlternative(alternative)
}
\arguments{
\item{test}{the alternative valuet used to Fisher exact's test must be one of
\code{"two.sided"}, \code{"greater"} or \code{"less"}. You can specify just
the initial letter.}
}
\description{
Checks whether the passed parameter is the alternative for detecting Co-methylation
}
\author{
Radu Zabet and Young Jun Kim
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/movingAverage.R
\name{.movingAverage}
\alias{.movingAverage}
\title{Moving average}
\usage{
.movingAverage(
  minPos,
  maxPos,
  pos,
  val,
  weights = 1,
  windowSizeHalf = 150,
  normalize = FALSE,
  kernelFunction = "triangular",
  lambda = 0.5
)
}
\arguments{
\item{minPos}{the first position of the input fragment}

\item{maxPos}{the last position of the input fragment}

\item{pos}{the positions where the vector has values}

\item{val}{the values of the input vector}

\item{weights}{the weights associated with the moving average}

\item{normalize}{\code{logical} value that indicates whether the moving 
average is normalised}

\item{kernelFunction}{a \code{character} indicating which kernel function to 
be used. Can be one of \code{"uniform"}, \code{"triangular"}, 
\code{"gaussian"} or \code{"epanechnicov"}.}

\item{lambda}{numeric value required for the Gaussian filter 
(\code{K(x) = exp(-lambda*x^2)})}

\item{windowSize}{the size of the moving average}
}
\value{
A numeric vector with the values computed by the moving average
}
\description{
This function computes the moving average for a vector that has missing 
values.
}
\author{
Radu Zabet
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/DMRcaller.R
\docType{data}
\name{PMDsNoiseFilterCG}
\alias{PMDsNoiseFilterCG}
\title{The PMDs between GM18501 and GM18876 using Noise_filter method}
\format{
The \code{\link[GenomicRanges]{GRanges}} element contain 5 metadata columns;
see \code{\link{computePMDs}}
}
\description{
A \code{GRangesList} object containing the PMDs between GM18501 and
GM18876 B-Lymphocyte cell lines in Homo sapiens
(see \code{\link{ontSampleGRangesList}}). The PMDs were computed on the
1.5 ~ 2 Mbp from Chromosome 1 with noise filter method using a triangular kernel
and a windowSize of 100 bp
}
\seealso{
\code{\link{filterPMDs}}, \code{\link{computePMDs}},
\code{\link{analyseReadsInsideRegionsForConditionPMD}}
and \code{\link{mergePMDsIteratively}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/regions.R
\name{.analyseReadsInsideRegionsOneSample}
\alias{.analyseReadsInsideRegionsOneSample}
\title{Analyse reads inside regions}
\usage{
.analyseReadsInsideRegionsOneSample(methylationData, regions)
}
\arguments{
\item{methylationData}{a \code{\link[GenomicRanges]{GRanges}} object with four metadata 
columns; see \code{\link{MethylationDataList}}.}

\item{regions}{a \code{\link[GenomicRanges]{GRanges}} object with the identified regions}
}
\value{
a \code{\link[GenomicRanges]{GRanges}} object with eaual sized tiles of the regions. 
The object consists of the following metadata.
\describe{
 \item{sumReadsM}{the number of methylated reads in condition 1}
 \item{sumReadsN}{the total number of reads in condition 1}
 \item{Proportion}{the proportion of methylated reads in condition 1}
}
}
\description{
Performs the analysis in all regions in a \code{\link[GenomicRanges]{GRanges}} object
}
\author{
Radu Zabet
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ONTdata.R
\name{selectCytosine}
\alias{selectCytosine}
\title{Select Cytosine Positions}
\usage{
selectCytosine(
  genome = BSgenome.Hsapiens.UCSC.hg38,
  context = c("CG", "CHG", "CHH"),
  chr = NULL,
  region = NULL
)
}
\arguments{
\item{genome}{A \code{BSgenome} object or the name (character) of a BSgenome package
to use as the reference genome. If a package name is given, it will be loaded
automatically (default: \code{BSgenome.Hsapiens.UCSC.hg38}).
\strong{Note:} When running on an HPC system, please ensure that the required
\code{BSgenome} package is already installed and loaded in advance.}

\item{context}{A character vector of one or more methylation contexts to include:
\code{"CG"}, \code{"CHG"}, and/or \code{"CHH"}.  Defaults to all three.}

\item{chr}{An optional character vector of chromosome names to restrict the
enumeration.  If \code{NULL}, all sequences in \code{genome} are used.}

\item{region}{An optional \code{\link[GenomicRanges]{GRanges}} object specifying
subregions to keep.  Requires \code{chr} to be non-\code{NULL}.}
}
\value{
A \code{\link[GenomicRanges]{GRanges}} object with one 1-bp range per
cytosine, and two metadata columns:
\describe{
  \item{context}{A factor indicating the context (\code{"CG"},
    \code{"CHG"}, or \code{"CHH"}).}
  \item{trinucleotide_context}{Character or factor giving the
    surrounding trinucleotide sequence (or \code{NA} for \code{"CG"}).}
}
}
\description{
Constructs a \code{\link[GenomicRanges]{GRanges}} of all cytosine positions
in the specified BSgenome (or BSgenome package name), optionally filtering
by methylation context (\code{"CG"}, \code{"CHG"}, \code{"CHH"}), by
chromosome, and by genomic region.
}
\details{
This function enumerate cytosine positions in a BSgenome reference
}
\examples{
\dontrun{
library(BSgenome.Hsapiens.UCSC.hg38)

# Only CpG sites on chr1 and chr2:
gr_chr1_2 <- selectCytosine(context="CG", chr=c("chr1","chr2"))

# CHH sites in a specific region on chr3:
my_region <- GRanges("chr3", IRanges(1e6, 1e6 + 1e5))
gr_region <- selectCytosine(context="CHH", chr="chr3", region=my_region)
}

}
\seealso{
\code{\link{readONTbam}}, \code{\link{ontSampleGRangesList}}
}
\author{
Nicolae Radu Zabet and Young Jun Kim
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Replicates.R
\name{joinReplicates}
\alias{joinReplicates}
\title{Joins together two GRange objects in a single containing all the
replicates}
\usage{
joinReplicates(methylationData1, methylationData2, usecomplete = FALSE)
}
\arguments{
\item{methylationData1}{the methylation data stored as a \code{\link[GenomicRanges]{GRanges}}
object with four metadata columns (see \code{\link{methylationDataList}}).}

\item{methylationData2}{the methylation data stored as a \code{\link[GenomicRanges]{GRanges}}
object with four metadata columns (see \code{\link{methylationDataList}}).}

\item{usecomplete}{Boolean, determine wheter, when the two dataset differ for
number of cytosines, if the smaller dataset should be added with zero reads
to match the bigger dataset.}
}
\value{
returns a \code{\link[GenomicRanges]{GRanges}} object containing multiple metadata
columns with the reads from each object passed as parameter
}
\description{
This function joins together data that come from biological replicates to
perform analysis
}
\examples{


# load the methylation data
data(methylationDataList)

# Joins the wildtype and the mutant in a single object
joined_data <- joinReplicates(methylationDataList[["WT"]],
                              methylationDataList[["met1-3"]], FALSE)


}
\author{
Alessandro Pio Greco and Nicolae Radu Zabet
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generalFunctions.R
\name{.printGenomicRanges}
\alias{.printGenomicRanges}
\title{Print GenomicRanges}
\usage{
.printGenomicRanges(gr)
}
\arguments{
\item{gr}{the GenomicRanges object}
}
\value{
a vector of type \code{character} with template chr:start..end
}
\description{
This function that prints a GenomicRanges object
}
\author{
Radu Zabet
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/movingAverage.R
\name{.movingSum}
\alias{.movingSum}
\title{Moving sum}
\usage{
.movingSum(
  minPos,
  maxPos,
  pos,
  val,
  weights = 1,
  windowSize = 150,
  normalize = FALSE
)
}
\arguments{
\item{minPos}{the first position of the input fragment}

\item{maxPos}{the last position of the input fragment}

\item{pos}{the positions where the vector has values}

\item{val}{the values of the input vector}

\item{weights}{the weights associated with the moving sum}

\item{windowSize}{the size of the moving sum}

\item{normalize}{\code{logical} value that indicates whether the moving sum is normalised}
}
\value{
A numeric vector with the values computed by the moving sum
}
\description{
This function computes the moving sum for a vector that has missing values.
}
\author{
Radu Zabet
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generalFunctions.R
\name{.validateBamfile}
\alias{.validateBamfile}
\title{Validate BAM Filename}
\usage{
.validateBamfile(bamfile)
}
\arguments{
\item{bamfile}{Character scalar. Path to a BAM file.}
}
\value{
Invisibly `TRUE` if file exists and extension is ".bam"; otherwise errors.
}
\description{
Checks that `bamfile` exists and has a ".bam" extension.
}
\author{
Radu Zabet and Young Jun Kim
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generalFunctions.R
\name{.validateContext}
\alias{.validateContext}
\title{Validate context}
\usage{
.validateContext(context)

.validateContext(context)
}
\arguments{
\item{context}{the context in which the DMRs are computed (\code{"CG"},
\code{"CHG"} or \code{"CHH"})}

\item{length}{the expected length of the vector. If NULL any length is
allowed}

\item{test}{the  modified context used to call methylation information from bamfile must be one of
sequence context for \code{selectCytosine()} (e.g. \code{"CG"}, \code{"CHG"}, \code{"CHH"})}
}
\description{
Checks whether the passed parameter is the context

Checks whether the passed parameter is the modified context
}
\author{
Radu Zabet

Radu Zabet and Young Jun Kim
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ONTdata.R
\name{readONTbam}
\alias{readONTbam}
\title{Load ONT BAM, decode MM/ML, and count modified vs. unmodified reads}
\usage{
readONTbam(
  bamfile,
  ref_gr = NULL,
  modif = "C+m?",
  prob_thresh = 0.5,
  genome = BSgenome.Hsapiens.UCSC.hg38,
  context = "CG",
  chr = NULL,
  region = NULL,
  synonymous = FALSE,
  parallel = FALSE,
  BPPARAM = NULL,
  cores = NULL
)
}
\arguments{
\item{bamfile}{Path to an indexed ONT BAM with MM/ML tags.}

\item{ref_gr}{A \code{GRanges} of genomic cytosine positions to annotate.
If \code{NULL}, will be created via \code{selectCytosine()}
using \code{context}, \code{chr}, \code{region}.}

\item{modif}{Character vector of MM codes to treat as “modified”
(e.g. \code{"C+m?"}, \code{"C+h?"}, \code{"C+m."}).}

\item{prob_thresh}{Numeric in \[0,1\] — minimum ML probability to call a read
“modified.”}

\item{genome}{A \code{BSgenome} object such as \code{BSgenome.Hsapiens.UCSC.hg38}.
This is used to extract sequence context and must be loaded in advance.
\strong{Note:} When running on an HPC system,
please ensure that the required
\code{BSgenome} package is already installed and loaded in advance.}

\item{context}{Sequence context for \code{selectCytosine()}
(e.g. \code{"CG"}, \code{"CHG"}, \code{"CHH"}).}

\item{chr}{Chromosome names to restrict \code{selectCytosine()}.}

\item{region}{A \code{GRanges} to further subset \code{selectCytosine()}.}

\item{synonymous}{Logical (default: FALSE). If TRUE, include modified calls that match the specified context sequence (e.g. CGG),
even if the site was previously excluded due to deletion or mismatch.
For example, if a deletion occurs at position 234523, but the surrounding context still forms CGG,
then the modified C at 234523 will be retained (Nmod=1).}

\item{parallel}{Logical. If TRUE, automatically detect your system condition and
decoding will use parallel threads via BiocParallel::.
If FALSE (default), decoding is done serially.}

\item{BPPARAM}{A \code{BiocParallelParam} object controlling parallel execution.
This value will automatically set when parallel is \code{TRUE}, also able to set as manually.}

\item{cores}{Integer number of workers (must not exceed BPPARAM$workers).
This value will automatically set as the maximum number of system workers,
also able to set as manually.}
}
\value{
A \code{GRanges} of the same length as \code{ref_gr}, with four
  additional metadata columns:
  \describe{
    \item{ONT_Cm}{comma-delimited read‐indices called modified}
    \item{ONT_C}{comma-delimited read‐indices covering but unmodified}
    \item{readsM}{integer count of modified reads per site}
    \item{readsN}{integer count of same‐strand reads covering each site}
  }
}
\description{
\code{readONTbam()} takes an indexed Nanopore BAM file with MM/ML tags,
decodes each read’s per-C modification probabilities, and overlays
them on a \code{GRanges} of candidate cytosine sites.
It returns a copy of \code{ref_gr} augmented with:
\itemize{
  \item \code{ONT_Cm} — comma-delimited read‐indices called “modified”
  \item \code{ONT_C}  — comma-delimited read‐indices covering but _not_ modified
  \item \code{readsM} — count of modified reads at each site
  \item \code{readsN} — total same-strand coverage at each site
}

You can either supply your own \code{ref_gr} (e.g.\ from \code{selectCytosine()})
or leave it \code{NULL} and pass \code{context}, \code{chr}, \code{region}
to build \code{ref_gr} on the fly.
}
\details{
This function read and annotate ONT MM/ML tags against a cytosine reference
}
\examples{
\dontrun{
library(DMRcaller)
library(BSgenome.Hsapiens.UCSC.hg38)

# set the bam file directory
bam_path <- system.file("extdata", "scanBamChr1Random5.bam", package="DMRcaller")

# read ONTbam file (chromosome 1 only) in CG context with BSgenome.Hsapiens.UCSC.hg38
ONTSampleGRanges <- readONTbam(bamfile = bam_path, ref_gr = NULL, modif = "C+m?",
                         prob_thresh = 0.50,genome = BSgenome.Hsapiens.UCSC.hg38,
                         context = "CG", chr = "chr1", region = NULL,
                         synonymous = FALSE, parallel = FALSE, BPPARAM = NULL)
}

}
\seealso{
\code{\link{selectCytosine}}, \code{\link{computeDMRs}},
\code{\link{computePMDs}}, \code{\link{computeCoMethylatedPositions}},
\code{\link{computeCoMethylatedRegions}}, \code{\link{filterVMRsONT}},
\code{\link{ontSampleGRangesList}}, \code{\link{scanBamChr1Random5}}
}
\author{
Nicolae Radu Zabet and Young Jun Kim
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/DMRcaller.R
\docType{data}
\name{ont_gr_GM18870_chr1_sorted_bins_1k}
\alias{ont_gr_GM18870_chr1_sorted_bins_1k}
\title{The ONT methylation data example}
\format{
The \code{GRanges} elements contain four
  additional metadata columns:
  \describe{
    \item{ONT_Cm}{comma-delimited read‐indices called modified}
    \item{ONT_C}{comma-delimited read‐indices covering but unmodified}
    \item{readsM}{integer count of modified reads per site}
    \item{readsN}{integer count of same‐strand reads covering each site}
  }
}
\source{
data from https://genome.cshlp.org/content/34/11/2061.
}
\description{
A \code{GRanges} object containing cytosine sites, annotated with
  per-site ONT methylation calls
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/loadData.R
\name{.joinMethylationData}
\alias{.joinMethylationData}
\title{Join methylation data}
\usage{
.joinMethylationData(
  cx1,
  cx2,
  sample1_name = "Sample1",
  sample2_name = "Sample2"
)
}
\arguments{
\item{cx1}{a \code{\link{GRanges}} object of CX reports (Bismark) with three
metadata columns the cytosine context (CG, CHG or CHH), the number of
methylated reads and the total number of reads at that position in condition
1}

\item{cx2}{a \code{\link{GRanges}} object of CX reports (Bismark) with three
metadata columns the cytosine context (CG, CHG or CHH), the number of
methylated reads and the total number of reads at that position in condition
2}
}
\value{
a the methylation data stored as a \code{\link{GRanges}}
object with six metadata columns (see \code{\link{methylationDataList}}).
}
\description{
This function takes as input two \code{\link{GRanges}} objects of CX reports
(Bismark) for two conditions and  returns a \code{\link{GRanges}} object with
five metadata columns
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generalFunctions.R
\name{.validateModif}
\alias{.validateModif}
\title{Validate Modification Tag in BAM}
\usage{
.validateModif(modif, bamfile, chr, genome)
}
\arguments{
\item{modif}{Character scalar. The MM tag code to validate (e.g. "C+m?").}

\item{bamfile}{Character path to BAM (must exist plus .bai index).}

\item{chr}{Optional character vector of chromosome(s) to restrict scan.}

\item{genome}{A BSgenome, used to determine seqlengths if `chr` is set.}
}
\value{
Invisibly `TRUE` if `modif` is found among the MM tags; otherwise errors.
}
\description{
Scans the BAM for MM/ML tags in the specified region (or whole file)
and checks that `modif` is one of the observed modification codes.
}
\author{
Radu Zabet and Young Jun Kim
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Replicates.R
\name{.computeBetaRegSingle}
\alias{.computeBetaRegSingle}
\title{Beta regression}
\usage{
.computeBetaRegSingle(y, condition)
}
\arguments{
\item{y}{vector containing the proportions (readsM/readsN) for a single cytosine.}

\item{condition}{vector containing the conditions of the experiment.}
}
\value{
The p-values of the beta regression test.
}
\description{
This function computes the p-values of the beta regression test
}
\author{
Alessandro Pio Greco and Nicolae Radu Zabet
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/regions.R
\name{.analyseReadsInsideRegionsForCondition}
\alias{.analyseReadsInsideRegionsForCondition}
\title{Analyse reads inside regions}
\usage{
.analyseReadsInsideRegionsForCondition(
  regions,
  methylationData,
  label = "",
  context = ""
)
}
\arguments{
\item{regions}{a \code{\link[GenomicRanges]{GRanges}} object with the identified regions}

\item{methylationData}{a \code{\link[GenomicRanges]{GRanges}} object with five metadata 
columns see \code{\link{methylationDataList}}}
}
\value{
a \code{\link[GenomicRanges]{GRanges}} object with eaual sized tiles of the regions. 
The object consists of the following metadata 
\describe{
 \item{sumReadsM}{the number of methylated reads in condition 1}
 \item{sumReadsN}{the total number of reads in condition 1}
 \item{proportion}{the proportion of methylated reads in condition 1}
 \item{cytosinesCount}{the number of cytosines in the correct context} 
}
}
\description{
Performs the analysis in all regions in a \code{\link[GenomicRanges]{GRanges}} object
}
\author{
Radu Zabet
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/regions.R
\name{.analyseReadsInsideRegionsPMDs}
\alias{.analyseReadsInsideRegionsPMDs}
\title{Analyse reads inside regions}
\usage{
.analyseReadsInsideRegionsPMDs(methylationData, regions)
}
\arguments{
\item{methylationData}{a \code{\link[GenomicRanges]{GRanges}} object with five metadata 
columns see \code{\link{methylationDataList}}}

\item{regions}{a \code{\link[GenomicRanges]{GRanges}} object with the identified regions}
}
\value{
a \code{\link[GenomicRanges]{GRanges}} object with eaual sized tiles of the regions. 
The object consists of the following metadata 
\describe{
 \item{sumReadsM}{the number of methylated reads}
 \item{sumReadsN}{the total number of reads}
 \item{proportion}{the proportion of methylated reads}
 \item{ONT_Cm}{comma-delimited modified read‐indices}
 \item{ONT_C}{comma-delimited read‐indices covering but unmodified}
 \item{cytosinesCount}{the number of cytosines in the correct context} 
}
}
\description{
Performs the analysis in all regions in a \code{\link[GenomicRanges]{GRanges}} object for 
PMD calculation
}
\author{
Radu Zabet and Young Jun Kim
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/loadData.R
\name{readBismark}
\alias{readBismark}
\title{Read Bismark}
\usage{
readBismark(file)
}
\arguments{
\item{file}{The filename (including path) of the methylation
(CX report generated by Bismark) to be read.}
}
\value{
the methylation data stored as a \code{\link{GRanges}}
object with four metadata columns (see \code{\link{methylationDataList}}).
}
\description{
This function takes as input a CX report file produced by Bismark
and returns a \code{\link{GRanges}} object with four metadata columns
The file represents the bisulfite sequencing methylation data.
}
\examples{

# load methylation data object
data(methylationDataList)

# save the one datasets into a file
saveBismark(methylationDataList[["WT"]], "chr3test_a_thaliana_wt.CX_report")

# load the data
methylationDataWT <- readBismark("chr3test_a_thaliana_wt.CX_report")

#check that the loading worked
all(methylationDataWT == methylationDataList[["WT"]])

}
\author{
Nicolae Radu Zabet and Jonathan Michael Foonlan Tsang
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/profile.R
\name{computeOverlapProfile}
\alias{computeOverlapProfile}
\title{Compute Overlaps Profile}
\usage{
computeOverlapProfile(
  subRegions,
  largeRegion,
  windowSize = floor(width(largeRegion)/500),
  binary = TRUE,
  cores = 1
)
}
\arguments{
\item{subRegions}{a \code{\link{GRanges}} object with the sub regions; e.g.
can be the DMRs.}

\item{largeRegion}{a \code{\link{GRanges}} object with the region where to
compute the overlaps; e.g. a chromosome}

\item{windowSize}{The \code{largeRegion} is partitioned into equal sized
tiles of width \code{windowSize}.}

\item{binary}{a value indicating whether to count 1 for each overlap or to
compute the width of the overlap}

\item{cores}{the number of cores used to compute the DMRs.}
}
\value{
a \code{\link{GRanges}} object with equal sized tiles of the regions.
The object has one metadata file \code{score} which represents: the number of
subRegions overlapping with the tile, in the case of \code{binary = TRUE},
and the width of the subRegions overlapping with the tile , in the case of
\code{binary = FALSE}.
}
\description{
This function computes the distribution of a subset of regions
(\code{\link{GRanges}} object)  over a large region (\code{\link{GRanges}}
object)
}
\examples{
# load the methylation data
data(methylationDataList)

# load the DMRs in CG context
data(DMRsNoiseFilterCG)

# the coordinates of the area to be plotted
largeRegion <- GRanges(seqnames = Rle("Chr3"), ranges = IRanges(1,1E5))

# compute overlaps distribution
hotspots <- computeOverlapProfile(DMRsNoiseFilterCG, largeRegion,
           windowSize = 10000, binary = FALSE)
}
\seealso{
\code{\link{plotOverlapProfile}}, \code{\link{filterDMRs}},
\code{\link{computeDMRs}} and \code{\link{mergeDMRsIteratively}}
}
\author{
Nicolae Radu Zabet
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/computePMDs.R
\name{mergePMDsIteratively}
\alias{mergePMDsIteratively}
\title{Merge PMDs iteratively}
\usage{
mergePMDsIteratively(
  PMDs,
  minGap = 200,
  respectSigns = TRUE,
  methylationData,
  context = "CG",
  minReadsPerCytosine = 4,
  minMethylation = 0.4,
  maxMethylation = 0.6,
  parallel = FALSE,
  BPPARAM = NULL,
  cores = NULL
)
}
\arguments{
\item{PMDs}{the list of PMDs as a \code{\link[GenomicRanges]{GRanges}} object; 
e.g. see \code{\link{computePMDs}}}

\item{minGap}{PMDs separated by a gap of at least \code{minGap} are not 
merged.}

\item{respectSigns}{logical value indicating whether to respect the sign when 
joining PMDs.}

\item{methylationData}{the methylation data in GRanges
(see \code{\link{ontSampleGRangesList}}).}

\item{context}{the context in which the PMDs are computed (\code{"CG"}, 
\code{"CHG"} 
or \code{"CHH"}).}

\item{minReadsPerCytosine}{two adjacent PMDs are merged only if the number of 
reads per cytosine of the new DMR is higher than \code{minReadsPerCytosine}.}

\item{minMethylation}{Numeric [0,1]; minimum mean methylation proportion.}

\item{maxMethylation}{Numeric [0,1]; maximum mean methylation proportion.}

\item{parallel}{Logical; run in parallel if \code{TRUE}.}

\item{BPPARAM}{A \code{BiocParallelParam} object controlling parallel execution.
This value will automatically set when parallel is \code{TRUE}, also able to set as manually.}

\item{cores}{Integer number of workers (must not exceed BPPARAM$workers).
This value will automatically set as the maximum number of system workers,
also able to set as manually.}
}
\value{
the reduced list of PMDs as a \code{\link[GenomicRanges]{GRanges}} object; 
e.g. see \code{\link{computePMDs}}
}
\description{
This function takes a list of PMDs and attempts to merge PMDs while keeping 
the new PMDs statistically significant.
}
\examples{
# load the ONT methylation data 
data(ontSampleGRangesList)

# load the PMDs in CG context they were computed with minGap = 200
data(PMDsNoiseFilterCG)


# merge the PMDs 
PMDsNoiseFilterCGLarger <- mergePMDsIteratively(PMDsNoiseFilterCG[1:100], 
                           minGap = 500, respectSigns = TRUE, 
                           ontSampleGRangesList[["GM18501"]], context = "CG", 
                           minReadsPerCytosine = 4, minMethylation = 0.4, 
                           maxMethylation = 0.6, cores = 1)

\dontrun{
# set genomic coordinates where to compute PMDs
chr1_ranges <- GRanges(seqnames = Rle("chr1"), ranges = IRanges(1E6+5E5,2E6))

# compute PMDs and remove gaps smaller than 200 bp
PMDsNoiseFilterCG200 <- computePMDs(ontSampleGRangesList[["GM18501"]], 
                       regions = chr1_ranges, context = "CG", method = "noise_filter", 
                       windowSize = 100, kernelFunction = "triangular",  
                       minCytosinesCount = 1, minMethylation = 0.4, 
                       maxMethylation = 0.6, minGap = 0, minSize = 200, 
                       minReadsPerCytosine = 1, cores = 1)
PMDsNoiseFilterCG0 <- computePMDs(ontSampleGRangesList[["GM18501"]], 
                       regions = chr1_ranges, context = "CG", method = "noise_filter", 
                       windowSize = 100, kernelFunction = "triangular", 
                       minCytosinesCount = 1, minMethylation = 0.4, 
                       maxMethylation = 0.6, minGap = 0, minSize = 0, 
                       minReadsPerCytosine = 1, cores = 1)
PMDsNoiseFilterCG0Merged200 <- mergePMDsIteratively(PMDsNoiseFilterCG0, 
                             minGap = 200, respectSigns = TRUE, 
                             ontSampleGRangesList[["GM18501"]], context = "CG",  
                             minReadsPerCytosine = 4, minMethylation = 0.4, 
                             maxMethylation = 0.6, cores = 1)                      

#check that all newley computed PMDs are identical
print(all(PMDsNoiseFilterCG200 == PMDsNoiseFilterCG0Merged200))

}

}
\seealso{
\code{\link{filterPMDs}}, \code{\link{computePMDs}}, 
\code{\link{analyseReadsInsideRegionsForCondition}} and 
\code{\link{PMDsNoiseFilterCG}}
}
\author{
Nicolae Radu Zabet and Young Jun Kim
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/computeDMRs.R
\name{.computeDMRsBins}
\alias{.computeDMRsBins}
\title{This function computes the differentially methylated regions between two conditions
using the bins method.}
\usage{
.computeDMRsBins(
  methylationData1,
  methylationData2,
  regions = NULL,
  context = "CG",
  binSize = 100,
  test = "fisher",
  pValueThreshold = 0.01,
  minCytosinesCount = 4,
  minProportionDifference = 0.4,
  minGap = 200,
  minSize = 50,
  minReadsPerCytosine = 4,
  cores = 1,
  BPPARAM = BPPARAM
)
}
\description{
This function computes the differentially methylated regions between two conditions
using the bins method.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/statisticalTests.R
\name{.computeaAjustedPValuesInDMRs}
\alias{.computeaAjustedPValuesInDMRs}
\title{Compute adjusted p-values in DMRs}
\usage{
.computeaAjustedPValuesInDMRs(test, DMRs, alternative = "two.sided")
}
\arguments{
\item{test}{a \code{character} indicating the statistical test used .
(\code{"fisher"} for Fisher's exact test or \code{"score"} for Score test).}

\item{DMRs}{a \code{\link{GRanges}} object with the DMRs.}

\item{alternative}{indicates the alternative hypothesis and must be one of
\code{"two.sided"}, \code{"greater"} or \code{"less"}. You can specify just
the initial letter. Only used in the 2 by 2 case. This is used only for
Fisher's test.}
}
\value{
The adjuested p-values of the statistical test.
}
\description{
This function computes the adjusted p-values (using Benjamini & Hochberg
method) using one the statistical tests
}
\author{
Radu Zabet
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/loadData.R
\name{readBismarkPool}
\alias{readBismarkPool}
\title{Read Bismark pool}
\usage{
readBismarkPool(files)
}
\arguments{
\item{files}{The filenames (including path) of the methylation
(CX report generated with Bismark) to be read}
}
\value{
the methylation data stored as a \code{\link{GRanges}}
object with four metadata columns (see \code{\link{methylationDataList}}).
}
\description{
This function takes as input a vector of CX report file produced by Bismark
and returns a \code{\link{GRanges}} object with four metadata columns
(see \code{\link{methylationDataList}}). The file represents the pooled
bisulfite sequencing data.
}
\examples{

# load methylation data object
data(methylationDataList)

# save the two datasets
saveBismark(methylationDataList[["WT"]],
           "chr3test_a_thaliana_wt.CX_report")
saveBismark(methylationDataList[["met1-3"]],
           "chr3test_a_thaliana_met13.CX_report")

# reload the two datasets and pool them
filenames <- c("chr3test_a_thaliana_wt.CX_report",
               "chr3test_a_thaliana_met13.CX_report")
methylationDataPool <- readBismarkPool(filenames)

}
\author{
Nicolae Radu Zabet and Jonathan Michael Foonlan Tsang
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/profile.R
\name{computeMethylationProfile}
\alias{computeMethylationProfile}
\title{Compute methylation profile}
\usage{
computeMethylationProfile(
  methylationData,
  region,
  windowSize = floor(width(region)/500),
  context = "CG"
)
}
\arguments{
\item{methylationData}{the methylation data stored as a \code{\link{GRanges}}
object with four metadata columns (see \code{\link{methylationDataList}}).}

\item{region}{a \code{\link{GRanges}} object with the regions where to
compute the DMRs.}

\item{windowSize}{a \code{numeric} value indicating the size of the window in
which methylation is averaged.}

\item{context}{the context in which the DMRs are computed (\code{"CG"},
\code{"CHG"} or \code{"CHH"}).}
}
\value{
a \code{\link{GRanges}} object with equal sized tiles of the
\code{region}. The object consists of the following metadata
\describe{
 \item{sumReadsM}{the number of methylated reads.}
 \item{sumReadsN}{the total number of reads.}
 \item{Proportion}{the proportion of methylated reads.}
 \item{cytosinesCount}{the number of cytosines in the regions}
 \item{context}{the context (\code{"CG"}, \code{"CHG"} or \code{"CHH"}).}
}
}
\description{
This function computes the low resolution profiles for the bisulfite
sequencing data.
}
\examples{

# load the methylation data
data(methylationDataList)

# the region where to compute the profile
region <- GRanges(seqnames = Rle("Chr3"), ranges = IRanges(1,1E6))

# compute low resolution profile in 20 Kb windows
lowResProfileWTCHH <- computeMethylationProfile(methylationDataList[["WT"]],
                     region, windowSize = 20000, context = "CHH")

\dontrun{
# compute low resolution profile in 10 Kb windows
lowResProfileWTCG <- computeMethylationProfile(methylationDataList[["WT"]],
                     region, windowSize = 10000, context = "CG")

lowResProfileMet13CG <- computeMethylationProfile(
                     methylationDataList[["met1-3"]],  region,
                     windowSize = 10000, context = "CG")
}

}
\seealso{
\code{\link{plotMethylationProfileFromData}},
\code{\link{plotMethylationProfile}}, \code{\link{methylationDataList}}
}
\author{
Nicolae Radu Zabet and Jonathan Michael Foonlan Tsang
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Replicates.R
\name{.analyseReadsInsideBinsReplicates}
\alias{.analyseReadsInsideBinsReplicates}
\title{Analyse reads inside regions}
\usage{
.analyseReadsInsideBinsReplicates(
  methylationData,
  bins,
  currentRegion,
  condition,
  pseudocountM,
  pseudocountN
)
}
\arguments{
\item{methylationData}{a \code{\link[GenomicRanges]{GRanges}} object with five metadata
columns; see \code{\link{methylationDataList}}}

\item{currentRegion}{a \code{\link[GenomicRanges]{GRanges}} object with the identified regions}

\item{condition}{The vector containing the two conditions for the experiment.}

\item{pseudocountM}{numerical value to be added to the methylated reads
before modelling beta regression.}

\item{pseudocountN}{numerical value to be added to the total reads
before modelling beta regression.}
}
\value{
a \code{\link[GenomicRanges]{GRanges}} object with eaual sized tiles of the regions.
The object consists of the following metadata
\describe{
 \item{sumReadsM1}{the number of methylated reads in condition 1}
 \item{sumReadsN1}{the total number of reads in condition 1}
 \item{proportion1}{the proportion of methylated reads in condition 1}
 \item{sumReadsM2}{the number of methylated reads in condition 2}
 \item{sumReadsN2}{the total number of reads in condition 2}
 \item{proportion2}{the proportion of methylated reads in condition 2}
 \item{cytosinesCount}{the number of cytosines in the correct context}
}
}
\description{
Performs the analysis in equal width regions of an \code{\link[GenomicRanges]{GRanges}}
object
}
\author{
Nicolae Radu Zabet and Alessandro Pio Greco
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/computePMDs.R
\name{.computePMDsBins}
\alias{.computePMDsBins}
\title{This function computes the partially methylated regions 
using the bins method.}
\usage{
.computePMDsBins(
  methylationData,
  regions = NULL,
  context = "CG",
  binSize = 100,
  minCytosinesCount = 4,
  minMethylation = 0.4,
  maxMethylation = 0.6,
  minGap = 200,
  minSize = 50,
  minReadsPerCytosine = 4,
  cores = 1,
  BPPARAM = BPPARAM
)
}
\description{
This function computes the partially methylated regions 
using the bins method.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/regions.R
\name{.splitGRangesEqualy}
\alias{.splitGRangesEqualy}
\title{.splitGRangesEqualy}
\usage{
.splitGRangesEqualy(regions, breaks = 1)
}
\arguments{
\item{regions}{a \code{\link[GenomicRanges]{GRanges}} object}

\item{breaks}{number of elements to break the GRanges}
}
\value{
a \code{list} consisting of GRanges objects of same total width
}
\description{
This function splits a set of GRanges into a list of equally width GRanges 
objects
}
\author{
Radu Zabet
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Replicates.R
\name{.computeAdjuestedPValuesReplicates}
\alias{.computeAdjuestedPValuesReplicates}
\title{Compute adjusted p-values}
\usage{
.computeAdjuestedPValuesReplicates(proportions, condition, cores)
}
\arguments{
\item{condition}{vector containing the conditions of the experiment.}

\item{cores}{the number of cores to be used to fit the model.}

\item{conditions}{matrix containing the proportions (readsM/readsN) for all the cytosines.}
}
\value{
The adjusted p-values of the statistical test.
}
\description{
This function computes the adjusted p-values (using Benjamini & Hochberg
method)
}
\author{
Nicolae Radu Zabet and Alessandro Pio Greco
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/computeVMDs.R
\name{filterVMDs}
\alias{filterVMDs}
\title{Filter VMDs}
\usage{
filterVMDs(
  methylationData,
  potentialVMDs,
  context = "CG",
  minCytosinesCount = 4,
  minReadsPerCytosine = 3,
  sdCutoffMethod = "per.high",
  percentage = 0.05,
  parallel = FALSE,
  BPPARAM = NULL,
  cores = NULL
)
}
\arguments{
\item{methylationData}{the methylation data in condition
(see \code{\link{ontSampleGRangesList}}).}

\item{potentialVMDs}{a \code{\link{GRanges}} object with potential VMDs 
where to compute the VMDs. This can be a a list of gene and/or transposable 
elements coordinates.}

\item{context}{the context in which the VMDs are computed (\code{"CG"}, 
\code{"CHG"} or \code{"CHH"}).}

\item{minCytosinesCount}{VMDs with less cytosines in the specified context 
than \code{minCytosinesCount} will be discarded.}

\item{minReadsPerCytosine}{VMDs with the average number of reads lower than 
\code{minReadsPerCytosine} are discarded.}

\item{sdCutoffMethod}{Character string specifying how to determine the cutoff
for filtering VMDs based on their methylation variance (weighted standard deviation).
Available options are:
\describe{
  \item{\code{"per.high"}}{Selects the top \code{percentage} of regions with the highest variance (standard deviation).}
  \item{\code{"per.low"}}{Selects the bottom \code{percentage} of regions with the lowest variance.}
  \item{\code{"EDE.high"}}{Uses the elbow point (inflection/knee) from the descendingly sorted variance values to determine a data-driven high-variance cutoff. Retains regions with SD above this elbow point.}
  \item{\code{"EDE.low"}}{Uses the elbow point from the ascendingly sorted variance values to define a low-variance cutoff. Retains regions with SD below this point.}
}
This allows either quantile-based filtering or automatic detection of variance thresholds based on distribution shape.}

\item{percentage}{Numeric cutoff used when \code{sdCutoffMethod} is set to
\code{"per.high"} or \code{"per.low"}. Represents the quantile threshold:
for example, \code{percentage = 0.05} keeps the top 5\% or bottom 5\% of
bins based on weighted standard deviation, depending on the selected method.}

\item{parallel}{Logical; run in parallel if \code{TRUE}.}

\item{BPPARAM}{A \code{BiocParallelParam} object controlling parallel execution.
This value will automatically set when parallel is \code{TRUE}, also able to set as manually.}

\item{cores}{Integer number of workers (must not exceed BPPARAM$workers).
This value will automatically set as the maximum number of system workers,
also able to set as manually.}
}
\value{
a \code{\link{GRanges}} object with 9 metadata columns that contain 
the VMDs; see \code{\link{computeVMDs}}.
}
\description{
This function verifies whether a set of potential VMDs (e.g. genes, 
transposons, CpG islands) are variance methylated or not.
}
\examples{
# load the ONT methylation data 
data(ontSampleGRangesList)
# load the gene annotation data
data(GEs_hg38)

# select the transcript
transcript <- GEs_hg38[which(GEs_hg38$type == "transcript")]

# the regions where to compute the VMDs
regions <- GRanges(seqnames = Rle("chr1"), ranges = IRanges(1E6+5E5,2E6))
transcript <- transcript[overlapsAny(transcript, regions)]

# filter genes that are variance methylated in the two conditions
VMDsGenesCG <- filterVMDs(ontSampleGRangesList[["GM18501"]], 
               potentialVMDs = transcript, 
               context = "CG", sdCutoffMethod = "per.high", percentage = 0.05,
               minCytosinesCount = 4, minReadsPerCytosine = 3, cores = 1)

}
\seealso{
\code{\link{computeVMDs}} 
and \code{\link{analyseReadsInsideRegionsForCondition}}
}
\author{
Nicolae Radu Zabet and Young Jun Kim
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spatialCorrelation.R
\name{.computeMethylationDataSpatialCorrelation}
\alias{.computeMethylationDataSpatialCorrelation}
\title{Compute methylation data spatial correlation}
\usage{
.computeMethylationDataSpatialCorrelation(methylationData, regions, distance)
}
\arguments{
\item{methylationData}{the methylation data stored as a \code{\link[GenomicRanges]{GRanges}}
object with four metadata columns (see \code{\link{methylationDataList}}).}

\item{regions}{a \code{\link[GenomicRanges]{GRanges}} object with the regions where to
compute the correlation If NULL, the correlation is computed genome-wide.}

\item{distance}{the distance used when computing the correlation of
methylation levels}
}
\value{
correlation of the methylation levels for Cytosines located at a
certain distance apart.
}
\description{
This function computes the correlation of the methylation levels for
Cytosines located at a certain distance apart. The function returns a the
correlation of methylation levels at distance equal to the specified
thresholds.
}
\author{
Radu Zabet
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generalFunctions.R
\name{.validateChromosome}
\alias{.validateChromosome}
\title{Validate Chromosome Names}
\usage{
.validateChromosome(chr, genome)
}
\arguments{
\item{chr}{Character vector of chromosome names.}

\item{genome}{A validated BSgenome object.}
}
\value{
Invisibly `TRUE` if all `chr` are in `seqnames(genome)`; otherwise errors.
}
\description{
Checks that `chr` entries are present in the supplied BSgenome.
}
\author{
Radu Zabet and Young Jun Kim
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/computePMDs.R
\name{.computePMDsNeighbourhood}
\alias{.computePMDsNeighbourhood}
\title{This function computes the partially methylated domains
using the neighbourhood method. This assumes the computation of partially methylated 
cytosines followed by smart merging of these cytosines while keeping the new PMDs 
statistically significant.}
\usage{
.computePMDsNeighbourhood(
  methylationData,
  regions = NULL,
  context = "CG",
  minCytosinesCount = 4,
  minMethylation = 0.4,
  maxMethylation = 0.6,
  minGap = 200,
  minSize = 50,
  minReadsPerCytosine = 4,
  cores = 1,
  BPPARAM = BPPARAM
)
}
\description{
This function computes the partially methylated domains
using the neighbourhood method. This assumes the computation of partially methylated 
cytosines followed by smart merging of these cytosines while keeping the new PMDs 
statistically significant.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generalFunctions.R
\name{.stopIfNotAll}
\alias{.stopIfNotAll}
\title{.stopIfNotAll}
\usage{
.stopIfNotAll(exprs, errorMsg)
}
\arguments{
\item{errorMsg}{the error message to be printed}

\item{expr}{an array of logical expressions}
}
\description{
This function asses an expression and if this is FALSE stops the execution
and prints a message
}
\author{
Radu Zabet
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/DMRcaller.R
\docType{data}
\name{ontSampleGRangesList}
\alias{ontSampleGRangesList}
\title{The ONT methylation data list}
\format{
The \code{GRanges} elements contain six metadata columns
\describe{
 \item{context}{the context in which the DMRs are computed \code{"CG"}.}
 \item{trinucleotide_context}{the specific context of the cytosine (H is
                             replaced by the actual nucleotide).}
 \item{ONT_Cm}{comma-delimited read‐indices called modified}
 \item{ONT_C}{comma-delimited read‐indices covering but unmodified}
 \item{readsM}{the number of methylated reads.}
 \item{readsN}{the total number of reads.}
}
}
\source{
Each element was created by calling \code{bam} files with \code{\link{readONTbam}}
function which in \code{\link{DMRcaller}} package.
The sample pod5 files were from the nanopore dataset from 1000 genome project
\url{https://pmc.ncbi.nlm.nih.gov/articles/PMC10942501/}.
\url{https://s3.amazonaws.com/1000g-ont/index.html?prefix=pod5_data/GM18501_R9/}
\code{.pod5} files in the case of GM18501 and
\url{https://s3.amazonaws.com/1000g-ont/index.html?prefix=pod5_data/GM18876_R9/}
\code{.pod5} files in the case of GM18876 cell line.
For base-calling and alignment, run \code{dorado, ver.0.9.6}
\url{https://github.com/nanoporetech/dorado?tab=readme-ov-file#dna-models}
to generate the \code{bam} files with \code{dna_r10.4.1_e8.2_400bps_hac@v5.2.0} as basecalling model.
}
\description{
A \code{GRangesList} object containing the methylation data at each cytosine
location in the genome in  GM18501 and  GM18876 B-Lymphocyte cell lines in
Homo sapiens. The data only contains the 1.5 ~ 2 Mbp from Chromosome 1.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generalFunctions.R
\name{.validateCoMethylationStatTest}
\alias{.validateCoMethylationStatTest}
\title{Validate statistial test for Co-methylation}
\usage{
.validateCoMethylationStatTest(test)
}
\arguments{
\item{test}{the statistical test used to call Co-methylation (\code{"fisher"} for
Fisher's exact test, \code{"score"} for Score test or \code{"permutation"} for Permutation test).}
}
\description{
Checks whether the passed parameter is the statistical test for detecting Co-methylation
}
\author{
Radu Zabet and Young Jun Kim
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generalFunctions.R
\name{.validateStatisticalTest}
\alias{.validateStatisticalTest}
\title{Validate statistial test}
\usage{
.validateStatisticalTest(test)
}
\arguments{
\item{test}{the statistical test used to call DMRs (\code{"fisher"} for
Fisher's exact test or \code{"score"} for Score test).}
}
\description{
Checks whether the passed parameter is the statistical test
}
\author{
Radu Zabet
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Replicates.R
\name{.convertResult}
\alias{.convertResult}
\title{This function extracts the p-value for each cytosine from the betareg object}
\usage{
.convertResult(x)
}
\description{
This function extracts the p-value for each cytosine from the betareg object
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/statisticalTests.R
\name{.fisherTestPValue}
\alias{.fisherTestPValue}
\title{Fisher test}
\usage{
.fisherTestPValue(x, alternative = c("two.sided", "less", "greater"))
}
\arguments{
\item{x}{a vector with 4 values}
}
\value{
The p-values of Fisher's exact test.
}
\description{
This function computes the p-values of the Fisher's exact test.
}
\author{
Radu Zabet
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/extractGC.R
\name{extractGC}
\alias{extractGC}
\title{Extract GC}
\usage{
extractGC(methylationData, genome, contexts = c("ALL", "CG", "CHG", "CHH"))
}
\arguments{
\item{methylationData}{the methylation data stored as a \code{\link{GRanges}}
object with four metadata columns (see \code{\link{methylationDataList}}).}

\item{genome}{a BSgenome with the DNA sequence of the organism}

\item{contexts}{the context in which the DMRs are computed (\code{"ALL"}, 
\code{"CG"}, \code{"CHG"} or \code{"CHH"}).}
}
\value{
the a subset of \code{methylationData} consisting of all GC sites.
}
\description{
This function extracts GC sites in the genome
}
\examples{

\dontrun{
# load the genome sequence
if(!require("BSgenome.Athaliana.TAIR.TAIR9", character.only = TRUE)){
  if (!requireNamespace("BiocManager", quietly=TRUE))
  install.packages("BiocManager")
  BiocManager::install("BSgenome.Athaliana.TAIR.TAIR9")
}
library(BSgenome.Athaliana.TAIR.TAIR9)

# load the methylation data
data(methylationDataList)

methylationDataWTGpCpG <- extractGC(methylationDataList[["WT"]], 
                                    BSgenome.Athaliana.TAIR.TAIR9,
                                    "CG")

}

}
\author{
Ryan Merritt
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/DMRcaller.R
\docType{data}
\name{ont_gr_GM18870_chr1_PMD_bins_1k}
\alias{ont_gr_GM18870_chr1_PMD_bins_1k}
\title{Partially Methilated Domains example}
\format{
The \code{GRanges} elements contain seven metadata columns:
\describe{
 \item{context}{the context in which the PMDs was computed (\code{"CG"},
 \code{"CHG"} or \code{"CHH"}).}
 \item{sumReadsM}{the number of methylated reads.}
 \item{sumReadsN}{the total number of reads.}
 \item{proportion}{the proportion methylated reads filtered between
 \code{minMethylation} and \code{maxMethylation}.}
 \item{cytosinesCount}{the number of cytosines in the PMDs.}
}
}
\source{
data from https://genome.cshlp.org/content/34/11/2061.
}
\description{
Partially methylated domains called on chr1 in GM18870 cells called in 1Kb
bins with \code{computePMDs} function.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/regions.R
\name{.mergePMDsIteratively}
\alias{.mergePMDsIteratively}
\title{This takes a list of PMDs and attempts to merge PMDs while keeping the new 
PMDs filtering by the preset parameter}
\usage{
.mergePMDsIteratively(
  PMDs,
  minGap,
  respectSigns = TRUE,
  methylationData,
  minReadsPerCytosine = 4,
  minMethylation = 0.4,
  maxMethylation = 0.6
)
}
\description{
This takes a list of PMDs and attempts to merge PMDs while keeping the new 
PMDs filtering by the preset parameter
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Replicates.R
\name{computeDMRsReplicates}
\alias{computeDMRsReplicates}
\title{Compute DMRs}
\usage{
computeDMRsReplicates(
  methylationData,
  condition = NULL,
  regions = NULL,
  context = "CG",
  method = "neighbourhood",
  binSize = 100,
  test = "betareg",
  pseudocountM = 1,
  pseudocountN = 2,
  pValueThreshold = 0.01,
  minCytosinesCount = 4,
  minProportionDifference = 0.4,
  minGap = 200,
  minSize = 50,
  minReadsPerCytosine = 4,
  parallel = FALSE,
  BPPARAM = NULL,
  cores = NULL
)
}
\arguments{
\item{methylationData}{the methylation data containing all the conditions
for all the replicates.}

\item{condition}{a vector of strings indicating the conditions for each
sample in \code{methylationData}. Two different values are allowed
(for the two conditions).}

\item{regions}{a \code{\link[GenomicRanges]{GRanges}} object with the regions where to
compute the DMRs. If \code{NULL}, the DMRs are computed genome-wide.}

\item{context}{the context in which the DMRs are computed (\code{"CG"},
\code{"CHG"} or \code{"CHH"}).}

\item{method}{the method used to compute the DMRs \code{"neighbourhood"}
or \code{"bins"}). The \code{"neighbourhood"} method computates
differentially methylated cytosines. Finally, the \code{"bins"}
method partiones the genome into equal sized tilling bins and performs the
statistical test between the two conditions in each bin. For all three
methods, the cytosines or bins are then merged into DMRs without affecting
the inital parameters used when calling the differentiall methylated
cytosines/bins (p-value, difference in methylation levels, minimum number of
reads per cytosine).}

\item{binSize}{the size of the tiling bins in nucleotides. This parameter is
required only if the selected method is \code{"bins"}.}

\item{test}{the statistical test used to call DMRs (\code{"betareg"} for
Beta regression).}

\item{pseudocountM}{numerical value to be added to the methylated reads
before modelling beta regression.}

\item{pseudocountN}{numerical value to be added to the total reads
before modelling beta regression.}

\item{pValueThreshold}{DMRs with p-values (when performing the statistical
test; see \code{test}) higher or equal than \code{pValueThreshold} are
discarded. Note that we adjust the p-values using the Benjamini and
Hochberg's method to control the false discovery rate.}

\item{minCytosinesCount}{DMRs with less cytosines in the specified context
than \code{minCytosinesCount} will be discarded.}

\item{minProportionDifference}{DMRs where the difference in methylation
proportion between the two conditions is lower than
\code{minProportionDifference} are discarded.}

\item{minGap}{DMRs separated by a gap of at least \code{minGap} are not
merged. Note that only DMRs where the change in methylation is in the same
direction are joined.}

\item{minSize}{DMRs with a size smaller than \code{minSize} are discarded.}

\item{minReadsPerCytosine}{DMRs with the average number of reads lower than
\code{minReadsPerCytosine} are discarded.}

\item{parallel}{Logical; run in parallel if \code{TRUE}.}

\item{BPPARAM}{A \code{BiocParallelParam} object controlling parallel execution.
This value will automatically set when parallel is \code{TRUE}, also able to set as manually.}

\item{cores}{Integer number of workers (must not exceed BPPARAM$workers).
This value will automatically set as the maximum number of system workers,
also able to set as manually.}
}
\value{
the DMRs stored as a \code{\link[GenomicRanges]{GRanges}} object with the following
metadata columns:
\describe{
 \item{direction}{a number indicating whether the region lost (-1)  or gain
 (+1) methylation in condition 2 compared to condition 1.}
 \item{context}{the context in which the DMRs was computed (\code{"CG"},
 \code{"CHG"} or \code{"CHH"}).}
 \item{sumReadsM1}{the number of methylated reads in condition 1.}
 \item{sumReadsN1}{the total number of reads in condition 1.}
 \item{proportion1}{the proportion methylated reads in condition 1.}
 \item{sumReadsM2}{the number of methylated reads in condition 2.}
 \item{sumReadsN2}{the total number reads in condition 2.}
 \item{proportion2}{the proportion methylated reads in condition 2.}
 \item{cytosinesCount}{the number of cytosines in the DMR.}
 \item{regionType}{a string indicating whether the region lost (\code{"loss"})
 or gained (\code{"gain"}) methylation in condition 2 compared to condition 1.}
 \item{pValue}{the p-value (adjusted to control the false discovery rate with
 the Benjamini and Hochberg's method) of the statistical test when the DMR was
 called.}
}
}
\description{
This function computes the differentially methylated regions between
replicates with two conditions.
}
\examples{

\dontrun{
# starting with data joined using joinReplicates
data("syntheticDataReplicates")

# compute the DMRs in CG context with neighbourhood method

# creating condition vector
condition <- c("a", "a", "b", "b")

# computing DMRs using the neighbourhood method
DMRsReplicatesNeighbourhood <- computeDMRsReplicates(methylationData = syntheticDataReplicates,
                                                     condition = condition,
                                                     regions = NULL,
                                                     context = "CHH",
                                                     method = "neighbourhood",
                                                     test = "betareg",
                                                     pseudocountM = 1,
                                                     pseudocountN = 2,
                                                     pValueThreshold = 0.01,
                                                     minCytosinesCount = 4,
                                                     minProportionDifference = 0.4,
                                                     minGap = 200,
                                                     minSize = 50,
                                                     minReadsPerCytosine = 4,
                                                     cores = 1)
}
}
\author{
Alessandro Pio Greco and Nicolae Radu Zabet
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/coverage.R
\name{.countMethylationDataCoverage}
\alias{.countMethylationDataCoverage}
\title{Count methylation data coverage}
\usage{
.countMethylationDataCoverage(methylationData, regions, threshold)
}
\arguments{
\item{methylationData}{the methylation data stored as a \code{\link{GRanges}}
object with four metadata columns (see \code{\link{methylationDataList}}).}

\item{regions}{a \code{\link{GRanges}} object with the regions where to
compute the coverage. If NULL, the coverage is computed genome-wide.}

\item{threshold}{the threshold used when counting cytosines that have at
least a certain number of reads}
}
\value{
a the number of cytosines with more reads than the threshold in the
specified region
}
\description{
This function computes the number of cytosines with more reads than a
threshold in the methylation data
}
\author{
Radu Zabet
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generalFunctions.R
\name{.validateBPPARAM}
\alias{.validateBPPARAM}
\title{Validate or Auto-Select a BiocParallelParam}
\usage{
.validateBPPARAM(BPPARAM = NULL, progressbar = FALSE)
}
\arguments{
\item{BPPARAM}{A \code{BiocParallelParam} or `NULL`.}

\item{progressbar}{Logical; whether to show a progress bar.}
}
\value{
A valid `BiocParallelParam` object.
}
\description{
If `BPPARAM` is `NULL`, picks `SnowParam` or `MulticoreParam` based on env vars
or OS; otherwise checks class and fork compatibility.
}
\author{
Radu Zabet and Young Jun Kim
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/computeCoMethylatedRegions.R
\name{.validateCorrelationStatTest}
\alias{.validateCorrelationStatTest}
\title{Validate statistial test for Correlation between two Regions}
\usage{
.validateCorrelationStatTest(test)
}
\arguments{
\item{test}{the statistical test used to call Correlation Between Methylated Regions (CMRs) (\code{"pearson"},
\code{"kendall"} or \code{"spearman"} for Correlation test).}
}
\description{
Checks whether the passed parameter is the statistical test for detecting Correlation in Regions
}
\author{
Radu Zabet and Young Jun Kim
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Replicates.R
\name{.computeBetaReg}
\alias{.computeBetaReg}
\title{This function applies the .computeBetaRegSingle to the entire dataset}
\usage{
.computeBetaReg(y, condition, cores)
}
\description{
This function applies the .computeBetaRegSingle to the entire dataset
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generalFunctions.R
\name{.validateGRanges}
\alias{.validateGRanges}
\title{Validate GRanges}
\usage{
.validateGRanges(
  regions,
  methylationData,
  length = NULL,
  generateGenomeWide = TRUE,
  variableName = "regions",
  minLength = 0
)
}
\arguments{
\item{regions}{a \code{\link{GRanges}} object. If \code{NULL} and
\code{generateGenomeWide} is true it uses the \code{methylationData} to
compute the regions and returns this \code{\link{GRanges}} object}

\item{methylationData}{the methylation data stored as a \code{\link{GRanges}}
object with six metadata columns (see \code{\link{methylationData}}).}

\item{length}{the expected length of the vector. If \code{NULL} any length is
allowed.}

\item{generateGenomeWide}{logical value to indicate whether to compute the
regions that span over all the \code{methylationData}}
}
\value{
a \code{\link{GRanges}} object representing the regions
}
\description{
Checks whether the passed parameter is a \code{\link{GRanges}} object
}
\author{
Radu Zabet
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/localProfile.R
\name{.plotGeneticElements}
\alias{.plotGeneticElements}
\title{Plot genetic elements}
\usage{
.plotGeneticElements(gff, region, col)
}
\arguments{
\item{gff}{a \code{\link{GRanges}} object with all elements usually imported
from a GFF3 file}

\item{region}{the location on the genome where to draw the genetic elements}

\item{col}{the colors used to draw the exons and the transposable elemnets}
}
\description{
This function draws the genetic elements over a plot
}
\author{
Radu Zabet
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/regions.R
\name{.analyseReadsInsideBinsOneSample}
\alias{.analyseReadsInsideBinsOneSample}
\title{Analyse reads inside regions for one sample}
\usage{
.analyseReadsInsideBinsOneSample(methylationData, bins, currentRegion)
}
\arguments{
\item{methylationData}{a \code{\link[GenomicRanges]{GRanges}} object with five metadata 
columns; see \code{\link{methylationDataList}}.}

\item{bins}{a \code{\link[GenomicRanges]{GRanges}} object with the bins}
}
\value{
a \code{\link[GenomicRanges]{GRanges}} object with eaual sized tiles of the regions. 
The object consists of the following metadata 
\describe{
 \item{sumReadsM}{the number of methylated reads}
 \item{sumReadsN}{the total number of reads}
 \item{Proportion}{the proportion of methylated reads}
}
}
\description{
Performs the analysis in equal width regions of an \code{\link[GenomicRanges]{GRanges}} 
object
}
\author{
Radu Zabet
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Replicates.R
\name{.validateStatisticalTestReplicates}
\alias{.validateStatisticalTestReplicates}
\title{Validate statistial test}
\usage{
.validateStatisticalTestReplicates(test)
}
\arguments{
\item{test}{the statistical test used to call DMRs (\code{"betareg"} for
Beta regression.}
}
\description{
Checks whether the passed parameter is the statistical test
}
\author{
Alessandro Pio Greco and Nicolae Radu Zabet
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/computePMDs.R
\name{.computePMDsNoiseFilter}
\alias{.computePMDsNoiseFilter}
\title{This function computes the partially methylated regions 
using the noise filter method.}
\usage{
.computePMDsNoiseFilter(
  methylationData,
  regions = NULL,
  context = "CG",
  windowSize = 100,
  kernelFunction = "triangular",
  lambda = 0.5,
  minCytosinesCount = 4,
  minMethylation = 0.4,
  maxMethylation = 0.6,
  minGap = 200,
  minSize = 50,
  minReadsPerCytosine = 4,
  cores = 1,
  BPPARAM = BPPARAM
)
}
\description{
This function computes the partially methylated regions 
using the noise filter method.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/profile.R
\name{plotOverlapProfile}
\alias{plotOverlapProfile}
\title{Plot overlap profile}
\usage{
plotOverlapProfile(
  overlapsProfiles1,
  overlapsProfiles2 = NULL,
  names = NULL,
  labels = NULL,
  col = NULL,
  title = "",
  logscale = FALSE,
  maxValue = NULL
)
}
\arguments{
\item{overlapsProfiles1}{a  \code{\link{GRanges}} object with the overlaps
profile; see \code{\link{computeOverlapProfile}}.}

\item{overlapsProfiles2}{a  \code{\link{GRanges}} object with the overlaps
profile; see \code{\link{computeOverlapProfile}}. This is optional. For
example, one can be use \code{overlapsProfiles1} to display hypomethylated
regions and \code{overlapsProfiles2} the hypermethylated regions.}

\item{names}{a \code{vector} of \code{character} to add labels for the two
overlapsProfiles. This is an optinal parameter.}

\item{labels}{a \code{vector} of \code{character} used to add a subfigure
character to the plot. If \code{NULL} nothing is added.}

\item{col}{a \code{character} vector with the colours. It needs to contain 2
colours. If not or if \code{NULL}, the defalut colours will be used.}

\item{title}{the title of the plot.}

\item{logscale}{a \code{logical} value indicating if the colours are on
logscale or not.}

\item{maxValue}{a maximum value in a region. Used for the colour scheme.}
}
\value{
Invisibly returns \code{NULL}.
}
\description{
This function plots the distribution of a set of subregions on a large
region.
}
\examples{

# load the methylation data
data(methylationDataList)

# load the DMRs in CG context
data(DMRsNoiseFilterCG)

# the coordinates of the area to be plotted
largeRegion <- GRanges(seqnames = Rle("Chr3"), ranges = IRanges(1,1E5))

# compute overlaps distribution
hotspotsHypo <- computeOverlapProfile(DMRsNoiseFilterCG, largeRegion,
                 windowSize = 10000, binary = FALSE)

plotOverlapProfile(GRangesList("Chr3"=hotspotsHypo),
                   names = c("hypomethylated"), title = "CG methylation")

\dontrun{

largeRegion <- GRanges(seqnames = Rle("Chr3"), ranges = IRanges(1,1E6))

hotspotsHypo <- computeOverlapProfile(
               DMRsNoiseFilterCG[(DMRsNoiseFilterCG$regionType == "loss")],
               largeRegion, windowSize=2000, binary=TRUE, cores=1)

hotspotsHyper <- computeOverlapProfile(
               DMRsNoiseFilterCG[(DMRsNoiseFilterCG$regionType == "gain")],
               largeRegion, windowSize=2000, binary=TRUE, cores=1)

plotOverlapProfile(GRangesList("Chr3"=hotspotsHypo),
                   GRangesList("Chr3"=hotspotsHyper),
                   names=c("loss", "gain"), title="CG methylation")
}

}
\seealso{
\code{\link{computeOverlapProfile}}, \code{\link{filterDMRs}},
\code{\link{computeDMRs}} and \code{\link{mergeDMRsIteratively}}
}
\author{
Nicolae Radu Zabet
}
